Package: nanoparquet
Title: Read and Write 'Parquet' Files
Version: 0.5.1
Authors@R: c(
    person("Gábor", "Csárdi", role = c("aut", "cre"), email = "csardi.gabor@gmail.com"),
    person("Hannes", "Mühleisen", role = c("aut", "cph"), comment = c(ORCID = "0000-0001-8552-0029")),
    person("Google Inc.", role = "cph"),
    person("Apache Software Foundation", role = "cph"),
    person("Posit Software, PBC", role = "cph"),
    person("RAD Game Tools", role = "cph"),
    person("Valve Software", role = "cph"),
    person("Tenacious Software LLC", role = "cph"),
    person("Facebook, Inc.", role = "cph"))
Description: Self-sufficient reader and writer for flat 'Parquet' files.
    Can read most 'Parquet' data types. Can write many 'R' data types,
    including factors and temporal types. See docs for limitations.
Depends: R (>= 4.0.0)
License: MIT + file LICENSE
URL: https://github.com/r-lib/nanoparquet,
        https://nanoparquet.r-lib.org/
BugReports: https://github.com/r-lib/nanoparquet/issues
Encoding: UTF-8
Suggests: arrow, bit64, blob, DBI, duckdb (>= 1.4.0), hms, mockery,
        pillar, processx, rprojroot, spelling, testthat, tzdb, withr
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Config/Needs/website: tidyverse/tidytemplate, r-lib/pkgdown, dplyr, gt,
        gtExtras, knitr, nycflights13, prettyunits, quarto, rmarkdown,
        sessioninfo, svglite
Language: en-US
Biarch: true
Config/roxygen2/version: 7.3.3.9000
NeedsCompilation: yes
Packaged: 2026-04-20 10:55:25 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Hannes Mühleisen [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-8552-0029>),
  Google Inc. [cph],
  Apache Software Foundation [cph],
  Posit Software, PBC [cph],
  RAD Game Tools [cph],
  Valve Software [cph],
  Tenacious Software LLC [cph],
  Facebook, Inc. [cph]
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-20 11:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 10:35:08 UTC; unix

Package: nanoparquet
Title: Read and Write 'Parquet' Files
Version: 0.5.1
Authors@R: c(
    person("Gábor", "Csárdi", role = c("aut", "cre"), email = "csardi.gabor@gmail.com"),
    person("Hannes", "Mühleisen", role = c("aut", "cph"), comment = c(ORCID = "0000-0001-8552-0029")),
    person("Google Inc.", role = "cph"),
    person("Apache Software Foundation", role = "cph"),
    person("Posit Software, PBC", role = "cph"),
    person("RAD Game Tools", role = "cph"),
    person("Valve Software", role = "cph"),
    person("Tenacious Software LLC", role = "cph"),
    person("Facebook, Inc.", role = "cph"))
Description: Self-sufficient reader and writer for flat 'Parquet' files.
    Can read most 'Parquet' data types. Can write many 'R' data types,
    including factors and temporal types. See docs for limitations.
Depends: R (>= 4.0.0)
License: MIT + file LICENSE
URL: https://github.com/r-lib/nanoparquet,
        https://nanoparquet.r-lib.org/
BugReports: https://github.com/r-lib/nanoparquet/issues
Encoding: UTF-8
Suggests: arrow, bit64, blob, DBI, duckdb (>= 1.4.0), hms, mockery,
        pillar, processx, rprojroot, spelling, testthat, tzdb, withr
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Config/Needs/website: tidyverse/tidytemplate, r-lib/pkgdown, dplyr, gt,
        gtExtras, knitr, nycflights13, prettyunits, quarto, rmarkdown,
        sessioninfo, svglite
Language: en-US
Biarch: true
Config/roxygen2/version: 7.3.3.9000
NeedsCompilation: yes
Packaged: 2026-04-20 10:55:25 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Hannes Mühleisen [aut, cph] (ORCID:
    <https://orcid.org/0000-0001-8552-0029>),
  Google Inc. [cph],
  Apache Software Foundation [cph],
  Posit Software, PBC [cph],
  RAD Game Tools [cph],
  Valve Software [cph],
  Tenacious Software LLC [cph],
  Facebook, Inc. [cph]
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-20 11:30:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 16:19:19 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(pillar::obj_sum,nanoparquet_logical_type)
export(append_parquet)
export(infer_parquet_schema)
export(parquet_column_types)
export(parquet_info)
export(parquet_metadata)
export(parquet_options)
export(parquet_schema)
export(read_parquet)
export(read_parquet_info)
export(read_parquet_metadata)
export(read_parquet_schema)
export(write_parquet)
useDynLib(nanoparquet, .registration=TRUE)

# nanoparquet

<!-- badges: start -->
[![R-CMD-check](https://github.com/r-lib/nanoparquet/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/nanoparquet/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/nanoparquet)](https://cran.r-project.org/package=nanoparquet)
[![](http://cranlogs.r-pkg.org/badges/nanoparquet)](https://r-pkg.org/pkg/nanoparquet)
<!-- badges: end -->

`nanoparquet` is a reader and writer for a common subset of Parquet files.

## Features:

* Read and write flat (i.e. non-nested) Parquet files.
* Can read most [Parquet data types](https://nanoparquet.r-lib.org/reference/nanoparquet-types.html).
* Can read a subset of columns from a Parquet file.
* Can write many R data types, including factors and temporal types
  to Parquet.
* Can append a data frame to a Parquet file without first reading and then
  rewriting the whole file.
* Completely dependency free.
* Supports Snappy, Gzip and Zstd compression.
* [Competitive](
  https://nanoparquet.r-lib.org/dev/articles/benchmarks.html) with other
  tools in terms of speed, memory use and file size.

## Limitations:

* Nested Parquet types are not supported.
* Some Parquet logical types are not supported: `INTERVAL`,
  `UNKNOWN`.
* Only Snappy, Gzip and Zstd compression is supported.
* Encryption is not supported.
* Reading files from URLs is not supported.
* nanoparquet always reads the data (or the selected subset of it) into
  memory. It does not work with out-of-memory data in Parquet files like
  Apache Arrow and DuckDB does.

## Installation

Install the R package from CRAN:

```r
install.packages("nanoparquet")
```

## Usage

### Read

Call `read_parquet()` to read a Parquet file:
```r
df <- nanoparquet::read_parquet("example.parquet")
```

To see the columns of a Parquet file and how their types are mapped to
R types by `read_parquet()`, call `read_parquet_schema()` first:
```r
nanoparquet::read_parquet_schema("example.parquet")
```

Folders of similar-structured Parquet files (e.g. produced by Spark)
can be read like this:

```r
df <- data.table::rbindlist(lapply(
  Sys.glob("some-folder/part-*.parquet"),
  nanoparquet::read_parquet
))
```

### Write

Call `write_parquet()` to write a data frame to a Parquet file:
```r
nanoparquet::write_parquet(mtcars, "mtcars.parquet")
```

To see how the columns of the data frame will be mapped to Parquet types
by `write_parquet()`, call `infer_parquet_schema()` first:
```r
nanoparquet::infer_parquet_schema(mtcars)
```

### Inspect

Call `read_parquet_info()`, `read_parquet_schema()`, or
`read_parquet_metadata()` to see various kinds of metadata from a Parquet
file:

* `read_parquet_info()` shows a basic summary of the file.
* `read_parquet_schema()` shows all columns, including non-leaf columns,
  and how they are mapped to R types by `read_parquet()`.
* `read_parquet_metadata()` shows the most complete metadata information:
  file meta data, the schema, the row groups and column chunks of the
  file.

```r
nanoparquet::read_parquet_info("mtcars.parquet")
nanoparquet::read_parquet_schema("mtcars.parquet")
nanoparquet::read_parquet_metadata("mtcars.parquet")
```

If you find a file that should be supported but isn't, please open an
issue here with a link to the file.

## Options

See also `?parquet_options()` for further details.

* `nanoparquet.class`: extra class to add to data frames returned by
  `read_parquet()`. If it is not defined, the default is `"tbl"`,
  which changes how the data frame is printed if the pillar package is
  loaded.
* `nanoparquet.compression_level`: See `?parquet_options()` for the
  defaults and the possible values for each compression method. `Inf`
  selects maximum compression for each method.
* `nanoparquet.num_rows_per_row_group`: The number of rows to put into a
  row group by `write_parquet()`, if row groups are not specified
  explicitly. It should be an integer scalar. Defaults to 10 million.
* `nanoparquet.use_arrow_metadata`: unless this is set to `FALSE`,
  `read_parquet()` will make use of Arrow metadata in the Parquet file.
  Currently this is used to detect factor columns.
* `nanoparquet.write_arrow_metadata`: unless this is set to `FALSE`,
  `write_parquet()` will add Arrow metadata to the Parquet file.
  This helps preserving classes of columns, e.g. factors will be read
  back as factors, both by nanoparquet and Arrow.
* `nanoparquet.write_data_page_version`: Data version to write by default.
  Possible values are 1 and 2. Default is 1.
* `nanoparquet.write_minmax_values`: Whether to write minimum and maximum
  values per row group, for data types that support this in
  `write_parquet()`.

## License

MIT

#' Nanoparquet options
#'
#' Create a list of nanoparquet options.
#' @param class The extra class or classes to add to data frames created
#'   in [read_parquet()]. By default nanoparquet adds the `"tbl"` class,
#'   so data frames are printed differently if the pillar package is
#'   loaded.
#' @param compression_level The compression level in [write_parquet()].
#'   `NA` is the default, and it specifies the default
#'   compression level of each method. `Inf` always selects the highest
#'   possible compression level. More details:
#'   * Snappy does not support compression levels currently.
#'   * GZIP supports levels from 0 (uncompressed), 1 (fastest), to 9 (best).
#'     The default is 6.
#'   * ZSTD allows positive levels up to 22 currently. 20 and above require
#'     more memory. Negative levels are also allowed, the lower the level,
#'     the faster the speed, at the cost of compression. Currently the
#'     smallest level is -131072. The default level is 3.
#' @param keep_row_groups This option is used when appending to a Parquet
#'   file with [append_parquet()]. If `TRUE` then the existing row groups
#'   of the file are always kept as is and nanoparquet creates new row
#'   groups for the new data. If `FALSE` (the default), then the last row
#'   group of the file will be overwritten if it is smaller than the
#'   default row group size, i.e. `num_rows_per_row_group`.
#' @param num_rows_per_row_group The number of rows to put into a row
#'   group, if row groups are not specified explicitly. It should be
#'   an integer scalar. Defaults to 10 million.
#' @param use_arrow_metadata `TRUE` or `FALSE`. If `TRUE`, then
#'   [read_parquet()] and [read_parquet_schema()] will make use of the Apache
#'   Arrow metadata to assign R classes to Parquet columns.
#'   This is currently used to detect factor columns, and to detect
#'   "difftime" columns.
#'
#'   If this option is `FALSE`:
#'   - "factor" columns are read as character vectors.
#'   - "difftime" columns are read as real numbers, meaning one
#'     of seconds, milliseconds, microseconds or nanoseconds. Impossible
#'     to tell which without using the Arrow metadata.
#' @param write_arrow_metadata Whether to add the Apache Arrow types as
#'   metadata to the file [write_parquet()].
#' @param write_data_page_version Data version to write by default.
#'   Possible values are 1 and 2. Default is 1.
#' @param read_int64_type How to represent INT64 columns with a 64-bit integer
#'   logical type in [read_parquet()] and [read_parquet_schema()].
#'   Possible values:
#'   * `"double"` (the default): read as a regular R `double` vector,
#'     which may lose precision for large values.
#'   * `"integer64"` or `"bit64::integer64"`: read as a `bit64::integer64`
#'     vector. Requires the bit64 package to be installed.
#' @param write_minmax_values Whether to write minimum and maximum values
#'   per row group, for data types that support this in [write_parquet()].
#'   However, nanoparquet currently does not support minimum and maximum
#'   values for the `DECIMAL`, `UUID` and `FLOAT16` logical types and the
#'   `BOOLEAN`, `BYTE_ARRAY` and `FIXED_LEN_BYTE_ARRAY` primitive types
#'   if they are writing without a logical type. Currently the default
#'   is `TRUE`.
#'
#' @return List of nanoparquet options.
#'
#' @export
#' @examplesIf FALSE
#' # the effect of using Arrow metadata
#' tmp <- tempfile(fileext = ".parquet")
#' d <- data.frame(
#'   fct = as.factor("a"),
#'   dft = as.difftime(10, units = "secs")
#' )
#' write_parquet(d, tmp)
#' read_parquet(tmp, options = parquet_options(use_arrow_metadata = TRUE))
#' read_parquet(tmp, options = parquet_options(use_arrow_metadata = FALSE))

parquet_options <- function(
  class = getOption("nanoparquet.class", "tbl"),
  compression_level = getOption("nanoparquet.compression_level", NA_integer_),
  read_int64_type = getOption("nanoparquet.read_int64_type", "double"),
  keep_row_groups = FALSE,
  num_rows_per_row_group = getOption(
    "nanoparquet.num_rows_per_row_group",
    10000000L
  ),
  use_arrow_metadata = getOption("nanoparquet.use_arrow_metadata", TRUE),
  write_arrow_metadata = getOption("nanoparquet.write_arrow_metadata", TRUE),
  write_data_page_version = getOption(
    "nanoparquet.write_data_page_version",
    1L
  ),
  write_minmax_values = getOption("nanoparquet.write_minmax_values", TRUE)
) {
  stopifnot(is.character(class))
  stopifnot(
    is_string(read_int64_type),
    read_int64_type %in% c("double", "integer64", "bit64::integer64")
  )
  stopifnot(is_flag(keep_row_groups))
  stopifnot(is_flag(use_arrow_metadata))
  stopifnot(is_flag(write_arrow_metadata))
  stopifnot(
    identical(write_data_page_version, 1) ||
      identical(write_data_page_version, 2) ||
      identical(write_data_page_version, 1L) ||
      identical(write_data_page_version, 2L)
  )
  stopifnot(is_flag(write_minmax_values))
  num_rows_per_row_group <- as_count(
    num_rows_per_row_group,
    "num_rows_per_row_group"
  )
  if (identical(compression_level, Inf)) {
    compression_level <- 100000L
  } else if (
    identical(compression_level, NA) ||
      identical(compression_level, NA_integer_) ||
      identical(compression_level, NA_real_)
  ) {
    compression_level <- NA_integer_
  } else {
    compression_level <- as_integer_scalar(
      compression_level,
      "compression_level"
    )
  }

  list(
    class = class,
    compression_level = compression_level,
    read_int64_type = read_int64_type,
    keep_row_groups = keep_row_groups,
    num_rows_per_row_group = num_rows_per_row_group,
    use_arrow_metadata = use_arrow_metadata,
    write_arrow_metadata = write_arrow_metadata,
    write_data_page_version = as.integer(write_data_page_version),
    write_minmax_values = write_minmax_values
  )
}

test_that("gzip compression levels", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- test_df()
  for (level in c(NA_integer_, 0:11)) {
    write_parquet(
      df,
      tmp,
      compression = "gzip",
      options = parquet_options(compression_level = level)
    )
    expect_equal(as.data.frame(read_parquet(tmp)), as.data.frame(df))
  }
})

test_that("zstd compression levels", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- test_df()
  for (level in c(NA_integer_, (-22):(22))) {
    write_parquet(
      df,
      tmp,
      compression = "zstd",
      options = parquet_options(compression_level = level)
    )
    expect_equal(as.data.frame(read_parquet(tmp)), as.data.frame(df))
  }
})

test_that("write_parquet -> INT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # integer -> INT32
  d <- data.frame(d = c(1:10, NA))
  write_parquet(d, tmp, schema = parquet_schema("INT32"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d <- data.frame(d = c(NA, rep(1:2, 50)))
  write_parquet(d, tmp, schema = parquet_schema("INT32"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  # logical -> INT32
  d <- data.frame(d = c(NA, (1:10 %% 2 == 0)))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT32"))
  })

  # string -> INT32
  d <- data.frame(d = c(NA, paste(1:10)))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT32"))
  })
})

test_that("write_parquet -> INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # integer -> INT64
  d <- data.frame(d = c(NA, 1:10))
  write_parquet(d, tmp, schema = parquet_schema("INT64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d <- data.frame(d = c(NA, rep(1:2, 50)))
  write_parquet(d, tmp, schema = parquet_schema("INT64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  # double -> INT64
  d <- data.frame(d = c(NA, as.double(1:10)))
  write_parquet(d, tmp, schema = parquet_schema("INT64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d <- data.frame(d = c(NA, as.double(rep(1:2, 50))))
  write_parquet(d, tmp, schema = parquet_schema("INT64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  # logical -> INT32
  d <- data.frame(d = (1:10 %% 2 == 0))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT64"))
  })

  # string -> INT32
  d <- data.frame(d = paste(1:10))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT64"))
  })

  # list -> INT32
  d <- data.frame(d = I(as.list(1:10)))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT64"))
  })
})

test_that("write_parquet -> INT96", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # integer -> INT96
  d <- data.frame(d = c(-5:5, NA))
  write_parquet(d, tmp, schema = parquet_schema("INT96"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    read_parquet_page(tmp, 4L)$data
  })
  d <- data.frame(d = c(rep(-1:1, 30), NA))
  write_parquet(d, tmp, schema = parquet_schema("INT96"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    read_parquet_page(tmp, 4L)$data
  })

  # double -> INT96
  d <- data.frame(d = as.double(c(-5:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT96"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    read_parquet_page(tmp, 4L)$data
  })
  d <- data.frame(d = as.double(c(rep(-1:1, 30), NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT96"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    read_parquet_page(tmp, 4L)$data
  })
})

test_that("write_parquet -> FLOAT", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # double -> FLOAT
  d <- data.frame(d = c(as.double(-5:5), NA))
  write_parquet(d, tmp, schema = parquet_schema("FLOAT"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d <- data.frame(d = c(as.double(rep(-1:1, 30)), NA))
  write_parquet(d, tmp, schema = parquet_schema("FLOAT"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("write_parquet -> BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # character -> BYTE_ARRAY
  d <- data.frame(d = c("foo", "bar", "foobar", NA))
  write_parquet(
    d,
    tmp,
    schema = parquet_schema("BYTE_ARRAY")
  )
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("write_parquet -> FIXED_LEN_BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # character -> FIXED_LEN_BYTE_ARRAY
  d <- data.frame(d = c("foo", "bar", "aaa", NA))
  write_parquet(
    d,
    tmp,
    schema = parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))
  )
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("write_parquet -> STRING", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # character -> STRING
  d <- data.frame(d = c("foo", "bar", "foobar", NA))
  write_parquet(d, tmp, schema = parquet_schema("STRING"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("write_parquet -> ENUM", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # character -> ENUM
  d <- data.frame(d = c("foo", "bar", "foobar", NA))
  write_parquet(d, tmp, schema = parquet_schema("ENUM"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  # factor -> ENUM
  d <- data.frame(d = as.factor(c("foo", "bar", "foobar", NA)))
  write_parquet(d, tmp, schema = parquet_schema("ENUM"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("write_parquet -> DECIMAL INT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # integer -> DECIMAL INT32
  d <- data.frame(d = c(-5:5, NA))
  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT32",
    scale = 0,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(999L, 1000L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(999L, 1000L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT32",
    scale = 2,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(9L, 10L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(9L, 10L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  # double -> DECIMAL INT32
  d <- data.frame(d = c(-5:5 / 2, NA))
  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT32",
    scale = 1,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(99.9, 100))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(99.9, 100))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT32",
    scale = 2,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(9.99, 10))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(9.99, 10))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
})

test_that("write_parquet -> DECIMAL INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # integer -> DECIMAL INT64
  d <- data.frame(d = c(-5:5, NA))
  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT64",
    scale = 0,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(999L, 1000L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(999L, 1000L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT64",
    scale = 2,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(9L, 10L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(9L, 10L))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  # double -> DECIMAL INT64
  d <- data.frame(d = c(-5:5 / 2, NA))
  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT64",
    scale = 1,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(99.9, 100))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(99.9, 100))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  schema <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT64",
    scale = 2,
    precision = 3
  ))
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- data.frame(d = c(9.99, 10))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
  d2 <- data.frame(d = -c(9.99, 10))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
})

test_that("smaller integers", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = c(-5:5, NA))
  write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = 127:128)
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  })
  d <- data.frame(d = -(127:129))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  })

  d <- data.frame(d = c(-5:5, NA))
  write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = 32767:32768)
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  })
  d <- data.frame(d = -(32767:32769))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  })

  d <- data.frame(d = c(0:5, NA))
  write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = 255:256)
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  })
  d <- data.frame(d = c(0L, -1L))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  })

  d <- data.frame(d = c(0:5, NA))
  write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = 65536:65536)
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  })
  d <- data.frame(d = c(0L, -1L))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  })
})

test_that("double to smaller integers", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = as.double(c(-5:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = as.double(127:128))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  })
  d <- data.frame(d = -as.double((127:129)))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_8"))
  })

  d <- data.frame(d = as.double(c(-5:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = as.double(32767:32768))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  })
  d <- data.frame(d = -as.double(32767:32769))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_16"))
  })

  d <- data.frame(d = as.double(c(-5:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT_32"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = c(2^31 - 1, 2^31))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_32"))
  })
  d <- data.frame(d = c(-2^31, -2^31 - 1))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT_32"))
  })

  d <- data.frame(d = as.double(c(0:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = as.double(255:256))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  })
  d <- data.frame(d = c(0, -1))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
  })

  d <- data.frame(d = as.double(c(0:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = as.double(65535:65536))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  })
  d <- data.frame(d = c(0, -1))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
  })

  d <- data.frame(d = as.double(c(0:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("UINT_32"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = c(2^32 - 1, 2^32))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_32"))
  })
  d <- data.frame(d = c(0, -1))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_32"))
  })
})

test_that("double to INT(64, *)", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = as.double(c(-5:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("INT_64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = c(9223372036854770000, 9223372036854774273))
  expect_snapshot(
    error = TRUE,
    {
      write_parquet(d, tmp, schema = parquet_schema("INT_64"))
    },
    transform = redact_maxint64
  )
  d <- data.frame(d = -c(9223372036854770000, 9223372036854774273))
  expect_snapshot(
    error = TRUE,
    {
      write_parquet(d, tmp, schema = parquet_schema("INT_64"))
    },
    transform = redact_maxint64
  )

  d <- data.frame(d = as.double(c(0:5, NA)))
  write_parquet(d, tmp, schema = parquet_schema("UINT_64"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = c(18446744073709450591.0, 18446744073709550592.0))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_64"))
  })
  d <- data.frame(d = c(0, -1))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UINT_64"))
  })
})

test_that("integer64 round-trip and read_int64_type option", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    x = bit64::as.integer64(c(1e15, -1e15, NA, 2^31))
  )
  write_parquet(d, tmp)

  res_dbl <- read_parquet(tmp)
  expect_equal(class(res_dbl$x), "numeric")
  expect_snapshot(as.data.frame(res_dbl))

  opts_i64 <- parquet_options(read_int64_type = "integer64")
  res_i64 <- read_parquet(tmp, options = opts_i64)
  expect_s3_class(res_i64$x, "integer64")
  expect_true(is.na(res_i64$x[3]))
  expect_equal(res_i64$x[c(1, 2, 4)], d$x[c(1, 2, 4)])

  opts_i64b <- parquet_options(read_int64_type = "bit64::integer64")
  res_i64b <- read_parquet(tmp, options = opts_i64b)
  expect_identical(res_i64b$x, res_i64$x)

  s_dbl <- read_parquet_schema(tmp)
  expect_equal(s_dbl$r_type[s_dbl$name == "x"], "double")

  s_i64 <- read_parquet_schema(tmp, options = opts_i64)
  expect_equal(s_i64$r_type[s_i64$name == "x"], "integer64")

  d2 <- data.frame(x = bit64::as.integer64(c(1L, 2L, NA)))
  write_parquet(d2, tmp, schema = parquet_schema("INT_64"))
  res2 <- read_parquet(tmp, options = opts_i64)
  expect_s3_class(res2$x, "integer64")
  expect_true(is.na(res2$x[3]))
  expect_equal(res2$x[c(1, 2)], d2$x[c(1, 2)])
})

test_that("JSON", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = c("foo", "bar", "foobar", NA))
  write_parquet(d, tmp, schema = parquet_schema("JSON"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("UUID", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = c(rep("00112233-4455-6677-8899-aabbccddeeff", 3), NA))
  write_parquet(d, tmp, schema = parquet_schema("UUID"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(d = c("00112233-4455-6677-8899-aabbccddeeff", "foo"))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UUID"))
  })
})

test_that("FLOAT16", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = c(0, 1, 2, NA, -1, NaN, -2, -Inf, Inf, 1 / 2))
  write_parquet(d, tmp, schema = parquet_schema("FLOAT16"))
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("NaN", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = c(1, 2, NaN, 4, 5))
  write_parquet(d, tmp)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("list of RAW to BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    d = I(list(
      charToRaw("foo"),
      charToRaw("bar"),
      charToRaw("foobar")
    ))
  )
  write_parquet(d, tmp)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(
    d = I(list(
      charToRaw("foo"),
      NULL,
      charToRaw("bar"),
      charToRaw("foobar"),
      NULL
    ))
  )
  write_parquet(d, tmp)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("list of RAW to FIXED_LEN_BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))

  d <- data.frame(
    d = I(list(
      charToRaw("foo"),
      charToRaw("bar"),
      charToRaw("aaa")
    ))
  )
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })

  d <- data.frame(
    d = I(list(
      charToRaw("foo"),
      NULL,
      charToRaw("bar"),
      charToRaw("aaa"),
      NULL
    ))
  )
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
})

test_that("blob::blob to BYTE_ARRAY", {
  skip_without("blob")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    d = blob::blob(
      charToRaw("foo"),
      charToRaw("bar"),
      charToRaw("foobar")
    )
  )
  write_parquet(d, tmp)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- read_parquet(tmp)
  expect_s3_class(d2$d, "blob")
  expect_equal(d2$d, d$d)

  # with NAs (NULL entries in blob)
  d_na <- data.frame(
    d = blob::blob(
      charToRaw("foo"),
      NULL,
      charToRaw("bar"),
      charToRaw("foobar"),
      NULL
    )
  )
  write_parquet(d_na, tmp)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d3 <- read_parquet(tmp)
  expect_s3_class(d3$d, "blob")
  expect_equal(d3$d, d_na$d)
})

test_that("blob::blob to FIXED_LEN_BYTE_ARRAY", {
  skip_without("blob")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))

  d <- data.frame(
    d = blob::blob(
      charToRaw("foo"),
      charToRaw("bar"),
      charToRaw("aaa")
    )
  )
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d2 <- read_parquet(tmp)
  expect_s3_class(d2$d, "blob")
  expect_equal(d2$d, d$d)

  # with NAs
  d_na <- data.frame(
    d = blob::blob(
      charToRaw("foo"),
      NULL,
      charToRaw("bar"),
      charToRaw("aaa"),
      NULL
    )
  )
  write_parquet(d_na, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp)[, -1])
    as.data.frame(read_parquet(tmp))
  })
  d3 <- read_parquet(tmp)
  expect_s3_class(d3$d, "blob")
  expect_equal(d3$d, d_na$d)
})

#' Infer Parquet schema of a data frame
#'
#' @param df Data frame.
#' @param options Return value of [parquet_options()], may modify the
#'   R to Parquet type mappings.
#' @return Data frame, the inferred schema. It has the same columns as
#'   the return value of [read_parquet_schema()]:
#'   `r paste0("\u0060", names(infer_parquet_schema(mtcars)), "\u0060", collapse = ", ")`.
#'
#' @seealso [read_parquet_schema()] to read the schema of a Parquet file,
#'   [parquet_schema()] to create a Parquet schema from scratch.
#' @export

infer_parquet_schema <- function(df, options = parquet_options()) {
  types <- .Call(rf_nanoparquet_map_to_parquet_types, df, options)

  # types[[cl]] is a list of N 4-tuples: list(type, rtype, logical_type, name)
  nsels <- lengths(types)
  col_idx <- rep(seq_along(df), nsels)
  pos_in_col <- sequence(nsels)

  all_sels <- unlist(types, recursive = FALSE)
  lt <- unname(lapply(all_sels, function(x) x[[3]]))
  ct <- lapply(lt, function(x) if (!is.null(x)) logical_to_converted(x))

  repetition_type <- mapply(
    infer_repetition_type,
    col = df[col_idx],
    nsels = nsels[col_idx],
    pos = pos_in_col
  )

  type_tab <- data.frame(
    file_name = rep(NA_character_, length(all_sels)),
    r_col = as.integer(col_idx),
    name = vapply(all_sels, function(x) x[[4]], ""),
    r_type = vapply(all_sels, function(x) x[[2]], ""),
    type = vapply(all_sels, function(x) x[[1]], ""),
    type_length = rep(NA_integer_, length(all_sels)),
    repetition_type = repetition_type,
    converted_type = map_chr(ct, function(x) {
      x[["converted_type"]] %||% NA_character_
    }),
    logical_type = I(lt),
    num_children = ifelse(
      nsels[col_idx] == 3L & pos_in_col <= 2L,
      1L,
      NA_integer_
    ),
    scale = map_int(ct, function(x) {
      x[["scale"]] %||% NA_integer_
    }),
    precision = map_int(ct, function(x) {
      x[["precision"]] %||% NA_integer_
    }),
    field_id = rep(NA_integer_, length(all_sels))
  )

  rownames(type_tab) <- NULL
  class(type_tab) <- c("tbl", class(type_tab))
  type_tab
}

infer_repetition_type <- function(col, nsels, pos) {
  if (nsels == 1L) {
    if (anyNA(col)) "OPTIONAL" else "REQUIRED"
  } else {
    # LIST column: OPTIONAL, REPEATED, OPTIONAL
    c("OPTIONAL", "REPEATED", "OPTIONAL")[pos]
  }
}

test_that("infer_parquet_schema", {
  expect_snapshot({
    as.data.frame(infer_parquet_schema(test_df(missing = FALSE)))
  })
  expect_snapshot({
    as.data.frame(infer_parquet_schema(test_df(missing = TRUE)))
  })
})

test_that("logical_to_converted", {
  expect_snapshot(error = TRUE, {
    logical_to_converted(list("DECIMAL"))
    logical_to_converted(list("TIME", TRUE, "SECS"))
    logical_to_converted(list("TIMESTAMP", TRUE, "SECS"))
    logical_to_converted(list("INT"))
    logical_to_converted(list("FOOBAR"))
  })
})

#' Read a Parquet file into a data frame
#'
#' Converts the contents of the named Parquet file to a R data frame.
#'
#' @param file Path to a Parquet file. It may also be an R connection,
#'   in which case it first reads all data from the connection, writes
#'   it into a temporary file, then reads the temporary file, and
#'   deletes it. The connection might be open, it which case it must be
#'   a binary connection. If it is not open, then `read_parquet()` will
#'   open it and also close it in the end.
#' @param col_select Columns to read. It can be a numeric vector of column
#'   indices, or a character vector of column names. It is an error to
#'   select the same column multiple times. The order of the columns in
#'   the result is the same as the order in `col_select`.
#' @param options Nanoparquet options, see [parquet_options()].
#' @return A `data.frame` with the file's contents.
#' @export
#' @seealso See [write_parquet()] to write Parquet files,
#'   [nanoparquet-types] for the R <-> Parquet type mapping.
#'   See [read_parquet_info()], for general information,
#'   [read_parquet_schema()] for information about the
#'   columns, and [read_parquet_metadata()] for the complete metadata.
#' @examples
#' file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
#' parquet_df <- nanoparquet::read_parquet(file_name)
#' print(str(parquet_df))

read_parquet <- function(file, col_select = NULL, options = parquet_options()) {
  if (inherits(file, "connection")) {
    tmp <- tempfile(fileext = ".parquet")
    on.exit(unlink(tmp), add = TRUE)
    dump_connection(file, tmp)
    file <- tmp
  }
  file <- path.expand(file)

  if (!is.null(col_select)) {
    stopifnot(is.numeric(col_select) || is.character(col_select))
    col_select <- col_select[!is.na(col_select)]
    col_select_mtpl <- col_select

    col_select <- unique(col_select)
    if (length(col_select) != length(col_select_mtpl)) {
      dpl <- sort(unique(col_select_mtpl[duplicated(col_select_mtpl)]))
      stop(
        "Column",
        if (length(dpl) > 1) "s",
        " ",
        paste(dpl, collapse = ", "),
        " selected multiple times in `read_parquet()`."
      )
    }

    if (is.numeric(col_select)) {
      col_select <- as.integer(col_select)
    } else {
      pq_col_names <- .Call(nanoparquet_read_col_names, file)
      col_match <- match(col_select, pq_col_names)
      if (anyNA(col_match)) {
        bad <- unique(col_select[is.na(col_match)])
        stop(
          "Column",
          if (length(bad) > 1) "s",
          " ",
          paste(bad, collapse = ", "),
          if (length(bad) == 1) " does " else " do ",
          "not exist in Parquet file"
        )
      }
      col_select <- col_match
    }
    stopifnot(all(col_select >= 1L))
  }

  if (
    options[["read_int64_type"]] != "double" &&
      !requireNamespace("bit64", quietly = TRUE)
  ) {
    stop(
      "The bit64 package is required when `read_int64_type` is not \"double\".\n"
    )
  }

  res <- .Call(nanoparquet_read2, file, options, col_select, sys.call())
  post_process_read_result(res, file, options, col_select)
}

post_process_read_result <- function(res, file, options, col_select) {
  dicts <- res[[2]]
  types <- res[[3]]
  arrow_schema <- res[[4]]
  repeats <- res[[5]]
  presents <- res[[6]]
  parents <- res[[7]]
  rep_types <- res[[8]]
  leaf_cols <- res[[9]]
  res <- res[[1]]
  if (options[["use_arrow_metadata"]] && !is.na(arrow_schema)) {
    res <- apply_arrow_schema(res, file, arrow_schema, dicts, types, col_select)
  }

  to_leaf_col <- which(leaf_cols >= 0)

  # fix up repeated columns, if any
  has_reps <- lengths(repeats) > 0
  for (idx in which(has_reps)) {
    leaf_idx <- to_leaf_col[idx]
    rep_type <- rep_types[leaf_idx]
    if (rep_type == 2) {
      # single REPEATED column
      bpos <- .Call(
        nanoparquet_repeated_positions,
        res[[idx]],
        repeats[[idx]],
        presents[[idx]],
        nrow(res)
      )
      res[[idx]] <- split_at(res[[idx]], bpos)
    } else {
      # LIST column, 3 layers
      lst_rep <- rep_types[parents[parents[leaf_idx] + 1] + 1]
      elt_rep <- rep_types[leaf_idx]
      if (lst_rep == 0 && elt_rep == 0) {
        res[[idx]] <- enlist_req_req(
          res[[idx]],
          repeats[[idx]],
          presents[[idx]],
          nrow(res)
        )
      } else if (lst_rep == 0 && elt_rep == 1) {
        res[[idx]] <- enlist_req_opt(
          res[[idx]],
          repeats[[idx]],
          presents[[idx]],
          nrow(res)
        )
      } else if (lst_rep == 1 && elt_rep == 0) {
        res[[idx]] <- enlist_opt_req(
          res[[idx]],
          repeats[[idx]],
          presents[[idx]],
          nrow(res)
        )
      } else if (lst_rep == 1 && elt_rep == 1) {
        res[[idx]] <- enlist_opt_opt(
          res[[idx]],
          repeats[[idx]],
          presents[[idx]],
          nrow(res)
        )
      }
    }
  }

  # add ptype attribute to blob columns for vctrs compatibility
  blobs <- which(vapply(res, "inherits", "blob", FUN.VALUE = logical(1)))
  for (idx in blobs) {
    attr(res[[idx]], "ptype") <- raw()
  }

  # convert hms from milliseconds to seconds, also integer -> double
  hmss <- which(vapply(res, "inherits", "hms", FUN.VALUE = logical(1)))
  for (idx in hmss) {
    res[[idx]] <- structure(
      unclass(res[[idx]]) / 1000,
      class = class(res[[idx]])
    )
  }

  # convert POSIXct from milliseconds to seconds
  posixcts <- which(vapply(res, "inherits", "POSIXct", FUN.VALUE = logical(1)))
  for (idx in posixcts) {
    res[[idx]][] <- structure(
      unclass(res[[idx]]) / 1000,
      class = class(res[[idx]])
    )
  }

  if (length(options[["class"]]) > 0) {
    class(res) <- unique(c(options[["class"]], class(res)))
  }

  res
}

enlist_req_req <- function(x, rep, def, nrows) {
  # def = 0 => empty list
  sizes <- .Call(nanoparquet_enlist_sizes_req_req, rep, def, as.double(nrows))
  split_at(x, sizes)
}

enlist_req_opt <- function(x, rep, def, nrows) {
  # def = 0 => empty list, 1 => NA
  .Call(nanoparquet_enlist_req_opt, x, rep, def, as.double(nrows))
}

enlist_opt_req <- function(x, rep, def, nrows) {
  # def = 0 => NULL, 1 => list()
  .Call(nanoparquet_enlist_opt_req, x, rep, def, as.double(nrows))
}

enlist_opt_opt <- function(x, rep, def, nrows) {
  # def = 0 => NULL ,1 => list(), 2 => NA
  .Call(nanoparquet_enlist_opt_opt, x, rep, def, as.double(nrows))
}

# dump the contents of a connection to path
dump_connection <- function(con, path) {
  if (!isOpen(con)) {
    on.exit(close(con), add = TRUE)
    open(con, "rb")
  }
  ocon <- file(path, open = "wb")
  # 10 MB buffer by default
  bs <- getOption("nanoparquet.con_buffer_size", 1024L * 1024L * 10)
  while (TRUE) {
    buf <- readBin(con, what = "raw", n = bs)
    if (length(buf) == 0) {
      break
    }
    writeBin(buf, ocon)
  }
  close(ocon)
}

read_parquet_row_group <- function(
  file,
  row_group,
  options = parquet_options()
) {
  if (inherits(file, "connection")) {
    tmp <- tempfile(fileext = ".parquet")
    on.exit(unlink(tmp), add = TRUE)
    dump_connection(file, tmp)
    file <- tmp
  }
  file <- path.expand(file)
  res <- .Call(nanoparquet_read_row_group, file, row_group, options, sys.call())
  post_process_read_result(res, file, options, col_select = NULL)
}

# TODO: this does not work currently
read_parquet_column_chunk <- function(
  file,
  row_group = 0L,
  column = 0L,
  options = parquet_options()
) {
  if (inherits(file, "connection")) {
    tmp <- tempfile(fileext = ".parquet")
    on.exit(unlink(tmp), add = TRUE)
    dump_connection(file, tmp)
    file <- tmp
  }
  file <- path.expand(file)
  res <- .Call(
    nanoparquet_read_column_chunk,
    file,
    row_group,
    column,
    options,
    sys.call()
  )
  post_process_read_result(res, file, options)
}

test_that("DECIMAL converted type", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = 100)
  sch <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "INT32",
    precision = 5,
    scale = 2
  ))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  d2 <- data.frame(d = 100L)
  sch$logical_type[1] <- list(NULL)
  write_parquet(d2, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("DECIMAL in BA dict", {
  pf <- test_path("data/decimals.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

return()

test_that("DECIMAL in BA", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = as.double(100:110))
  sch <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "BYTE_ARRAY",
    precision = 5,
    scale = 2
  ))
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  d2 <- data.frame(d = 100:110)
  sch <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "BYTE_ARRAY",
    precision = 5,
    scale = 2
  ))
  write_parquet(d2, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("DECIMAL in BA, dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(d = as.double(100:110))
  sch <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "BYTE_ARRAY",
    precision = 5,
    scale = 2
  ))
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  d2 <- data.frame(d = 100:110)
  sch <- parquet_schema(list(
    "DECIMAL",
    primitive_type = "BYTE_ARRAY",
    precision = 5,
    scale = 2
  ))
  write_parquet(d2, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("write large file", {
  if (Sys.getenv("NANOPARQUET_ALL_TESTS") != "true") {
    expect_true(TRUE)
    return()
  }
  mt <- test_df(missing = TRUE, factor = TRUE)
  larger <- do.call("rbind", replicate(1000, mt, simplify = FALSE))
  big <- do.call("rbind", replicate(100, larger, simplify = FALSE))
  large <- do.call("rbind", replicate(10, big, simplify = FALSE))

  tmp <- tempfile(fileext = ".parquet")
  system.time(write_parquet(large, tmp))
  system.time(large2 <- read_parquet(tmp))
  expect_equal(large, large2)
})

test_that("write >4GB file: uint32_t/int32_t overflow (issue #143)", {
  if (Sys.getenv("NANOPARQUET_ALL_TESTS") != "true") {
    expect_true(TRUE)
    return()
  }

  # Reproduces issue #143: file offsets and column sizes were stored as
  # uint32_t/int32_t, overflowing once the file exceeded 4 GB (2^32 bytes).
  # 270M rows x 2 double columns x 8 bytes = ~4.3 GB uncompressed, crossing
  # both the 2^31 threshold (int32_t column_bytes) and the 2^32 threshold
  # (uint32_t write_columns end).

  n <- 270e6L
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(a = as.double(seq_len(n)), b = as.double(seq_len(n)) * 2.0)
  write_parquet(df, tmp, compression = "uncompressed")
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df, df2)
})

test_that("not open", {
  pf <- test_path("data/factor.parquet")
  con <- file(pf)
  expect_equal(
    read_parquet(con),
    read_parquet(pf)
  )
  # A closed (=invalid in R) connection will error here
  expect_error(isOpen(con))
})

test_that("open", {
  pf <- test_path("data/factor.parquet")
  con <- file(pf, open = "rb")
  on.exit(close(con), add = TRUE)
  expect_equal(
    read_parquet(con),
    read_parquet(pf)
  )
  expect_true(isOpen(con))
})

test_that("raw, opened", {
  pf <- test_path("data/factor.parquet")
  bts <- readBin(pf, what = "raw", n = file.size(pf))
  con <- rawConnection(bts, open = "rb")
  on.exit(close(con), add = TRUE)
  expect_equal(
    read_parquet(con),
    read_parquet(pf)
  )
  expect_true(isOpen(con))
})

test_that("read_parquet_row_group", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  un <- function(x) {
    x <- as.data.frame(x)
    rownames(x) <- NULL
    x
  }

  do <- function(df, encoding = NULL) {
    write_parquet(
      df,
      tmp,
      row_groups = c(1L, 11L, 21L, 31L)
    )
    expect_snapshot(as.data.frame(read_parquet_metadata(tmp)$row_groups)[-1])
    expect_equal(un(read_parquet_row_group(tmp, 0L)), un(df[1:10, ]))
    expect_equal(un(read_parquet_row_group(tmp, 1L)), un(df[11:20, ]))
    expect_equal(un(read_parquet_row_group(tmp, 2L)), un(df[21:30, ]))
    expect_equal(un(read_parquet_row_group(tmp, 3L)), un(df[31:32, ]))
  }

  do(test_df(), encoding = "PLAIN")
  do(test_df(missing = TRUE), encoding = "PLAIN")
  do(test_df(), encoding = c(large = "RLE", "RLE_DICTIONARY"))
  do(test_df(missing = TRUE), encoding = c(large = "RLE", "RLE_DICTIONARY"))
})

#' @exportS3Method pillar::obj_sum
obj_sum.nanoparquet_logical_type <- function(x, ...) {
  type <- x$type
  paste0(
    type,
    if (type == "INT") sprintf("(%d, %s)", x$bit_width, x$is_signed),
    if (type == "DECIMAL") sprintf("(%d, %d)", x$scale %||% 0, x$precision),
    if (type %in% c("TIME", "TIMESTAMP")) {
      sprintf("(%s, %s)", x$is_adjusted_to_utc, x$unit)
    }
  )
}

#' Create a Parquet schema
#'
#' You can use this schema to specify how to write out a data frame to
#' a Parquet file with [write_parquet()].
#'
#' @param ... Parquet type specifications, see below.
#'   For backwards compatibility, you can supply a file name
#'   here, and then `parquet_schema` behaves as [read_parquet_schema()].
#' @return Data frame with the same columns as [read_parquet_schema()]:
#'   `r paste0("\u0060", names(parquet_schema("INT32")), "\u0060", collapse = ", ")`.
#' @details
#' A schema is a list of potentially named type specifications. A schema
#' is stored in a data frame. Each (potentially named) argument of
#' `parquet_schema` may be a character scalar, or a list. Parameterized
#' types need to be specified as a list. Primitive Parquet types may be
#' specified as a string or a list.
#'
#' # Possible types:
#'
#' Special type:
#'
#' * `"AUTO"`: this is not a Parquet type, but it tells [write_parquet()]
#'   to map the R type to Parquet automatically, using the default mapping
#'   rules.
#'
#' Primitive Parquet types:
#'
#' * `"BOOLEAN"`
#' * `"INT32"`
#' * `"INT64"`
#' * `"INT96"`
#' * `"FLOAT"`
#' * `"DOUBLE"`
#' * `"BYTE_ARRAY"`
#' * `"FIXED_LEN_BYTE_ARRAY"`: fixed-length byte array. It needs a
#'   `type_length` parameter, an integer between 0 and 2^31-1.
#'
#' Parquet logical types:
#'
#' * `"STRING"`
#' * `"ENUM"`
#' * `"UUID"`
#' * `"INTEGER"`: signed or unsigned integer. It needs a `bit_width` and
#'   an `is_signed` parameter. `bit_width` must be 8, 16, 32 or 64.
#'   `is_signed` must be `TRUE` or `FALSE`.
#' * `"INT"`: same as `"INTEGER"`. The Parquet documentation uses `"INT"`,
#'   but the actual specification uses `"INTEGER"`. Both are supported in
#'   nanoparquet.
#' * `"DECIMAL"`: decimal number of specified scale and precision.
#'   It needs the `precision` and `primitive_type` parameters. Also
#'   supports the `scale` parameter, it defaults to zero if not specified.
#' * `"FLOAT16"`
#' * `"DATE"`
#' * `"TIME"`: needs an `is_adjusted_utc` (`TRUE` or `FALSE`) and a
#'   `unit` parameter. `unit` must be `"MILLIS"`, `"MICROS"` or `"NANOS"`.
#' * `"TIMESTAMP"`: needs an `is_adjusted_utc` (`TRUE` or `FALSE`) and a
#'   `unit` parameter. `unit` must be `"MILLIS"`, `"MICROS"` or `"NANOS"`.
#' * `"JSON"`
#' * `"BSON"`
#' * `"LIST"`: list of some other type. It needs an `element` parameter,
#'   which is a type specification for the list elements. Currently
#'   `element` can be only `"INT32"`, `"DOUBLE"` or `"STRING"`.
#'   Also, currently nanoparquet always write LIST columns that are
#'   `"OPTIONAL"`, both for the list elements and the list itself.
#'
#' Logical types `MAP`, and `UNKNOWN` are not supported currently.
#'
#' Converted types are deprecated in the Parquet specification in favor of
#' logical types, but `parquet_schema()` accepts some converted types as a
#' syntactic shortcut for the corresponding logical types:
#'
#' * `INT_8` mean `list("INT", bit_width = 8, is_signed = TRUE)`.
#' * `INT_16` mean `list("INT", bit_width = 16, is_signed = TRUE)`.
#' * `INT_32` mean `list("INT", bit_width = 32, is_signed = TRUE)`.
#' * `INT_64` mean `list("INT", bit_width = 64, is_signed = TRUE)`.
#' * `TIME_MICROS` means `list("TIME", is_adjusted_utc = TRUE, unit = "MICROS")`.
#' * `TIME_MILLIS` means `list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS")`.
#' * `TIMESTAMP_MICROS` means `list("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")`.
#' * `TIMESTAMP_MILLIS` means `list("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MILLIS")`.
#' * `UINT_8` means `list("INT", bit_width = 8, is_signed = FALSE)`.
#' * `UINT_16` means `list("INT", bit_width = 16, is_signed = FALSE)`.
#' * `UINT_32` means `list("INT", bit_width = 32, is_signed = FALSE)`.
#' * `UINT_64` means `list("INT", bit_width = 64, is_signed = FALSE)`.
#'
#' ## Missing values
#'
#' Each type might also have a `repetition_type` parameter, with possible
#' values `"REQUIRED"`, `"OPTIONAL"` or `"REPEATED"`. `"REQUIRED"` columns
#' do not allow missing values. Missing values are allowed in `"OPTIONAL"`
#' columns. `"REPEATED"` columns are currently not supported in
#' [write_parquet()].
#'
#' @export
#' @examples
#' parquet_schema(
#'   c1 = "INT32",
#'   c2 = list("INT", bit_width = 64, is_signed = TRUE),
#'   c3 = list("STRING", repetition_type = "OPTIONAL"),
#'   l = list("LIST", element = "DOUBLE")
#' )

parquet_schema <- function(...) {
  args <- list(...)
  if (
    length(args) == 1 &&
      is_string(args[[1]]) &&
      !is.null(args[[1]]) &&
      file.exists(args[[1]]) &&
      (is.null(names(args)) || names(args)[1] %in% c("", "file"))
  ) {
    warning(
      "Using `parquet_schema()` to read the schema from a file is deprecated. ",
      "Use `read_parquet_schema()` instead."
    )
    read_parquet_schema(args[[1]])
  } else {
    parquet_schema_create(args)
  }
}

parquet_schema_create <- function(types) {
  if (length(types) == 0) {
    ptdf <- data.frame(
      file_name = character(),
      r_col = integer(),
      name = character(),
      r_type = character(),
      type = character(),
      type_length = integer(),
      repetition_type = character(),
      converted_type = character(),
      logical_type = I(list()),
      num_children = integer(),
      scale = integer(),
      precision = integer(),
      field_id = integer()
    )
    class(ptdf) <- c("nanoparquet_schema", "tbl", class(ptdf))
    return(ptdf)
  }
  ptypes <- lapply(types, function(t) do.call(parquet_type, as.list(t)))
  nms <- names(types) %||% rep("", length(types))
  nms[nms == ""] <- NA_character_

  normalize <- function(x, nm) {
    na <- rep(NA_character_, length(x$type))
    nai <- rep(NA_integer_, length(x$type))
    x$file_name <- na
    if (is.null(x$name)) {
      x$name <- nm
    } else {
      x$name[1] <- nm
    }
    x$type <- x$type %||% NA_character_
    x$r_type <- na
    x$type_length <- x$type_length %||% nai
    x$repetition_type <- x$repetition_type %||% na
    x$converted_type <- x$converted_type %||% na
    x$num_children <- x$num_children %||% nai
    x$scale <- x$scale %||% nai
    x$precision <- x$precision %||% nai
    x$field_id <- nai
    x
  }

  ptypes <- mapply(FUN = normalize, ptypes, nms, SIMPLIFY = FALSE)

  r_col <- unlist(mapply(
    function(x, idx) rep(as.integer(idx), length(x$type)),
    ptypes,
    seq_along(ptypes),
    SIMPLIFY = FALSE
  ))

  ptdf <- data.frame(
    file_name = unlist(lapply(ptypes, "[[", "file_name")),
    r_col = r_col,
    name = unlist(lapply(ptypes, "[[", "name")),
    r_type = unlist(lapply(ptypes, "[[", "r_type")),
    type = unlist(lapply(ptypes, "[[", "type")),
    type_length = unlist(lapply(ptypes, "[[", "type_length")),
    repetition_type = unlist(lapply(ptypes, "[[", "repetition_type")),
    converted_type = unlist(lapply(ptypes, "[[", "converted_type")),
    logical_type = if (length(ptypes) == 0) {
      I(list())
    } else {
      I(unlist(
        lapply(ptypes, function(x) {
          lt <- x$logical_type
          if (is.null(lt) || inherits(lt, "nanoparquet_logical_type")) {
            list(lt)
          } else {
            lt
          }
        }),
        recursive = FALSE,
        use.names = FALSE
      ))
    },
    num_children = unlist(lapply(ptypes, "[[", "num_children")),
    scale = unlist(lapply(ptypes, "[[", "scale")),
    precision = unlist(lapply(ptypes, "[[", "precision")),
    field_id = unlist(lapply(ptypes, "[[", "field_id"))
  )

  class(ptdf) <- c("nanoparquet_schema", "tbl", class(ptdf))
  ptdf
}

parquet_type <- function(
  type,
  type_length = NULL,
  bit_width = NULL,
  is_signed = NULL,
  precision = NULL,
  scale = NULL,
  is_adjusted_utc = NULL,
  unit = NULL,
  primitive_type = NULL,
  repetition_type = NULL,
  encoding = NULL,
  element = NULL
) {
  fixed_len_byte_array <- function() {
    stopifnot(
      !is.null(type_length),
      is_uint32(type_length)
    )
    r <- list(
      type = "FIXED_LEN_BYTE_ARRAY",
      type_length = as.integer(type_length)
    )
    type_length <<- NULL
    r
  }

  int <- function() {
    stopifnot(
      !is.null(bit_width),
      bit_width %in% c(8L, 16L, 32L, 64L),
      !is.null(is_signed),
      is_flag(is_signed)
    )
    r <- list(
      type = if (bit_width <= 32L) "INT32" else "INT64",
      logical_type = nanoparquet_logical_type(list(
        type = "INT",
        bit_width = as.integer(bit_width),
        is_signed = is_signed
      ))
    )
    bit_width <<- NULL
    is_signed <<- NULL
    r
  }

  decimal <- function() {
    stopifnot(
      !is.null(precision),
      is_uint32(precision),
      precision > 0,
      is.null(scale) || (is_uint32(scale) && scale <= precision),
      !is.null(primitive_type),
      is_string(primitive_type),
      primitive_type %in%
        c("INT32", "INT64", "BYTE_ARRAY", "FIXED_LEN_BYTE_ARRAY")
    )
    if (primitive_type == "INT32") {
      stopifnot(precision <= 9)
    } else if (primitive_type == "INT64") {
      stopifnot(precision <= 18)
    } else if (primitive_type == "FIXED_LEN_BYTE_ARRAY") {
      stopifnot(
        !is.null(type_length),
        is_uint32(type_length),
        precision <= floor(log10(2^(8 * type_length - 1) - 1))
      )
    }
    r <- list(
      type = primitive_type,
      logical_type = nanoparquet_logical_type(list(
        type = "DECIMAL",
        scale = if (!is.null(scale)) as.integer(scale),
        precision = as.integer(precision)
      )),
      scale = as.integer(scale %||% 0),
      precision = as.integer(precision)
    )
    precision <<- NULL
    scale <<- NULL
    if (primitive_type == "FIXED_LEN_BYTE_ARRAY") {
      r[["type_length"]] <- as.integer(type_length)
      type_length <<- NULL
    }
    primitive_type <<- NULL
    r
  }

  time <- function() {
    stopifnot(
      !is.null(is_adjusted_utc),
      is_flag(is_adjusted_utc),
      !is.null(unit),
      is_string(unit),
      unit %in% c("MILLIS", "MICROS", "NANOS")
    )
    r <- list(
      type = if (unit == "MILLIS") "INT32" else "INT64",
      logical_type = nanoparquet_logical_type(list(
        type = "TIME",
        is_adjusted_utc = is_adjusted_utc,
        unit = unit
      ))
    )
    is_adjusted_utc <<- NULL
    unit <<- NULL
    r
  }

  timestamp <- function() {
    stopifnot(
      !is.null(is_adjusted_utc),
      is_flag(is_adjusted_utc),
      !is.null(unit),
      is_string(unit),
      unit %in% c("MILLIS", "MICROS", "NANOS")
    )
    r <- list(
      type = "INT64",
      logical_type = nanoparquet_logical_type(list(
        type = "TIMESTAMP",
        is_adjusted_utc = is_adjusted_utc,
        unit = unit
      ))
    )
    is_adjusted_utc <<- NULL
    unit <<- NULL
    r
  }

  nested_list <- function() {
    stopifnot(!is.null(element))
    if (is.list(element) && !is.null(element$repetition_type)) {
      stop("Custom repetition_type is not yet supported for LIST columns")
    }
    r <- do.call("parquet_type", as.list(element))
    element <<- NULL
    r
  }

  err <- function(typename) {
    # nocov start this is probably a covr bug that this is not covered
    stop("Parquet type '", typename, "' is not supported by nanoparquet")
    # nocov end
  }

  ptype <- switch(
    type,
    # dummy type to denote auto-detection
    AUTO = list(type = NA_character_),

    # primitive types
    BOOLEAN = list(type = "BOOLEAN"),
    INT32 = list(type = "INT32"),
    INT64 = list(type = "INT64"),
    INT96 = list(type = "INT96"),
    FLOAT = list(type = "FLOAT"),
    DOUBLE = list(type = "DOUBLE"),
    BYTE_ARRAY = list(type = "BYTE_ARRAY"),
    FIXED_LEN_BYTE_ARRAY = fixed_len_byte_array(),

    # logical types
    STRING = list(
      type = "BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "STRING"))
    ),
    ENUM = list(
      type = "BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "ENUM"))
    ),
    UUID = list(
      type = "FIXED_LEN_BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "UUID")),
      type_length = 16L
    ),
    INTEGER = int(),
    INT = int(),
    DECIMAL = decimal(),
    FLOAT16 = list(
      type = "FIXED_LEN_BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "FLOAT16")),
      type_length = 2L
    ),
    DATE = list(
      type = "INT32",
      logical_type = nanoparquet_logical_type(list(type = "DATE"))
    ),
    TIME = time(),
    TIMESTAMP = timestamp(),
    # There is no INTERVAL logical type? What's going on here?
    # INTERVAL = list(
    #   type = "FIXED_LEN_BYTE_ARRAY",
    #   logical_type = nanoparquet_logical_type(list(type = "INTERVAL")),
    #   type_length = 12L
    # ),
    JSON = list(
      type = "BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "JSON"))
    ),
    BSON = list(
      type = "BYTE_ARRAY",
      logical_type = nanoparquet_logical_type(list(type = "BSON"))
    ),
    LIST = {
      nl <- nested_list()
      list(
        name = c(NA_character_, "list", "element"),
        type = c(NA_character_, NA_character_, nl$type),
        logical_type = list(
          nanoparquet_logical_type(list(type = "LIST")),
          NULL,
          nl$logical_type
        ),
        converted_type = c(
          "LIST",
          NA_character_,
          nl$converted_type %||% NA_character_
        ),
        repetition_type = c("OPTIONAL", "REPEATED", "OPTIONAL"),
        num_children = c(1L, 1L, NA_integer_)
      )
    },
    MAP = err("MAP"),
    UNKNOWN = err("UNKNOWN"),

    # some converted types as shortcuts
    "INT_8" = {
      bit_width <- 8
      is_signed <- TRUE
      int()
    },
    "INT_16" = {
      bit_width <- 16
      is_signed <- TRUE
      int()
    },
    "INT_32" = {
      bit_width <- 32
      is_signed <- TRUE
      int()
    },
    "INT_64" = {
      bit_width <- 64
      is_signed <- TRUE
      int()
    },

    "TIME_MICROS" = {
      is_adjusted_utc <- TRUE
      unit <- "MICROS"
      time()
    },
    "TIME_MILLIS" = {
      is_adjusted_utc <- TRUE
      unit <- "MILLIS"
      time()
    },
    "TIMESTAMP_MICROS" = {
      is_adjusted_utc <- TRUE
      unit <- "MICROS"
      timestamp()
    },
    "TIMESTAMP_MILLIS" = {
      is_adjusted_utc <- TRUE
      unit <- "MILLIS"
      timestamp()
    },

    "UINT_8" = {
      bit_width <- 8
      is_signed <- FALSE
      int()
    },
    "UINT_16" = {
      bit_width <- 16
      is_signed <- FALSE
      int()
    },
    "UINT_32" = {
      bit_width <- 32
      is_signed <- FALSE
      int()
    },
    "UINT_64" = {
      bit_width <- 64
      is_signed <- FALSE
      int()
    },

    # bail
    err(type)
  )

  if (!is.null(repetition_type)) {
    stopifnot(
      is_string(repetition_type),
      repetition_type %in% c("REQUIRED", "OPTIONAL", "REPEATED")
    )
    ptype[["repetition_type"]] <- repetition_type
  }

  # check that every non-NULL parameter is used
  stopifnot(
    "Unused Parquet type parameter 'type_length'." = is.null(type_length),
    "Unused Parquet type parameter 'bit_width'." = is.null(bit_width),
    "Unused Parquet type parameter 'is_signed'." = is.null(is_signed),
    "Unused Parquet type parameter 'precision'." = is.null(precision),
    "Unused Parquet type parameter 'scale'." = is.null(scale),
    "Unused Parquet type parameter 'is_adjusted_utc'." = is.null(
      is_adjusted_utc
    ),
    "Unused Parquet type parameter 'unit'." = is.null(unit),
    "Unused Parquet type parameter 'primitive_type'." = is.null(primitive_type)
  )

  if (!is.null(ptype[["logical_type"]])) {
    if (is.null(ptype[["converted_type"]])) {
      ct <- logical_to_converted(ptype[["logical_type"]])
      ptype[names(ct)] <- ct
    }
  }

  ptype
}

nanoparquet_logical_type <- function(x) {
  class(x) <- "nanoparquet_logical_type"
  x
}

logical_to_converted <- function(logical_type) {
  res <- .Call(rf_nanoparquet_logical_to_converted, logical_type)
  res[["converted_type"]] <- names(ctype_names)[res[["converted_type"]] + 1]
  res
}

map_schema_to_df <- function(schema, df, options) {
  stopifnot(
    is.null(schema) || is.data.frame(schema),
    is.data.frame(df)
  )
  # Remove root node (r_col = NA, present only in file-read schemas).
  schema <- schema[!is.na(schema[["r_col"]]), ]

  # All remaining rows are passed to C++. Use r_col to find the top-level row
  # of each R column (first row per r_col group), which carries the column name.
  rcol <- schema[["r_col"]]
  top <- !duplicated(rcol)
  nms <- schema$name[top]

  if (is.null(schema)) {
    # no schema at all, everything auto-detected
    schema <- do.call(
      "parquet_schema",
      structure(as.list(rep("AUTO", ncol(df))), names = names(df))
    )
  } else if (!is.null(nms) && !anyNA(nms) && !any(nms == "")) {
    # all properly named — match against df columns
    dfmiss <- setdiff(nms, names(df))
    if (length(dfmiss) > 0) {
      stop(
        "Parquet schema column",
        if (length(dfmiss) > 1) "s",
        " missing from the data: ",
        paste(dfmiss, collapse = ", ")
      )
    }
    # add AUTO rows for df columns not in the schema
    smiss <- setdiff(names(df), nms)
    auto <- do.call(
      "parquet_schema",
      structure(as.list(rep("AUTO", length(smiss))), names = smiss)
    )
    # Reorder all rows (grouped by r_col) to match df column order.
    top_rcols <- rcol[top]
    schema_nms <- setdiff(names(df), smiss)
    col_order <- match(schema_nms, nms)
    ordered_rows <- unlist(lapply(col_order, function(i) {
      which(rcol == top_rcols[i])
    }))
    schema <- schema[ordered_rows, ]
    # Renumber r_col: each column gets its position in the df.
    new_rcol <- unlist(lapply(seq_along(col_order), function(i) {
      rep(match(schema_nms[i], names(df)), sum(rcol == top_rcols[col_order[i]]))
    }))
    schema[["r_col"]] <- new_rcol
    # Append AUTO rows (one row each, r_col = position in df).
    if (nrow(auto) > 0) {
      auto[["r_col"]] <- match(smiss, names(df))
      schema <- rbind(schema, auto)
    }
    schema <- schema[order(schema[["r_col"]]), ]
  } else {
    # unnamed / partially named
    n_top <- sum(top)
    if (n_top != ncol(df)) {
      stop(
        "Parquet schema size does not match data size. ",
        "They must match, unless the schema is fully named."
      )
    }
    if (is.null(nms) || all(is.na(nms) | nms == "")) {
      # no names at all, ok
    } else {
      # some names, those must match
      if (is.na(nms) | nms == names(df)) {
        stop(
          "Parquet schema names do not fully match data frame names."
        )
      }
    }
  }

  schema[["type"]] <- type_names[schema[["type"]]]
  schema[["repetition_type"]] <- repetition_types[schema[["repetition_type"]]]
  schema[["converted_type"]] <- ctype_names[schema[["converted_type"]]]

  schema
}

test_that("errors", {
  options <- parquet_options()
  expect_snapshot(error = TRUE, {
    .Call(
      rf_nanoparquet_write,
      mtcars,
      tempfile(),
      dim(mtcars),
      0L,
      list(character(), character()),
      rep(FALSE, ncol(mtcars)),
      options,
      map_schema_to_df(NULL, mtcars),
      rep(10L, ncol(mtcars)),
      1L,
      sys.call()
    )
  })
})

test_that("force PLAIN / RLE", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  d <- data.frame(d1 = 1:100, d2 = rep(1L, 100))

  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "true")
  write_parquet(d, tmp)
  expect_snapshot({
    read_parquet_pages(tmp)[["encoding"]]
  })

  withr::local_envvar(
    NANOPARQUET_FORCE_PLAIN = NA_character_,
    NANOPARQUET_FORCE_RLE = "true"
  )
  write_parquet(d, tmp)
  expect_snapshot({
    read_parquet_pages(tmp)[["encoding"]]
  })
})

test_that("cutoff for dict encoding decision", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  d <- data.frame(d1 = c(1:10000, rep(1L, 100000)))
  write_parquet(d, tmp)
  expect_snapshot({
    read_parquet_pages(tmp)[["encoding"]]
  })
  d <- data.frame(d1 = c(rep(c(TRUE, FALSE), 5000), rep(TRUE, 100000)))
  write_parquet(d, tmp)
  expect_snapshot({
    read_parquet_pages(tmp)[["encoding"]]
  })
})

test_that("write broken DECIMAL INT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(dec = 1:5)
  d2 <- data.frame(dec = as.double(1:5))

  schema <- parquet_schema(
    list("DECIMAL", precision = 5, scale = 2, primitive_type = "INT32")
  )
  schema$logical_type[1] <- list(NULL)

  # works w/o logical type
  write_parquet(d, tmp, schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
  })

  # no scale?
  schema2 <- schema
  schema2$scale <- NA_integer_
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema2)
  })

  # no precision?
  schema2 <- schema
  schema2$precision <- NA_integer_
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema2)
  })

  # bad precision?
  schema2 <- schema
  schema2$precision <- 10L
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema2)
  })
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema2)
  })
})

test_that("write broken DECIMAL INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(dec = 1:5)
  d2 <- data.frame(dec = as.double(1:5))

  schema <- parquet_schema(
    list("DECIMAL", precision = 5, scale = 2, primitive_type = "INT64")
  )
  schema$logical_type[1] <- list(NULL)
  schema2 <- schema
  schema2$precision <- 19L
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema2)
  })
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema2)
  })
})

test_that("write broken INT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(dec = 1:5)
  d2 <- data.frame(dec = as.double(1:5))

  schema <- parquet_schema("INT_32")
  schema$logical_type[[1]]$bit_width <- 64L

  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema)
  })
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
})

test_that("write broken UINT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(dec = 1:5)
  d2 <- data.frame(dec = as.double(1:5))

  schema <- parquet_schema("UINT_32")
  schema$logical_type[[1]]$bit_width <- 64L

  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema)
  })
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
})

test_that("write broken INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d2 <- data.frame(dec = as.double(1:5))

  schema <- parquet_schema("INT_64")
  schema$logical_type[[1]]$bit_width <- 32

  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })
})

test_that("INT96 errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = c(TRUE, FALSE))

  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("INT96"))
  })
})

test_that("FLOAT errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = c(TRUE, FALSE))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("FLOAT"))
  })
})

test_that("DOUBLE errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = c(TRUE, FALSE))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("DOUBLE"))
  })
})

test_that("BYTE_ARRAY errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = I(list(c(TRUE, FALSE))))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("BYTE_ARRAY"))
  })

  d2 <- data.frame(c = c(TRUE, FALSE))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = parquet_schema("BYTE_ARRAY"))
  })
})

test_that("FIXED_LEN_BYTE_ARRAY errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = I(list(c(TRUE, FALSE))))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("UUID"))
  })

  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("FLOAT16"))
  })

  d2 <- data.frame(c = c("foo", "bar", "foobar"))
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = schema)
  })

  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema)
  })

  d3 <- data.frame(c = I(list(charToRaw("foo"), charToRaw("foo2"))))
  expect_snapshot(error = TRUE, {
    write_parquet(d3, tmp, schema = schema)
  })

  d4 <- data.frame(c = c(TRUE, FALSE))
  expect_snapshot(error = TRUE, {
    write_parquet(d4, tmp, schema = schema)
  })
})

test_that("BOOLEAN errors", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = 1:5)
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = parquet_schema("BOOLEAN"))
  })

  schema <- parquet_schema(list("BOOLEAN", repetition_type = "OPTIONAL"))
  expect_snapshot(error = TRUE, {
    write_parquet(d, tmp, schema = schema)
  })
})

test_that("Errors when writing a dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(c = 1:5)
  expect_snapshot(error = TRUE, {
    write_parquet(
      d,
      tmp,
      schema = parquet_schema("DOUBLE"),
      encoding = "RLE_DICTIONARY"
    )
  })

  d2 <- data.frame(c = as.double(1:5))
  expect_snapshot(error = TRUE, {
    write_parquet(
      d2,
      tmp,
      schema = parquet_schema("BYTE_ARRAY"),
      encoding = "RLE_DICTIONARY"
    )
  })

  d3 <- data.frame(c = as.factor(letters))
  expect_snapshot(error = TRUE, {
    write_parquet(
      d3,
      tmp,
      schema = parquet_schema("DOUBLE"),
      encoding = "RLE_DICTIONARY"
    )
  })

  # too small value for DECIMAL INT32
  d4 <- data.frame(c = -101L)
  schema4 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT32"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
  })
  # too larse values for DECIMAL INT32
  d5 <- data.frame(c = 101L)
  schema5 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT32"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
  })

  # too small value for DECIMAL INT64
  d4 <- data.frame(c = -101L)
  schema4 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT64"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
  })
  # too larse values for DECIMAL INT64
  d5 <- data.frame(c = 101L)
  schema5 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT64"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
  })

  # no INTSXP -> DOUBLE conversion
  d5 <- data.frame(c = 1:5)
  expect_snapshot(error = TRUE, {
    write_parquet(
      d5,
      tmp,
      schema = parquet_schema("DOUBLE"),
      encoding = "RLE_DICTIONARY"
    )
  })
})

test_that("POSIXct dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    d = structure(1724157480.12919, class = c("POSIXct", "POSIXt"))
  )
  write_parquet(d, tmp, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    read_parquet_metadata(tmp)[["column_chunks"]]$encodings
  })

  expect_snapshot(error = TRUE, {
    write_parquet(
      d,
      tmp,
      schema = parquet_schema("INT32"),
      encoding = "RLE_DICTIONARY"
    )
  })

  d2 <- data.frame(x = as.difftime(1, units = "secs"))
  write_parquet(d2, tmp, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    read_parquet_metadata(tmp)[["column_chunks"]]$encodings
  })

  expect_snapshot(error = TRUE, {
    write_parquet(
      d2,
      tmp,
      schema = parquet_schema("INT32"),
      encoding = "RLE_DICTIONARY"
    )
  })
})

test_that("more dictionaries", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # too small value for DECIMAL INT32
  d4 <- data.frame(c = as.double(-101L))
  schema4 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT32"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
  })
  # too larse values for DECIMAL INT32
  d5 <- data.frame(c = as.double(101L))
  schema5 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT32"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
  })

  # invalid bit width
  schema <- parquet_schema("INT_32")
  schema$logical_type[[1]]$bit_width <- 64L
  expect_snapshot(error = TRUE, {
    write_parquet(d5, tmp, schema = schema, encoding = "RLE_DICTIONARY")
  })
  schema <- parquet_schema("UINT_32")
  schema$logical_type[[1]]$bit_width <- 64L
  expect_snapshot(error = TRUE, {
    write_parquet(d5, tmp, schema = schema, encoding = "RLE_DICTIONARY")
  })

  # too large value
  d6 <- data.frame(c = 128.0)
  expect_snapshot(error = TRUE, {
    write_parquet(
      d6,
      tmp,
      schema = parquet_schema("INT_8"),
      encoding = "RLE_DICTIONARY"
    )
  })

  # too large value
  d7 <- data.frame(c = 256.0)
  expect_snapshot(error = TRUE, {
    write_parquet(
      d7,
      tmp,
      schema = parquet_schema("UINT_8"),
      encoding = "RLE_DICTIONARY"
    )
  })

  # too small value
  d8 <- data.frame(c = -1)
  expect_snapshot(error = TRUE, {
    write_parquet(
      d8,
      tmp,
      schema = parquet_schema("UINT_8"),
      encoding = "RLE_DICTIONARY"
    )
  })

  # too small value for DECIMAL INT64
  d9 <- data.frame(c = as.double(-101L))
  schema4 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT64"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d9, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
  })
  # too larse values for DECIMAL INT32
  d10 <- data.frame(c = as.double(101L))
  schema5 <- parquet_schema(list(
    "DECIMAL",
    precision = 2,
    primitive_type = "INT64"
  ))
  expect_snapshot(error = TRUE, {
    write_parquet(d10, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
  })

  # no DOUBLE -> BYTE_ARRAY conversion
  expect_snapshot(error = TRUE, {
    write_parquet(
      d10,
      tmp,
      schema = parquet_schema("BYTE_ARRAY"),
      encoding = "RLE_DICTIONARY"
    )
  })
})

test_that("Even more dictionaries", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # bad UUID value
  d <- data.frame(x = "not-a-uuid")
  expect_snapshot(error = TRUE, {
    write_parquet(
      d,
      tmp,
      schema = parquet_schema("UUID"),
      encoding = "RLE_DICTIONARY"
    )
  })

  # invalid string length in FIXED_LEN_BYTE_ARRAY
  d2 <- data.frame(x = c("foo", "thisisbad"))
  sch2 <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))
  expect_snapshot(error = TRUE, {
    write_parquet(d2, tmp, schema = sch2, encoding = "RLE_DICTIONARY")
  })

  # no CHARSXP -> DOUBLE conversion
  expect_snapshot(error = TRUE, {
    write_parquet(
      d2,
      tmp,
      schema = parquet_schema("DOUBLE"),
      encoding = "RLE_DICTIONARY"
    )
  })
})

test_that("R -> Parquet mapping error", {
  d <- data.frame(x = raw(10))
  expect_snapshot(error = TRUE, {
    infer_parquet_schema(d)
  })
})

test_that("argument errors", {
  expect_snapshot(error = TRUE, {
    write_parquet(mtcars, 1:10)
    write_parquet(mtcars, letters)
  })
})

test_that("edge case: empty dict", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(address_id = c(NA_integer_, NA_integer_, NA_integer_))
  write_parquet(df, tmp)
  expect_equal(as.data.frame(read_parquet(tmp)), df)
})

test_that("character -> ENUM", {
  schema <- "ENUM"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }
  do(data.frame(s = c("foo", "bar", "foobar")))
  do(data.frame(s = c("foo", "bar", NA, "foobar")))
  do(data.frame(s = rep("foo", 10)))
  d <- data.frame(s = rep("foo", 11))
  d[["s"]][5] <- NA
  do(d)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("factor -> ENUM", {
  # a factor is always a dictionary
  schema <- "ENUM"

  d <- data.frame(s = as.factor(c("foo", "bar", "foobar")))
  test_write(d, schema)

  d <- data.frame(s = as.factor(c("foo", "bar", NA, "foobar")))
  test_write(d, schema)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> DECOMAL INT32", {
  schema <- list("DECIMAL", precision = 2, scale = 1, primitive_type = "INT32")

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> DECIMAL INT64", {
  schema <- list("DECIMAL", precision = 2, scale = 1, primitive_type = "INT64")

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }

  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> DECOMAL INT32", {
  schema <- list("DECIMAL", precision = 2, scale = 1, primitive_type = "INT32")

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = rep(1, 10)))
  do(d <- data.frame(d = c(rep(1, 5), NA, rep(1, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> DECIMAL INT64", {
  schema <- list("DECIMAL", precision = 2, scale = 1, primitive_type = "INT64")

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = rep(1, 10)))
  do(d <- data.frame(d = c(rep(1, 5), NA, rep(1, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> INT(8, *)", {
  schema <- "INT_8"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> INT(16, *)", {
  schema <- "INT_16"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> INT(64, *)", {
  schema <- "INT_64"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT64
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(8, TRUE)", {
  schema <- "INT_8"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(8, FALSE)", {
  schema <- "UINT_8"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(16, TRUE)", {
  schema <- "INT_16"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(16, FALSE)", {
  schema <- "UINT_16"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(32, TRUE)", {
  schema <- "INT_32"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(32, FALSE)", {
  schema <- "UINT_32"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(64, TRUE)", {
  schema <- "INT_64"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT(64, FALSE)", {
  schema <- "UINT_64"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, encoding = "PLAIN")
    test_write(d, schema, encoding = "RLE_DICTIONARY")
  }

  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("character -> UUID", {
  schema <- "UUID"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }

  do(data.frame(
    u = c(
      "00112233-4455-6677-8899-aabbccddeeff",
      "01112233-4455-6677-8899-aabbccddeeff",
      "02112233-4455-6677-8899-aabbccddeeff"
    )
  ))
  do(data.frame(
    u = c(
      "00112233-4455-6677-8899-aabbccddeeff",
      "01112233-4455-6677-8899-aabbccddeeff",
      NA,
      "02112233-4455-6677-8899-aabbccddeeff"
    )
  ))
  do(data.frame(u = rep("00112233-4455-6677-8899-aabbccddeeff", 10)))
  d <- data.frame(u = rep("00112233-4455-6677-8899-aabbccddeeff", 10))
  d[[1]][5] <- NA
  do(d)
})

test_that("obj_sum.nanoparquet_logical_type", {
  sch <- parquet_schema(
    "INT_8",
    list("DECIMAL", scale = 1, precision = 3, primitive_type = "INT32"),
    list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS"),
    list("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")
  )
  expect_snapshot({
    lapply(sch$logical_type, obj_sum.nanoparquet_logical_type)
  })
})

test_that("parquet_type", {
  expect_snapshot({
    parquet_type("AUTO")
    parquet_type("BOOLEAN")
    parquet_type("INT32")
    parquet_type("INT64")
    parquet_type("INT96")
    parquet_type("FLOAT")
    parquet_type("DOUBLE")
    parquet_type("BYTE_ARRAY")
    parquet_type("FIXED_LEN_BYTE_ARRAY", type_length = 10)

    parquet_type("STRING")
    parquet_type("ENUM")
    parquet_type("UUID")
    parquet_type("INT", bit_width = 8, is_signed = TRUE)
    parquet_type("INT", bit_width = 16, is_signed = TRUE)
    parquet_type("INT", bit_width = 32, is_signed = FALSE)
    parquet_type("INT", bit_width = 64, is_signed = FALSE)
    parquet_type("DECIMAL", precision = 5, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 5, scale = 0, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 5, scale = 5, primitive_type = "INT64")
    parquet_type("DECIMAL", precision = 5, primitive_type = "BYTE_ARRAY")
    parquet_type(
      "DECIMAL",
      precision = 5,
      primitive_type = "FIXED_LEN_BYTE_ARRAY",
      type_length = 5
    )
    parquet_type("FLOAT16")
    parquet_type("DATE")
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MILLIS")
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MICROS")
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "NANOS")
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MILLIS")
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "NANOS")
    parquet_type("JSON")
    parquet_type("BSON")
  })

  expect_snapshot({
    parquet_type("INT32", repetition_type = "OPTIONAL")
    parquet_type("STRING", repetition_type = "REQUIRED")
    parquet_type(
      "TIME",
      repetition_type = "REPEATED",
      is_adjusted_utc = TRUE,
      unit = "MILLIS"
    )
  })

  expect_snapshot(error = TRUE, {
    parquet_type("FOO")
    parquet_type("FIXED_LEN_BYTE_ARRAY")
    parquet_type("INT", bit_width = 8)
    parquet_type("INT", is_signed = TRUE)
    parquet_type("INT", bit_width = 10, is_signed = TRUE)
    parquet_type("INT", bit_width = 16, is_signed = 1)
    parquet_type("DECIMAL", precision = 5)
    parquet_type("DECIMAL", primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 5 / 2, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 0, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 5, scale = 6, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 10, primitive_type = "INT32")
    parquet_type("DECIMAL", precision = 19, primitive_type = "INT64")
    parquet_type(
      "DECIMAL",
      precision = 12,
      primitive_type = "FIXED_LEN_BYTE_ARRAY",
      type_length = 5
    )
    parquet_type("TIME", is_adjusted_utc = TRUE)
    parquet_type("TIME", unit = "MILLIS")
    parquet_type("TIME", is_adjusted_utc = 1, unit = "MILLIS")
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "FOO")
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE)
    parquet_type("TIMESTAMP", unit = "MILLIS")
    parquet_type("TIMESTAMP", is_adjusted_utc = 1, unit = "MILLIS")
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "FOO")
    parquet_type("MAP")
    parquet_type("UNKNOWN")
    parquet_type("INT32", repetition_type = TRUE)
    parquet_type("INT32", repetition_type = "FOO")
  })

  # Need this as well for covr, which does not handle stop()
  # in snapshots, apparently
  expect_error(parquet_type("FOO"), "not supported by nanoparquet")
  expect_error(parquet_type("MAP"), "not supported by nanoparquet")
  expect_error(parquet_type("UNKNOWN"), "not supported by nanoparquet")
})

test_that("parquet_type converted type shortcuts", {
  expect_equal(
    parquet_type("INT_8"),
    parquet_type("INT", bit_width = 8, is_signed = TRUE)
  )
  expect_equal(
    parquet_type("INT_16"),
    parquet_type("INT", bit_width = 16, is_signed = TRUE)
  )
  expect_equal(
    parquet_type("INT_32"),
    parquet_type("INT", bit_width = 32, is_signed = TRUE)
  )
  expect_equal(
    parquet_type("INT_64"),
    parquet_type("INT", bit_width = 64, is_signed = TRUE)
  )

  expect_equal(
    parquet_type("UINT_8"),
    parquet_type("INT", bit_width = 8, is_signed = FALSE)
  )
  expect_equal(
    parquet_type("UINT_16"),
    parquet_type("INT", bit_width = 16, is_signed = FALSE)
  )
  expect_equal(
    parquet_type("UINT_32"),
    parquet_type("INT", bit_width = 32, is_signed = FALSE)
  )
  expect_equal(
    parquet_type("UINT_64"),
    parquet_type("INT", bit_width = 64, is_signed = FALSE)
  )

  expect_equal(
    parquet_type("TIME_MICROS"),
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MICROS")
  )
  expect_equal(
    parquet_type("TIME_MILLIS"),
    parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MILLIS")
  )

  expect_equal(
    parquet_type("TIMESTAMP_MICROS"),
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")
  )
  expect_equal(
    parquet_type("TIMESTAMP_MILLIS"),
    parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MILLIS")
  )
})

test_that("parquet_schema", {
  sch <- parquet_schema(
    "INT32",
    "BOOLEAN",
    "INT32",
    "INT64",
    "INT96",
    "FLOAT",
    "BYTE_ARRAY",
    list("FIXED_LEN_BYTE_ARRAY", type_length = 10)
  )
  expect_snapshot(as.data.frame(sch))

  sch2 <- parquet_schema(
    a = "INT32",
    b = "BOOLEAN",
    c = "INT32",
    d = "INT64",
    e = "INT96",
    f = "FLOAT",
    g = "BYTE_ARRAY",
    list("FIXED_LEN_BYTE_ARRAY", type_length = 10)
  )
  expect_snapshot(as.data.frame(sch2))

  sch3 <- parquet_schema(
    a = "INT32",
    b = "BOOLEAN",
    c = "INT32",
    d = "INT64",
    e = "INT96",
    f = "FLOAT",
    g = "BYTE_ARRAY",
    h = list("FIXED_LEN_BYTE_ARRAY", type_length = 10)
  )
  expect_snapshot(as.data.frame(sch3))

  sch4 <- parquet_schema(
    "STRING",
    "ENUM",
    "UUID",
    list("INTEGER", bit_width = 8, is_signed = TRUE),
    list("INTEGER", bit_width = 64, is_signed = FALSE),
    list("DECIMAL", precision = 5, primitive_type = "INT64"),
    "FLOAT16",
    "DATE",
    "JSON",
    "BSON"
  )
  expect_snapshot(as.data.frame(sch4))

  sch5 <- parquet_schema(
    foo = "AUTO",
    bar = "INT32"
  )
  expect_snapshot(as.data.frame(sch5))
})

#' Map between R and Parquet data types
#'
#' Note that this function is now deprecated. Please use
#' [read_parquet_schema()] for files, and [infer_parquet_schema()] for
#' data frames.
#'
#' This function works two ways. It can map the R types of a data frame to
#' Parquet types, to see how [write_parquet()] would write out the data
#' frame. It can also map the types of a Parquet file to R types, to see
#' how [read_parquet()] would read the file into R.
#'
#' @param x Path to a Parquet file, or a data frame.
#' @param options Nanoparquet options, see [parquet_options()].
#' @return Data frame with columns:
#'   * `file_name`: file name.
#'   * `name`: column name.
#'   * `type`: (low level) Parquet data type.
#'   * `r_type`: the R type that corresponds to the Parquet type.
#'     Might be `NA` if [read_parquet()] cannot read this column. See
#'     [nanoparquet-types] for the type mapping rules.
#'   * `repetition_type`: whether the column in `REQUIRED` (cannot be
#'     `NA`) or `OPTIONAL` (may be `NA`). `REPEATED` columns are not
#'     currently supported by nanoparquet.
#'   * `logical_type`: Parquet logical type in a list column.
#'      An element has at least an entry called `type`, and potentially
#'      additional entries, e.g. `bit_width`, `is_signed`, etc.
#'
#' @seealso [read_parquet_metadata()] to read more metadata,
#'   [read_parquet_info()] for a very short summary.
#'   [read_parquet_schema()] for the complete Parquet schema.
#'   [read_parquet()], [write_parquet()], [nanoparquet-types].
#' @export

parquet_column_types <- function(x, options = parquet_options()) {
  warning(
    "`parquet_column_types()` is deprecated, please use ",
    "`read_parquet_schema()` or `parquet_schema()` instead."
  )
  if (is.character(x)) {
    parquet_column_types_file(x, options)
  } else if (is.data.frame(x)) {
    infer_parquet_schema(x, options)
  } else {
    stop("`x` must be a file name or a data frame in `parquet_column_types()`")
  }
}

parquet_column_types_file <- function(file, options) {
  mtd <- read_parquet_metadata(file)
  sch <- mtd$schema
  add_r_type_to_schema(mtd, sch, options)
}

add_r_type_to_schema <- function(mtd, sch, options, col_select = NULL) {
  kv <- mtd$file_meta_data$key_value_metadata[[1]]

  type_map <- c(
    BOOLEAN = "logical",
    INT32 = "integer",
    INT64 = "double",
    DOUBLE = "double",
    FLOAT = "double",
    INT96 = "POSIXct",
    FIXED_LEN_BYTE_ARRAY = "blob",
    BYTE_ARRAY = "blob"
  )

  sch$r_type <- unname(type_map[sch$type])

  sch$r_type[
    sch$type == "FIXED_LEN_BYTE_ARRAY" &
      sch$converted_type == "DECIMAL"
  ] <- "double"
  sch$r_type[
    vapply(
      sch$logical_type,
      function(x) {
        !is.null(x$type) && x$type %in% c("STRING", "ENUM", "UUID")
      },
      logical(1)
    ) |
      sch$converted_type == "UTF8"
  ] <- "character"

  # detected from Arrow schema
  if (options[["use_arrow_metadata"]]) {
    spec <- if ("ARROW:schema" %in% kv$key) {
      kv <- mtd$file_meta_data$key_value_metadata[[1]]
      arrow_find_special(
        kv$value[match("ARROW:schema", kv$key)],
        file
      )
    }
    if (length(spec$factor)) {
      sch$r_type[spec$factor] <- "factor"
    }
    if (length(spec$difftime)) sch$r_type[spec$difftime] <- "difftime"
  }

  # TODO: this is duplicated in the C++ code
  # our own conversions
  dates <- vapply(
    sch$logical_type,
    function(lt) !is.null(lt$type) && lt$type == "DATE",
    logical(1)
  ) |
    sch$converted_type == "DATE"
  sch$r_type[dates] <- "Date"

  int64s <- vapply(
    sch$logical_type,
    function(lt) {
      !is.null(lt$type) &&
        lt$type == "INT" &&
        !is.null(lt$bit_width) &&
        lt$bit_width == 64L
    },
    logical(1)
  ) |
    (!is.na(sch$converted_type) & sch$converted_type == "INT_64")
  if (options[["read_int64_type"]] != "double") {
    sch$r_type[int64s] <- "integer64"
  }

  hmss <- vapply(
    sch$logical_type,
    function(lt) !is.null(lt$type) && lt$type == "TIME",
    logical(1)
  ) |
    sch$converted_type == "TIME_MILLIS" |
    sch$converted_type == "TIME_MICROS"
  sch$r_type[hmss] <- "hms"

  poscts <- vapply(
    sch$logical_type,
    function(lt) !is.null(lt) && lt$type == "TIMESTAMP",
    logical(1)
  ) |
    sch$converted_type == "TIMESTAMP_MICROS"
  sch$r_type[poscts] <- "POSIXct"

  sch$r_col <- compute_r_col(sch)
  cols <- c(
    "file_name",
    "r_col",
    "name",
    "r_type",
    setdiff(colnames(sch), c("file_name", "r_col", "name", "r_type"))
  )
  sch <- sch[, cols]

  sch <- process_nested_columns(sch)

  sch
}

compute_r_col <- function(sch) {
  n <- nrow(sch)
  r_col <- rep(NA_integer_, n)
  if (n <= 1L) {
    return(r_col)
  }

  subtree_end <- function(i) {
    nc <- sch$num_children[i]
    if (is.na(nc) || nc == 0L) {
      return(i)
    }
    end <- i
    for (k in seq_len(nc)) {
      end <- subtree_end(end + 1L)
    }
    end
  }

  col_idx <- 0L
  i <- 2L
  while (i <= n) {
    col_idx <- col_idx + 1L
    end <- subtree_end(i)
    r_col[i:end] <- col_idx
    i <- end + 1L
  }

  r_col
}

process_nested_columns <- function(sch) {
  is_group <- !is.na(sch$num_children) & sch$num_children > 0
  is_group[1] <- FALSE
  wh_groups <- which(is_group)

  # first work out the child columns for each group
  # this makes it easier to create the correct nested types
  children <- replicate(nrow(sch), list(integer()))
  parents <- rep(NA_integer_, nrow(sch))
  for (gr in rev(wh_groups)) {
    nc <- sch$num_children[gr]
    chidx <- gr
    while (nc > 0) {
      chidx <- chidx + 1L
      children[[gr]] <- c(children[[gr]], chidx)
      parents[chidx] <- gr
      # if this child is a group, skip its children
      if (!is.na(sch$num_children[chidx])) {
        chidx <- chidx + sch$num_children[chidx]
      }
      nc <- nc - 1L
    }
  }
  sch$children <- children

  is_list <- (!is.na(sch$converted_type) & sch$converted_type == "LIST") |
    vapply(
      sch$logical_type,
      function(lt) {
        !is.null(lt$type) && lt$type == "LIST"
      },
      logical(1)
    )

  is_repeated <- !is.na(sch$repetition_type) &
    sch$repetition_type == "REPEATED" &
    !is_list &
    (is.na(parents) | !is_list[parents])

  wh_nested <- which(is_group | is_repeated)
  for (nst in rev(wh_nested)) {
    is_middle <- FALSE
    if (is_group[nst]) {
      # is it a LIST middle node? Rules from
      # https://github.com/apache/parquet-format/blob/master/LogicalTypes.md
      # #backward-compatibility-rules
      is_middle <- !is.na(sch$num_children[nst]) &&
        sch$num_children[nst] == 1L &&
        sch$repetition_type[sch$children[[nst]]] != "REPEATED" &&
        sch$name[sch$children[[nst]]] != "array" &&
        sch$name[sch$children[[nst]]] !=
          paste0(sch$name[parents[nst]], "_tuple")
      if (is_middle) {
        sch$r_type[nst] <- sch$r_type[sch$children[[nst]]]
      } else {
        sch$r_type[nst] <- paste0(
          "list(",
          paste(sch$r_type[sch$children[[nst]]], collapse = ", "),
          ")"
        )
      }
      sch$r_type[sch$children[[nst]]] <- NA_character_
    }
    if (is_repeated[nst] && !is_middle) {
      sch$r_type[nst] <- paste0("list(", sch$r_type[nst], ")")
    }
  }

  sch
}

`%||%` <- function(l, r) if (is.null(l)) r else l

`%&&%` <- function(l, r) if (is.null(l)) NULL else r

is_rcmd_check <- function() {
  if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    FALSE
  } else {
    Sys.getenv("_R_CHECK_PACKAGE_NAME_", "") != ""
  }
}

mkdirp <- function(x) {
  dir.create(x, showWarnings = FALSE, recursive = TRUE)
}

capitalize <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

is_asan <- function() {
  .Call(is_asan_)
}

is_ubsan <- function() {
  .Call(is_ubsan_)
}

is_flag <- function(x) {
  is.logical(x) && length(x) == 1 && !is.na(x)
}

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}

is_icount <- function(x, zero = FALSE) {
  is.integer(x) &&
    length(x) == 1 &&
    !is.na(x) &&
    ((zero && x >= 0L) || (!zero && x >= 1L))
}

is_dcount <- function(x, zero = FALSE) {
  is.double(x) &&
    length(x) == 1 &&
    !is.na(x) &&
    as.integer(x) == x &&
    ((zero && x >= 0) || (!zero && x >= 1))
}

as_count <- function(x, name = "x", zero = FALSE) {
  if (is_icount(x, zero)) {
    return(x)
  }
  if (is_dcount(x, zero)) {
    return(as.integer(x))
  }
  stop(name, " must be a count, i.e. an integer scalar")
}

as_integer_scalar <- function(x, name = "x") {
  if (is.integer(x) && length(x) == 1 && !is.na(x)) {
    return(x)
  }
  if (is.double(x) && length(x) == 1 && !is.na(x) && as.integer(x) == x) {
    return(as.integer(x))
  }
  stop(name, " must be an integer scalar")
}

is_uint32 <- function(x) {
  is.numeric(x) &&
    length(x) == 1 &&
    !is.na(x) &&
    round(x) == x &&
    x >= 0 &&
    x <= 4294967295
}

map_chr <- function(.x, .f, ...) {
  vapply(.x, .f, FUN.VALUE = character(1), ...)
}

map_int <- function(.x, .f, ...) {
  vapply(.x, .f, FUN.VALUE = integer(1), ...)
}

any_na <- function(x) {
  t <- typeof(x)
  if (t == "list") {
    .Call(rf_nanoparquet_any_null, x)
  } else if (inherits(x, "integer64")) {
    .Call(rf_nanoparquet_any_na_int64, x)
  } else if (t == "double") {
    .Call(rf_nanoparquet_any_na, x)
  } else {
    anyNA(x)
  }
}

is_named <- function(x) {
  nm <- names(x)
  !is.null(nm) && !anyNA(nm)
}

split_at <- function(x, sizes) {
  starts <- c(1, cumsum(sizes[-length(sizes)]) + 1)
  Map(function(s, n) x[seq_len(n) + s - 1], starts, sizes)
}

test_that("is_rcmd_check", {
  withr::local_envvar(NOT_CRAN = "true")
  expect_false(is_rcmd_check())

  withr::local_envvar(
    NOT_CRAN = NA_character_,
    "_R_CHECK_PACKAGE_NAME_" = NA_character_
  )
  expect_false(is_rcmd_check())

  withr::local_envvar(
    NOT_CRAN = NA_character_,
    "_R_CHECK_PACKAGE_NAME_" = "foo"
  )
  expect_true(is_rcmd_check())
})

test_that("is_flag", {
  expect_true(is_flag(TRUE))
  expect_true(is_flag(FALSE))

  expect_false(is_flag(c(TRUE, TRUE)))
  expect_false(is_flag(1))
  expect_false(is_flag(NA))
})

test_that("is_string", {
  expect_true(is_string("a"))

  expect_false(is_string(c("a", "b")))
  expect_false(is_string(NA_character_))
  expect_false(is_string(1))
  expect_false(is_string(NULL))
})

test_that("is_uint32", {
  expect_true(is_uint32(0))
  expect_true(is_uint32(0L))
  expect_true(is_uint32(100))
  expect_true(is_uint32(100L))
  expect_true(is_uint32(4294967295))
  expect_true(is_uint32(4000000000))
  expect_true(is_uint32(2147483647))
  expect_true(is_uint32(2147483647L))

  expect_false(is_uint32(1 / 2))
  expect_false(is_uint32(-1))
  expect_false(is_uint32(4294967296))
  expect_false(is_uint32("a"))
  expect_false(is_uint32(NA_integer_))
  expect_false(is_uint32(NA_real_))
  expect_false(is_uint32("foo"))
})

test_that("is_icount", {
  expect_true(is_icount(1L))
  expect_true(is_icount(100L))
  expect_true(is_icount(2147483647L))

  expect_false(is_icount(NA_integer_))
  expect_false(is_icount(1:2))
  expect_false(is_icount(1))
  expect_false(is_icount(0L))
  expect_false(is_icount(-100L))
})

test_that("is_dcount", {
  expect_true(is_dcount(1))
  expect_true(is_dcount(100))
  expect_true(is_dcount(2147483647))

  expect_false(is_dcount(NA_real_))
  expect_false(is_dcount(1:2))
  expect_false(is_dcount(1L))
  expect_false(is_dcount(0))
  expect_false(is_dcount(-100))
})

test_that("as_count", {
  expect_equal(as_count(1), 1L)
  expect_equal(as_count(100), 100L)
  expect_snapshot(error = TRUE, {
    as_count(1:2)
    as_count(0)
    as_count(NA_real_)
    as_count(-100)
    as_count(-100L)
  })
})

#' Write a data frame to a Parquet file
#'
#' Writes the contents of an R data frame into a Parquet file.
#'
#' `write_parquet()` converts string columns to UTF-8 encoding by calling
#' [base::enc2utf8()]. It does the same for factor levels.
#'
#' @param x Data frame to write.
#' @param file Path to the output file. If this is the string `":raw:"`,
#'   then the data frame is written to a memory buffer, and the memory
#'   buffer is returned as a raw vector. If `":stdout:"`, then it is
#'   written to the standard output. (When writing to the standard output,
#'   special care is needed to make sure no regular R output gets mixed
#'   up with the Parquet bytes!)
#' @param schema Parquet schema. Specify a schema to tweak the default
#'   nanoparquet R -> Parquet type mappings. Use [parquet_schema()] to
#'   create a schema that you can use here, or [read_parquet_schema()] to
#'   use the schema of a Parquet file.
#' @param compression Compression algorithm to use. Currently `"snappy"`
#'   (the default), `"gzip"`, `"zstd"`, and `"uncompressed"` are supported.
#' @param encoding Encoding to use. Possible values:
#'   * If `NULL`, the appropriate encoding is selected automatically:
#'     `RLE` or `PLAIN` for `BOOLEAN` columns, `RLE_DICTIONARY` for other
#'     columns with many repeated values, and `PLAIN` otherwise.
#'   * If It is a single (unnamed) character string, then it'll be used
#'     for all columns.
#'   * If it is an unnamed character vector of encoding names of the same
#'     length as the number of columns in the data frame, then those
#'     encodings will be used for each column.
#'   * If it is a named character vector, then the named must be unique
#'     and each name must match a column name, to specify the encoding of
#'     that column. The special empty name (`""`) applies to the rest of
#'     the columns. If there is no empty name, the rest of the columns
#'     will use the default encoding.
#'
#'   If `NA_character_` is specified for a column, the default encoding is
#'   used for the column.
#'
#'   If a specified encoding is invalid for a certain column type,
#'   or nanoparquet does not implement it, `write_parquet()` throws an
#'   error.
#'
#'   Currently `write_parquet()` supports the following encodings:
#'
#'   - `PLAIN` for all column types,
#'   - `PLAIN_DICTIONARY` and `RLE_DICTIONARY` for all column types,
#'   - `RLE` for BOOLEAN columns.
#'
#'   See [parquet-encodings] for more about encodings.
#' @param metadata Additional key-value metadata to add to the file.
#'   This must be a named character vector, or a data frame with columns
#'   character columns called `key` and `value`.
#' @param row_groups Row groups of the Parquet file. If `NULL`, then the
#'   `num_rows_per_row_group` option is used from the `options` argument,
#'   see [parquet_options()]. Otherwise it must be an integer vector,
#'   specifying the starts of the row groups.
#' @param options Nanoparquet options, see [parquet_options()].
#' @return `NULL`, unless `file` is `":raw:"`, in which case the Parquet
#'   file is returned as a raw vector.
#'
#' @export
#' @seealso [read_parquet_metadata()], [read_parquet()].
#' @examplesIf FALSE
#' # add row names as a column, because `write_parquet()` ignores them.
#' mtcars2 <- cbind(name = rownames(mtcars), mtcars)
#' write_parquet(mtcars2, "mtcars.parquet")

write_parquet <- function(
  x,
  file,
  schema = NULL,
  compression = c("snappy", "gzip", "zstd", "uncompressed"),
  encoding = NULL,
  metadata = NULL,
  row_groups = NULL,
  options = parquet_options()
) {
  file <- path.expand(file)

  compression <- parse_compression(compression, options)

  dim <- as.integer(dim(x))

  x <- prepare_write_df(x)

  schema <- map_schema_to_df(schema, x, options)

  if (is.null(metadata)) {
    metadata <- list(character(), character())
  } else if (is.data.frame(metadata)) {
    stopifnot(ncol(metadata) == 2)
  } else {
    stopifnot(
      is.character(metadata),
      length(names(metadata)) == length(metadata)
    )
    metadata <- list(names(metadata), unname(metadata))
  }

  if (options[["write_arrow_metadata"]]) {
    if (!"ARROW:schema" %in% metadata[[1]]) {
      metadata[[1]] <- c(metadata[[1]], "ARROW:schema")
      metadata[[2]] <- c(metadata[[2]], encode_arrow_schema(x))
    }
  }

  schema_top <- schema[!duplicated(schema[["r_col"]]), ]
  schema_top <- check_schema_required_cols(x, schema_top)
  required <- schema_top[["repetition_type"]] == "REQUIRED"

  encoding <- parse_encoding(encoding, x)

  row_groups <- row_groups %||%
    default_row_groups(x, schema, compression, encoding, options)

  row_group_starts <- parse_row_groups(x, row_groups)
  x <- row_group_starts[[1]]
  row_group_starts <- row_group_starts[[2]]

  res <- .Call(
    rf_nanoparquet_write,
    x,
    file,
    dim,
    compression,
    metadata,
    required,
    options,
    schema,
    encodings[encoding],
    row_group_starts,
    sys.call()
  )

  if (is.null(res)) {
    invisible()
  } else {
    res
  }
}

parse_compression <- function(
  compression = c("snappy", "gzip", "zstd", "uncompressed"),
  options
) {
  codecs <- c("uncompressed" = 0L, "snappy" = 1L, "gzip" = 2L, "zstd" = 6L)
  compression <- codecs[match.arg(compression)]
  if (is.na(options[["compression_level"]])) {
    # -1 is an allowed value for zstd, so we set the default here
    if (compression == "zstd") {
      options[["compression_level"]] <- 3L
    } else {
      options[["compression_level"]] <- -1L
    }
  }
  compression
}

prepare_write_df <- function(x) {
  # convert strings to UTF-8
  strs <- which(vapply(x, is.character, logical(1)))
  for (idx in strs) {
    x[[idx]] <- enc2utf8(x[[idx]])
  }
  # factor levels as well
  fctrs <- which(vapply(x, "inherits", "factor", FUN.VALUE = logical(1)))
  for (idx in fctrs) {
    levels(x[[idx]]) <- enc2utf8(levels(x[[idx]]))
  }

  # Date must be integer
  double_dates <- which(vapply(x, FUN.VALUE = logical(1), function(x) {
    inherits(x, "Date") && is.double(x)
  }))
  for (idx in double_dates) {
    cls <- class(x[[idx]])
    xi <- as.integer(floor(as.numeric(x[[idx]])))
    class(xi) <- cls
    x[[idx]] <- xi
  }

  # Convert hms to double
  hmss <- which(vapply(x, "inherits", "hms", FUN.VALUE = logical(1)))
  for (idx in hmss) {
    # this keeps the class
    mode(x[[idx]]) <- "double"
  }

  # Make sure POSIXct is double
  posixcts <- which(vapply(x, "inherits", "POSIXct", FUN.VALUE = logical(1)))
  for (idx in posixcts) {
    # keeps the class
    mode(x[[idx]]) <- "double"
  }

  # difftime must be saved in seconds
  difftimes <- which(vapply(x, "inherits", "difftime", FUN.VALUE = logical(1)))
  for (idx in setdiff(difftimes, hmss)) {
    x[[idx]] <- as.difftime(
      as.double(x[[idx]], units = "secs"),
      units = "secs"
    )
  }

  x
}

check_schema_required_cols <- function(x, schema) {
  # if schema has REQUIRED, but the column has NAs, that's an error
  rt <- schema[["repetition_type"]]
  req <- !is.na(rt) & (rt == "REQUIRED" | rt == 0L)
  hasna <- vapply(x, any_na, logical(1))
  bad <- which(req & hasna)
  if (length(bad) > 0) {
    stop(
      "Parquet schema does not allow missing values for column",
      if (length(bad) > 1) "s",
      ":",
      paste(names(x)[bad], collapse = ", ")
    )
  }
  schema[["repetition_type"]][is.na(rt)] <-
    ifelse(hasna[is.na(rt)], "OPTIONAL", "REQUIRED")
  schema
}

parse_encoding <- function(encoding, x) {
  stopifnot(
    "`encoding` must be `NULL` or a character vector" = is.null(encoding) ||
      is.character(encoding),
    "`encoding` contains at least one unknown encoding" = all(
      is.na(encoding) | encoding %in% names(encodings)
    )
  )

  if (is.null(encoding)) {
    structure(rep(NA_character_, length(x)), names = names(x))
  } else if (is_named(encoding)) {
    stopifnot(
      "names of `encoding` must be unique" = !anyDuplicated(names(encoding)),
      "names of `encoding` must match names of `x`" = all(
        names(encoding) %in% c(names(x), "")
      )
    )
    def <- c(encoding[names(encoding) == ""], NA_character_)[1]
    encoding <- encoding[names(encoding) != ""]
    res <- structure(rep(def, length(x)), names = names(x))
    res[names(encoding)] <- encoding
    res
  } else if (length(encoding) == 1) {
    structure(rep(encoding, length(x)), names = names(x))
  } else {
    stopifnot(length(encoding) == length(x))
    structure(encoding, names = names(x))
  }
}

# we should refine this later
default_row_groups <- function(x, schema, compression, encoding, options) {
  if (nrow(x) == 0L) {
    return(integer(0))
  }
  default_size <- options[["num_rows_per_row_group"]]
  seq(1L, nrow(x), by = default_size)
}

# this one as well
default_append_row_groups <- function(
  x,
  crnt_metadata,
  schema,
  compression,
  encoding,
  options
) {
  default_size <- options[["num_rows_per_row_group"]]
  crnt_sizes <- crnt_metadata$row_groups$num_rows
  last_size <- utils::tail(crnt_sizes, 1)
  crnt <- utils::head(c(1L, cumsum(crnt_sizes) + 1L), -1)
  if (last_size + nrow(x) > default_size) {
    # create new row group(s)
    crnt_rows <- crnt_metadata$file_meta_data$num_rows
    new <- seq(1L, nrow(x) + last_size, by = default_size)[-1] +
      crnt_rows -
      last_size
    crnt <- c(crnt, new)
  }
  as.integer(crnt)
}

parse_row_groups <- function(x, rg) {
  if (length(rg) == 0L) {
    if (!is.integer(rg)) {
      stop(
        "Row groups must be specified as a growing positive integer ",
        "vector, starting with 1."
      )
    }
    return(list(x = x, row_groups = rg))
  }
  if (
    !is.integer(rg) ||
      anyNA(rg) ||
      any(rg <= 0) ||
      any(diff(rg) <= 0) ||
      rg[1] != 1L
  ) {
    stop(
      "Row groups must be specified as a growing positive integer ",
      "vector, starting with 1."
    )
  }
  list(x = x, row_groups = rg)
}

#' Append a data frame to an existing Parquet file
#'
#' The schema of the data frame must be compatible with the schema of
#' the file.
#'
#' @section Warning:
#' This function is **not** atomic! If it is interrupted, it may leave
#' the file in a corrupt state. To work around this create a copy of the
#' original file, append the new data to the copy, and then rename the
#' new, extended file to the original one.
#'
#' @section About row groups:
#' A Parquet file may be partitioned into multiple row groups, and indeed
#' most large Parquet files are. `append_parquet()` is only able to update
#' the existing file along the row group boundaries. There are two
#' possibilities:
#'
#' - `append_parquet()` keeps all existing row groups in `file`, and
#'   creates new row groups for the new data. This mode can be forced by
#'   the `keep_row_groups` option in `options`, see [parquet_options()].
#' - Alternatively, `write_parquet` will overwrite the _last_ row group in
#'   file, with its existing contents plus the (beginning of) the new data.
#'   This mode makes more sense if the last row group is small, because
#'   many small row groups are inefficient.
#'
#' By default `append_parquet` chooses between the two modes automatically,
#' aiming to create row groups with at least `num_rows_per_row_group`
#' (see [parquet_options()]) rows. You can customize this behavior with
#' the `keep_row_groups` options and the `row_groups` argument.
#'
#' @param x Data frame to append.
#' @param file Path to the output file.
#' @param compression Compression algorithm to use for the newly written
#'   data. See [write_parquet()].
#' @param encoding Encoding to use for the newly written data. It does not
#'   have to be the same as the encoding of data in `file`. See
#'   [write_parquet()] for possible values.
#' @param row_groups Row groups of the new, extended Parquet file.
#'   [append_parquet()] can only change the last existing row group, and
#'   if `row_groups` is specified, it has respect this. I.e. if the
#'   existing file has `n` rows, and the last row group starts at `k`
#'   (`k <= n`), then the first row group in `row_groups` that refers to
#'   the new data must start at `k` or `n+1`.
#'   (It is simpler to specify `num_rows_per_row_group` in `options`, see
#'   [parquet_options()] instead of `row_groups`. Only use `row_groups` if
#'   you need complete control.)
#' @param options Nanoparquet options, for the new data, see
#'   [parquet_options()]. The `keep_row_groups` option also affects whether
#'   `append_parquet()` overwrites existing row groups in `file`.
#'
#' @export
#' @seealso [write_parquet()].

append_parquet <- function(
  x,
  file,
  compression = c("snappy", "gzip", "zstd", "uncompressed"),
  encoding = NULL,
  row_groups = NULL,
  options = parquet_options()
) {
  file <- path.expand(file)

  if (!file.exists(file)) {
    return(write_parquet(
      x,
      file,
      compression = compression,
      encoding = encoding,
      row_groups = row_groups,
      options = options
    ))
  }

  compression <- parse_compression(compression, options)

  x <- prepare_write_df(x)

  mtd <- read_parquet_metadata(file)
  schema <- read_parquet_schema(file)
  schema <- map_schema_to_df(schema, x, list())
  schema_top <- schema[!duplicated(schema[["r_col"]]), ]
  schema_top <- check_schema_required_cols(x, schema_top)
  required <- schema_top[["repetition_type"]] == "REQUIRED"
  encoding <- parse_encoding(encoding, x)

  nrow_file <- as.integer(mtd$file_meta_data$num_rows)
  row_groups <- row_groups %||%
    if (options[["keep_row_groups"]]) {
      c(
        1L,
        nrow_file +
          default_row_groups(
            x,
            schema,
            compression,
            encoding,
            options
          )
      )
    } else {
      default_append_row_groups(
        x,
        mtd,
        schema,
        compression,
        encoding,
        options
      )
    }
  row_group_starts <- parse_row_groups(x, row_groups)
  x <- row_group_starts[[1]]
  row_group_starts <- row_group_starts[[2]]

  # check if we are extending the last row group of the original file
  nrow_total <- nrow_file + nrow(x)
  extend_last_row_group <- !(nrow_file + 1L) %in% row_group_starts

  # if yes, prepend the last row group of the file to the new data
  if (extend_last_row_group) {
    num_rgs_file <- nrow(mtd$row_groups)
    x_prep <- read_parquet_row_group(file, num_rgs_file - 1L)
    x <- rbind(x_prep, x)
    nrow_keep <- nrow_file - nrow(x_prep)
  } else {
    nrow_keep <- nrow_file
  }

  # these indices refer to the new data
  row_group_starts <- row_group_starts[row_group_starts > nrow_keep] -
    nrow_keep

  dim <- c(dim(x), nrow_total)

  res <- .Call(
    rf_nanoparquet_append,
    x,
    file,
    as.integer(dim),
    compression,
    required,
    options,
    schema,
    encodings[encoding],
    as.integer(row_group_starts),
    extend_last_row_group,
    sys.call()
  )

  if (is.null(res)) {
    invisible()
  } else {
    res
  }
}

test_that("parquet CLI can read files written by write_parquet()", {
  skip_without_parquet_cli()

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mtcars, tmp)
  res <- processx::run("parquet", c("cat", tmp), error_on_status = FALSE)
  expect_equal(res$status, 0L)
})

test_that("append_parquet creates file if it doesn't exist", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df1 <- test_df()
  append_parquet(df1, tmp)
  expect_true(file.exists(tmp))
  expect_equal(read_parquet(tmp), df1)
})

test_that("keep all row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df1 <- test_df(missing = TRUE)
  write_parquet(df1, tmp)
  mtd1 <- read_parquet_metadata(tmp)

  df2 <- test_df()
  append_parquet(
    df2,
    tmp,
    options = parquet_options(keep_row_groups = TRUE)
  )
  df3 <- read_parquet(tmp)
  mtd2 <- read_parquet_metadata(tmp)

  df21 <- utils::head(df3, nrow(df1))
  df22 <- utils::tail(df3, nrow(df2))
  row.names(df22) <- NULL
  expect_equal(df21, df1)
  expect_equal(df22, df2)
  expect_equal(mtd1$file_meta_data[-3], mtd2$file_meta_data[-3])

  expect_equal(nrow(mtd2$row_groups), 2L)
  expect_equal(
    mtd1$column_chunks,
    utils::head(mtd2$column_chunks, nrow(mtd1$column_chunks))
  )
})

test_that("overwrite last row group", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df1 <- test_df(missing = TRUE)
  write_parquet(df1, tmp, row_groups = c(1L, 31L))
  mtd1 <- read_parquet_metadata(tmp)

  df2 <- test_df()
  append_parquet(
    df2,
    tmp,
    options = parquet_options(num_rows_per_row_group = 30)
  )
  df3 <- read_parquet(tmp)
  mtd2 <- read_parquet_metadata(tmp)

  df21 <- utils::head(df3, nrow(df1))
  df22 <- utils::tail(df3, nrow(df2))
  row.names(df22) <- NULL
  expect_equal(df21, df1)
  expect_equal(df22, df2)
  expect_equal(mtd1$file_meta_data[-3], mtd2$file_meta_data[-3])

  # the first column chunk of row group 1 is at the same position as before
  ccs1 <- mtd1$column_chunks
  ccs2 <- mtd2$column_chunks
  idx1 <- which(ccs1$row_group == 1)[1]
  idx2 <- which(ccs2$row_group == 1)[1]
  expect_equal(idx1, idx2)
  expect_equal(ccs1$file_offset[idx1], ccs2$file_offset[idx2])
})

test_that("appending all-NA rows to REQUIRED columns gives a clear error", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df_valid <- data.frame(
    id = as.integer(1:5),
    value = as.numeric(1:5),
    label = c("A", "B", "C", "D", "E")
  )
  df_nas <- data.frame(
    id = as.integer(c(NA, NA, NA)),
    value = c(NA_real_, NA_real_, NA_real_),
    label = c(NA_character_, NA_character_, NA_character_)
  )

  write_parquet(df_valid, tmp)
  expect_snapshot(error = TRUE, {
    append_parquet(df_nas, tmp)
  })
})

test_that("overwrite a single row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df1 <- test_df(missing = TRUE)
  write_parquet(df1, tmp)
  mtd1 <- read_parquet_metadata(tmp)

  df2 <- test_df()
  append_parquet(df2, tmp)
  df3 <- read_parquet(tmp)
  mtd2 <- read_parquet_metadata(tmp)

  df21 <- utils::head(df3, nrow(df1))
  df22 <- utils::tail(df3, nrow(df2))
  row.names(df22) <- NULL
  expect_equal(df21, df1)
  expect_equal(df22, df2)
  expect_equal(mtd1$file_meta_data[-3], mtd2$file_meta_data[-3])
})

test_that("parse_encoding", {
  expect_snapshot({
    names(mtcars)
    parse_encoding(NULL, mtcars)
    parse_encoding("PLAIN", mtcars)
    parse_encoding(c(disp = "RLE_DICTIONARY"), mtcars)
    parse_encoding(c(disp = "RLE_DICTIONARY", vs = "PLAIN"), mtcars)
    parse_encoding(c(disp = "RLE", "PLAIN"), mtcars)
    parse_encoding(c(disp = "RLE", "PLAIN", vs = "PLAIN"), mtcars)
  })

  expect_snapshot(error = TRUE, {
    parse_encoding(1:2, mtcars)
    parse_encoding(c("PLAIN", "foobar"), mtcars)
    parse_encoding(c(foo = "PLAIN", foo = "RLE"), mtcars)
    parse_encoding(c(disp = "PLAIN", foo = "RLE"), mtcars)
  })
})

test_that("BOOLEAN", {
  do <- function(d) {
    test_write(d)
    test_write(d, encoding = "PLAIN")
    test_write(d, encoding = "RLE")
  }
  do(data.frame(l = c(TRUE, FALSE, TRUE)))
  do(data.frame(l = c(TRUE, FALSE, NA, TRUE)))
  do(data.frame(l = rep(TRUE, 16)))
  do(data.frame(l = c(rep(TRUE, 8), NA, rep(TRUE, 8))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  d <- data.frame(l = c(rep(TRUE, 8), NA, rep(TRUE, 8)))
  expect_snapshot(error = TRUE, {
    # not implemented
    write_parquet(d, tmp, encoding = "RLE_DICTIONARY")
    write_parquet(d, tmp, encoding = "BIT_PACKED")
    # invalid for BOOLEAN
    write_parquet(d, tmp, encoding = "BYTE_STREAM_SPLIT")
  })
})

test_that("BOOLEAN multi-page round-trip (issue #142)", {
  # With NANOPARQUET_PAGE_SIZE=1024 (1KB), a BOOLEAN column needs ~8192 rows per
  # page. Use 20000 rows to force 3 pages, exposing a bug where the second page
  # was decoded to the same memory offset as the first (psize was 0 for BOOLEAN).
  set.seed(42)
  n <- 20000L
  x <- sample(c(TRUE, FALSE, NA), n, replace = TRUE, prob = c(0.9, 0.05, 0.05))
  d <- data.frame(l = x)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  withr::with_envvar(c(NANOPARQUET_PAGE_SIZE = "1024"), {
    write_parquet(d, tmp, encoding = "RLE")
    d2 <- as.data.frame(read_parquet(tmp))
  })
  expect_equal(d, d2)

  withr::with_envvar(c(NANOPARQUET_PAGE_SIZE = "1024"), {
    write_parquet(d, tmp, encoding = "PLAIN")
    d2 <- as.data.frame(read_parquet(tmp))
  })
  expect_equal(d, d2)
})

test_that("INT32", {
  do <- function(d) {
    test_write(d)
    test_write(d, encoding = "PLAIN")
    test_write(d, encoding = "RLE_DICTIONARY")
  }
  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5)))
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT32
    write_parquet(d, tmp, encoding = "RLE")
  })
})

test_that("integer -> INT64", {
  schema <- "INT64"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = 1:5))
  do(data.frame(d = c(1:2, NA, 3:5)))
  do(data.frame(d = rep(1L, 10)))
  do(d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT64
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT64", {
  schema <- "INT64"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for INT64
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("integer -> INT96", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema("INT96")

  d <- data.frame(d = 1:5)
  expect_snapshot({
    write_parquet(d, tmp, schema = schema)
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
  })

  d <- data.frame(d = c(1:2, NA, 3:5))
  expect_snapshot({
    write_parquet(d, tmp, schema = schema)
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
  })

  d <- data.frame(d = rep(1L, 10))
  expect_snapshot({
    write_parquet(d, tmp, schema = schema)
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
  })

  d <- data.frame(d = c(rep(1L, 5), NA, rep(1L, 5)))
  expect_snapshot({
    write_parquet(d, tmp, schema = schema)
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]

    write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
  })

  expect_snapshot(error = TRUE, {
    # unsupported for INT96
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("double -> INT96", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- "INT96"

  # TODO: fix tests
  if (.Platform$OS.type == "windows" && getRversion() < "4.2.0") {
    skip("Needs INT96 read w/o converting to time")
  }

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = as.double(1:5)))
  do(data.frame(d = as.double(c(1:2, NA, 3:5))))
  do(data.frame(d = as.double(rep(1L, 10))))
  do(d <- data.frame(d = as.double(c(rep(1L, 5), NA, rep(1L, 5)))))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # unsupported for INT96
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("FLOAT", {
  schema <- "FLOAT"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = 1:5 / 2))
  do(data.frame(d = c(1:2 / 2, NA, 3:5 / 2)))
  do(data.frame(d = rep(1, 10) / 2))
  do(d <- data.frame(d = c(rep(1, 5) / 2, NA, rep(1, 5) / 2)))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for FLOAT
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("DOUBLE", {
  schema <- "DOUBLE"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = 1:5 / 2))
  do(data.frame(d = c(1:2 / 2, NA, 3:5 / 2)))
  do(data.frame(d = rep(1, 10) / 2))
  do(d <- data.frame(d = c(rep(1, 5) / 2, NA, rep(1, 5) / 2)))

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    # unsupported for DOUBLE
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("BYTE_ARRAY, string", {
  schema <- "STRING"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(s = c("foo", "bar", "foobar")))
  do(data.frame(s = c("foo", "bar", NA, "foobar")))
  do(data.frame(d = rep("foo", 10)))
  d <- data.frame(d = rep("foo", 10))
  d[["d"]][5] <- NA
  do(d)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("BYTE_ARRAY, RAW", {
  schema <- "BYTE_ARRAY"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
  }

  do(data.frame(s = I(lapply(c("foo", "bar", "foobar"), charToRaw))))
  do(data.frame(
    s = I(list(
      charToRaw("foo"),
      charToRaw("bar"),
      NULL,
      charToRaw("foobar")
    ))
  ))
  do(data.frame(d = I(lapply(rep("foo", 10), charToRaw))))
  d <- data.frame(d = I(lapply(rep("foo", 10), charToRaw)))
  d[["d"]][5] <- list(NULL)
  do(d)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("FIXED_LEN_BYTE_ARRAY, RAW", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- list("FIXED_LEN_BYTE_ARRAY", type_length = 3)

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
  }
  do(data.frame(s = I(lapply(c("foo", "bar", "aaa"), charToRaw))))
  d <- data.frame(s = I(lapply(c("foo", "bar", "aaa", "aaa"), charToRaw)))
  d[["s"]][3] <- list(NULL)
  do(d)
  do(data.frame(d = I(lapply(rep("foo", 10), charToRaw))))
  d <- data.frame(d = I(lapply(rep("foo", 10), charToRaw)))
  d[["d"]][5] <- list(NULL)
  do(d)

  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "PLAIN_DICTIONARY")
    # unsupported for FIXED_LEN_BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("FIXED_LEN_BYTE_ARRAY, FLOAT16", {
  schema <- "FLOAT16"

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(d = 1:5 / 2))
  d <- data.frame(d = 1:6 / 2)
  d[["d"]][3] <- NA
  do(d)
  do(data.frame(d = rep(1 / 2, 10)))
  d <- data.frame(d = rep(1 / 2, 10))
  d[["d"]][5] <- NA
  do(d)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for FIXED_LEN_BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("FIXED_LEN_BYTE_ARRAY, character", {
  schema <- list("FIXED_LEN_BYTE_ARRAY", type_length = 3)

  do <- function(d) {
    test_write(d, schema)
    test_write(d, schema, "PLAIN")
    test_write(d, schema, "RLE_DICTIONARY")
  }
  do(data.frame(s = c("foo", "bar", "aaa")))
  do(data.frame(s = c("foo", "bar", NA, "aaa")))
  do(data.frame(d = rep("foo", 10)))
  d <- data.frame(d = rep("foo", 10))
  d[["d"]][5] <- NA
  do(d)

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema(schema)
  expect_snapshot(error = TRUE, {
    # not implemented yet
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    # unsupported for FIXED_LEN_BYTE_ARRAY
    write_parquet(d, tmp, schema = schema, encoding = "RLE")
  })
})

test_that("polars can read temporal types", {
  skip_without_polars()

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  do <- function(df, path = tmp) {
    write_parquet(df, path)
    pyscript <- sprintf(
      r"[
import polars as pl
pl.read_parquet("%s")
    ]",
      normalizePath(path, winslash = "/")
    )
    pytmp <- tempfile(fileext = ".py")
    on.exit(unlink(pytmp), add = TRUE)
    writeLines(pyscript, pytmp)
    py <- if (Sys.which("python3") != "") "python3" else "python"
    processx::run(py, pytmp, stderr = "2>&1")
  }

  # Date
  df_date <- data.frame(x = Sys.Date())
  expect_silent(do(df_date))

  # hms, integer
  # it is unclear if this ever comes up in practice
  df_hmsi <- data.frame(
    x = structure(0L, units = "secs", class = c("hms", "difftime"))
  )
  expect_silent(do(df_hmsi))

  # hms, double
  df_hmsd <- data.frame(x = hms::hms(0))
  expect_silent(do(df_hmsd))

  # difftime
  df_difftime <- data.frame(x = as.difftime(1, units = "secs"))
  expect_silent(do(df_difftime))

  # POSIXct
  df_posixct <- data.frame(x = Sys.time())
  expect_silent(do(df_posixct))

  # factor
  df_factor <- data.frame(x = as.factor(c("a", "a")))
  expect_silent(do(df_factor))
})

read_arrow_schema <- function(file) {
  mtd <- read_parquet_metadata(file)
  kvm <- mtd[["file_meta_data"]][["key_value_metadata"]][[1]]
  if ("ARROW:schema" %in% kvm[["key"]]) {
    as <- kvm[["value"]][match("ARROW:schema", kvm[["key"]])]
    parse_arrow_schema(as)
  }
}

apply_arrow_schema <- function(
  tab,
  file,
  arrow_schema,
  dicts,
  types,
  col_select
) {
  if (is.na(arrow_schema)) {
    return(tab)
  }
  spec <- arrow_find_special(arrow_schema, file, col_select)
  for (idx in spec$factor) {
    clevels <- Reduce(union, dicts[[idx]])
    tab[[idx]] <- factor(tab[[idx]], levels = clevels)
  }
  for (idx in spec$difftime) {
    # only if INT64, otherwise hms, probably
    if (types[[idx]] != 2) {
      next
    }
    mult <- switch(
      spec$columns$type[[idx]]$unit,
      SECOND = 1,
      MILLISECOND = 1000,
      MICROSECOND = 1000 * 1000,
      NANOSECOND = 1000 * 1000 * 1000,
      stop("Unknown Arrow time unit")
    )
    tab[[idx]] <- as.difftime(tab[[idx]] / mult, units = "secs")
  }
  tab
}

arrow_find_special <- function(asch, file, col_select = NULL) {
  amd <- tryCatch(
    parse_arrow_schema(asch)$columns,
    error = function(e) {
      warning(
        sprintf(
          "Failed to parse Arrow schema from parquet file at '%s'",
          file
        ),
        call. = TRUE
      )
      NULL
    }
  )
  if (is.null(amd)) {
    return(list())
  }
  # Subset of columns?
  if (!is.null(col_select)) {
    amd <- amd[col_select, , drop = FALSE]
  }
  # If the type is Utf8 and it is a dictionary, then it is a factor
  fct <- which(
    amd$type_type == "Utf8" & !vapply(amd$dictionary, is.null, logical(1))
  )
  dft <- which(amd$type_type == "Duration")
  list(factor = fct, difftime = dft, columns = amd)
}

arrow_types <- c(
  NONE = 0L,
  Null = 1L,
  Int = 2L,
  FloatingPoint = 3L,
  Binary = 4L,
  Utf8 = 5L,
  Bool = 6L,
  Decimal = 7L,
  Date = 8L,
  Time = 9L,
  Timestamp = 10L,
  Interval = 11L,
  List = 12L,
  Struct_ = 13L,
  Union = 14L,
  FixedSizeBinary = 15L,
  FixedSizeList = 16L,
  Map = 17L,
  Duration = 18L,
  LargeBinary = 19L,
  LargeUtf8 = 20L,
  LargeList = 21L,
  RunEndEncoded = 22L,
  BinaryView = 23L,
  Utf8View = 24L,
  ListView = 25L,
  LargeListView = 26L
)

float_precision_names <- c(
  HALF = 0L,
  SINGLE = 1L,
  DOUBLE = 2L
)

date_unit_names <- c(
  DAY = 0L,
  MILLISECOND = 1L
)

time_unit_names <- c(
  SECOND = 0L,
  MILLISECOND = 1L,
  MICROSECOND = 2L,
  NANOSECOND = 3L
)

interval_unit_names <- c(
  YEAR_MONTH = 0L,
  DAY_TIME = 1L,
  MONTH_DAY_NANO = 2L
)

union_mode_names <- c(
  Sparse = 0L,
  Dense = 1L
)

endianness_names <- c(
  Little = 0L,
  Big = 1L
)

features_names <- c(
  UNUSED = 0L,
  DICTIONARY_REPLACEMENT = 1L,
  COMPRESSED_BODY = 2L
)

dict_kind_names <- c(
  "DenseArray" = 0L
)

parse_arrow_schema <- function(schema) {
  ret <- .Call(nanoparquet_parse_arrow_schema, schema)

  columns <- ret[[1]]
  columns$type_type <- names(arrow_types)[columns$type_type + 1L]
  columns$type <- I(columns$type)
  columns$dictionary <- I(columns$dictionary)
  columns$custom_metadata <- I(columns$custom_metadata)
  columns <- as.data.frame(columns)
  class(columns) <- c("tbl", class(columns))

  kv <- ret[[2]]
  kv <- as.data.frame(kv)
  class(kv) <- c("tbl", class(kv))

  endianness <- ret[[3]]
  endianness <- names(endianness_names)[endianness + 1L]

  features <- ret[[4]]
  features <- names(features_names)[features + 1L]

  list(
    columns = columns,
    custom_metadata = kv,
    endianness = endianness,
    features = features
  )
}

# factors -> Utf8 with dictionary
# INTSXP -> Int(32, true)
# REALSXP -> FloatingPoint(DOUBLE)
# STRSXP -> Utf8
# LGLSXP -> Bool

encode_arrow_schema_r <- function(df, schema) {
  endianness <- capitalize(.Platform$endian)
  dates <- vapply(df, function(c) inherits(c, "Date"), logical(1))
  hmss <- vapply(df, function(c) inherits(c, "hms"), logical(1))
  psxcts <- vapply(df, function(c) inherits(c, "POSIXct"), logical(1))
  fctrs <- vapply(df, function(c) inherits(c, "factor"), logical(1))
  int64s <- vapply(df, function(c) inherits(c, "integer64"), logical(1))
  dfts <- vapply(
    df,
    function(c) !inherits(c, "hms") && inherits(c, "difftime"),
    logical(1)
  )
  typemap <- c(
    "integer" = "Int",
    "double" = "FloatingPoint",
    "character" = "Utf8",
    "logical" = "Bool",
    "list" = "Binary"
  )
  dftypes <- vapply(df, typeof, character(1))
  artypes <- typemap[dftypes]
  artypes[dates] <- "Date"
  artypes[hmss] <- "Time"
  artypes[psxcts] <- "Timestamp"
  artypes[fctrs] <- "Utf8"
  artypes[int64s] <- "Int"
  artypes[dfts] <- "Duration"
  if (anyNA(artypes)) {
    stop(
      "Unsuppoted types when writing Parquet file: ",
      paste(unique(dftypes[is.na(artypes)]), collapse = ", ")
    )
  }
  typeattr <- list(
    "Int" = list(bit_width = 32L, is_signed = TRUE),
    "FloatingPoint" = list(precision = "DOUBLE"),
    "Utf8" = NULL,
    "Bool" = NULL,
    "Duration" = list(unit = "NANOSECOND"),
    "Time" = list(unit = "SECOND", bit_width = 32L),
    "Date" = list(unit = "DAY"),
    "Timestamp" = list(unit = "MICROSECOND", timezone = "UTC")
  )
  schema <- list(
    columns = data.frame(
      stringsAsFactors = FALSE,
      name = names(df),
      type_type = artypes,
      type = I(unname(typeattr[artypes])),
      nullable = rep(TRUE, ncol(df)),
      dictionary = I(replicate(ncol(df), NULL, simplify = FALSE)),
      custom_metadata = I(replicate(ncol(df), list(), simplify = FALSE))
    ),
    custom_metadata = list(NULL, NULL),
    endianness = endianness,
    features = character()
  )
  for (idx in which(int64s)) {
    schema$columns$type[[idx]] <- list(bit_width = 64L, is_signed = TRUE)
  }
  for (idx in which(fctrs)) {
    schema$columns$dictionary[[idx]] <- list(
      id = 0,
      index_type = list(bit_width = factor_bits(df[[idx]]), is_signed = TRUE),
      is_ordered = is.ordered(df[[idx]]),
      dictionary_kind = "DenseArray"
    )
  }
  schema
}

# Replace strings with numeric IDs, so we can use them in C++
fill_arrow_schema_enums_type <- function(type_type, type) {
  switch(
    type_type,
    "FloatingPoint" = {
      type$precision <- float_precision_names[type$precision]
    },
    "Date" = {
      type$unit <- date_unit_names[type$unit]
    },
    "Time" = ,
    "Timestamp" = ,
    "Duration" = {
      type$unit <- time_unit_names[type$unit]
    },
    "Interval" = {
      type$unit <- interval_unit_names[type$unit]
    },
    "Union" = {
      type$mode <- union_mode_names[type$mode]
    },
    # otherwise nothing to do
    NULL
  )

  type
}

fill_arrow_schema_enums_dict <- function(dict) {
  if (!is.null(dict)) {
    dict$dictionary_kind <- dict_kind_names[dict$dictionary_kind]
  }
  dict
}

fill_arrow_schema_enums <- function(schema) {
  schema$columns$type[] <- mapply(
    schema$columns$type_type,
    schema$columns$type,
    FUN = fill_arrow_schema_enums_type,
    SIMPLIFY = FALSE,
    USE.NAMES = FALSE
  )
  schema$columns$type_type <- arrow_types[schema$columns$type_type]
  schema$columns$dictionary[] <- lapply(
    schema$columns$dictionary,
    fill_arrow_schema_enums_dict
  )
  schema$endianness <- endianness_names[schema$endianness]
  schema$features <- features_names[schema$features]
  schema
}

encode_arrow_schema <- function(df) {
  schema <- encode_arrow_schema_r(df)
  schema <- fill_arrow_schema_enums(schema)
  rawenc <- .Call(nanoparquet_encode_arrow_schema, schema)
  enc <- base64_encode(rawenc)
  enc
}

# how many bits do we need for this factor in a signed integer
# Arrow only supports 8, 16, 32 and 64.
factor_bits <- function(x) {
  l <- length(levels(x))
  if (l < 2^(8 - 1)) {
    8L
  } else if (l < 2^(16 - 1)) {
    16L
  } else if (l < 2^(32 - 1)) {
    32L
  } else {
    64L
  }
}

base64_decode <- function(x) {
  .Call(nanoparquet_base64_decode, x)
}

base64_encode <- function(x) {
  .Call(nanoparquet_base64_encode, x)
}

alltypes_plain <- structure(
  list(
    id = c(4L, 5L, 6L, 7L, 2L, 3L, 0L, 1L),
    bool_col = c(TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, FALSE, FALSE),
    tinyint_col = c(0L, 1L, 0L, 1L, 0L, 1L, 0L, 1L),
    smallint_col = c(0L, 1L, 0L, 1L, 0L, 1L, 0L, 1L),
    int_col = c(0L, 1L, 0L, 1L, 0L, 1L, 0L, 1L),
    bigint_col = c(0, 10, 0, 10, 0, 10, 0, 10),
    float_col = c(
      0,
      1.10000002384186,
      0,
      1.10000002384186,
      0,
      1.10000002384186,
      0,
      1.10000002384186
    ),
    double_col = c(0, 10.1, 0, 10.1, 0, 10.1, 0, 10.1),
    date_string_col = c(
      "03/01/09",
      "03/01/09",
      "04/01/09",
      "04/01/09",
      "02/01/09",
      "02/01/09",
      "01/01/09",
      "01/01/09"
    ),
    string_col = c("0", "1", "0", "1", "0", "1", "0", "1"),
    timestamp_col = structure(
      c(
        1235865600,
        1235865660,
        1238544000,
        1238544060,
        1233446400,
        1233446460,
        1230768000,
        1230768060
      ),
      class = c("POSIXct", "POSIXt"),
      tzone = "UTC"
    )
  ),
  row.names = c(NA, -8L),
  class = "data.frame"
)

alltypes_plain_snappy <- structure(
  list(
    id = 6:7,
    bool_col = c(TRUE, TRUE),
    tinyint_col = 0:1,
    smallint_col = 0:1,
    int_col = 0:1,
    bigint_col = c(0, 10),
    float_col = c(0, 1.10000002384186),
    double_col = c(0, 10.1),
    date_string_col = c("04/01/09", "04/01/09"),
    string_col = c("0", "1"),
    timestamp_col = structure(
      c(1238544000, 1238544060),
      class = c("POSIXct", "POSIXt"),
      tzone = "UTC"
    )
  ),
  row.names = c(NA, -2L),
  class = "data.frame"
)

data_comparable <- function(df1, df2, dlt = .0001) {
  df1 <- as.data.frame(df1, stringsAsFactors = F)
  df2 <- as.data.frame(df2, stringsAsFactors = F)
  if (!identical(dim(df1), dim(df2))) {
    return(FALSE)
  }
  for (col_i in length(df1)) {
    col1 <- df1[[col_i]]
    col2 <- df2[[col_i]]
    if (is.numeric(col1)) {
      # reference answers are rounded to two decimals
      col1 <- round(col1, 2)
      col2 <- round(col2, 2)
      if (any(abs(col1 - col2) > col1 * dlt)) {
        return(FALSE)
      }
    } else {
      col1 <- trimws(as.character(col1))
      col2 <- trimws(as.character(col2))
      if (any(col1 != col2)) {
        return(FALSE)
      }
    }
  }
  return(TRUE)
}

test_that("various error cases", {
  # https://github.com/llvm/llvm-project/issues/59432
  if (is_asan()) {
    skip("ASAN bug")
  }
  expect_error(res <- read_parquet(""))
  expect_error(res <- read_parquet("DONTEXIST"))
  tf <- tempfile()
  expect_error(res <- read_parquet(tf))
  expect_error(res <- read_parquet(c(tf, tf)))
})

test_that("basic reading works", {
  res <- read_parquet(test_path("data/alltypes_plain.parquet"))
  expect_true(data_comparable(alltypes_plain, res))
})

test_that("basic reading works with snappy", {
  res <- read_parquet(test_path("data/alltypes_plain.snappy.parquet"))
  expect_true(data_comparable(alltypes_plain_snappy, res))
})

test_that("read factors, marked by Arrow", {
  res <- read_parquet(test_path("data/factor.parquet"))
  expect_snapshot({
    as.data.frame(res[1:5, ])
    sapply(res, class)
  })
})

test_that("Can't parse Arrow schema", {
  expect_snapshot(
    arrow_find_special(base64_encode("foobar"), "myfile")
  )
})

test_that("round trip with arrow", {
  # Don't want to skip on the parquet capability missing, because then
  # this might not be tested on the CI. So rather we skip on CRAN.
  skip_on_cran()
  skip_without("arrow")
  mt <- test_df(factor = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  arrow::write_parquet(mt, tmp, compression = "uncompressed")
  expect_equal(read_parquet(tmp), mt)
  unlink(tmp)

  arrow::write_parquet(mt, tmp, compression = "snappy")
  expect_equal(read_parquet(tmp), mt)
})

test_that("round trip with arrow, empty data frame", {
  # https://github.com/r-lib/nanoparquet/issues/160
  skip_on_cran()
  skip_without("arrow")
  df <- data.frame(
    a = character(0),
    b = numeric(0),
    c = logical(0)
  )
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  arrow::write_parquet(df, tmp)
  result <- read_parquet(tmp)
  expect_equal(nrow(result), 0L)
  expect_equal(ncol(result), 3L)
  expect_equal(names(result), c("a", "b", "c"))
})

test_that("round trip with duckdb", {
  skip_on_cran()
  skip_without("duckdb")
  skip_without("arrow")
  # https://github.com/llvm/llvm-project/issues/59432
  if (is_asan()) {
    skip("ASAN bug")
  }
  mt <- test_df()
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  drv <- duckdb::duckdb()
  con <- DBI::dbConnect(drv)
  on.exit(DBI::dbDisconnect(con), add = TRUE)
  DBI::dbWriteTable(con, "mtcars", as.data.frame(mt))

  DBI::dbExecute(
    con,
    DBI::sqlInterpolate(
      con,
      "COPY mtcars TO ?filename (FORMAT 'parquet', COMPRESSION 'uncompressed')",
      filename = tmp
    )
  )
  expect_equal(read_parquet(tmp), mt)
  unlink(tmp)

  DBI::dbExecute(
    con,
    DBI::sqlInterpolate(
      con,
      "COPY mtcars TO ?filename (FORMAT PARQUET, COMPRESSION 'snappy')",
      filename = tmp
    )
  )
  arrow::write_parquet(mt, tmp, compression = "snappy")
  expect_equal(read_parquet(tmp), mt)
})

test_that("read Date", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    d = c(Sys.Date() - 1, Sys.Date(), Sys.Date() + 1)
  )
  write_parquet(d, tmp)

  d2 <- read_parquet(tmp)
  expect_s3_class(d2$d, "Date")
  expect_equal(d$d, d2$d)
})

test_that("read hms", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    h = hms::hms(1, 2, 3)
  )
  write_parquet(d, tmp)

  d2 <- read_parquet(tmp)
  expect_s3_class(d2$h, "hms")
  expect_equal(d2$h, d$h)
})

test_that("read hms in MICROS", {
  pf <- test_path("data/timetz.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("read POSIXct", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    h = .POSIXct(Sys.time(), tz = "UTC")
  )
  write_parquet(d, tmp)

  d2 <- read_parquet(tmp)
  expect_s3_class(d$h, "POSIXct")
  expect_equal(d$h, d2$h)
})

test_that("read POSIXct in MILLIS", {
  skip_on_cran()
  skip_without("arrow")
  # This file has UTC = FALSE, so the exact result depends on the current
  # time zone. But it should match Arrow.
  pf <- test_path("data/timestamp-ms.parquet")
  d1 <- read_parquet(pf)
  d2 <- arrow::read_parquet(pf)
  expect_equal(
    as.data.frame(d1),
    as.data.frame(d2)
  )
})

test_that("read difftime", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # Fractional seconds are kept
  d <- data.frame(
    h = as.difftime(10 + 1 / 9, units = "secs")
  )
  write_parquet(d, tmp)

  d2 <- read_parquet(tmp)
  expect_s3_class(d2$h, "difftime")
  expect_equal(d$h, d2$h)

  # Other units are converted to secs
  d <- data.frame(
    h = as.difftime(10, units = "mins")
  )
  write_parquet(d, tmp)
  d2 <- read_parquet(tmp)
  expect_snapshot({
    as.data.frame(d2)
  })
})

test_that("RLE BOOLEAN", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    l = c(
      logical(30),
      !logical(5),
      logical(20),
      !logical(30)
    )
  )

  write_parquet(d, tmp)
  expect_equal(
    unclass(read_parquet_metadata(tmp)$column_chunks$encodings),
    list("RLE")
  )

  expect_equal(as.data.frame(read_parquet(tmp)), d)

  # larger DF

  d <- data.frame(
    l = c(
      logical(runif(1) * 3000),
      !logical(runif(1) * 50),
      logical(runif(1) * 2000),
      !logical(runif(1) * 3000)
    )
  )

  write_parquet(d, tmp)
  expect_equal(
    unclass(read_parquet_metadata(tmp)$column_chunks$encodings),
    list("RLE")
  )

  expect_equal(as.data.frame(read_parquet(tmp)), d)
})

test_that("read GZIP compressed files", {
  pf <- test_path("data/gzip.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("V2 data pages", {
  pf <- test_path("data/parquet_go.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("V2 data page with missing values", {
  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/duckdb-bug1589.parquet")
  expect_equal(
    as.data.frame(read_parquet(pf)),
    as.data.frame(arrow::read_parquet(pf))
  )
})

test_that("Tricky V2 data page", {
  # has repetition levels to be ignored and uncompressed
  # definition levels
  pf <- test_path("data/rle_boolean_encoding.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("zstd", {
  pf <- test_path("data/zstd.parquet")
  expect_true(all(read_parquet_metadata(pf)$column_chunks$codec == "ZSTD"))
  pf2 <- test_path("data/gzip.parquet")
  expect_equal(read_parquet(pf), read_parquet(pf2))
})

test_that("zstd with data page v2", {
  pf <- test_path("data/zstd-v2.parquet")
  expect_true(all(read_parquet_metadata(pf)$column_chunks$codec == "ZSTD"))
  expect_true(
    all(
      read_parquet_pages(pf)$page_type %in% c("DICTIONARY_PAGE", "DATA_PAGE_V2")
    )
  )
  pf2 <- test_path("data/gzip.parquet")
  expect_equal(read_parquet(pf), read_parquet(pf2))
})

test_that("DELTA_BIANRY_PACKED encoding", {
  suppressPackageStartupMessages(library(bit64))
  pf <- test_path("data/dbp-int32.parquet")
  expect_snapshot({
    read_parquet_metadata(pf)$column_chunks$encodings
    read_parquet(pf)
  })

  pf2 <- test_path("data/dbp-int32-missing.parquet")
  expect_snapshot({
    read_parquet_metadata(pf2)$column_chunks$encodings
    read_parquet(pf2)
  })

  pf3 <- test_path("data/dbp-int64.parquet")
  expect_snapshot({
    read_parquet_metadata(pf3)$column_chunks$encodings
    read_parquet(pf3)
  })
})

test_that("UUID columns", {
  pf <- test_path("data/uuid-arrow.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("DELTA_LENGTH_BYTE_ARRAY encoding", {
  pf <- test_path("data/delta_length_byte_array.parquet")
  dlba <- read_parquet(pf)
  expect_snapshot({
    as.data.frame(dlba)[1:10, ]
    rle(nchar(dlba$FRUIT))
  })
})

test_that("DELTA_BYTE_ARRAY encoding", {
  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/delta_byte_array.parquet")
  dba <- read_parquet(pf)
  expect_snapshot({
    as.data.frame(dba)[1:5, ]
  })
  expect_equal(
    as.data.frame(arrow::read_parquet(pf)),
    as.data.frame(dba)
  )
})

test_that("BYTE_STREAM_SPLIT encoding", {
  skip_on_cran()
  skip_without("arrow")
  if (getRversion() < "4.2.0" && .Platform$OS.type == "windows") {
    skip("arrow fails here")
  }
  pf <- test_path("data/byte_stream_split.parquet")
  bss <- read_parquet(pf)
  expect_snapshot({
    as.data.frame(bss)[1:5, ]
  })
  expect_equal(
    as.data.frame(arrow::read_parquet(pf)),
    as.data.frame(bss)
  )
})

test_that("More BYTE_STREAM_SPLIT", {
  skip_on_cran()
  pf <- test_path("data/byte_stream_split_extended.gzip.parquet")
  bss <- read_parquet(pf)
  expect_snapshot({
    as.data.frame(bss)[1:5, ]
  })
  for (i in 1:7) {
    expect_equal(bss[[2 * i - 1]], bss[[2 * i]])
  }
})

test_that("DECIMAL in FIXED_LEN_BYTE_ARRAY (128-bit)", {
  # 16-byte / 128-bit decimals; previously only low 64 bits were read (#148)
  pf <- test_path("data/decimal128.parquet")
  df <- read_parquet(pf)
  expect_equal(typeof(df$Value), "double")
  expect_equal(typeof(df$DataCapture), "double")
  expect_equal(
    df$Value,
    c(51.70559, 49.52529, 51.69562, 49.90946, 50.06762, 54.69116)
  )
  expect_equal(df$DataCapture, rep(-99, 6))
})

test_that("DECIMAL in INT32, INT64", {
  pf <- test_path("data/int32_decimal.parquet")
  expect_equal(typeof(read_parquet(pf)[[1]]), "double")
  expect_snapshot(as.data.frame(read_parquet(pf)))

  pf <- test_path("data/int64_decimal.parquet")
  expect_equal(typeof(read_parquet(pf)[[1]]), "double")
  expect_snapshot(as.data.frame(read_parquet(pf)))
})

test_that("FLOAT16", {
  pf <- test_path("data/float16_nonzeros_and_nans.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf))
    as.data.frame(read_parquet(pf))
  })
})

test_that("base64", {
  x <- serialize(mtcars, connection = NULL)
  x64 <- base64_encode(x)
  expect_equal(base64_decode(x64), x)
  expect_equal(base64_decode(charToRaw(x64)), x)

  # spaces are allowed
  x64s <- paste0(
    substr(x64, 1, 1000),
    "\n ",
    substr(x64, 1001, nchar(x64))
  )
  expect_equal(base64_decode(x64s), x)

  # strings are ok as well
  expect_equal(base64_decode(base64_encode("foo")), charToRaw("foo"))

  # errors, the `expect_error()`s are included because covr dor not work
  # correctly for `expect_snapshot()`, it seems.
  expect_error(base64_encode(1:10))
  expect_error(base64_decode(1:10))
  expect_error(base64_decode("foobar;;;"))

  expect_snapshot(error = TRUE, base64_encode(1:10))
  expect_snapshot(error = TRUE, base64_decode(1:10))
  expect_snapshot(error = TRUE, base64_decode("foobar;;;"))
})

test_that("factor_bits", {
  fct <- as.factor(letters)
  mockery::stub(factor_bits, "length", 100)
  expect_equal(factor_bits(fct), 8L)
  mockery::stub(factor_bits, "length", 127)
  expect_equal(factor_bits(fct), 8L)
  mockery::stub(factor_bits, "length", 128)
  expect_equal(factor_bits(fct), 16L)
  mockery::stub(factor_bits, "length", 32767)
  expect_equal(factor_bits(fct), 16L)
  mockery::stub(factor_bits, "length", 32768)
  expect_equal(factor_bits(fct), 32L)
  mockery::stub(factor_bits, "length", 2147483647)
  expect_equal(factor_bits(fct), 32L)
  mockery::stub(factor_bits, "length", 2147483648)
  expect_equal(factor_bits(fct), 64L)
})

test_that("temporal types", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  withr::local_options(nanoparquet.write_arrow_metadata = TRUE)

  # Date
  df_date <- data.frame(x = Sys.Date())
  write_parquet(df_date, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })

  # hms, integer
  # it is unclear if this ever comes up in practice
  df_hmsi <- data.frame(
    x = structure(0L, units = "secs", class = c("hms", "difftime"))
  )
  write_parquet(df_hmsi, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })

  # hms, double
  df_hmsd <- data.frame(x = hms::hms(0))
  write_parquet(df_hmsd, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })

  # difftime
  df_difftime <- data.frame(x = as.difftime(1, units = "secs"))
  write_parquet(df_difftime, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })

  # POSIXct
  df_posixct <- data.frame(x = Sys.time())
  write_parquet(df_posixct, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })

  # factor
  df_factor <- data.frame(x = as.factor(c("a", "a")))
  write_parquet(df_factor, tmp)
  np <- read_arrow_schema(tmp)
  expect_snapshot({
    as.data.frame(np[["columns"]])
    np[["columns"]][["type"]]
  })
})

test_that("arrow-rs can read ARROW:schema (issue #152)", {
  skip_on_cran()
  skip_without_cargo()

  crate <- normalizePath(
    testthat::test_path("fixtures/arrow-rs-reader"),
    mustWork = TRUE
  )
  processx::run(
    "cargo",
    c("build", "--manifest-path", file.path(crate, "Cargo.toml")),
    error_on_status = TRUE
  )
  bin <- file.path(crate, "target", "debug", "arrow-rs-reader")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  withr::local_options(nanoparquet.write_arrow_metadata = TRUE)

  write_parquet(data.frame(a = c(1, 2), b = c(3, 4)), tmp)
  result <- processx::run(bin, tmp, error_on_status = TRUE)
  expect_equal(trimws(result$stdout), "OK")
})

#' Metadata of all pages of a Parquet file
#'
#' @details
#' Reading all the page headers might be slow for large files, especially
#' if the file has many small pages.
#'
#' @param file Path to a Parquet file.
#' @return Data frame with columns:
#'   * `file_name`: file name.
#'   * `row_group`: id of the row group the page belongs to,
#'     an integer between 0 and the number of row groups
#'     minus one.
#'   * `column`: id of the column. An integer between the
#'     number of leaf columns minus one. Note that only leaf
#'     columns are considered, as non-leaf columns do not
#'     have any pages.
#'   * `page_type`: `DATA_PAGE`, `INDEX_PAGE`, `DICTIONARY_PAGE` or
#'     `DATA_PAGE_V2`.
#'   * `page_header_offset`: offset of the data page (its header) in the
#'     file.
#'   * `uncompressed_page_size`: does not include the page header, as per
#'     Parquet spec.
#'   * `compressed_page_size`: without the page header.
#'   * `crc`: integer, checksum, if present in the file, can be `NA`.
#'   * `num_values`: number of data values in this page, include
#'     `NULL` (`NA` in R) values.
#'   * `encoding`: encoding of the page, current possible encodings:
#'     `r paste0('"', names(encodings), '"', collapse = ", ")`.
#'   * `definition_level_encoding`: encoding of the definition levels,
#'     see `encoding` for possible values. This can be missing in V2 data
#'     pages, where they are always RLE encoded.
#'   * `repetition_level_encoding`: encoding of the repetition levels,
#'     see `encoding` for possible values. This can be missing in V2 data
#'     pages, where they are always RLE encoded.
#'   * `data_offset`: offset of the actual data in the file.
#'   * `page_header_length`: size of the page header, in bytes.
#'
#' @keywords internal
#' @seealso [read_parquet_page()] to read a page.
#' @examples
#' file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
#' nanoparquet:::read_parquet_pages(file_name)

read_parquet_pages <- function(file) {
  file <- path.expand(file)
  res <- .Call(nanoparquet_read_pages, file)
  res$encoding <- names(encodings)[res$encoding + 1L]
  res$definition_level_encoding <-
    names(encodings)[res$definition_level_encoding + 1L]
  res$repetition_level_encoding <-
    names(encodings)[res$repetition_level_encoding + 1L]
  res$page_type <- names(page_types)[res$page_type + 1L]
  res <- as.data.frame(res)
  class(res) <- c("tbl", class(res))
  res
}

#' Read a page from a Parquet file
#'
#' @param file Path to a Parquet file.
#' @param offset Integer offset of the start of the page in the file.
#'   See [read_parquet_pages()] for a list of all pages and their offsets.
#' @return Named list. Many entries correspond to the columns of
#'   the result of [read_parquet_pages()]. Additional entries are:
#'   * `codec`: compression codec. Possible values:
#'   * `has_repetition_levels`: whether the page has repetition levels.
#'   * `has_definition_levels`: whether the page has definition levels.
#'   * `schema_column`: which schema column the page corresponds to. Note
#'     that only leaf columns have pages.
#'   * `data_type`: low level Parquet data type. Possible values:
#'   * `repetition_type`: whether the column the page belongs to is
#'     `REQUIRED`, `OPTIONAL` or `REPEATED`.
#'   * `page_header`: the bytes of the page header in a raw vector.
#'   * `num_null`: number of missing (`NA`) values. Only set in V2 data
#'     pages.
#'   * `num_rows`: this is the same as `num_values` for flat tables, i.e.
#'     files without repetition levels.
#'   * `compressed_data`: the data of the page in a raw vector. It includes
#'     repetition and definition levels, if any.
#'   * `data`: the uncompressed data, if nanoparquet supports the
#'     compression codec of the file (GZIP and SNAPPY at the time of
#'     writing), or if the file is not compressed. In the latter case it
#'     is the same as `compressed_data`.
#'
#' @keywords internal
#' @seealso [read_parquet_pages()] for a summary of all pages.
#' @examplesIf Sys.getenv("IN_PKGDOWN") == "true"
#' file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
#' nanoparquet:::read_parquet_pages(file_name)
#' options(max.print = 100)  # otherwise long raw vector
#' nanoparquet:::read_parquet_page(file_name, 4L)

read_parquet_page <- function(file, offset) {
  file <- path.expand(file)
  res <- .Call(nanoparquet_read_page, file, as.double(offset))
  res$page_type <- names(page_types)[res$page_type + 1L]
  res$codec <- names(codecs)[res$codec + 1L]
  res$encoding <- names(encodings)[res$encoding + 1L]
  res$definition_level_encoding <-
    names(encodings)[res$definition_level_encoding + 1L]
  res$repetition_level_encoding <-
    names(encodings)[res$repetition_level_encoding + 1L]
  res$data_type <- names(type_names)[res$data_type + 1L]
  res$repetition_type <- names(repetition_types)[res$repetition_type + 1L]
  res$compressed_data <- res$data
  copy <- 0L
  if (res$page_type == "DATA_PAGE_V2") {
    if (!is.na(res$repetition_levels_byte_length)) {
      copy <- copy + res$repetition_levels_byte_length
    }
    if (!is.na(res$definition_levels_byte_length)) {
      copy <- copy + res$definition_levels_byte_length
    }
  }
  if (res$codec == "SNAPPY") {
    res$data <- c(
      if (copy > 0) res$data[1:copy],
      snappy_uncompress(res$data[(copy + 1L):length(res$data)])
    )
  } else if (res$codec == "GZIP") {
    res$compressed_data <- res$data
    res$data <- c(
      if (copy > 0) res$data[1:copy],
      gzip_uncompress(
        res$data[(copy + 1L):length(res$data)],
        res$uncompressed_page_size - copy
      )
    )
  } else if (res$codec == "ZSTD") {
    res$compressed_data <- res$data
    res$data <- c(
      if (copy > 0) res$data[1:copy],
      zstd_uncompress(
        res$data[(copy + 1L):length(res$data)],
        res$uncompressed_page_size - copy
      )
    )
  } else if (res$codec == "UNCOMPRESSED") {
    # keep data
  } else {
    res$data <- NULL
  }

  res
}

snappy_compress <- function(buffer) {
  .Call(snappy_compress_raw, buffer)
}

snappy_uncompress <- function(buffer) {
  .Call(snappy_uncompress_raw, buffer)
}

gzip_compress <- function(buffer) {
  .Call(gzip_compress_raw, buffer)
}

gzip_uncompress <- function(buffer, uncompressed_length) {
  .Call(gzip_uncompress_raw, buffer, uncompressed_length)
}

zstd_compress <- function(buffer) {
  .Call(zstd_compress_raw, buffer)
}

zstd_uncompress <- function(buffer, uncompressed_length) {
  .Call(zstd_uncompress_raw, buffer, uncompressed_length)
}

#' RLE encode integers
#'
#' @param x Integer vector.
#' @return Raw vector, the encoded integers. It has two attributes:
#'   * `bit_length`: the number of bits needed to encode the input, and
#'   * `length`: length of the original integer input.
#'
#' @keywords internal
#' @seealso [rle_decode_int()]
#' @family encodings

rle_encode_int <- function(x) {
  bw <- if (length(x)) {
    max(as.integer(ceiling(log2(max(x) + 1L))), 1L)
  } else {
    1L
  }
  res <- .Call(nanoparquet_rle_encode_int, x, bw)
  attr(res, "bit_width") <- bw
  attr(res, "length") <- length(x)
  res
}

#' RLE decode integers
#'
#' @param x Raw vector of the encoded integers.
#' @param bit_width Bit width used for the encoding.
#' @param length Length of the output. If `NA` then we assume that `x`
#'   starts with length of the output, encoded as a 4 byte integer.
#' @return The decoded integer vector.
#'
#' @keywords internal
#' @seealso [rle_encode_int()]
#' @family encodings

rle_decode_int <- function(
  x,
  bit_width = attr(x, "bit_width"),
  length = attr(x, "length") %||% NA
) {
  .Call(nanoparquet_rle_decode_int, x, bit_width, is.na(length), length)
}

# dbp_encode_int <- function(x) {
# 	.Call(nanoparquet_dbp_encode_int32, x)
# }

dbp_decode_int <- function(x) {
  .Call(nanoparquet_dbp_decode_int32, x)
}

# dbp_encode_int64 <- function(x) {
# 	.Call(nanoparquet_dbp_encode_int64, x)
# }

dbp_decode_int64 <- function(x) {
  .Call(nanoparquet_dbp_decode_int64, x)
}

unpack_bits <- function(x, bit_width, n) {
  .Call(nanoparquet_unpack_bits_int32, x, bit_width, n)
}

pack_bits <- function(x, bit_width = NULL) {
  bit_width <- bit_width %||%
    if (length(x)) {
      max(as.integer(ceiling(log2(max(x) + 1L))), 1L)
    } else {
      0L
    }
  .Call(nanoparquet_pack_bits_int32, x, bit_width)
}

dict_encode <- function(x, n = length(x)) {
  .Call(nanoparquet_create_dict, x, n)
}

dict_encode_idx <- function(x, from = 1L, until = length(x) + 1L) {
  .Call(nanoparquet_create_dict_idx, x, from - 1L, until - 1L, sys.call())
}

lgl_avg_run_length <- function(x, n = length(x)) {
  .Call(nanoparquet_avg_run_length, x, n)
}

test_that("factors are written as strings", {
  mt <- test_df(factor = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mt, tmp)
  expect_snapshot(
    as.data.frame(read_parquet_schema(tmp))[
      c("name", "type", "converted_type", "logical_type")
    ]
  )
})

test_that("round trip", {
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  mt <- test_df()
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mt, tmp, compression = "uncompressed")
  expect_true(all(read_parquet(tmp) == mt))
  unlink(tmp)

  write_parquet(mt, tmp, compression = "snappy")
  expect_true(all(read_parquet(tmp) == mt))
})

test_that("round trip with arrow", {
  # Don't want to skip on the parquet capability missing, because then
  # this might not be tested on the CI. So rather we skip on CRAN.
  skip_on_cran()
  skip_without("arrow")
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  mt <- test_df(tibble = TRUE, factor = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mt, tmp, compression = "uncompressed")
  mt2 <- arrow::read_parquet(tmp)
  expect_equal(mt2, mt)
  unlink(tmp)

  write_parquet(mt, tmp, compression = "snappy")
  mt2 <- arrow::read_parquet(tmp)
  expect_equal(mt2, mt)
})

test_that("round trip with duckdb", {
  skip_without("duckdb")
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  mt <- as.data.frame(test_df())
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mt, tmp, compression = "uncompressed")
  df <- duckdb::sql_query(sprintf("FROM '%s'", tmp))
  expect_equal(df, mt)
  unlink(tmp)

  write_parquet(mt, tmp, compression = "snappy")
  df <- duckdb::sql_query(sprintf("FROM '%s'", tmp))
  expect_equal(df, mt)
})

test_that("round trip with pandas/pyarrow", {
  skip_on_cran()
  skip_without_pyarrow()
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  mt <- test_df()
  tmp1 <- tempfile(fileext = ".parquet")
  tmp2 <- tempfile(fileext = ".parquet")
  on.exit(unlink(c(tmp1, tmp2)), add = TRUE)
  # need to create to be able to call normalizePath()
  file.create(tmp1)
  file.create(tmp2)

  py_read <- function(input, output) {
    pyscript <- sprintf(
      r"[
import pyarrow
import pandas
pandas.set_option("display.width", 150)
pandas.set_option("display.max_columns", 150)
pandas.set_option("display.max_colwidth", 150)
df = pandas.read_parquet("%s", engine = "pyarrow")
print(df)
print(df.dtypes)
df.to_parquet("%s", engine = "pyarrow")
]",
      normalizePath(input, winslash = "/"),
      normalizePath(output, winslash = "/")
    )
    pytmp <- tempfile(fileext = ".py")
    on.exit(unlink(pytmp), add = TRUE)
    writeLines(pyscript, pytmp)
    py <- if (Sys.which("python3") != "") "python3" else "python"
    processx::run(py, pytmp, stderr = "2>&1", error_on_status = FALSE)
  }

  write_parquet(mt, tmp1, compression = "uncompressed")
  pyout <- py_read(tmp1, tmp2)
  if (pyout$status != 0) {
    stop("Failed to run Python, are all packages installed?")
  }
  expect_snapshot(writeLines(pyout$stdout))

  mt2 <- read_parquet(tmp2)
  expect_equal(mt2, mt)
})

test_that("errors", {
  # https://github.com/llvm/llvm-project/issues/59432
  if (is_asan()) {
    skip("ASAN bug")
  }

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  mt <- mt2 <- test_df(factor = TRUE)
  mt$list <- I(replicate(nrow(mt), list(1:4), simplify = FALSE))

  expect_error(write_parquet(mt, tmp))
  expect_snapshot(error = TRUE, write_parquet(mt, tmp))

  expect_error(write_parquet(mt, 1:10))
  expect_snapshot(error = TRUE, write_parquet(mt, 1:10))

  expect_error(write_parquet(mt2, tmp, metadata = "bad"))
  expect_snapshot(error = TRUE, write_parquet(mt2, tmp, metadata = "bad"))
  expect_error(write_parquet(mt2, tmp, metadata = mtcars))
  expect_snapshot(error = TRUE, write_parquet(mt2, tmp, metadata = mtcars))
})

test_that("writing metadata", {
  mt <- mt2 <- test_df()
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(mt, tmp, metadata = c("foo" = "bar"))
  kvm <- read_parquet_metadata(tmp)$file_meta_data$key_value_metadata[[1]]
  expect_snapshot(as.data.frame(kvm)[1, ])
})

test_that("strings are converted to UTF-8", {
  utf8 <- paste(
    "\u0043\u0073\u00e1\u0072\u0064\u0069",
    "\u0047\u00e1\u0062\u006f\u0072"
  )
  df <- data.frame(
    stringsAsFactors = FALSE,
    utf8 = utf8,
    latin1 = iconv(utf8, "UTf-8", "latin1")
  )
  expect_snapshot({
    charToRaw(utf8)
    charToRaw(df$utf8)
    charToRaw(df$latin1)
  })

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  write_parquet(df, tmp)
  df2 <- read_parquet(tmp)

  expect_snapshot({
    charToRaw(utf8)
    charToRaw(df2$utf8)
    charToRaw(df2$latin1)
  })
})

test_that("REQ PLAIN", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    stringsAsFactors = FALSE,
    i = c(1L, 2L, 2L, 2L, 3L, 3L, 3L, 3L),
    r = c(1, 1, 1, 1, 1, 1, 2, 10),
    l = c(TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, TRUE, TRUE),
    s = c("A", "A", "A", "B", "C", "D", "D", "D")
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    read_parquet_page(tmp, pgs$page_header_offset[2])$data
    read_parquet_page(tmp, pgs$page_header_offset[3])$data
    read_parquet_page(tmp, pgs$page_header_offset[4])$data
  })

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    read_parquet_page(tmp, pgs$page_header_offset[2])$data
    read_parquet_page(tmp, pgs$page_header_offset[3])$data
    read_parquet_page(tmp, pgs$page_header_offset[4])$data
  })
})

test_that("OPT PLAIN", {
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    stringsAsFactors = FALSE,
    i = c(1L, NA, 2L, NA, NA, 3L, 3L, 3L),
    r = c(1, 1, 1, 1, 1, 1, NA, NA),
    l = c(TRUE, FALSE, NA, NA, TRUE, FALSE, TRUE, NA),
    s = c(NA, NA, "A", "B", "C", NA, "D", NA)
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[1])$data)
    readBin(data[-(1:6)], "integer", sum(!is.na(d$i)))
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[2])$data)
    readBin(data[-(1:6)], "double", sum(!is.na(d$r)))
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[3])$data)
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[4])$data)
  })

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[1])$data)
    readBin(data[-(1:6)], "integer", sum(!is.na(d$i)))
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[2])$data)
    readBin(data[-(1:6)], "double", sum(!is.na(d$r)))
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[3])$data)
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[4])$data)
  })
})

test_that("REQ RLE_DICT", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    stringsAsFactors = FALSE,
    f = as.factor(c("A", "A", "B", "B", "C", "C", "D", "E"))
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[2])$data)
    rle_decode_int(data[-1], bit_width = as.integer(data[1]), nrow(d))
  })

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    read_parquet_page(tmp, pgs$page_header_offset[2])$data
  })
})

test_that("OPT RLE_DICT", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    stringsAsFactors = FALSE,
    f = as.factor(c(NA, "A", "B", NA, NA, "C", "D", "E"))
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[2])$data)
    rle_decode_int(
      data[-(1:7)],
      bit_width = as.integer(data[7]),
      sum(!is.na(d$f))
    )
  })

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
    read_parquet_page(tmp, pgs$page_header_offset[2])$data
  })
})

test_that("REQ RLE", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    l = c(logical(30), !logical(5), logical(20), !logical(30))
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_metadata(tmp)$column_chunks$encodings
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[1])$data)
  })
  expect_equal(
    rle_decode_int(data[-(1:4)], bit_width = 1L, nrow(d)),
    d$l + 0L
  )

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_metadata(tmp)$column_chunks$encodings
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
  })
})

test_that("OPT RLE", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    l = c(NA, logical(30), !logical(5), logical(20), NA, !logical(30))
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_metadata(tmp)$column_chunks$encodings
    data <- print(read_parquet_page(tmp, pgs$page_header_offset[1])$data)
  })
  expect_equal(
    rle_decode_int(data[-(1:4)], bit_width = 1L, nrow(d)),
    0L + !is.na(d$l)
  )
  on.exit(close(con), add = TRUE)
  def_len <- readBin(con <- rawConnection(data), "integer", 1)
  expect_equal(
    rle_decode_int(
      data[-(1:(8 + def_len))],
      bit_width = 1L,
      length(na.omit(d$l))
    ),
    0L + c(na.omit(d$l))
  )

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_metadata(tmp)$column_chunks$encodings
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
  })
})

test_that("Factor levels not in the data", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    f = factor(sample(letters[1:5], 100, replace = TRUE), levels = letters)
  )

  write_parquet(d, tmp)
  d2 <- read_parquet(tmp)
  expect_s3_class(d2$f, "factor")
  expect_equal(levels(d2$f), letters)
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    read_parquet_page(tmp, pgs$page_header_offset[1])$data
  })
  data <- read_parquet_page(tmp, pgs$page_header_offset[2])$data
  expect_equal(
    rle_decode_int(data[-1], as.integer(data[1]), nrow(d)),
    as.integer(d$f) - 1L
  )
})

test_that("write Date", {
  skip_on_cran()
  skip_without("arrow")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    d = c(Sys.Date() - 1, Sys.Date(), Sys.Date() + 1)
  )
  write_parquet(d, tmp)

  d2 <- arrow::read_parquet(tmp)
  expect_equal(d$d, d2$d)
})

test_that("write hms", {
  skip_on_cran()
  skip_without("arrow")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    h = hms::hms(1, 2, 3)
  )
  write_parquet(d, tmp)

  d2 <- arrow::read_parquet(tmp)
  expect_equal(d2$h, d$h)
})

test_that("write POSIXct", {
  skip_on_cran()
  skip_without("arrow")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    h = .POSIXct(Sys.time(), tz = "UTC")
  )
  write_parquet(d, tmp)

  d2 <- arrow::read_parquet(tmp)
  expect_equal(d$h, d2$h)
})

test_that("write difftime", {
  skip_on_cran()
  skip_without("arrow")

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # Fractional seconds are kept
  d <- data.frame(
    h = as.difftime(10 + 1 / 9, units = "secs")
  )
  write_parquet(d, tmp)

  d2 <- arrow::read_parquet(tmp)
  expect_equal(d$h, d2$h)

  # Other units are converted to secs
  tmp2 <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp2), add = TRUE)

  d <- data.frame(
    h = as.difftime(10, units = "mins")
  )
  write_parquet(d, tmp2)
  d2 <- arrow::read_parquet(tmp2)

  expect_snapshot({
    as.data.frame(d2)
  })
})

test_that("parquet_pages", {
  pp <- read_parquet_pages(test_path("data/mtcars-arrow.parquet"))
  expect_snapshot(as.data.frame(pp))
})

test_that("read_parquet_page", {
  pf <- test_path("data/mtcars-arrow.parquet")
  page <- read_parquet_page(pf, 4)
  expect_snapshot(page)
  page2 <- read_parquet_page(pf, 444)
  expect_snapshot(page2)
})

test_that("read_parquet_page for trick v2 data page", {
  pf <- test_path("data/rle_boolean_encoding.parquet")
  expect_snapshot({
    read_parquet_page(pf, 4L)
  })
})

test_that("read_parquet_page error", {
  # https://github.com/llvm/llvm-project/issues/59432
  if (is_asan()) {
    skip("ASAN bug")
  }
  pf <- test_path("data/mtcars-arrow.parquet")
  expect_error(read_parquet_page(pf, 5))
})

test_that("snappy", {
  for (i in 1:10) {
    x <- as.raw(as.integer(runif(1000) * 100))
    comp <- snappy_compress(x)
    uncp <- snappy_uncompress(comp)
    expect_false(length(x) == length(comp) && all(x == comp))
    expect_equal(uncp, x)
  }
})

test_that("snappy error", {
  expect_error(snappy_uncompress(charToRaw("foobar")))
})

test_that("MemStream", {
  expect_snapshot({
    .Call(test_memstream)
    cat(rawToChar(.Call(test_memstream)))
  })
})

test_that("DELTA_BIT_PACKED decoding", {
  skip_on_cran()
  skip_without_pyarrow()
  if (is_asan()) {
    skip("ASAN false positive")
  }

  dodbp <- function(df, path) {
    write_parquet(df, path)
    pyscript <- sprintf(
      r"[
import pyarrow
import pyarrow.parquet as pq
path = "%s"
df = pq.read_table(path)
writer = pq.ParquetWriter(path, df.schema, use_dictionary = False, column_encoding = 'DELTA_BINARY_PACKED')
writer.write_table(df)
writer.close()
]",
      normalizePath(path, winslash = "/")
    )
    pytmp <- tempfile(fileext = ".py")
    on.exit(unlink(pytmp), add = TRUE)
    writeLines(pyscript, pytmp)
    py <- if (Sys.which("python3") != "") "python3" else "python"
    processx::run(py, pytmp, stderr = "2>&1", error_on_status = FALSE)
  }

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(x = c(1:100, 500:600 * 2L))
  stat <- dodbp(d, tmp)
  expect_snapshot(stat)
  dbp <- read_parquet_page(tmp, 4L)$data
  expect_equal(dbp_decode_int(dbp), d$x)

  d <- data.frame(x = c(-(101:200) * 2L + 1:10))
  stat <- dodbp(d, tmp)
  expect_snapshot(stat)
  dbp <- read_parquet_page(tmp, 4L)$data
  expect_equal(dbp_decode_int(dbp), d$x)

  d <- data.frame(
    x = c(-(101:200) * 2L + 1:10, as.integer(rnorm(123) * 100L + 1000L))
  )
  stat <- dodbp(d, tmp)
  expect_snapshot(stat)
  dbp <- read_parquet_page(tmp, 4L)$data
  expect_equal(dbp_decode_int(dbp), d$x)

  d <- data.frame(x = 0L)
  stat <- dodbp(d, tmp)
  expect_snapshot(stat)
  dbp <- read_parquet_page(tmp, 4L)$data
  expect_equal(dbp_decode_int(dbp), d$x)

  # large integers
  d <- data.frame(
    x = c(
      as.integer(2^31 - 1 - runif(100) * 10L),
      as.integer(-2^31 + 1 + runif(100) * 10L)
    )
  )
  stat <- dodbp(d, tmp)
  expect_snapshot(stat)
  dbp <- read_parquet_page(tmp, 4L)$data
  expect_equal(dbp_decode_int(dbp), d$x)
})

test_that("DELTA_BINARY_PACKED edge cases", {
  skip_on_cran()
  if (is_ubsan()) {
    skip("UBSAN false positive")
  }
  if (is_asan()) {
    skip("ASAN false positive")
  }
  pf <- test_path("data/issue10279_delta_encoding.parquet")
  pgs <- read_parquet_pages(pf)
  expect_snapshot({
    # this is BOOLEAN in the file, which should not happen AFAICT
    p1 <- read_parquet_page(pf, pgs$page_header_offset[1])
    dbp1 <- p1$data[
      (p1$definition_levels_byte_length +
        p1$repetition_levels_byte_length +
        1):length(p1$data)
    ]
    dbp_decode_int(dbp1)

    p3 <- read_parquet_page(pf, pgs$page_header_offset[3])
    dbp3 <- p3$data
    dbp_decode_int(dbp3)

    p4 <- read_parquet_page(pf, pgs$page_header_offset[4])
    dbp4 <- p4$data
    dbp_decode_int(dbp4)

    # unfortunately the minimum int32 value is NA in R
    # not sure what to do about this...
    p5 <- read_parquet_page(pf, pgs$page_header_offset[5])
    dbp5 <- p5$data
    dbp_decode_int(dbp5)
  })
})

test_that("DELTA_BIANRY_PACKED INT64", {
  suppressPackageStartupMessages(library(bit64))
  pf <- test_path("data/dbp-int64.parquet")
  dt <- read_parquet_page(pf, 4L)$data
  on.exit(close(con), add = TRUE)
  len <- readBin(con <- rawConnection(dt), "integer", 1)
  dt <- dt[(len + 4 + 1):length(dt)]
  expect_snapshot({
    dbp_decode_int(dt)
  })
})

test_that("rle_decode", {
  # arrow might be broken on CRAN
  skip_on_cran()
  skip_without("arrow")
  tmp <- tempfile(fileext = ".parquet")
  d <- data.frame(
    x = as.factor(sample(letters, 100, replace = TRUE))
  )
  arrow::write_parquet(d, tmp)
  pages <- read_parquet_pages(tmp)
  data_page <- read_parquet_page(tmp, pages$page_header_offset[2])
  data <- data_page$data
  def_len <- readBin(data, "int", n = 1)
  data <- data[-(1:(def_len + 4L))]
  bw <- as.integer(data[1])
  data <- data[-1]
  idx <- rle_decode_int(data, bit_width = bw, length = nrow(d))
  # parquet indices are zero based, hence the -1L
  expect_equal(idx, as.integer(d$x) - 1L)
})

test_that("rle_encode", {
  chk <- function(x) {
    r <- rle_encode_int(x)
    x2 <- rle_decode_int(r, attr(r, "bit_width"), length(x))
    expect_equal(x2, x)
  }
  chk(c(0:16, rep(1L, 20), 0:16, rep(2L, 20)))
})

test_that("edge cases", {
  chk <- function(x) {
    r <- rle_encode_int(x)
    x2 <- rle_decode_int(r, attr(r, "bit_width"), length(x))
    expect_equal(x2, x)
  }
  chk(integer())
  chk(0L)
  chk(1L)
  chk(7L)
  chk(8L)
  chk(100L)
  chk(1:2)
  chk(1:3)
  chk(1:4)
  chk(1:5)
  chk(1:6)
  chk(1:7)
  chk(1:8)
  chk(1:9)
  chk(0:1)
  chk(0:2)
  chk(0:3)
  chk(0:4)
  chk(0:5)
  chk(0:6)
  chk(0:7)
  chk(0:8)
  chk(0:9)
  for (l in c(2, 7, 8, 9, 16, 256, 512, 513, 100000)) {
    chk(rep(0L, l))
    chk(rep(1L, l))
  }
})

test_that("REPEATED columns, no LIST", {
  pf <- test_path("data/repeated_primitive_no_list_no_nest.parquet")
  expect_snapshot({
    as.data.frame(read_parquet(pf))
  })
})

test_that("Only 3-layer LIST is supported for now", {
  pf <- test_path("data/old_list_structure.parquet")
  expect_snapshot(error = TRUE, {
    as.data.frame(read_parquet(pf))
  })
})

test_that("LIST", {
  expect_snapshot({
    as.data.frame(read_parquet(test_path("data/list-req-req.parquet")))
    as.data.frame(read_parquet(test_path("data/list-req-opt.parquet")))
    as.data.frame(read_parquet(test_path("data/list-opt-req.parquet")))
    as.data.frame(read_parquet(test_path("data/list-opt-opt.parquet")))
  })

  expect_snapshot({
    as.data.frame(read_parquet(test_path("data/list-v2-req-req.parquet")))
    as.data.frame(read_parquet(test_path("data/list-v2-req-opt.parquet")))
    as.data.frame(read_parquet(test_path("data/list-v2-opt-req.parquet")))
    as.data.frame(read_parquet(test_path("data/list-v2-opt-opt.parquet")))
  })

  elts <- c(
    "has_repetition_levels",
    "has_definition_levels",
    "num_values",
    "num_rows"
  )

  pf <- test_path("data/repeated_primitive_no_list_no_nest.parquet")
  pgoff <- read_parquet_pages(pf)$page_header_offset[4]
  expect_snapshot({
    read_parquet_page(pf, pgoff)[elts]
  })

  pf2 <- test_path("data/list-req-req.parquet")
  pgoff2 <- read_parquet_pages(pf2)$page_header_offset[2]
  expect_snapshot({
    read_parquet_page(pf2, pgoff2)[elts]
  })

  pf3 <- test_path("data/list-req-opt.parquet")
  pgoff3 <- read_parquet_pages(pf3)$page_header_offset[2]
  expect_snapshot({
    read_parquet_page(pf3, pgoff3)[elts]
  })

  pf4 <- test_path("data/list-opt-req.parquet")
  pgoff4 <- read_parquet_pages(pf4)$page_header_offset[2]
  expect_snapshot({
    read_parquet_page(pf4, pgoff4)[elts]
  })

  pf5 <- test_path("data/list-opt-opt.parquet")
  pgoff5 <- read_parquet_pages(pf5)$page_header_offset[2]
  expect_snapshot({
    read_parquet_page(pf5, pgoff5)[elts]
  })
})

test_that("write and read list(integer())", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list(1L, c(2L, 3L), NULL, c(4L, NA_integer_, 6L))
  write_parquet(df, tmp)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("write and read list(double())", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list(1.5, c(2.5, 3.5), NULL, c(4.5, NA_real_, 6.5))
  write_parquet(df, tmp)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("write and read list(character())", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list("a", c("b", "c"), NULL, c("d", NA_character_, "f"))
  write_parquet(df, tmp)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(integer()) multiple data pages", {
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "1024")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  N <- 1000
  pattern <- list(1L, c(2L, 3L), NULL, c(4L, NA_integer_, 6L))
  df <- data.frame(id = seq_len(N))
  df$x <- rep_len(pattern, N)

  write_parquet(df, tmp)
  pgs <- read_parquet_pages(tmp)
  expect_gt(sum(pgs$page_type == "DATA_PAGE"), 1)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)

  # data page v2
  write_parquet(df, tmp, options = parquet_options(write_data_page_version = 2))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(double()) multiple data pages", {
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "1024")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  N <- 1000
  pattern <- list(1.5, c(2.5, 3.5), NULL, c(4.5, NA_real_, 6.5))
  df <- data.frame(id = seq_len(N))
  df$x <- rep_len(pattern, N)

  write_parquet(df, tmp)
  pgs <- read_parquet_pages(tmp)
  expect_gt(sum(pgs$page_type == "DATA_PAGE"), 1)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)

  # data page v2
  write_parquet(df, tmp, options = parquet_options(write_data_page_version = 2))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(character()) multiple data pages", {
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "1024")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  N <- 1000
  pattern <- list("a", c("b", "c"), NULL, c("d", NA_character_, "f"))
  df <- data.frame(id = seq_len(N))
  df$x <- rep_len(pattern, N)

  write_parquet(df, tmp)
  pgs <- read_parquet_pages(tmp)
  expect_gt(sum(pgs$page_type == "DATA_PAGE"), 1)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)

  # data page v2
  write_parquet(df, tmp, options = parquet_options(write_data_page_version = 2))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(integer()) multiple row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:8)
  df$x <- list(
    1L,
    c(2L, 3L),
    NULL,
    c(4L, NA_integer_, 6L),
    integer(0),
    c(7L, 8L),
    NULL,
    9L
  )
  write_parquet(df, tmp, row_groups = c(1L, 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 2L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(double()) multiple row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:8)
  df$x <- list(
    1.5,
    c(2.5, 3.5),
    NULL,
    c(4.5, NA_real_, 6.5),
    double(0),
    c(7.5, 8.5),
    NULL,
    9.5
  )
  write_parquet(df, tmp, row_groups = c(1L, 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 2L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(character()) multiple row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:8)
  df$x <- list(
    "a",
    c("b", "c"),
    NULL,
    c("d", NA_character_, "f"),
    character(0),
    c("g", "h"),
    NULL,
    "i"
  )
  write_parquet(df, tmp, row_groups = c(1L, 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 2L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(integer()) many row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  pattern <- list(
    1L,
    c(2L, 3L),
    NULL,
    c(4L, NA_integer_, 6L),
    integer(0),
    c(7L, 8L),
    NULL,
    9L
  )
  df <- data.frame(id = 1:40)
  df$x <- rep_len(pattern, 40)
  write_parquet(df, tmp, row_groups = seq(1L, 40L, by = 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 10L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(double()) many row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  pattern <- list(
    1.5,
    c(2.5, 3.5),
    NULL,
    c(4.5, NA_real_, 6.5),
    double(0),
    c(7.5, 8.5),
    NULL,
    9.5
  )
  df <- data.frame(id = 1:40)
  df$x <- rep_len(pattern, 40)
  write_parquet(df, tmp, row_groups = seq(1L, 40L, by = 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 10L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(character()) many row groups", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  pattern <- list(
    "a",
    c("b", "c"),
    NULL,
    c("d", NA_character_, "f"),
    character(0),
    c("g", "h"),
    NULL,
    "i"
  )
  df <- data.frame(id = 1:40)
  df$x <- rep_len(pattern, 40)
  write_parquet(df, tmp, row_groups = seq(1L, 40L, by = 4L))
  expect_equal(nrow(read_parquet_metadata(tmp)[["row_groups"]]), 10L)
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
})

test_that("list(integer()) dictionary encoding", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list(1L, c(2L, 3L), NULL, c(4L, NA_integer_, 6L))
  df$y <- c("a", "b", "a", "b")

  write_parquet(df, tmp, encoding = c(x = "RLE_DICTIONARY"))
  pgs <- read_parquet_pages(tmp)
  expect_true(any(pgs$page_type == "DICTIONARY_PAGE"))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
  expect_equal(df2$y, df$y)
})

test_that("list(double()) dictionary encoding", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list(1.5, c(2.5, 3.5), NULL, c(4.5, NA_real_, 6.5))
  df$y <- c("a", "b", "a", "b")

  write_parquet(df, tmp, encoding = c(x = "RLE_DICTIONARY"))
  pgs <- read_parquet_pages(tmp)
  expect_true(any(pgs$page_type == "DICTIONARY_PAGE"))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
  expect_equal(df2$y, df$y)
})

test_that("list(character()) dictionary encoding", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(id = 1:4)
  df$x <- list("a", c("b", "c"), NULL, c("d", NA_character_, "f"))
  df$y <- c("a", "b", "a", "b")

  write_parquet(df, tmp, encoding = c(x = "RLE_DICTIONARY"))
  pgs <- read_parquet_pages(tmp)
  expect_true(any(pgs$page_type == "DICTIONARY_PAGE"))
  df2 <- as.data.frame(read_parquet(tmp))
  expect_equal(df2$id, df$id)
  expect_equal(df2$x, df$x)
  expect_equal(df2$y, df$y)
})

test_that("write and read zero columns, zero rows", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame()
  write_parquet(df, tmp)
  df2 <- read_parquet(tmp)
  expect_equal(ncol(df2), 0L)
  expect_equal(nrow(df2), 0L)
  expect_equal(
    read_parquet_metadata(tmp)$file_meta_data$num_rows,
    0L
  )
})

test_that("write and read zero rows", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(
    i = integer(),
    d = double(),
    s = character(),
    l = logical()
  )
  write_parquet(df, tmp)
  df2 <- read_parquet(tmp)
  expect_equal(nrow(df2), 0L)
  expect_equal(ncol(df2), 4L)
  expect_equal(names(df2), c("i", "d", "s", "l"))
  expect_equal(
    read_parquet_metadata(tmp)$file_meta_data$num_rows,
    0L
  )
})

test_that("read file with zero-row row group (issue #162)", {
  # File has two row groups: one with 1 row and one with 0 rows.
  # The zero-row row group has no dictionary page, so data_page_offset is 0.
  # nanoparquet must not seek to offset 0 in this case.
  f <- test_path("data/diann_minimal.parquet")
  d <- read_parquet(f)
  expect_equal(nrow(d), 1L)
  expect_equal(ncol(d), 1L)
})

test_that("REQ PLAIN", {
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "8192") # 8k pages
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  N <- 10000
  d <- data.frame(
    stringsAsFactors = FALSE,
    i = rep(c(1L, 2L, 2L, 2L, 3L, 3L, 3L, 3L), N / 8),
    r = rep(c(1, 1, 1, 1, 1, 1, 2, 10), N / 8),
    l = rep(c(TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, TRUE, TRUE), N / 8),
    s = rep(c("A", "A", "A", "B", "C", "D", "D", "D"), N / 8)
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    tapply(pgs$num_values, pgs$column + 1, sum)
  })
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    tapply(pgs$num_values, pgs$column + 1, sum)
  })
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  # data page v2
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "uncompressed"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(d, tmp, options = parquet_options(write_data_page_version = 2))
  expect_equal(as.data.frame(read_parquet(tmp)), d)
})

test_that("OPT PLAIN", {
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "8192") # 8k pages
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  N <- 10000
  d <- data.frame(
    stringsAsFactors = FALSE,
    i = rep(c(1L, NA, 2L, NA, NA, 3L, 3L, 3L), N / 8),
    r = rep(c(1, 1, 1, 1, 1, 1, NA, NA), N / 8),
    l = rep(c(TRUE, FALSE, NA, NA, TRUE, FALSE, TRUE, NA), N / 8),
    s = rep(c(NA, NA, "A", "B", "C", NA, "D", NA), N / 8)
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    tapply(pgs$num_values, pgs$column + 1, sum)
  })
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_snapshot({
    tapply(pgs$num_values, pgs$column + 1, sum)
  })
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  # data page v2
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "uncompressed"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(d, tmp, options = parquet_options(write_data_page_version = 2))
  expect_equal(as.data.frame(read_parquet(tmp)), d)
})

test_that("REQ RLE_DICT", {
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "8192") # 8k pages
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # the index is bit-packed, so we need a bit more of this
  N <- 100000
  d <- data.frame(
    stringsAsFactors = FALSE,
    f = rep(as.factor(c("A", "A", "B", "B", "C", "C", "D", "E")), N / 8)
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_equal(sum(pgs$num_values[pgs$page_type == "DATA_PAGE"]), nrow(d))
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_equal(sum(pgs$num_values[pgs$page_type == "DATA_PAGE"]), nrow(d))
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  # data page v2
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "uncompressed"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "snappy"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "gzip"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "zstd"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
})

test_that("OPT RLE_DICT", {
  withr::local_envvar(NANOPARQUET_PAGE_SIZE = "8192") # 8k pages
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  # the index is bit-packed, so we need a bit more of this
  N <- 100000
  d <- data.frame(
    stringsAsFactors = FALSE,
    f = rep(as.factor(c(NA, "A", "B", NA, NA, "C", "D", "E")), N / 8)
  )

  write_parquet(d, tmp, compression = "uncompressed")
  pgs <- read_parquet_pages(tmp)
  expect_equal(sum(pgs$num_values[pgs$page_type == "DATA_PAGE"]), nrow(d))
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  write_parquet(d, tmp, compression = "snappy")
  pgs <- read_parquet_pages(tmp)
  expect_equal(sum(pgs$num_values[pgs$page_type == "DATA_PAGE"]), nrow(d))
  expect_equal(as.data.frame(read_parquet(tmp)), d)

  # data page v2
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "uncompressed"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "snappy"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "gzip"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
  write_parquet(
    d,
    tmp,
    options = parquet_options(write_data_page_version = 2),
    compression = "zstd"
  )
  expect_equal(as.data.frame(read_parquet(tmp)), d)
})

test_that("write_parquet() to memory", {
  d <- test_df(missing = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  pm <- write_parquet(d, ":raw:")
  write_parquet(d, tmp)
  pm2 <- readBin(tmp, "raw", file.size(tmp))

  expect_equal(pm, pm2)
})

test_that("write_parquet() to memory 2", {
  # https://github.com/r-lib/nanoparquet/issues/77
  data <- data.frame(
    TEST_1 = seq(as.Date("2000/1/1"), by = "day", length.out = 100000),
    TEST_2 = seq(1:100000),
    TEST_3 = seq(as.Date("1990/1/1"), by = "day", length.out = 100000),
    TEST_2 = seq(100000:1)
  )

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  r1 <- write_parquet(data, ":raw:")
  write_parquet(data, tmp)
  r2 <- readBin(tmp, what = "raw", file.size(tmp))
  expect_equal(r1, r2)
})

test_that("write_parquet() to stdout", {
  skip_on_cran()
  if (packageVersion("processx") < "3.8.7.9000") {
    skip("processx 3.8.7.9000 is required")
  }

  tmp1 <- tempfile(fileext = ".parquet")
  tmp2 <- tempfile(fileext = ".parquet")
  script <- tempfile(fileext = ".R")
  on.exit(unlink(c(tmp1, tmp2, script)), add = TRUE)

  libpaths <- paste(sprintf('"%s"', .libPaths()), collapse = ", ")
  txt <- c(
    sprintf(".libPaths(c(%s))", libpaths),
    "nanoparquet::write_parquet(mtcars, \":stdout:\")"
  )
  writeLines(txt, script)
  processx::run(
    rscript(),
    c("--quiet", script),
    stdout = tmp1,
    stderr = NULL,
    encoding = "binary"
  )
  r1 <- readBin(tmp1, "raw", file.size(tmp1))
  write_parquet(mtcars, tmp2)
  r2 <- readBin(tmp2, "raw", file.size(tmp2))
  expect_equal(file.size(tmp1), file.size(tmp2))
  expect_equal(r1, r2)
})

test_that("gzip compression", {
  d <- test_df(missing = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(d, tmp, compression = "gzip")
  expect_equal(read_parquet_page(tmp, 4L)$codec, "GZIP")
  expect_equal(read_parquet(tmp), d)
})

test_that("zstd compression", {
  d <- test_df(missing = TRUE)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(d, tmp, compression = "zstd")
  expect_equal(read_parquet_page(tmp, 4L)$codec, "ZSTD")
  expect_equal(read_parquet(tmp), d)
})

test_that("Conversion of sub-dates prior Posix origin is correct", {
  data <- data.frame(
    days = as.Date(c(-1.1, -0.1, 0, 0.1, 1.1), origin = "1970-01-01")
  )

  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  write_parquet(data, tmp)
  expect_equal(
    as.character(as.data.frame(read_parquet(tmp))$date),
    as.character(data$date)
  )
})

test_that("errors", {
  expect_snapshot(error = TRUE, {
    parquet_options(num_rows_per_row_group = "foobar")
  })

  df <- test_df()
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  expect_snapshot(error = TRUE, {
    write_parquet(df, tmp, row_groups = "foobar")
    write_parquet(df, tmp, row_groups = c(100L, 1L))
    write_parquet(df, tmp, row_groups = c(1L, 100L))
  })
})

test_that("row groups", {
  tmp1 <- tempfile(fileext = ".parquet")
  tmp2 <- tempfile(fileext = ".parquet")
  on.exit(unlink(c(tmp1, tmp2)), add = TRUE)

  df <- test_df()
  write_parquet(df, tmp1, row_groups = 1L)
  write_parquet(df, tmp2, row_groups = c(1L, 16L))
  expect_equal(read_parquet(tmp1), read_parquet(tmp2))
  expect_equal(nrow(read_parquet_metadata(tmp2)[["row_groups"]]), 2L)

  unlink(tmp2)
  write_parquet(df, tmp2, row_groups = seq_len(nrow(df)))
  expect_equal(read_parquet(tmp1), read_parquet(tmp2))
  expect_equal(nrow(read_parquet_metadata(tmp2)[["row_groups"]]), nrow(df))

  unlink(tmp2)
  withr::local_options(nanoparquet.num_rows_per_row_group = 10L)
  write_parquet(df, tmp2)
  expect_equal(read_parquet(tmp1), read_parquet(tmp2))
  expect_equal(nrow(read_parquet_metadata(tmp2)[["row_groups"]]), 4L)
})

test_that("factors & factor levels", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(
    f = factor(c(rep("a", 100), rep("b", 100), rep("c", 100)))
  )
  withr::local_options(nanoparquet.num_rows_per_row_group = 50L)
  write_parquet(df, tmp)
  expect_equal(as.data.frame(read_parquet(tmp)), df)
  # the same dict is written into every dicitonary page
  pgs <- read_parquet_pages(tmp)
  dict_ofs <- pgs[["page_header_offset"]][
    pgs[["page_type"]] == "DICTIONARY_PAGE"
  ]
  dict_data <- read_parquet_page(tmp, dict_ofs[1])[["data"]]
  for (do in dict_ofs) {
    expect_equal(dict_data, read_parquet_page(tmp, do)[["data"]])
  }
})

test_that("non-factors write local dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- data.frame(
    stringsAsFactors = FALSE,
    f = c(rep("a", 100), rep("b", 100), rep("c", 100))
  )
  withr::local_options(nanoparquet.num_rows_per_row_group = 40L)
  write_parquet(df, tmp)
  expect_equal(as.data.frame(read_parquet(tmp)), df)
  pgs <- read_parquet_pages(tmp)
  dict_ofs <- pgs[["page_header_offset"]][
    pgs[["page_type"]] == "DICTIONARY_PAGE"
  ]
  expect_snapshot({
    for (do in dict_ofs) {
      print(read_parquet_page(tmp, do)[["data"]])
    }
  })
})

test_that("strings in a dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  df <- test_df()
  write_parquet(
    df,
    tmp,
    encoding = c(large = "RLE", "RLE_DICTIONARY"),
    options = parquet_options(num_rows_per_row_group = 10)
  )
  expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
})

type_names <- c(
  BOOLEAN = 0L,
  INT32 = 1L,
  INT64 = 2L,
  INT96 = 3L,
  FLOAT = 4L,
  DOUBLE = 5L,
  BYTE_ARRAY = 6L,
  FIXED_LEN_BYTE_ARRAY = 7L
)
ctype_names <- c(
  UTF8 = 0L,
  MAP = 1L,
  MAP_KEY_VALUE = 2L,
  LIST = 3L,
  ENUM = 4L,
  DECIMAL = 5L,
  DATE = 6L,
  TIME_MILLIS = 7L,
  TIME_MICROS = 8L,
  TIMESTAMP_MILLIS = 9L,
  TIMESTAMP_MICROS = 10L,
  UINT_8 = 11L,
  UINT_16 = 12L,
  UINT_32 = 13L,
  UINT_64 = 14L,
  INT_8 = 15L,
  INT_16 = 16L,
  INT_32 = 17L,
  INT_64 = 18L,
  JSON = 19L,
  BSON = 20L,
  INTERVAL = 21L
)

repetition_types <- c(
  REQUIRED = 0L,
  OPTIONAL = 1L,
  REPEATED = 2L
)

encodings <- c(
  PLAIN = 0L,
  GROUP_VAR_INT = 1L, # was never used, now deprecated
  PLAIN_DICTIONARY = 2L,
  RLE = 3L,
  BIT_PACKED = 4L,
  DELTA_BINARY_PACKED = 5L,
  DELTA_LENGTH_BYTE_ARRAY = 6L,
  DELTA_BYTE_ARRAY = 7L,
  RLE_DICTIONARY = 8L,
  BYTE_STREAM_SPLIT = 9L
)

codecs <- c(
  UNCOMPRESSED = 0L,
  SNAPPY = 1L,
  GZIP = 2L,
  LZO = 3L,
  BROTLI = 4L,
  LZ4 = 5L,
  ZSTD = 6L
)

page_types <- c(
  DATA_PAGE = 0L,
  INDEX_PAGE = 1L,
  DICTIONARY_PAGE = 2L,
  DATA_PAGE_V2 = 3
)

format_schema_result <- function(mtd, sch, options) {
  sch$type <- names(type_names)[sch$type + 1L]
  sch$converted_type <-
    names(ctype_names)[sch$converted_type + 1L]
  sch$repetition_type <-
    names(repetition_types)[sch$repetition_type + 1L]
  sch$logical_type <- I(sch$logical_type)
  sch <- as.data.frame(sch)
  sch <- add_r_type_to_schema(mtd, sch, options)
  class(sch) <- c("nanoparquet_schema", "tbl", class(sch))

  sch
}

#' Read the metadata of a Parquet file
#'
#' This function should work on all files, even if [read_parquet()] is
#' unable to read them, because of an unsupported schema, encoding,
#' compression or other reason.
#'
#' @param file Path to a Parquet file.
#' @param options Options that potentially alter the default Parquet to R
#'   type mappings, see [parquet_options()].
#' @return A named list with entries:
#'   * `file_meta_data`: a data frame with file meta data:
#'     - `file_name`: file name.
#'     - `version`: Parquet version, an integer.
#'     - `num_rows`: total number of rows.
#'     - `key_value_metadata`: list column of a data frames with two
#'       character columns called `key` and `value`. This is the key-value
#'       metadata of the file. Arrow stores its schema here.
#'     - `created_by`: A string scalar, usually the name of the software
#'       that created the file.
#'   * `schema`:
# -- If YOU UPDATE THIS, ALSO UPDATE parquet_schema BELOW -----------------
#'     data frame, the schema of the file. It has one row for
#'     each node (inner node or leaf node). For flat files this means one
#'     root node (inner node), always the first one, and then one row for
#'     each "real" column. For nested schemas, the rows are in depth-first
#'     search order. Most important columns are:
#'     - `file_name`: file name.
#'     - `r_col`: integer, the column index in R, starting from one.
#'       `NA` for the root node. For Parquet files without list columns
#'       this is simply the column index in the file. A list column has
#'       multiple schema rows, and they all have the same `r_col`.
#'     - `name`: column name.
#'     - `r_type`: the R type that corresponds to the Parquet type.
#'       Might be `NA` if [read_parquet()] cannot read this column. See
#'       [nanoparquet-types] for the type mapping rules.
#'     - `r_type`:
#'     - `type`: data type. One of the low level data types.
#'     - `type_length`: length for fixed length byte arrays.
#'     - `repettion_type`: character, one of `REQUIRED`, `OPTIONAL` or
#'       `REPEATED`.
#'     - `logical_type`: a list column, the logical types of the columns.
#'       An element has at least an entry called `type`, and potentially
#'       additional entries, e.g. `bit_width`, `is_signed`, etc.
#'     - `num_children`: number of child nodes. Should be a non-negative
#'       integer for the root node, and `NA` for a leaf node.
# -------------------------------------------------------------------------
#'   * `$row_groups`: a data frame, information about the row groups.
#'     Some important columns:
#'     - `file_name`: file name.
#'     - `id`: row group id, integer from zero to number of row groups
#'       minus one.
#'     - `total_byte_size`: total uncompressed size of all column data.
#'     - `num_rows`: number of rows.
#'     - `file_offset`: where the row group starts in the file. This is
#'       optional, so it might be `NA`.
#'     - `total_compressed_size`: total byte size of all compressed
#'       (and potentially encrypted) column data in this row group.
#'       This is optional, so it might be `NA`.
#'     - `ordinal`: ordinal position of the row group in the file, starting
#'       from zero. This is optional, so it might be `NA`. If `NA`, then
#'       the order of the row groups is as they appear in the metadata.
# -------------------------------------------------------------------------
#'   * `$column_chunks`: a data frame, information about all column chunks,
#'     across all row groups. Some important columns:
#'     - `file_name`: file name.
#'     - `row_group`: which row group this chunk belongs to.
#'     - `column`: which leaf column this chunks belongs to. The order is
#'       the same as in `$schema`, but only leaf columns (i.e. columns with
#'       `NA` children) are counted.
#'     - `file_path`: which file the chunk is stored in. `NA` means the
#'       same file.
#'     - `file_offset`: where the column chunk begins in the file.
#'     - `type`: low level parquet data type.
#'     - `encodings`: encodings used to store this chunk. It is a list
#'       column of character vectors of encoding names. Current possible
#'       encodings: `r paste0('"', names(encodings), '"', collapse = ", ")`.
#'     - `path_in_scema`: list column of character vectors. It is simply
#'       the path from the root node. It is simply the column name for
#'       flat schemas.
#'     - `codec`: compression codec used for the column chunk. Possible
#'       values are: `r paste0('"', names(codecs), '"', collapse = ", ")`.
#'     - `num_values`: number of values in this column chunk.
#'     - `total_uncompressed_size`: total uncompressed size in bytes.
#'     - `total_compressed_size`: total compressed size in bytes.
#'     - `data_page_offset`: absolute position of the first data page of
#'       the column chunk in the file.
#'     - `index_page_offset`: absolute position of the first index page of
#'       the column chunk in the file, or `NA` if there are no index pages.
#'     - `dictionary_page_offset`: absolute position of the first
#'       dictionary page of the column chunk in the file, or `NA` if there
#'       are no dictionary pages.
#'     - `null_count`: the number of missing values in the column chunk.
#'       It may be `NA`.
#'     - `min_value`: list column of raw vectors, the minimum value of the
#'       column, in binary. If `NULL`, then then it is not specified.
#'       This column is experimental.
#'     - `max_value`: list column of raw vectors, the maximum value of the
#'       column, in binary. If `NULL`, then then it is not specified.
#'       This column is experimental.
#'     - `is_min_value_exact`: whether the minimum value is an actual
#'       value of a column, or a bound. It may be `NA`.
#'     - `is_max_value_exact`: whether the maximum value is an actual
#'       value of a column, or a bound. It may be `NA`.
#'
#' @export
#' @seealso [read_parquet_info()] for a much shorter summary.
#'   [read_parquet_schema()] for column information.
#'   [read_parquet()] to read, [write_parquet()] to write Parquet files,
#'   [nanoparquet-types] for the R <-> Parquet type mappings.
#' @examples
#' file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
#' nanoparquet::read_parquet_metadata(file_name)

read_parquet_metadata <- function(file, options = parquet_options()) {
  file <- path.expand(file)
  res <- .Call(nanoparquet_read_metadata, file)

  res$file_meta_data$key_value_metadata <-
    as.data.frame(res$file_meta_data$key_value_metadata)
  class(res$file_meta_data$key_value_metadata) <-
    c("tbl", class(res$file_meta_data$key_value_metadata))
  res$file_meta_data$key_value_metadata <-
    I(list(res$file_meta_data$key_value_metadata))
  res$file_meta_data <- as.data.frame(res$file_meta_data)
  class(res$file_meta_data) <- c("tbl", class(res$file_meta_data))

  res$schema <- format_schema_result(res, res$schema, options)

  res$row_groups <- as.data.frame(res$row_groups)
  class(res$row_groups) <- c("tbl", class(res$row_groups))

  res$column_chunks$type <- names(type_names)[res$column_chunks$type + 1L]
  res$column_chunks$encodings <- lapply(
    res$column_chunks$encodings,
    function(ec) {
      names(encodings)[ec + 1L]
    }
  )
  res$column_chunks$codec <- names(codecs)[res$column_chunks$codec + 1L]
  res$column_chunks$encodings <- I(res$column_chunks$encodings)
  res$column_chunks$path_in_schema <- I(res$column_chunks$path_in_schema)
  res$column_chunks$min_value <- I(res$column_chunks$min_value)
  res$column_chunks$max_value <- I(res$column_chunks$max_value)
  res$column_chunks <- as.data.frame(res$column_chunks)
  class(res$column_chunks) <- c("tbl", class(res$column_chunks))

  res
}

#' @export
#' @rdname read_parquet_metadata

parquet_metadata <- function(file) {
  warning(
    "`parquet_metadata()` is deprecated. ",
    "Please use `read_parquet_metadata()` instead."
  )
  read_parquet_metadata(file)
}

#' Read the schema of a Parquet file
#'
#' This function should work on all files, even if [read_parquet()] is
#' unable to read them, because of an unsupported schema, encoding,
#' compression or other reason.
#'
#' @param file Path to a Parquet file.
#' @param options Return value of [parquet_options()], options that
#'   potentially modify the Parquet to R type mappings.
#' @return
# -- If YOU UPDATE THIS, ALSO UPDATE read_parquet_metadata ABOVE ----------
#'     Data frame, the schema of the file. It has one row for
#'     each node (inner node or leaf node). For flat files this means one
#'     root node (inner node), always the first one, and then one row for
#'     each "real" column. For nested schemas, the rows are in depth-first
#'     search order. Most important columns are:
#'     - `file_name`: file name.
#'     - `r_col`: integer, the column index in R, starting from one.
#'       `NA` for the root node. For Parquet files without list columns
#'       this is simply the column index in the file. A list column has
#'       multiple schema rows, and they all have the same `r_col`.
#'     - `name`: column name.
#'     - `r_type`: the R type that corresponds to the Parquet type.
#'       Might be `NA` if [read_parquet()] cannot read this column. See
#'       [nanoparquet-types] for the type mapping rules.
#'     - `type`: data type. One of the low level data types.
#'     - `type_length`: length for fixed length byte arrays.
#'     - `repettion_type`: character, one of `REQUIRED`, `OPTIONAL` or
#'       `REPEATED`.
#'     - `logical_type`: a list column, the logical types of the columns.
#'       An element has at least an entry called `type`, and potentially
#'       additional entries, e.g. `bit_width`, `is_signed`, etc.
#'     - `num_children`: number of child nodes. Should be a non-negative
#'       integer for the root node, and `NA` for a leaf node.
# -------------------------------------------------------------------------
#'
#' @seealso [read_parquet_metadata()] to read more metadata,
#'   [read_parquet_info()] to show only basic information.
#'   [read_parquet()], [write_parquet()], [nanoparquet-types].
#' @export

read_parquet_schema <- function(file, options = parquet_options()) {
  file <- path.expand(file)
  mtd <- read_parquet_metadata(file, options = options)
  mtd$schema
}

#' Short summary of a Parquet file
#'
#' @param file Path to a Parquet file.
#' @return Data frame with columns:
#'   * `file_name`: file name.
#'   * `num_cols`: number of columns. (The number of child nodes of the
#'      root node in the schema.)
#'   * `num_rows`: number of rows.
#'   * `num_row_groups`: number of row groups.
#'   * `file_size`: file size in bytes.
#'   * `parquet_version`: Parquet version.
#'   * `created_by`: A string scalar, usually the name of the software
#'       that created the file. `NA` if not available.
#'
#' @seealso [read_parquet_metadata()] to read more metadata,
#'   [read_parquet_schema()] for column information.
#'   [read_parquet()], [write_parquet()], [nanoparquet-types].
#' @export

read_parquet_info <- function(file) {
  file <- path.expand(file)
  mtd <- read_parquet_metadata(file)
  info <- data.frame(
    stringsAsFactors = FALSE,
    file_name = file,
    num_cols = mtd$schema$num_children[1],
    num_rows = mtd$file_meta_data$num_rows,
    num_row_groups = nrow(mtd$row_groups),
    file_size = file.size(file),
    parquet_version = mtd$file_meta_data$version,
    created_by = mtd$file_meta_data$created_by
  )
  class(info) <- c("tbl", class(info))
  info
}

#' @export
#' @rdname read_parquet_info

parquet_info <- function(file) {
  warning(
    "`parquet_info()` is deprecated, please use `read_parquet_info() ",
    "instead."
  )
  read_parquet_info(file)
}

test_that("null_count is written", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- test_df(missing = TRUE)
  write_parquet(
    df,
    tmp,
    options = parquet_options(num_rows_per_row_group = 10)
  )
  expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
  expect_snapshot(
    as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]][,
      c("row_group", "column", "null_count")
    ])
  )
})

test_that("min/max for integers", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = c(
      sample(1:5),
      sample(c(1:3, -100L, 100L)),
      sample(c(-1000L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    )
  )

  as_int <- function(x) {
    sapply(x, function(xx) {
      xx %&&% readBin(xx, what = "integer") %||% NA_integer_
    })
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int(mtd[["min_value"]]),
      as_int(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for DATEs", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  do <- function(...) {
    df <- data.frame(
      day = rep(as.Date("2024-09-16") - 10:1, each = 10),
      count = 1:100
    )
    df$day[c(1, 20, 25, 40)] <- as.Date(NA_character_)
    write_parquet(
      df,
      tmp,
      options = parquet_options(num_rows_per_row_group = 20),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    minv <- mtd[mtd$column == 0, "min_value"]
    maxv <- mtd[mtd$column == 0, "max_value"]
    list(
      as.data.frame(read_parquet_schema(tmp)[, -1]),
      as.Date(
        map_int(minv, readBin, what = "integer", n = 1),
        origin = "1970-01-01"
      ),
      as.Date(
        map_int(maxv, readBin, what = "integer", n = 1),
        origin = "1970-01-01"
      )
    )
  }

  expect_snapshot(do())
})

test_that("min/max for double -> signed integers", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = as.double(c(
      sample(1:5),
      sample(c(1:3, -100L, 100L)),
      sample(c(-1000L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    ))
  )

  as_int <- function(x) {
    sapply(x, function(xx) {
      xx %&&% readBin(xx, what = "integer") %||% NA_integer_
    })
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "INT32"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int(mtd[["min_value"]]),
      as_int(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for double -> unsigned integers", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = as.double(c(
      sample(1:5),
      sample(c(1:3, 1L, 100L)),
      sample(c(0L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    ))
  )

  as_int <- function(x) {
    sapply(x, function(xx) {
      xx %&&% readBin(xx, what = "integer") %||% NA_integer_
    })
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "UINT_32"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int(mtd[["min_value"]]),
      as_int(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("minmax for double -> INT32 TIME(MULLIS)", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  # IDK what's the point of signed TIME, but it seems to be allowed, the
  # sort order is signed
  df <- data.frame(
    x = hms::as_hms(c(
      sample(1:5),
      sample(c(1:3, -100L, 100L)),
      sample(c(-1000L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    ))
  )

  as_int <- function(x) {
    sapply(x, function(xx) {
      xx %&&% readBin(xx, what = "integer") %||% NA_integer_
    })
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "TIME_MILLIS"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int(mtd[["min_value"]]),
      as_int(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for DOUBLE", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = as.double(c(
      sample(1:5),
      sample(c(1:3, -100, 100)),
      sample(c(-1000, NA_real_, 1000, NA_real_, NA_real_)),
      rep(NA_real_, 3)
    ))
  )

  as_dbl <- function(x) {
    sapply(x, function(xx) xx %&&% readBin(xx, what = "double") %||% NA_real_)
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_dbl(mtd[["min_value"]]),
      as_dbl(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for FLOAT", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = as.double(c(
      sample(1:5),
      sample(c(1:3, -100, 100)),
      sample(c(-1000, NA_real_, 1000, NA_real_, NA_real_)),
      rep(NA_real_, 3)
    ))
  )

  as_flt <- function(x) {
    sapply(x, function(xx) xx %&&% .Call(read_float, xx) %||% NA_real_)
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "FLOAT"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(read_parquet(tmp)), as.data.frame(df))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_flt(mtd[["min_value"]]),
      as_flt(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for integer -> INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = c(
      sample(1:5),
      sample(c(1:3, -100L, 100L)),
      sample(c(-1000L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    )
  )

  as_int64 <- function(x) {
    sapply(x, function(xx) xx %&&% .Call(read_int64, xx) %||% NA_real_)
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "INT64"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int64(mtd[["min_value"]]),
      as_int64(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for REALSXP -> INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = as.double(c(
      sample(1:5),
      sample(c(1:3, -100L, 100L)),
      sample(c(-1000L, NA_integer_, 1000L, NA_integer_, NA_integer_)),
      rep(NA_integer_, 3)
    ))
  )

  as_int64 <- function(x) {
    sapply(x, function(xx) xx %&&% .Call(read_int64, xx) %||% NA_real_)
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = "INT64"),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int64(mtd[["min_value"]]),
      as_int64(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("min/max for STRING", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- data.frame(
    x = c(
      sample(letters[1:5]),
      sample(c(letters[1:3], "!!!", "~~~")),
      sample(c("!", NA_character_, "~", NA_character_, NA_character_)),
      rep(NA_character_, 3)
    )
  )

  as_str <- function(x) {
    sapply(x, function(xx) xx %&&% rawToChar(xx) %||% NA_character_)
  }

  do <- function(encoding = "PLAIN", type = "STRING", ...) {
    write_parquet(
      df,
      tmp,
      schema = parquet_schema(x = type),
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    expect_equal(read_parquet_schema(tmp)$logical_type[[2]]$type, type)
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_str(mtd[["min_value"]]),
      as_str(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))

  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))

  expect_snapshot(do(compression = "snappy", type = "JSON"))
  expect_snapshot(do(compression = "uncompressed", type = "JSON"))

  # dictionary
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "snappy",
    type = "JSON"
  ))
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "uncompressed",
    type = "JSON"
  ))

  expect_snapshot(do(compression = "snappy", type = "BSON"))
  expect_snapshot(do(compression = "uncompressed", type = "BSON"))

  # dictionary
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "snappy",
    type = "BSON"
  ))
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "uncompressed",
    type = "BSON"
  ))

  expect_snapshot(do(compression = "snappy", type = "ENUM"))
  expect_snapshot(do(compression = "uncompressed", type = "ENUM"))

  # dictionary
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "snappy",
    type = "ENUM"
  ))
  expect_snapshot(do(
    encoding = "RLE_DICTIONARY",
    compression = "uncompressed",
    type = "ENUM"
  ))
})

test_that("min/max for REALSXP -> TIMESTAMP (INT64)", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  now <- 0L
  df <- data.frame(
    x = .POSIXct(
      as.double(c(
        sample(now + 1:5),
        sample(c(now + c(1:3, -100L, 100L))),
        sample(c(
          now - 1000L,
          NA_integer_,
          now + 1000L,
          NA_integer_,
          NA_integer_
        )),
        rep(NA_integer_, 3)
      )),
      tz = "UTC"
    )
  )

  as_int64 <- function(x) {
    sapply(x, function(xx) xx %&&% .Call(read_int64, xx) %||% NA_real_)
  }

  do <- function(encoding = "PLAIN", ...) {
    write_parquet(
      df,
      tmp,
      encoding = encoding,
      options = parquet_options(num_rows_per_row_group = 5),
      ...
    )
    expect_equal(as.data.frame(df), as.data.frame(read_parquet(tmp)))
    mtd <- as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]])
    list(
      as_int64(mtd[["min_value"]]),
      as_int64(mtd[["max_value"]]),
      mtd[["is_min_value_exact"]],
      mtd[["is_max_value_exact"]]
    )
  }
  expect_snapshot(do(compression = "snappy"))
  expect_snapshot(do(compression = "uncompressed"))
  return()
  # dictionary
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "snappy"))
  expect_snapshot(do(encoding = "RLE_DICTIONARY", compression = "uncompressed"))
})

test_that("parquet_metadata", {
  withr::local_envvar(NANOPARQUET_FORCE_PLAIN = "1")
  df <- test_df()
  mkdirp(tmpdir <- tempfile())
  tmp <- file.path(tmpdir, "test.parquet")
  on.exit(unlink(tmpdir, recursive = TRUE), add = TRUE)
  write_parquet(df, tmp, compression = "uncompressed")

  wd <- getwd()
  on.exit(setwd(wd), add = TRUE)
  setwd(tmpdir)

  mtd <- read_parquet_metadata("test.parquet")
  mtd$file_meta_data$key_value_metadata <-
    as.data.frame(mtd$file_meta_data$key_value_metadata)

  expect_snapshot({
    as.data.frame(mtd$file_meta_data)
    as.data.frame(mtd$schema)
    as.data.frame(mtd$row_groups)
    as.data.frame(mtd$column_chunks[, 1:20])
  })

  sch <- read_parquet_schema("test.parquet")
  expect_snapshot(as.data.frame(sch))
})

test_that("ENUM type", {
  pf <- test_path("data/enum.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
})

test_that("UUID type", {
  pf <- test_path("data/uuid-arrow.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
})

test_that("DATE type", {
  pf <- test_path("data/date.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
})

test_that("DECIMAL type", {
  pf <- test_path("data/decimals.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
  expect_snapshot(sch$logical_type)
})

test_that("TIME type", {
  pf <- test_path("data/timetz.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
  expect_snapshot(sch$logical_type)
})

test_that("TIMESTAMP type", {
  pf <- test_path("data/timestamp.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
  expect_snapshot(sch$logical_type)
})

test_that("LIST type", {
  pf <- test_path("data/nested_lists.snappy.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
})

test_that("MAP type", {
  pf <- test_path("data/map.parquet")
  sch <- read_parquet_schema(pf)
  expect_snapshot(as.data.frame(sch))
})

test_that("key-value metadata", {
  pf <- test_path("data/mtcars-arrow.parquet")
  mtd <- read_parquet_metadata(pf)
  expect_snapshot(
    as.data.frame(mtd$file_meta_data$key_value_metadata[[1]])
  )
})

test_that("parquet_info", {
  library(pillar)
  expect_snapshot({
    read_parquet_info(test_path("data/enum.parquet"))
    read_parquet_info(test_path("data/factor.parquet"))
    read_parquet_info(test_path("data/decimals.parquet"))
  })
})

test_that("TIMESTAMP", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  ts <- structure(1724186255.52948, class = c("POSIXct", "POSIXt"))
  d <- data.frame(ts = ts)
  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MILLIS"
  ))
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  sch2 <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MICROS"
  ))
  write_parquet(d, tmp, schema = sch2)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  sch3 <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "NANOS"
  ))
  write_parquet(d, tmp, schema = sch3)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("TIMESTAMP, dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  ts <- structure(1724186255.52948, class = c("POSIXct", "POSIXt"))
  d <- data.frame(ts = ts)
  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MILLIS"
  ))
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch2 <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MICROS"
  ))
  write_parquet(d, tmp, schema = sch2, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch3 <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "NANOS"
  ))
  write_parquet(d, tmp, schema = sch3, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })
})

test_that("TIMESTAMP_MILLIS, TIMESTAMP_MICROS", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  ts <- structure(1724186255.52948, class = c("POSIXct", "POSIXt"))
  d <- data.frame(ts = ts)
  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MILLIS"
  ))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
  })

  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MICROS"
  ))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
  })
})

test_that("TIMESTAMP_MILLIS, TIMESTAMP_MICROS, dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  ts <- structure(1724186255.52948, class = c("POSIXct", "POSIXt"))
  d <- data.frame(ts = ts)
  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MILLIS"
  ))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch <- parquet_schema(list(
    "TIMESTAMP",
    is_adjusted_utc = TRUE,
    unit = "MICROS"
  ))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })
})

test_that("TIME", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  dt <- hms::hms(1, 2, 3)
  d <- data.frame(dt = dt)
  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS"))
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  sch2 <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MICROS"))
  write_parquet(d, tmp, schema = sch2)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  sch3 <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "NANOS"))
  write_parquet(d, tmp, schema = sch3)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("TIME_MILLIS, TIME_MICROS", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  dt <- hms::hms(1, 2, 3)
  d <- data.frame(dt = dt)
  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS"))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })

  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MICROS"))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch)
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
  })
})

test_that("TIME, dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  dt <- hms::hms(1, 2, 3)
  d <- data.frame(dt = dt)
  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS"))
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch2 <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MICROS"))
  write_parquet(d, tmp, schema = sch2, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch3 <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "NANOS"))
  write_parquet(d, tmp, schema = sch3, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })
})

test_that("TIME_MILLIS, TIME_MICROS, dictionary", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  dt <- hms::hms(1, 2, 3)
  d <- data.frame(dt = dt)
  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS"))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })

  sch <- parquet_schema(list("TIME", is_adjusted_utc = TRUE, unit = "MICROS"))
  sch$logical_type[1] <- list(NULL)
  write_parquet(d, tmp, schema = sch, encoding = "RLE_DICTIONARY")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
  })
})

test_that("deprecated warnings", {
  pf <- test_path("data/alltypes_plain.parquet")
  expect_snapshot(
    sch <- parquet_schema(pf)
  )
  expect_equal(sch, read_parquet_schema(pf))

  expect_snapshot(
    pct <- parquet_column_types(mtcars)
  )

  expect_snapshot(
    mtd <- parquet_metadata(pf)
  )
  expect_equal(mtd, read_parquet_metadata(pf))

  expect_snapshot(
    info <- parquet_info(pf)
  )
  expect_equal(info, read_parquet_info(pf))
})

skip_without_parquet_cli <- function() {
  skip_on_cran()
  if (Sys.which("parquet") == "") {
    skip("parquet CLI not found on PATH")
  }
}

skip_without_cargo <- function() {
  skip_on_cran()
  if (Sys.which("cargo") == "") {
    skip("cargo not found on PATH")
  }
}

skip_without_pyarrow <- function() {
  skip_on_cran()
  if (tolower(Sys.getenv("_R_CHECK_FORCE_SUGGESTS_")) != "false") {
    return()
  }
  pyscript <- r"[
    import pyarrow
    import pyarrow.parquet as pq
  ]"
  pytmp <- tempfile(fileext = ".py")
  on.exit(unlink(pytmp), add = TRUE)
  writeLines(pyscript, pytmp)
  py <- if (Sys.which("python3") != "") "python3" else "python"
  res <- tryCatch(
    processx::run(py, pytmp, stderr = "2>&1"),
    error = function(err) err
  )
  if (inherits(res, "error")) {
    skip("missing pyarrow")
  }
}

skip_without_polars <- function() {
  skip_on_cran()
  if (tolower(Sys.getenv("_R_CHECK_FORCE_SUGGESTS_")) != "false") {
    return()
  }
  pyscript <- r"[
    import polars
  ]"
  pytmp <- tempfile(fileext = ".py")
  on.exit(unlink(pytmp), add = TRUE)
  writeLines(pyscript, pytmp)
  py <- if (Sys.which("python3") != "") "python3" else "python"
  res <- tryCatch(
    processx::run(py, pytmp, stderr = "2>&1"),
    error = function(err) err
  )
  if (inherits(res, "error")) {
    skip("missing polars in Python")
  }
}

skip_without <- function(pkgs) {
  if (any(c("arrow", "duckdb") %in% pkgs)) {
    skip_on_cran()
  }
  if (
    "duckdb" %in%
      pkgs &&
      getRversion() < "4.2.0" &&
      .Platform$OS.type == "windows"
  ) {
    skip("duckdb requires R 4.2.0 on Windows")
  }
  if (tolower(Sys.getenv("_R_CHECK_FORCE_SUGGESTS_")) != "false") {
    return()
  }
  ok <- vapply(pkgs, requireNamespace, logical(1), quietly = TRUE)
  if (any(!ok)) {
    skip(paste0("missing ", paste(pkgs[!ok], collapse = ", ")))
  }
}

test_df <- function(tibble = FALSE, factor = FALSE, missing = FALSE) {
  df <- cbind(nam = rownames(mtcars), mtcars)
  df$cyl <- as.integer(df$cyl)
  df$large <- df$cyl >= 6
  if (factor) {
    df$fac <- as.factor(tolower(substr(df$nam, 1, 1)))
  }
  rownames(df) <- NULL

  if (missing) {
    for (i in seq_len(ncol(df))) {
      if (i <= nrow(df)) {
        df[i, i] <- NA
      } else {
        df[1, i] <- NA
      }
    }
  }

  if (tibble) {
    class(df) <- c("tbl_df", "tbl", "data.frame")
  } else {
    class(df) <- c("tbl", "data.frame")
  }
  df
}

test_write <- function(d, schema = NULL, encoding = NULL) {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema1 <- if (!is.null(schema)) parquet_schema(schema)
  write_parquet(d, tmp, schema = schema1, encoding = encoding)

  expect_snapshot({
    schema
    encoding
    read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    as.data.frame(read_parquet(tmp))
  })
}

rscript <- function() {
  if (.Platform$OS.type == "windows") {
    file.path(R.home("bin"), "Rscript.exe")
  } else {
    file.path(R.home("bin"), "Rscript")
  }
}

redact_maxint64 <- function(x) {
  gsub("922337203685477[0-9][0-9][0-9][0-9]", "922337203685477xxxx", x)
}

utcts <- function(x) {
  as.POSIXct(as.POSIXlt(as.Date(x), tz = "UTC"))
}

make_nested_list_parquet <- function(filename, depth, rows = NULL, ...) {
  # Write a Parquet file with a single column 'a' that is a list nested to
  # `depth` levels (depth=1 -> list<int32>, depth=2 -> list<list<int32>>).
  #
  # Args:
  #   filename  output .parquet path
  #   depth     nesting depth, must be >= 1
  #   rows      optional row data; if NULL a small default example is used.
  #             Must be an R list of rows, each nested to `depth` levels with
  #             integer values at the leaves.

  if (depth < 1L) {
    stop("depth must be >= 1")
  }

  # Build the Arrow type recursively: list_of(list_of(...(int32())...))
  make_type <- function(d) {
    if (d == 0L) arrow::int32() else arrow::list_of(make_type(d - 1L))
  }

  # Default data: 3 rows that exercise empty lists at every level.
  # At depth 1: list(1:3, integer(0), 4L)
  # At depth d: wraps depth d-1 rows as [normal+empty, empty, singleton].
  make_rows <- function(d) {
    if (d == 1L) {
      list(1:3, integer(0), 4L)
    } else {
      inner <- make_rows(d - 1L)
      list(
        list(inner[[1]], inner[[2]]),
        list(),
        list(inner[[3]])
      )
    }
  }

  if (is.null(rows)) {
    rows <- make_rows(depth)
  }

  arr <- arrow::Array$create(rows, type = make_type(depth))
  arrow::write_parquet(arrow::arrow_table(a = arr), filename, ...)
}

# Write a Parquet file with a single list<int32> column 'a', with controllable
# repetition types for the outer list and its elements.
#
# list_nullable    = TRUE  -> outer list field is OPTIONAL (may be NULL)
#                  = FALSE -> outer list field is REQUIRED (never NULL)
# element_nullable = TRUE  -> list elements are OPTIONAL (may be NULL)
#                  = FALSE -> list elements are REQUIRED (never NULL)
#
# rows: optional data; must be an R list of integer vectors (no NAs when
#       element_nullable = FALSE, no NULLs when list_nullable = FALSE).
make_list_parquet <- function(
  filename,
  list_nullable = TRUE,
  element_nullable = TRUE,
  rows = NULL
) {
  elem_field <- arrow::field(
    "item",
    arrow::int32(),
    nullable = element_nullable
  )
  list_type <- arrow::list_of(elem_field)

  if (is.null(rows)) {
    rows <- list(1:3, integer(0), 4L)
  }

  arr <- arrow::Array$create(rows, type = list_type)

  col_field <- arrow::field("a", list_type, nullable = list_nullable)
  tbl <- arrow::arrow_table(a = arr, schema = arrow::schema(col_field))

  arrow::write_parquet(tbl, filename)
}

read_parquet_duckdb <- function(file) {
  duckdb::sql_query(sprintf("FROM '%s'", file))
}

test_that("read a subset", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- as.data.frame(test_df(missing = TRUE))
  write_parquet(df, tmp)

  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 1:3)),
    df[, 1:3]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = c(1:3, 13))),
    df[, c(1:3, 13)]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 10:13)),
    df[, 10:13]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = c("cyl", "mpg", "nam"))),
    df[, c("cyl", "mpg", "nam")]
  )
  expect_equal(
    as.data.frame(read_parquet(
      tmp,
      col_select = c("am", "gear", "carb", "large")
    )),
    df[, c("am", "gear", "carb", "large")]
  )
})

test_that("read a subset, edge cases", {
  loadNamespace("pillar")
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- as.data.frame(test_df(missing = TRUE))
  write_parquet(df, tmp)

  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 1)),
    df[, 1, drop = FALSE]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 13)),
    df[, 13, drop = FALSE]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = "nam")),
    df[, "nam", drop = FALSE]
  )
  expect_snapshot({
    read_parquet(tmp, col_select = integer())
    read_parquet(tmp, col_select = character())
  })
})

test_that("read a subset, factor to test Arrow metadata", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- as.data.frame(test_df(missing = TRUE, factor = TRUE))
  write_parquet(df, tmp)
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 10:14)),
    df[, 10:14]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 14)),
    df[, 14, drop = FALSE]
  )
})

test_that("subset column order", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- as.data.frame(test_df(missing = TRUE, factor = TRUE))
  write_parquet(df, tmp)

  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = 3:1)),
    df[, 3:1]
  )
  expect_equal(
    as.data.frame(read_parquet(tmp, col_select = c("cyl", "mpg", "nam"))),
    df[, c("cyl", "mpg", "nam")]
  )
})

test_that("error if a column is requested multiple times", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  df <- as.data.frame(test_df(missing = TRUE, factor = TRUE))
  write_parquet(df, tmp)

  expect_snapshot(error = TRUE, {
    read_parquet(tmp, col_select = c(1, 1))
    read_parquet(tmp, col_select = c(3, 4, 5, 3))
    read_parquet(tmp, col_select = c(3:4, 4:3))
    read_parquet(tmp, col_select = "foo")
    read_parquet(tmp, col_select = c("foo", "bar"))
    read_parquet(tmp, col_select = c("nam", "nam"))
    read_parquet(tmp, col_select = c("cyl", "disp", "hp", "cyl"))
    read_parquet(tmp, col_select = c("cyl", "disp", "disp", "cyl"))
  })
})

test_that("class", {
  withr::local_options(nanoparquet.class = NULL)
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  write_parquet(test_df(), tmp)
  expect_equal(class(read_parquet(tmp)), c("tbl", "data.frame"))
  expect_equal(
    class(read_parquet(
      tmp,
      options = parquet_options(class = c("foo", "bar", "data.frame"))
    )),
    c("foo", "bar", "data.frame")
  )
  withr::local_options(nanoparquet.class = "foobar")
  expect_equal(class(read_parquet(tmp)), c("foobar", "data.frame"))
})

test_that("mixing RLE_DICTIONARY and PLAIN", {
  # https://github.com/r-lib/nanoparquet/issues/110
  pf <- test_path("data/mixed.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  tab <- read_parquet(pf)
  expect_equal(tab$x, rep(0:399, 6))
  expect_equal(tab$y, rep(0:399, 6))
  expect_equal(tab$s, as.character(rep(0:399, 6)))
  expect_equal(tab$f, rep(0:399, 6))
  expect_equal(tab$d, rep(0:399, 6))
  expect_equal(tab$i96, rep(utcts(sprintf('%d-01-01', 1800:2199)), 6))

  pf <- test_path("data/mixed2.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  tab <- read_parquet(pf)
  expect_equal(tab$x, rep(0:399, 6))
  expect_equal(tab$y, rep(0:399, 6))
  expect_equal(tab$s, as.character(rep(0:399, 6)))
  expect_equal(tab$f, rep(0:399, 6))
  expect_equal(tab$d, rep(0:399, 6))
  expect_equal(tab$i96, rep(utcts(sprintf('%d-01-01', 1800:2199)), 6))

  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/mixed-miss.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  d1 <- as.data.frame(read_parquet(pf))
  d2 <- as.data.frame(arrow::read_parquet(pf))
  expect_equal(d1[, 1:5], d2[, 1:5])
  # arrow does not read INT86 into a time stamp, so compare manually
  expect_equal(is.na(d1[, 6]), is.na(d2[, 6]))
  bs6 <- utcts(sprintf('%d-01-01', 1:2400))
  bs6[is.na(d1[, 6])] <- NA
  expect_equal(d1[, 6], bs6)
})

test_that("mixing RLE_DICTIONARY and PLAIN, DECIMAL", {
  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/decimal.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  t1 <- read_parquet(pf)
  t2 <- arrow::read_parquet(pf)
  expect_equal(
    as.data.frame(t1),
    as.data.frame(t2)
  )

  pf <- test_path("data/decimal2.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  t1 <- as.data.frame(read_parquet(pf))
  t2 <- as.data.frame(arrow::read_parquet(pf))
  expect_equal(t1[, 1], t2[, 1])
  expect_equal(t1[, 2], t2[, 2])
  expect_equal(t1[, 3], t2[, 3])
  expect_equal(t1[, 4], t2[, 4])
})

test_that("mixing RLE_DICTIONARY and PLAIN, BYTE_ARRAY", {
  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/binary.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  t1 <- as.data.frame(read_parquet(pf))
  t2 <- as.data.frame(arrow::read_parquet(pf))
  expect_equal(unclass(t1[, 1]), unclass(t2[, 1]), ignore_attr = TRUE)
  expect_equal(unclass(t1[, 2]), unclass(t2[, 2]), ignore_attr = TRUE)
  expect_s3_class(t1[, 1], "blob")
  expect_s3_class(t1[, 2], "blob")
  expect_identical(attr(t1[, 1], "ptype"), raw())
  expect_identical(attr(t1[, 2], "ptype"), raw())
})

test_that("mixing RLE_DICTIONARY and PLAIN, FLOAT16", {
  skip_on_cran()
  skip_without("arrow")
  if (getRversion() < "4.2.0" && .Platform$OS.type == "windows") {
    skip("arrow fails here")
  }
  pf <- test_path("data/float16.parquet")
  expect_snapshot({
    as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    as.data.frame(read_parquet_pages(pf)[, c(
      "page_type",
      "num_values",
      "encoding"
    )])
  })
  t1 <- as.data.frame(read_parquet(pf))
  t2 <- as.data.frame(arrow::read_parquet(pf))
  # arrow is buggy, even the missingness pattern is wrong :(
  expect_equal(t1[, 1], rep(0:399, 3))
  expect_equal(
    which(is.na(t1[, 2])),
    c(30, 66, 422, 568, 878, 947, 988, 1006, 1170, 1183) + 1
  )
  bs2 <- rep(0:399, 3)
  bs2[is.na(t1[, 2])] <- NA
  expect_equal(t1[, 2], bs2)
})

# https://github.com/r-lib/nanoparquet/issues/132
test_that("dict page w/o dict offset set", {
  pf <- test_path("data/broken/polars-no-dict-offset.parquet")
  expect_equal(
    as.data.frame(read_parquet(pf)),
    data.frame(a = c(1, 2, 3), b = c(4, 5, 6))
  )
})

# -------------------------------------------------------------------------
# Primitive types

test_that("BOOLEAN", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(i = rep(c(TRUE, FALSE), 3))
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(1, 3, 5)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })

  # longer
  d <- data.frame(i = rep(c(TRUE, FALSE), 100))
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    read_parquet(tmp)[["i"]]
  })

  d[["i"]][c(1, 3, 5, 50, 99)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    read_parquet(tmp)[["i"]]
  })
})

test_that("INT32", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(i = 1:10)
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(2, 6, 10)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(i = rep(1:2, 20))
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(1, 5, 10, 20, 39)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })
})

test_that("INT64", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema("INT64")

  d <- data.frame(i = 1:10)
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(2, 6, 10)] <- NA
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(i = rep(1:2, 20))
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(1, 5, 10, 20, 39)] <- NA
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })
})

# TODO: INT96, once we have a way to not convert it to POSIXct

test_that("FLOAT", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)
  schema <- parquet_schema("FLOAT")

  d <- data.frame(i = 1:10 / 2)
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(2, 6, 10)] <- NA
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(i = rep(1:2, 20) / 2)
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(1, 5, 10, 20, 39)] <- NA
  write_parquet(d, tmp, compression = "uncompressed", schema = schema)
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })
})

test_that("FLOAT from multiple row groups and pages", {
  skip_on_cran()
  skip_without("arrow")
  pf <- test_path("data/float.parquet")
  expect_equal(
    as.data.frame(arrow::read_parquet(pf)),
    as.data.frame(read_parquet(pf))
  )
})

test_that("DOUBLE", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(i = 1:10 / 2)
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(2, 6, 10)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(i = rep(1:2, 20) / 2)
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })

  d[["i"]][c(1, 5, 10, 20, 39)] <- NA
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet(tmp))
  })
})

test_that("BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    l = I(list(
      as.raw(1:10),
      as.raw(11:20),
      as.raw(1:3),
      as.raw(1:1)
    ))
  )
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # missing
  d[["l"]][3] <- list(NULL)
  write_parquet(d, tmp, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(s = c("foo", "bar", "foobar"))
  write_parquet(
    d,
    tmp,
    compression = "uncompressed",
    encoding = "RLE_DICTIONARY"
  )
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # missing
  d <- data.frame(s = c("foo", "bar", NA, "foobar"))
  write_parquet(
    d,
    tmp,
    compression = "uncompressed",
    encoding = "RLE_DICTIONARY"
  )
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })
})

test_that("FIXED_LEN_BYTE_ARRAY", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- data.frame(
    l = I(list(
      as.raw(1:10),
      as.raw(11:20),
      as.raw(1:10),
      as.raw(21:30)
    ))
  )
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 10))
  write_parquet(d, tmp, schema = schema, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # missing
  d[["l"]][3] <- list(NULL)
  write_parquet(d, tmp, schema = schema, compression = "uncompressed")
  expect_snapshot({
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # dict
  d <- data.frame(s = c("foo", "bar", "aaa"))
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))
  write_parquet(
    d,
    tmp,
    schema = schema,
    compression = "uncompressed",
    encoding = "RLE_DICTIONARY"
  )
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })

  # missing
  d <- data.frame(s = c("foo", "bar", NA, "aaa"))
  schema <- parquet_schema(list("FIXED_LEN_BYTE_ARRAY", type_length = 3))
  write_parquet(
    d,
    tmp,
    schema = schema,
    compression = "uncompressed",
    encoding = "RLE_DICTIONARY"
  )
  expect_snapshot({
    as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    as.data.frame(read_parquet_schema(tmp))[, -1]
    as.data.frame(read_parquet(tmp))
  })
})

# -------------------------------------------------------------------------

test_that("compression", {
  tmp <- tempfile(fileext = ".parquet")
  on.exit(unlink(tmp), add = TRUE)

  d <- test_df(missing = TRUE)
  d2 <- test_df(missing = TRUE)[, 1:11]
  v2 <- parquet_options(write_data_page_version = 2)
  for (comp in c("uncompressed", "snappy", "gzip", "zstd")) {
    write_parquet(d, tmp, compression = comp, encoding = "PLAIN")
    expect_equal(read_parquet(tmp), d, info = comp)
    write_parquet(d2, tmp, compression = comp, encoding = "RLE_DICTIONARY")
    expect_equal(read_parquet(tmp), d2, info = comp)

    write_parquet(d, tmp, compression = comp, encoding = "PLAIN", options = v2)
    expect_equal(read_parquet(tmp), d, info = comp)
    write_parquet(
      d2,
      tmp,
      compression = comp,
      encoding = "RLE_DICTIONARY",
      options = v2
    )
    expect_equal(read_parquet(tmp), d2, info = comp)
  }
})

#' @useDynLib nanoparquet, .registration=TRUE
#' @details
#' ```{r include = FALSE}
#' lines <- readLines("README.md")
#' end <- which(lines == "<!-- badges: end -->")
#' lines <- lines[-(1:end)]
#' readme <- tempfile()
#' writeLines(lines, readme)
#' ```
#' ```{r child = readme}
#' ```
"_PACKAGE"

#' @name nanoparquet-types
#' @title nanoparquet's type maps
#' @description
#' How nanoparquet maps R types to Parquet types.
#'
#' @details
#' ```{r, child = "tools/types.Rmd"}
#' ```
#' @seealso [nanoparquet-package] for options that modify the type
#' mappings.
NULL

#' @name parquet-encodings
#' @title Parquet encodings
#' @description
#' Various Parquet encodings
#'
#' @details
#' ```{r, child = "tools/encodings.Rmd"}
#' ```
#' @seealso [write_parquet()] on how to select a non-default encoding when
#' writing Parquet files.
NULL

# R's data types

When writing out a data frame, nanoparquet maps R's data types to Parquet
logical types. The following table is a summary of the mapping. For the
details see below.

R type           | Parquet type            | Default | Notes
:----------------|:------------------------|:-------:|:----------------------------------------
bit64::integer64 | INT64                   | x       | `NA_integer64_` marks missing values.
blob::blob       | BYTE_ARRAY              | x       | Missing values are `NULL`.
"                | FIXED_LEN_BYTE_ARRAY    |         | All entries must have the same length. Missing values are `NULL`.
character        | STRING(BYTE_ARRAY)      | x       | I.e. STRSXP. Converted to UTF-8.
"                | BYTE_ARRAY              |         |
"                | FIXED_LEN_BYTE_ARRAY    |         |
"                | ENUM                    |         |
"                | UUID                    |         |
Date             | DATE                    | x       |
difftime         | INT64                   | x       | If not hms::hms. Arrow metadata marks it as Duration(NS).
factor           | STRING                  | x       | Arrow metadata marks it as a factor.
"                | ENUM                    |         |
hms::hms         | TIME(true, MILLIS)      | x       | Sub-milliseconds precision is lost.
integer          | INT(32, true)           | x       | I.e. INTSXP.
"                | INT64                   |         |
"                | INT96                   |         |
"                | DECIMAL(INT32)          |         |
"                | DECIMAL(INT64)          |         |
"                | INT(8, *)               |         |
"                | INT(16, *)              |         |
"                | INT(32, signed)         |         |
list             | LIST(INT32 elements)    | x       | List of integer vectors. `NULL` entries and `NA` elements are supported.
"                | LIST(DOUBLE elements)   | x       | List of double vectors. `NULL` entries and `NA` elements are supported.
"                | LIST(STRING elements)   | x       | List of character vectors. `NULL` entries and `NA` elements are supported.
"                | BYTE_ARRAY              |         | Must be a list of raw vectors. Missing values are `NULL`.
"                | FIXED_LEN_BYTE_ARRAY    |         | Must be a list of raw vectors of the same length. Missing values are `NULL`.
logical          | BOOLEAN                 | x       | I.e. LGLSXP.
numeric          | DOUBLE                  | x       | I.e. REALSXP.
"                | INT96                   |         |
"                | FLOAT                   |         |
"                | DECIMAL(INT32)          |         |
"                | DECIMAL(INT64)          |         |
"                | INT(*, *)               |         |
"                | FLOAT16                 |         |
POSIXct          | TIMESTAMP(true, MICROS) | x       | Sub-microsecond precision is lost.

The non-default mappings can be selected via the `schema` argument. E.g.
to write out a factor column called 'name' as `ENUM`, use
```r
write_parquet(..., schema = parquet_schema(name = "ENUM"))
```

The detailed mapping rules are listed below, in order of preference.
These rules will likely change until nanoparquet reaches version 1.0.0.

1. `bit64::integer64` objects (from the bit64 package) are written as
   `INT64`. nanoparquet handles any object that inherits the `integer64`
   class this way. `NA_integer64_` (i.e. `INT64_MIN`) marks missing values.
1. `blob::blob` objects (from the blob package) are written as `BYTE_ARRAY`.
   `blob::blob` is a list of raw vectors, and nanoparquet handles any object
   that inherits the `blob` class this way, even if the blob package is not
   installed. Missing values (i.e. `NULL` list entries) are supported.
1. Factors (i.e. vectors that inherit the *factor* class) are converted
   to character vectors using `as.character()`, then written as a
   `STRSXP` (character vector) type. The fact that a column is a factor
   is stored in the Arrow metadata (see below), unless the
   `nanoparquet.write_arrow_metadata` option is set to `FALSE`.
1. Dates (i.e. the `Date` class) is written as `DATE` logical type, which
   is an `INT32` type internally.
1. `hms` objects (from the hms package) are written as `TIME(true, MILLIS)`.
   logical type, which is internally the `INT32` Parquet type.
   Sub-milliseconds precision is lost.
1. `POSIXct` objects are written as `TIMESTAMP(true, MICROS)` logical type,
   which is internally the `INT64` Parquet type.
   Sub-microsecond precision is lost.
1. `difftime` objects (that are not `hms` objects, see above), are
   written as an `INT64` Parquet type, and noting in the Arrow metadata
   (see below) that this column has type `Duration` with `NANOSECONDS`
   unit.
1. Integer vectors (`INTSXP`) are written as `INT(32, true)` logical type,
   which corresponds to the `INT32` type.
1. Real vectors (`REALSXP`) are written as the `DOUBLE` type.
1. Character vectors (`STRSXP`) are written as the `STRING` logical type,
   which has the `BYTE_ARRAY` type. They are always converted to UTF-8
   before writing.
1. Logical vectors (`LGLSXP`) are written as the `BOOLEAN` type.
1. Other vectors error currently.

You can use [infer_parquet_schema()] on a data frame to map R data types
to Parquet data types.

To change the default R to Parquet mapping, use [parquet_schema()] and
the `schema` argument of [write_parquet()]. Currently supported
non-default mappings are:

- `integer` to `INT64`,
- `integer` to `INT96`,
- `double` to `INT96`,
- `double` to `FLOAT`,
- `character` to `BYTE_ARRAY`,
- `character` to `FIXED_LEN_BYTE_ARRAY`,
- `character` to `ENUM`,
- `factor` to `ENUM`,
- `integer` to `DECIAML` & `INT32`,
- `integer` to `DECIAML` & `INT64`,
- `double` to `DECIAML` & `INT32`,
- `double` to `DECIAML` & `INT64`,
- `integer` to `INT(8, *)`, `INT(16, *)`, `INT(32, signed)`,
- `double` to `INT(*, *)`,
- `character` to `UUID`,
- `double` to `FLOAT16`,
- `list` of `integer` vectors to `LIST` with `INT32` elements,
- `list` of `double` vectors to `LIST` with `DOUBLE` elements,
- `list` of `character` vectors to `LIST` with `STRING` elements,
- `list` of `raw` vectors to `BYTE_ARRAY`,
- `list` of `raw` vectors to `FIXED_LEN_BYTE_ARRAY`,
- `blob::blob` to `FIXED_LEN_BYTE_ARRAY`.

# Parquet's data types

When reading a Parquet file nanoparquet also relies on logical types and
the Arrow metadata (if present, see below) in addition to the low level
data types. The following table summarizes the mappings. See more details
below.

Parquet type         | R type     | Notes
:--------------------|:-----------|:---------------------------------------------
*Logical types*      |            |
BSON                 | character  |
DATE                 | Date       |
DECIMAL              | numeric    | REALSXP, potentially losing precision.
ENUM                 | character  |
FLOAT16              | numeric    | REALSXP
INT(8, *)            | integer    |
INT(16, *)           | integer    |
INT(32, *)           | integer    | Large unsigned values may overflow!
INT(64, *)           | numeric    | REALSXP, or `integer64` if `read_int64_type` option is set.
INTERVAL             | list(raw)  | Missing values are `NULL`.
JSON                 | character  |
LIST                 | list       | Elements are read as their corresponding R type.
MAP                  |            | Not supported.
STRING               | factor     | If Arrow metadata says it is a factor. Also UTF8.
"                    | character  | Otherwise. Also UTF8.
TIME                 | hms::hms   | Also TIME_MILLIS and TIME_MICROS.
TIMESTAMP            | POSIXct    | Also TIMESTAMP_MILLIS and TIMESTAMP_MICROS.
UUID                 | character  | In `00112233-4455-6677-8899-aabbccddeeff` form.
UNKNOWN              |            | Not supported.
*Primitive types*    |            |
BOOLEAN              | logical    |
BYTE_ARRAY           | factor     | If Arrow metadata says it is a factor.
"                    | blob::blob | Otherwise. Missing values are `NULL`.
DOUBLE               | numeric    | REALSXP
FIXED_LEN_BYTE_ARRAY | blob::blob | Missing values are `NULL`.
FLOAT                | numeric    | REALSXP
INT32                | integer    |
INT64                | numeric    | REALSXP, or `integer64` if `read_int64_type` option is set.
INT96                | POSIXct    |

The exact rules are below. These rules will likely change until nanoparquet
reaches version 1.0.0.

1. The `BOOLEAN` type is read as a logical vector (`LGLSXP`).
1. The `STRING` logical type and the `UTF8` converted type is read as
   a character vector with UTF-8 encoding.
1. The `DATE` logical type and the `DATE` converted type are read as a
   `Date` R object.
1. The `TIME` logical type and the `TIME_MILLIS` and `TIME_MICROS`
   converted types are read as an `hms` object, see the hms package.
1. The `TIMESTAMP` logical type and the `TIMESTAMP_MILLIS` and
   `TIMESTAMP_MICROS` converted types are read as `POSIXct` objects.
   If the logical type has the `UTC` flag set, then the time zone of the
   `POSIXct` object is set to `UTC`.
1. `INT32` is read as an integer vector (`INTSXP`).
1. `INT64` is read as a real vector (`REALSXP`) by default. If the
   `read_int64_type` option in `parquet_options()` is set to `"integer64"`
   or `"bit64::integer64"`, it is read as a `bit64::integer64` vector
   instead. `NA_integer64_` (i.e. `INT64_MIN`) marks missing values.
1. `DOUBLE` and `FLOAT` are read as real vectors (`REALSXP`).
1. `INT96` is read as a `POSIXct` read vector with the `tzone` attribute
   set to `"UTC"`. It was an old convention to store time stamps as
   `INT96` objects.
1. The `DECIMAL` converted type (`FIXED_LEN_BYTE_ARRAY` or `BYTE_ARRAY`
   type) is read as a real vector (`REALSXP`), potentially losing
   precision.
1. The `ENUM` logical type is read as a character vector.
1. The `UUID` logical type is read as a character vector that uses the
   `00112233-4455-6677-8899-aabbccddeeff` form.
1. The `FLOAT16` logical type is read as a real vector (`REALSXP`).
1. `BYTE_ARRAY` is read as a *factor* object if the file was written
   by Arrow and the original data type of the column was a factor.
   (See 'The Arrow metadata below.)
1. Otherwise `BYTE_ARRAY` and `FIXED_LEN_BYTE_ARRAY` are read as a
   `blob::blob` object. `blob::blob` is a list of raw vectors with class
   `blob` (and related vctrs classes). The blob package is not required to
   use this object; it is simply a list of raw vectors. Missing values are
   denoted by `NULL`.

Other logical and converted types are read as their annotated low level
types:

1. `INT(8, true)`, `INT(16, true)` and `INT(32, true)` are read as
   integer vectors because they are `INT32` internally in Parquet.
1. `INT(64, true)` is read as a real vector (`REALSXP`), unless the
   `read_int64_type` option is set (see above).
1. Unsigned integer types `INT(8, false)`, `INT(16, false)` and
   `INT(32, false)` are read as integer vectors (`INTSXP`). Large
   positive values may overflow into negative values, this is a known
   issue that we will fix.
1. `INT(64, false)` is read as a real vector (`REALSXP`), unless the
   `read_int64_type` option is set (see above). Large positive values may
   overflow into negative values, this is a known issue that we will fix.
1. `INTERVAL` is a fixed length byte array, and nanoparquet reads it as
   a list of raw vectors. Missing values are denoted by `NULL`.
1. `JSON` columns are read as character vectors (`STRSXP`).
1. `BSON` columns are read as raw vectors (`RAWSXP`).

These types are not yet supported:

1. Nested `LIST` types (lists of lists) are not supported.
1. The `MAP` logical type is not supported.
1. The `UNKNOWN` logical type is not supported.

You can use the [read_parquet_schema()] function to see how R would read
the columns of a Parquet file. Look at the `r_type` column.

# The Arrow metadata

Apache Arrow (i.e. the arrow R package) adds additional metadata to
Parquet files when writing them in `arrow::write_parquet()`. Then,
when reading the file in `arrow::read_parquet()`, it uses this metadata
to recreate the same Arrow and R data types as before writing.

`nanoparquet::write_parquet()` also adds the Arrow metadata to Parquet
files, unless the `nanoparquet.write_arrow_metadata` option is set to
`FALSE`.

Similarly, `nanoparquet::read_parquet()` uses the Arrow metadata in the
Parquet file (if present), unless the `nanoparquet.use_arrow_metadata`
option is set to FALSE.

The Arrow metadata is stored in the file level key-value metadata, with
key `ARROW:schema`.

Currently nanoparquet uses the Arrow metadata for two things:
- It uses it to detect factors. Without the Arrow metadata factors are
  read as string vectors.
- It uses it to detect `difftime` objects. Without the arrow metadata
  these are read as `INT64` columns, containing the time difference in
  nanoseconds.

# write_parquet -> INT32

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   2
      3   3
      4   4
      5   5
      6   6
      7   7
      8   8
      9   9
      10 10
      11 NA

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1   NA
      2    1
      3    2
      4    1
      5    2
      6    1
      7    2
      8    1
      9    2
      10   1
      11   2
      12   1
      13   2
      14   1
      15   2
      16   1
      17   2
      18   1
      19   2
      20   1
      21   2
      22   1
      23   2
      24   1
      25   2
      26   1
      27   2
      28   1
      29   2
      30   1
      31   2
      32   1
      33   2
      34   1
      35   2
      36   1
      37   2
      38   1
      39   2
      40   1
      41   2
      42   1
      43   2
      44   1
      45   2
      46   1
      47   2
      48   1
      49   2
      50   1
      51   2
      52   1
      53   2
      54   1
      55   2
      56   1
      57   2
      58   1
      59   2
      60   1
      61   2
      62   1
      63   2
      64   1
      65   2
      66   1
      67   2
      68   1
      69   2
      70   1
      71   2
      72   1
      73   2
      74   1
      75   2
      76   1
      77   2
      78   1
      79   2
      80   1
      81   2
      82   1
      83   2
      84   1
      85   2
      86   1
      87   2
      88   1
      89   2
      90   1
      91   2
      92   1
      93   2
      94   1
      95   2
      96   1
      97   2
      98   1
      99   2
      100  1
      101  2

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT32"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet INT32 type.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT32"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a character vector as a Parquet INT32 type.

# write_parquet -> INT64

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  NA
      2   1
      3   2
      4   3
      5   4
      6   5
      7   6
      8   7
      9   8
      10  9
      11 10

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1   NA
      2    1
      3    2
      4    1
      5    2
      6    1
      7    2
      8    1
      9    2
      10   1
      11   2
      12   1
      13   2
      14   1
      15   2
      16   1
      17   2
      18   1
      19   2
      20   1
      21   2
      22   1
      23   2
      24   1
      25   2
      26   1
      27   2
      28   1
      29   2
      30   1
      31   2
      32   1
      33   2
      34   1
      35   2
      36   1
      37   2
      38   1
      39   2
      40   1
      41   2
      42   1
      43   2
      44   1
      45   2
      46   1
      47   2
      48   1
      49   2
      50   1
      51   2
      52   1
      53   2
      54   1
      55   2
      56   1
      57   2
      58   1
      59   2
      60   1
      61   2
      62   1
      63   2
      64   1
      65   2
      66   1
      67   2
      68   1
      69   2
      70   1
      71   2
      72   1
      73   2
      74   1
      75   2
      76   1
      77   2
      78   1
      79   2
      80   1
      81   2
      82   1
      83   2
      84   1
      85   2
      86   1
      87   2
      88   1
      89   2
      90   1
      91   2
      92   1
      93   2
      94   1
      95   2
      96   1
      97   2
      98   1
      99   2
      100  1
      101  2

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  NA
      2   1
      3   2
      4   3
      5   4
      6   5
      7   6
      8   7
      9   8
      10  9
      11 10

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1   NA
      2    1
      3    2
      4    1
      5    2
      6    1
      7    2
      8    1
      9    2
      10   1
      11   2
      12   1
      13   2
      14   1
      15   2
      16   1
      17   2
      18   1
      19   2
      20   1
      21   2
      22   1
      23   2
      24   1
      25   2
      26   1
      27   2
      28   1
      29   2
      30   1
      31   2
      32   1
      33   2
      34   1
      35   2
      36   1
      37   2
      38   1
      39   2
      40   1
      41   2
      42   1
      43   2
      44   1
      45   2
      46   1
      47   2
      48   1
      49   2
      50   1
      51   2
      52   1
      53   2
      54   1
      55   2
      56   1
      57   2
      58   1
      59   2
      60   1
      61   2
      62   1
      63   2
      64   1
      65   2
      66   1
      67   2
      68   1
      69   2
      70   1
      71   2
      72   1
      73   2
      74   1
      75   2
      76   1
      77   2
      78   1
      79   2
      80   1
      81   2
      82   1
      83   2
      84   1
      85   2
      86   1
      87   2
      88   1
      89   2
      90   1
      91   2
      92   1
      93   2
      94   1
      95   2
      96   1
      97   2
      98   1
      99   2
      100  1
      101  2

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT64"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet INT64 type.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT64"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a character vector as a Parquet INT64 type.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT64"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a list as a Parquet INT64 type.

# write_parquet -> INT96

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d POSIXct INT96          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_page(tmp, 4L)$data
    Output
        [1] 04 00 00 00 16 01 03 00 fb ff ff ff ff ff ff ff ff ff ff ff fc ff ff ff ff
       [26] ff ff ff ff ff ff ff fd ff ff ff ff ff ff ff ff ff ff ff fe ff ff ff ff ff
       [51] ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff 00 00 00 00 00 00 00
       [76] 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 02 00 00 00 00 00 00 00
      [101] 00 00 00 00 03 00 00 00 00 00 00 00 00 00 00 00 04 00 00 00 00 00 00 00 00
      [126] 00 00 00 05 00 00 00 00 00 00 00 00 00 00 00

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d POSIXct INT96          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_page(tmp, 4L)$data
    Output
       [1] ff ff ff ff ff ff ff ff ff ff ff ff 00 00 00 00 00 00 00 00 00 00 00 00 01
      [26] 00 00 00 00 00 00 00 00 00 00 00

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d POSIXct INT96          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_page(tmp, 4L)$data
    Output
        [1] 04 00 00 00 16 01 03 00 fb ff ff ff ff ff ff ff ff ff ff ff fc ff ff ff ff
       [26] ff ff ff ff ff ff ff fd ff ff ff ff ff ff ff ff ff ff ff fe ff ff ff ff ff
       [51] ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff ff 00 00 00 00 00 00 00
       [76] 00 00 00 00 00 01 00 00 00 00 00 00 00 00 00 00 00 02 00 00 00 00 00 00 00
      [101] 00 00 00 00 03 00 00 00 00 00 00 00 00 00 00 00 04 00 00 00 00 00 00 00 00
      [126] 00 00 00 05 00 00 00 00 00 00 00 00 00 00 00

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d POSIXct INT96          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_page(tmp, 4L)$data
    Output
       [1] ff ff ff ff ff ff ff ff ff ff ff ff 00 00 00 00 00 00 00 00 00 00 00 00 01
      [26] 00 00 00 00 00 00 00 00 00 00 00

# write_parquet -> FLOAT

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double FLOAT          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double FLOAT          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -1
      2   0
      3   1
      4  -1
      5   0
      6   1
      7  -1
      8   0
      9   1
      10 -1
      11  0
      12  1
      13 -1
      14  0
      15  1
      16 -1
      17  0
      18  1
      19 -1
      20  0
      21  1
      22 -1
      23  0
      24  1
      25 -1
      26  0
      27  1
      28 -1
      29  0
      30  1
      31 -1
      32  0
      33  1
      34 -1
      35  0
      36  1
      37 -1
      38  0
      39  1
      40 -1
      41  0
      42  1
      43 -1
      44  0
      45  1
      46 -1
      47  0
      48  1
      49 -1
      50  0
      51  1
      52 -1
      53  0
      54  1
      55 -1
      56  0
      57  1
      58 -1
      59  0
      60  1
      61 -1
      62  0
      63  1
      64 -1
      65  0
      66  1
      67 -1
      68  0
      69  1
      70 -1
      71  0
      72  1
      73 -1
      74  0
      75  1
      76 -1
      77  0
      78  1
      79 -1
      80  0
      81  1
      82 -1
      83  0
      84  1
      85 -1
      86  0
      87  1
      88 -1
      89  0
      90  1
      91 NA

# write_parquet -> BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             d
      1             66, 6f, 6f
      2             62, 61, 72
      3 66, 6f, 6f, 62, 61, 72
      4                   NULL

# write_parquet -> FIXED_LEN_BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      d   blob FIXED_LEN_BYTE_ARRAY           3        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 d
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61
      4       NULL

# write_parquet -> STRING

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name    r_type       type type_length repetition_type converted_type
      1    NA schema      <NA>       <NA>          NA            <NA>           <NA>
      2     1      d character BYTE_ARRAY          NA        OPTIONAL           UTF8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2       STRING           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             d
      1    foo
      2    bar
      3 foobar
      4   <NA>

# write_parquet -> ENUM

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name    r_type       type type_length repetition_type converted_type
      1    NA schema      <NA>       <NA>          NA            <NA>           <NA>
      2     1      d character BYTE_ARRAY          NA        OPTIONAL           ENUM
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2         ENUM           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             d
      1    foo
      2    bar
      3 foobar
      4   <NA>

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name    r_type       type type_length repetition_type converted_type
      1    NA schema    factor       <NA>          NA            <NA>           <NA>
      2     1      d character BYTE_ARRAY          NA        OPTIONAL           ENUM
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2         ENUM           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             d
      1    foo
      2    bar
      3 foobar
      4   <NA>

# write_parquet -> DECIMAL INT32

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     0         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 3, scale 0: 1000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 3, scale 0: -1000

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     2         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 3, scale 2: 10

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 3, scale 2: -10

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     1         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1  -2.5
      2  -2.0
      3  -1.5
      4  -1.0
      5  -0.5
      6   0.0
      7   0.5
      8   1.0
      9   1.5
      10  2.0
      11  2.5
      12   NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 3, scale 1: 100.000000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 3, scale 1: -100.000000

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     2         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1  -2.5
      2  -2.0
      3  -1.5
      4  -1.0
      5  -0.5
      6   0.0
      7   0.5
      8   1.0
      9   1.5
      10  2.0
      11  2.5
      12   NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 3, scale 2: 10.000000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 3, scale 2: -10.000000

# write_parquet -> DECIMAL INT64

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     0         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 3, scale 0: 1000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 3, scale 0: -1000

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     2         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 3, scale 2: 10

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 3, scale 2: -10

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     1         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1  -2.5
      2  -2.0
      3  -1.5
      4  -1.0
      5  -0.5
      6   0.0
      7   0.5
      8   1.0
      9   1.5
      10  2.0
      11  2.5
      12   NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 3, scale 1: 100.000000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 3, scale 1: -100.000000

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 DECIMAL,....           NA     2         3       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1  -2.5
      2  -2.0
      3  -1.5
      4  -1.0
      5  -0.5
      6   0.0
      7   0.5
      8   1.0
      9   1.5
      10  2.0
      11  2.5
      12   NA

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 3, scale 2: 10.000000

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 3, scale 2: -10.000000

# smaller integers

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL          INT_8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 8, TRUE           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 8: 128 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 8: -129 at column 1, row 3:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL         INT_16
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 16,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 16: 32768 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 16: -32769 at column 1, row 3:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL         UINT_8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 8, ....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for UINT with bit width 8: 256 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for UINT with bit width 8: -1 at column 1, row 2:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        UINT_16
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 16,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for UINT with bit width 16: 65536 at column 1, row 1:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for UINT with bit width 16: -1 at column 1, row 2:

# double to smaller integers

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL          INT_8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 8, TRUE           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 8: 128.000000 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 8: -129.000000 at column 1, row 3:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL         INT_16
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 16,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 16: 32768.000000 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 16: -32769.000000 at column 1, row 3:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL         INT_32
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 32,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_32"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 32: 2147483648.000000 at column 1, row 2:

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_32"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 32: -2147483649.000000 at column 1, row 2:

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL         UINT_8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 8, ....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 8: 256.000000 at column 1, row 2.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_8"))
    Condition
      Error in `write_parquet()`:
      ! Negative values are not allowed in unsigned INT column:-1.000000 at column 1, row 2.

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        UINT_16
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 16,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 16: 65536.000000 at column 1, row 2.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_16"))
    Condition
      Error in `write_parquet()`:
      ! Negative values are not allowed in unsigned INT column:-1.000000 at column 1, row 2.

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        OPTIONAL        UINT_32
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 32,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_32"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 32: 4294967296.000000 at column 1, row 2.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_32"))
    Condition
      Error in `write_parquet()`:
      ! Negative values are not allowed in unsigned INT column:-1.000000 at column 1, row 2.

# double to INT(64, *)

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL         INT_64
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 64,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1  -5
      2  -4
      3  -3
      4  -2
      5  -1
      6   0
      7   1
      8   2
      9   3
      10  4
      11  5
      12 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_64"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 64: 922337203685477xxxx.000000 at column 1, row 2.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT_64"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too small for INT with bit width 64: -922337203685477xxxx.000000 at column 1, row 2.

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      d double INT64          NA        OPTIONAL        UINT_64
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 INT, 64,....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  0
      2  1
      3  2
      4  3
      5  4
      6  5
      7 NA

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_64"))
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for unsigned INT with bit width 64: 18446744073709551616.000000 at column 1, row 2.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UINT_64"))
    Condition
      Error in `write_parquet()`:
      ! Negative values are not allowed in unsigned INT column:-1.000000 at column 1, row 2.

# integer64 round-trip and read_int64_type option

    Code
      as.data.frame(res_dbl)
    Output
                    x
      1  1.000000e+15
      2 -1.000000e+15
      3            NA
      4  2.147484e+09

# JSON

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        OPTIONAL           JSON
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2         JSON           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             d
      1    foo
      2    bar
      3 foobar
      4   <NA>

# UUID

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name    r_type                 type type_length repetition_type
      1    NA schema      <NA>                 <NA>          NA            <NA>
      2     1      d character FIXED_LEN_BYTE_ARRAY          16        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>         UUID           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           d
      1 00112233-4455-6677-8899-aabbccddeeff
      2 00112233-4455-6677-8899-aabbccddeeff
      3 00112233-4455-6677-8899-aabbccddeeff
      4                                 <NA>

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("UUID"))
    Condition
      Error in `write_parquet()`:
      ! Invalid UUID value in column 1, row 2

# FLOAT16

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      c   blob FIXED_LEN_BYTE_ARRAY           2        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>      FLOAT16           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
            c
      1   0.0
      2   1.0
      3   2.0
      4    NA
      5  -1.0
      6   NaN
      7  -2.0
      8  -Inf
      9   Inf
      10  0.5

# NaN

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type   type type_length repetition_type converted_type
      1    NA schema   <NA>   <NA>          NA            <NA>           <NA>
      2     1      d double DOUBLE          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   2
      3 NaN
      4   4
      5   5

# list of RAW to BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             d
      1             66, 6f, 6f
      2             62, 61, 72
      3 66, 6f, 6f, 62, 61, 72

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             d
      1             66, 6f, 6f
      2                   NULL
      3             62, 61, 72
      4 66, 6f, 6f, 62, 61, 72
      5                   NULL

# list of RAW to FIXED_LEN_BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      d   blob FIXED_LEN_BYTE_ARRAY           3        REQUIRED
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 d
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      d   blob FIXED_LEN_BYTE_ARRAY           3        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 d
      1 66, 6f, 6f
      2       NULL
      3 62, 61, 72
      4 61, 61, 61
      5       NULL

# blob::blob to BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                d
      1 blob[3 B]
      2 blob[3 B]
      3 blob[6 B]

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      d   blob BYTE_ARRAY          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                d
      1 blob[3 B]
      2      <NA>
      3 blob[3 B]
      4 blob[6 B]
      5      <NA>

# blob::blob to FIXED_LEN_BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      d   blob FIXED_LEN_BYTE_ARRAY           3        REQUIRED
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                d
      1 blob[3 B]
      2 blob[3 B]
      3 blob[3 B]

---

    Code
      as.data.frame(read_parquet_schema(tmp)[, -1])
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      d   blob FIXED_LEN_BYTE_ARRAY           3        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                d
      1 blob[3 B]
      2      <NA>
      3 blob[3 B]
      4 blob[3 B]
      5      <NA>

# parse_encoding

    Code
      names(mtcars)
    Output
       [1] "mpg"  "cyl"  "disp" "hp"   "drat" "wt"   "qsec" "vs"   "am"   "gear"
      [11] "carb"
    Code
      parse_encoding(NULL, mtcars)
    Output
       mpg  cyl disp   hp drat   wt qsec   vs   am gear carb 
        NA   NA   NA   NA   NA   NA   NA   NA   NA   NA   NA 
    Code
      parse_encoding("PLAIN", mtcars)
    Output
          mpg     cyl    disp      hp    drat      wt    qsec      vs      am    gear 
      "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" 
         carb 
      "PLAIN" 
    Code
      parse_encoding(c(disp = "RLE_DICTIONARY"), mtcars)
    Output
                   mpg              cyl             disp               hp 
                    NA               NA "RLE_DICTIONARY"               NA 
                  drat               wt             qsec               vs 
                    NA               NA               NA               NA 
                    am             gear             carb 
                    NA               NA               NA 
    Code
      parse_encoding(c(disp = "RLE_DICTIONARY", vs = "PLAIN"), mtcars)
    Output
                   mpg              cyl             disp               hp 
                    NA               NA "RLE_DICTIONARY"               NA 
                  drat               wt             qsec               vs 
                    NA               NA               NA          "PLAIN" 
                    am             gear             carb 
                    NA               NA               NA 
    Code
      parse_encoding(c(disp = "RLE", "PLAIN"), mtcars)
    Output
          mpg     cyl    disp      hp    drat      wt    qsec      vs      am    gear 
      "PLAIN" "PLAIN"   "RLE" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" 
         carb 
      "PLAIN" 
    Code
      parse_encoding(c(disp = "RLE", "PLAIN", vs = "PLAIN"), mtcars)
    Output
          mpg     cyl    disp      hp    drat      wt    qsec      vs      am    gear 
      "PLAIN" "PLAIN"   "RLE" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" "PLAIN" 
         carb 
      "PLAIN" 

---

    Code
      parse_encoding(1:2, mtcars)
    Condition
      Error in `parse_encoding()`:
      ! `encoding` must be `NULL` or a character vector
    Code
      parse_encoding(c("PLAIN", "foobar"), mtcars)
    Condition
      Error in `parse_encoding()`:
      ! `encoding` contains at least one unknown encoding
    Code
      parse_encoding(c(foo = "PLAIN", foo = "RLE"), mtcars)
    Condition
      Error in `parse_encoding()`:
      ! names of `encoding` must be unique
    Code
      parse_encoding(c(disp = "PLAIN", foo = "RLE"), mtcars)
    Condition
      Error in `parse_encoding()`:
      ! names of `encoding` must match names of `x`

# BOOLEAN

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE      RLE
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3    NA
      4  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3    NA
      4  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE      RLE
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2 FALSE
      3    NA
      4  TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE      RLE
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9  TRUE
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9  TRUE
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE      RLE
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9  TRUE
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9    NA
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE
      17 TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9    NA
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE
      17 TRUE

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE      RLE
    Code
      as.data.frame(read_parquet(tmp))
    Output
            l
      1  TRUE
      2  TRUE
      3  TRUE
      4  TRUE
      5  TRUE
      6  TRUE
      7  TRUE
      8  TRUE
      9    NA
      10 TRUE
      11 TRUE
      12 TRUE
      13 TRUE
      14 TRUE
      15 TRUE
      16 TRUE
      17 TRUE

---

    Code
      write_parquet(d, tmp, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BOOLEAN column: RLE_DICTIONARY
    Code
      write_parquet(d, tmp, encoding = "BIT_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BOOLEAN column: BIT_PACKED
    Code
      write_parquet(d, tmp, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for BOOLEAN column: BYTE_STREAM_SPLIT

# INT32

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      NULL
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# integer -> INT64

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# double -> INT64

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# integer -> INT96

    Code
      write_parquet(d, tmp, schema = schema)
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY

---

    Code
      write_parquet(d, tmp, schema = schema)
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY

---

    Code
      write_parquet(d, tmp, schema = schema)
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY

---

    Code
      write_parquet(d, tmp, schema = schema)
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "PLAIN")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE_DICTIONARY")
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT96 column: RLE

# double -> INT96

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3 -4713-11-24
      4 -4713-11-24
      5 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3 -4713-11-24
      4 -4713-11-24
      5 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3 -4713-11-24
      4 -4713-11-24
      5 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3        <NA>
      4 -4713-11-24
      5 -4713-11-24
      6 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3        <NA>
      4 -4713-11-24
      5 -4713-11-24
      6 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1 -4713-11-24
      2 -4713-11-24
      3        <NA>
      4 -4713-11-24
      5 -4713-11-24
      6 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6  -4713-11-24
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6  -4713-11-24
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6  -4713-11-24
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6         <NA>
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24
      11 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6         <NA>
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24
      11 -4713-11-24

---

    Code
      schema
    Output
      [1] "INT96"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                   d
      1  -4713-11-24
      2  -4713-11-24
      3  -4713-11-24
      4  -4713-11-24
      5  -4713-11-24
      6         <NA>
      7  -4713-11-24
      8  -4713-11-24
      9  -4713-11-24
      10 -4713-11-24
      11 -4713-11-24

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT96 column: RLE

# FLOAT

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      schema
    Output
      [1] "FLOAT"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for FLOAT column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for FLOAT column: RLE

# DOUBLE

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 1.5
      5 2.0
      6 2.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      schema
    Output
      [1] "DOUBLE"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6   NA
      7  0.5
      8  0.5
      9  0.5
      10 0.5
      11 0.5

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for DOUBLE column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for DOUBLE column: RLE

# BYTE_ARRAY, string

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo

---

    Code
      schema
    Output
      [1] "STRING"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
            d
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for BYTE_ARRAY column: RLE

# BYTE_ARRAY, RAW

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             s
      1             66, 6f, 6f
      2             62, 61, 72
      3 66, 6f, 6f, 62, 61, 72

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             s
      1             66, 6f, 6f
      2             62, 61, 72
      3 66, 6f, 6f, 62, 61, 72

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             s
      1             66, 6f, 6f
      2             62, 61, 72
      3                   NULL
      4 66, 6f, 6f, 62, 61, 72

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                             s
      1             66, 6f, 6f
      2             62, 61, 72
      3                   NULL
      4 66, 6f, 6f, 62, 61, 72

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [1] "BYTE_ARRAY"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for list(raw) BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for list(raw) BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for list(raw) BYTE_ARRAY column: RLE

# FIXED_LEN_BYTE_ARRAY, RAW

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for list(raw) FIXED_LEN_BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for list(raw) FIXED_LEN_BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "PLAIN_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for list(raw) FIXED_LEN_BYTE_ARRAY column: RLE_DICTIONARY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for list(raw) FIXED_LEN_BYTE_ARRAY column: RLE

# FIXED_LEN_BYTE_ARRAY, FLOAT16

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3 1.5
      4 2.0
      5 2.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 2.0
      5 2.5
      6 3.0

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 2.0
      5 2.5
      6 3.0

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 0.5
      2 1.0
      3  NA
      4 2.0
      5 2.5
      6 3.0

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5  0.5
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5   NA
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5   NA
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      schema
    Output
      [1] "FLOAT16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           d
      1  0.5
      2  0.5
      3  0.5
      4  0.5
      5   NA
      6  0.5
      7  0.5
      8  0.5
      9  0.5
      10 0.5

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for FIXED_LEN_BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for FIXED_LEN_BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for FIXED_LEN_BYTE_ARRAY column: RLE

# FIXED_LEN_BYTE_ARRAY, character

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5  66, 6f, 6f
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      schema
    Output
      [[1]]
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 3
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                  d
      1  66, 6f, 6f
      2  66, 6f, 6f
      3  66, 6f, 6f
      4  66, 6f, 6f
      5        NULL
      6  66, 6f, 6f
      7  66, 6f, 6f
      8  66, 6f, 6f
      9  66, 6f, 6f
      10 66, 6f, 6f

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for FIXED_LEN_BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for FIXED_LEN_BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for FIXED_LEN_BYTE_ARRAY column: RLE

# null_count is written

    Code
      as.data.frame(read_parquet_metadata(tmp)[["column_chunks"]][, c("row_group",
        "column", "null_count")])
    Output
         row_group column null_count
      1          0      0          1
      2          0      1          1
      3          0      2          1
      4          0      3          1
      5          0      4          1
      6          0      5          1
      7          0      6          1
      8          0      7          1
      9          0      8          1
      10         0      9          1
      11         0     10          0
      12         0     11          0
      13         0     12          0
      14         1      0          0
      15         1      1          0
      16         1      2          0
      17         1      3          0
      18         1      4          0
      19         1      5          0
      20         1      6          0
      21         1      7          0
      22         1      8          0
      23         1      9          0
      24         1     10          1
      25         1     11          1
      26         1     12          1
      27         2      0          0
      28         2      1          0
      29         2      2          0
      30         2      3          0
      31         2      4          0
      32         2      5          0
      33         2      6          0
      34         2      7          0
      35         2      8          0
      36         2      9          0
      37         2     10          0
      38         2     11          0
      39         2     12          0
      40         3      0          0
      41         3      1          0
      42         3      2          0
      43         3      3          0
      44         3      4          0
      45         3      5          0
      46         3      6          0
      47         3      7          0
      48         3      8          0
      49         3      9          0
      50         3     10          0
      51         3     11          0
      52         3     12          0

# min/max for integers

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for DATEs

    Code
      do()
    Output
      [[1]]
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1    day    Date INT32          NA        OPTIONAL           DATE
      3     2  count integer INT32          NA        REQUIRED         INT_32
        logical_type num_children scale precision field_id children
      1                         2    NA        NA       NA         
      2         DATE           NA    NA        NA       NA         
      3 INT, 32,....           NA    NA        NA       NA         
      
      [[2]]
      [1] "2024-09-06" "2024-09-08" "2024-09-10" "2024-09-12" "2024-09-14"
      
      [[3]]
      [1] "2024-09-07" "2024-09-09" "2024-09-11" "2024-09-13" "2024-09-15"
      

# min/max for double -> signed integers

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for double -> unsigned integers

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]  1  1  0 NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]  1  1  0 NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]  1  1  0 NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]  1  1  0 NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# minmax for double -> INT32 TIME(MULLIS)

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1000  -100000 -1000000       NA
      
      [[2]]
      [1]    5000  100000 1000000      NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1000  -100000 -1000000       NA
      
      [[2]]
      [1]    5000  100000 1000000      NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1000  -100000 -1000000       NA
      
      [[2]]
      [1]    5000  100000 1000000      NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1000  -100000 -1000000       NA
      
      [[2]]
      [1]    5000  100000 1000000      NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for DOUBLE

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for FLOAT

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for integer -> INT64

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for REALSXP -> INT64

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1]     1  -100 -1000    NA
      
      [[2]]
      [1]    5  100 1000   NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for STRING

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "snappy", type = "JSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed", type = "JSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy", type = "JSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed", type = "JSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "snappy", type = "BSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed", type = "BSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy", type = "BSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed", type = "BSON")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "snappy", type = "ENUM")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed", type = "ENUM")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "snappy", type = "ENUM")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(encoding = "RLE_DICTIONARY", compression = "uncompressed", type = "ENUM")
    Output
      [[1]]
      [1] "a"   "!!!" "!"   NA   
      
      [[2]]
      [1] "e"   "~~~" "~"   NA   
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# min/max for REALSXP -> TIMESTAMP (INT64)

    Code
      do(compression = "snappy")
    Output
      [[1]]
      [1]  1e+06 -1e+08 -1e+09     NA
      
      [[2]]
      [1] 5e+06 1e+08 1e+09    NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

---

    Code
      do(compression = "uncompressed")
    Output
      [[1]]
      [1]  1e+06 -1e+08 -1e+09     NA
      
      [[2]]
      [1] 5e+06 1e+08 1e+09    NA
      
      [[3]]
      [1] TRUE TRUE TRUE   NA
      
      [[4]]
      [1] TRUE TRUE TRUE   NA
      

# parquet_type

    Code
      parquet_type("AUTO")
    Output
      $type
      [1] NA
      
    Code
      parquet_type("BOOLEAN")
    Output
      $type
      [1] "BOOLEAN"
      
    Code
      parquet_type("INT32")
    Output
      $type
      [1] "INT32"
      
    Code
      parquet_type("INT64")
    Output
      $type
      [1] "INT64"
      
    Code
      parquet_type("INT96")
    Output
      $type
      [1] "INT96"
      
    Code
      parquet_type("FLOAT")
    Output
      $type
      [1] "FLOAT"
      
    Code
      parquet_type("DOUBLE")
    Output
      $type
      [1] "DOUBLE"
      
    Code
      parquet_type("BYTE_ARRAY")
    Output
      $type
      [1] "BYTE_ARRAY"
      
    Code
      parquet_type("FIXED_LEN_BYTE_ARRAY", type_length = 10)
    Output
      $type
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $type_length
      [1] 10
      
    Code
      parquet_type("STRING")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "STRING"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "UTF8"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("ENUM")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "ENUM"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "ENUM"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("UUID")
    Output
      $type
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "UUID"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $type_length
      [1] 16
      
      $converted_type
      [1] NA
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("INT", bit_width = 8, is_signed = TRUE)
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "INT"
      
      $bit_width
      [1] 8
      
      $is_signed
      [1] TRUE
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "INT_8"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("INT", bit_width = 16, is_signed = TRUE)
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "INT"
      
      $bit_width
      [1] 16
      
      $is_signed
      [1] TRUE
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "INT_16"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("INT", bit_width = 32, is_signed = FALSE)
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "INT"
      
      $bit_width
      [1] 32
      
      $is_signed
      [1] FALSE
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "UINT_32"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("INT", bit_width = 64, is_signed = FALSE)
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "INT"
      
      $bit_width
      [1] 64
      
      $is_signed
      [1] FALSE
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "UINT_64"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("DECIMAL", precision = 5, primitive_type = "INT32")
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "DECIMAL"
      
      $scale
      NULL
      
      $precision
      [1] 5
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $scale
      [1] 0
      
      $precision
      [1] 5
      
      $converted_type
      [1] "DECIMAL"
      
    Code
      parquet_type("DECIMAL", precision = 5, scale = 0, primitive_type = "INT32")
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 0
      
      $precision
      [1] 5
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $scale
      [1] 0
      
      $precision
      [1] 5
      
      $converted_type
      [1] "DECIMAL"
      
    Code
      parquet_type("DECIMAL", precision = 5, scale = 5, primitive_type = "INT64")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 5
      
      $precision
      [1] 5
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $scale
      [1] 5
      
      $precision
      [1] 5
      
      $converted_type
      [1] "DECIMAL"
      
    Code
      parquet_type("DECIMAL", precision = 5, primitive_type = "BYTE_ARRAY")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "DECIMAL"
      
      $scale
      NULL
      
      $precision
      [1] 5
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $scale
      [1] 0
      
      $precision
      [1] 5
      
      $converted_type
      [1] "DECIMAL"
      
    Code
      parquet_type("DECIMAL", precision = 5, primitive_type = "FIXED_LEN_BYTE_ARRAY",
        type_length = 5)
    Output
      $type
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "DECIMAL"
      
      $scale
      NULL
      
      $precision
      [1] 5
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $scale
      [1] 0
      
      $precision
      [1] 5
      
      $type_length
      [1] 5
      
      $converted_type
      [1] "DECIMAL"
      
    Code
      parquet_type("FLOAT16")
    Output
      $type
      [1] "FIXED_LEN_BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "FLOAT16"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $type_length
      [1] 2
      
      $converted_type
      [1] NA
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("DATE")
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "DATE"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "DATE"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MILLIS")
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "TIME"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "TIME_MILLIS"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIME", is_adjusted_utc = TRUE, unit = "MICROS")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "TIME"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "TIME_MICROS"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIME", is_adjusted_utc = TRUE, unit = "NANOS")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "TIME"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] NA
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MILLIS")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "TIMESTAMP_MILLIS"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "TIMESTAMP_MICROS"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "NANOS")
    Output
      $type
      [1] "INT64"
      
      $logical_type
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] NA
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("JSON")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "JSON"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "JSON"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("BSON")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "BSON"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $converted_type
      [1] "BSON"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      

---

    Code
      parquet_type("INT32", repetition_type = "OPTIONAL")
    Output
      $type
      [1] "INT32"
      
      $repetition_type
      [1] "OPTIONAL"
      
    Code
      parquet_type("STRING", repetition_type = "REQUIRED")
    Output
      $type
      [1] "BYTE_ARRAY"
      
      $logical_type
      $type
      [1] "STRING"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $repetition_type
      [1] "REQUIRED"
      
      $converted_type
      [1] "UTF8"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      
    Code
      parquet_type("TIME", repetition_type = "REPEATED", is_adjusted_utc = TRUE,
        unit = "MILLIS")
    Output
      $type
      [1] "INT32"
      
      $logical_type
      $type
      [1] "TIME"
      
      $is_adjusted_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      $repetition_type
      [1] "REPEATED"
      
      $converted_type
      [1] "TIME_MILLIS"
      
      $scale
      [1] NA
      
      $precision
      [1] NA
      

---

    Code
      parquet_type("FOO")
    Condition
      Error in `err()`:
      ! Parquet type 'FOO' is not supported by nanoparquet
    Code
      parquet_type("FIXED_LEN_BYTE_ARRAY")
    Condition
      Error in `fixed_len_byte_array()`:
      ! !is.null(type_length) is not TRUE
    Code
      parquet_type("INT", bit_width = 8)
    Condition
      Error in `int()`:
      ! !is.null(is_signed) is not TRUE
    Code
      parquet_type("INT", is_signed = TRUE)
    Condition
      Error in `int()`:
      ! !is.null(bit_width) is not TRUE
    Code
      parquet_type("INT", bit_width = 10, is_signed = TRUE)
    Condition
      Error in `int()`:
      ! bit_width %in% c(8L, 16L, 32L, 64L) is not TRUE
    Code
      parquet_type("INT", bit_width = 16, is_signed = 1)
    Condition
      Error in `int()`:
      ! is_flag(is_signed) is not TRUE
    Code
      parquet_type("DECIMAL", precision = 5)
    Condition
      Error in `decimal()`:
      ! !is.null(primitive_type) is not TRUE
    Code
      parquet_type("DECIMAL", primitive_type = "INT32")
    Condition
      Error in `decimal()`:
      ! !is.null(precision) is not TRUE
    Code
      parquet_type("DECIMAL", precision = 5 / 2, primitive_type = "INT32")
    Condition
      Error in `decimal()`:
      ! is_uint32(precision) is not TRUE
    Code
      parquet_type("DECIMAL", precision = 0, primitive_type = "INT32")
    Condition
      Error in `decimal()`:
      ! precision > 0 is not TRUE
    Code
      parquet_type("DECIMAL", precision = 5, scale = 6, primitive_type = "INT32")
    Condition
      Error in `decimal()`:
      ! is.null(scale) || (is_uint32(scale) && scale <= precision) is not TRUE
    Code
      parquet_type("DECIMAL", precision = 10, primitive_type = "INT32")
    Condition
      Error in `decimal()`:
      ! precision <= 9 is not TRUE
    Code
      parquet_type("DECIMAL", precision = 19, primitive_type = "INT64")
    Condition
      Error in `decimal()`:
      ! precision <= 18 is not TRUE
    Code
      parquet_type("DECIMAL", precision = 12, primitive_type = "FIXED_LEN_BYTE_ARRAY",
        type_length = 5)
    Condition
      Error in `decimal()`:
      ! precision <= floor(log10(2^(8 * type_length - 1) - 1)) is not TRUE
    Code
      parquet_type("TIME", is_adjusted_utc = TRUE)
    Condition
      Error in `time()`:
      ! !is.null(unit) is not TRUE
    Code
      parquet_type("TIME", unit = "MILLIS")
    Condition
      Error in `time()`:
      ! !is.null(is_adjusted_utc) is not TRUE
    Code
      parquet_type("TIME", is_adjusted_utc = 1, unit = "MILLIS")
    Condition
      Error in `time()`:
      ! is_flag(is_adjusted_utc) is not TRUE
    Code
      parquet_type("TIME", is_adjusted_utc = TRUE, unit = "FOO")
    Condition
      Error in `time()`:
      ! unit %in% c("MILLIS", "MICROS", "NANOS") is not TRUE
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = TRUE)
    Condition
      Error in `timestamp()`:
      ! !is.null(unit) is not TRUE
    Code
      parquet_type("TIMESTAMP", unit = "MILLIS")
    Condition
      Error in `timestamp()`:
      ! !is.null(is_adjusted_utc) is not TRUE
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = 1, unit = "MILLIS")
    Condition
      Error in `timestamp()`:
      ! is_flag(is_adjusted_utc) is not TRUE
    Code
      parquet_type("TIMESTAMP", is_adjusted_utc = TRUE, unit = "FOO")
    Condition
      Error in `timestamp()`:
      ! unit %in% c("MILLIS", "MICROS", "NANOS") is not TRUE
    Code
      parquet_type("MAP")
    Condition
      Error in `err()`:
      ! Parquet type 'MAP' is not supported by nanoparquet
    Code
      parquet_type("UNKNOWN")
    Condition
      Error in `err()`:
      ! Parquet type 'UNKNOWN' is not supported by nanoparquet
    Code
      parquet_type("INT32", repetition_type = TRUE)
    Condition
      Error in `parquet_type()`:
      ! is_string(repetition_type) is not TRUE
    Code
      parquet_type("INT32", repetition_type = "FOO")
    Condition
      Error in `parquet_type()`:
      ! repetition_type %in% c("REQUIRED", "OPTIONAL", "REPEATED") is not TRUE

# parquet_schema

    Code
      as.data.frame(sch)
    Output
        file_name r_col name r_type                 type type_length repetition_type
      1      <NA>     1 <NA>   <NA>                INT32          NA            <NA>
      2      <NA>     2 <NA>   <NA>              BOOLEAN          NA            <NA>
      3      <NA>     3 <NA>   <NA>                INT32          NA            <NA>
      4      <NA>     4 <NA>   <NA>                INT64          NA            <NA>
      5      <NA>     5 <NA>   <NA>                INT96          NA            <NA>
      6      <NA>     6 <NA>   <NA>                FLOAT          NA            <NA>
      7      <NA>     7 <NA>   <NA>           BYTE_ARRAY          NA            <NA>
      8      <NA>     8 <NA>   <NA> FIXED_LEN_BYTE_ARRAY          10            <NA>
        converted_type logical_type num_children scale precision field_id
      1           <NA>                        NA    NA        NA       NA
      2           <NA>                        NA    NA        NA       NA
      3           <NA>                        NA    NA        NA       NA
      4           <NA>                        NA    NA        NA       NA
      5           <NA>                        NA    NA        NA       NA
      6           <NA>                        NA    NA        NA       NA
      7           <NA>                        NA    NA        NA       NA
      8           <NA>                        NA    NA        NA       NA

---

    Code
      as.data.frame(sch2)
    Output
        file_name r_col name r_type                 type type_length repetition_type
      a      <NA>     1    a   <NA>                INT32          NA            <NA>
      b      <NA>     2    b   <NA>              BOOLEAN          NA            <NA>
      c      <NA>     3    c   <NA>                INT32          NA            <NA>
      d      <NA>     4    d   <NA>                INT64          NA            <NA>
      e      <NA>     5    e   <NA>                INT96          NA            <NA>
      f      <NA>     6    f   <NA>                FLOAT          NA            <NA>
      g      <NA>     7    g   <NA>           BYTE_ARRAY          NA            <NA>
             <NA>     8 <NA>   <NA> FIXED_LEN_BYTE_ARRAY          10            <NA>
        converted_type logical_type num_children scale precision field_id
      a           <NA>                        NA    NA        NA       NA
      b           <NA>                        NA    NA        NA       NA
      c           <NA>                        NA    NA        NA       NA
      d           <NA>                        NA    NA        NA       NA
      e           <NA>                        NA    NA        NA       NA
      f           <NA>                        NA    NA        NA       NA
      g           <NA>                        NA    NA        NA       NA
                  <NA>                        NA    NA        NA       NA

---

    Code
      as.data.frame(sch3)
    Output
        file_name r_col name r_type                 type type_length repetition_type
      a      <NA>     1    a   <NA>                INT32          NA            <NA>
      b      <NA>     2    b   <NA>              BOOLEAN          NA            <NA>
      c      <NA>     3    c   <NA>                INT32          NA            <NA>
      d      <NA>     4    d   <NA>                INT64          NA            <NA>
      e      <NA>     5    e   <NA>                INT96          NA            <NA>
      f      <NA>     6    f   <NA>                FLOAT          NA            <NA>
      g      <NA>     7    g   <NA>           BYTE_ARRAY          NA            <NA>
      h      <NA>     8    h   <NA> FIXED_LEN_BYTE_ARRAY          10            <NA>
        converted_type logical_type num_children scale precision field_id
      a           <NA>                        NA    NA        NA       NA
      b           <NA>                        NA    NA        NA       NA
      c           <NA>                        NA    NA        NA       NA
      d           <NA>                        NA    NA        NA       NA
      e           <NA>                        NA    NA        NA       NA
      f           <NA>                        NA    NA        NA       NA
      g           <NA>                        NA    NA        NA       NA
      h           <NA>                        NA    NA        NA       NA

---

    Code
      as.data.frame(sch4)
    Output
         file_name r_col name r_type                 type type_length repetition_type
      1       <NA>     1 <NA>   <NA>           BYTE_ARRAY          NA            <NA>
      2       <NA>     2 <NA>   <NA>           BYTE_ARRAY          NA            <NA>
      3       <NA>     3 <NA>   <NA> FIXED_LEN_BYTE_ARRAY          16            <NA>
      4       <NA>     4 <NA>   <NA>                INT32          NA            <NA>
      5       <NA>     5 <NA>   <NA>                INT64          NA            <NA>
      6       <NA>     6 <NA>   <NA>                INT64          NA            <NA>
      7       <NA>     7 <NA>   <NA> FIXED_LEN_BYTE_ARRAY           2            <NA>
      8       <NA>     8 <NA>   <NA>                INT32          NA            <NA>
      9       <NA>     9 <NA>   <NA>           BYTE_ARRAY          NA            <NA>
      10      <NA>    10 <NA>   <NA>           BYTE_ARRAY          NA            <NA>
         converted_type logical_type num_children scale precision field_id
      1            UTF8       STRING           NA    NA        NA       NA
      2            ENUM         ENUM           NA    NA        NA       NA
      3            <NA>         UUID           NA    NA        NA       NA
      4           INT_8 INT, 8, TRUE           NA    NA        NA       NA
      5         UINT_64 INT, 64,....           NA    NA        NA       NA
      6         DECIMAL DECIMAL,....           NA     0         5       NA
      7            <NA>      FLOAT16           NA    NA        NA       NA
      8            DATE         DATE           NA    NA        NA       NA
      9            JSON         JSON           NA    NA        NA       NA
      10           BSON         BSON           NA    NA        NA       NA

---

    Code
      as.data.frame(sch5)
    Output
          file_name r_col name r_type  type type_length repetition_type
      foo      <NA>     1  foo   <NA>  <NA>          NA            <NA>
      bar      <NA>     2  bar   <NA> INT32          NA            <NA>
          converted_type logical_type num_children scale precision field_id
      foo           <NA>                        NA    NA        NA       NA
      bar           <NA>                        NA    NA        NA       NA

# BOOLEAN

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DATA_PAGE"
    Code
      as.data.frame(read_parquet(tmp))
    Output
            i
      1  TRUE
      2 FALSE
      3  TRUE
      4 FALSE
      5  TRUE
      6 FALSE

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DATA_PAGE"
    Code
      as.data.frame(read_parquet(tmp))
    Output
            i
      1    NA
      2 FALSE
      3    NA
      4 FALSE
      5    NA
      6 FALSE

---

    Code
      read_parquet(tmp)[["i"]]
    Output
        [1]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [13]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [25]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [37]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [49]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [61]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [73]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [85]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [97]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [109]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [121]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [133]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [145]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [157]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [169]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [181]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [193]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE

---

    Code
      read_parquet(tmp)[["i"]]
    Output
        [1]    NA FALSE    NA FALSE    NA FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [13]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [25]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [37]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [49]  TRUE    NA  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [61]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [73]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [85]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
       [97]  TRUE FALSE    NA FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [109]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [121]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [133]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [145]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [157]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [169]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [181]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE
      [193]  TRUE FALSE  TRUE FALSE  TRUE FALSE  TRUE FALSE

# INT32

    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1   1
      2   2
      3   3
      4   4
      5   5
      6   6
      7   7
      8   8
      9   9
      10 10

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1   1
      2  NA
      3   3
      4   4
      5   5
      6  NA
      7   7
      8   8
      9   9
      10 NA

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
         i
      1  1
      2  2
      3  1
      4  2
      5  1
      6  2
      7  1
      8  2
      9  1
      10 2
      11 1
      12 2
      13 1
      14 2
      15 1
      16 2
      17 1
      18 2
      19 1
      20 2
      21 1
      22 2
      23 1
      24 2
      25 1
      26 2
      27 1
      28 2
      29 1
      30 2
      31 1
      32 2
      33 1
      34 2
      35 1
      36 2
      37 1
      38 2
      39 1
      40 2

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1  NA
      2   2
      3   1
      4   2
      5  NA
      6   2
      7   1
      8   2
      9   1
      10 NA
      11  1
      12  2
      13  1
      14  2
      15  1
      16  2
      17  1
      18  2
      19  1
      20 NA
      21  1
      22  2
      23  1
      24  2
      25  1
      26  2
      27  1
      28  2
      29  1
      30  2
      31  1
      32  2
      33  1
      34  2
      35  1
      36  2
      37  1
      38  2
      39 NA
      40  2

# INT64

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1   1
      2   2
      3   3
      4   4
      5   5
      6   6
      7   7
      8   8
      9   9
      10 10

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1   1
      2  NA
      3   3
      4   4
      5   5
      6  NA
      7   7
      8   8
      9   9
      10 NA

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
         i
      1  1
      2  2
      3  1
      4  2
      5  1
      6  2
      7  1
      8  2
      9  1
      10 2
      11 1
      12 2
      13 1
      14 2
      15 1
      16 2
      17 1
      18 2
      19 1
      20 2
      21 1
      22 2
      23 1
      24 2
      25 1
      26 2
      27 1
      28 2
      29 1
      30 2
      31 1
      32 2
      33 1
      34 2
      35 1
      36 2
      37 1
      38 2
      39 1
      40 2

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double INT64          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
          i
      1  NA
      2   2
      3   1
      4   2
      5  NA
      6   2
      7   1
      8   2
      9   1
      10 NA
      11  1
      12  2
      13  1
      14  2
      15  1
      16  2
      17  1
      18  2
      19  1
      20 NA
      21  1
      22  2
      23  1
      24  2
      25  1
      26  2
      27  1
      28  2
      29  1
      30  2
      31  1
      32  2
      33  1
      34  2
      35  1
      36  2
      37  1
      38  2
      39 NA
      40  2

# FLOAT

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double FLOAT          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2  1.0
      3  1.5
      4  2.0
      5  2.5
      6  3.0
      7  3.5
      8  4.0
      9  4.5
      10 5.0

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double FLOAT          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2   NA
      3  1.5
      4  2.0
      5  2.5
      6   NA
      7  3.5
      8  4.0
      9  4.5
      10  NA

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double FLOAT          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2  1.0
      3  0.5
      4  1.0
      5  0.5
      6  1.0
      7  0.5
      8  1.0
      9  0.5
      10 1.0
      11 0.5
      12 1.0
      13 0.5
      14 1.0
      15 0.5
      16 1.0
      17 0.5
      18 1.0
      19 0.5
      20 1.0
      21 0.5
      22 1.0
      23 0.5
      24 1.0
      25 0.5
      26 1.0
      27 0.5
      28 1.0
      29 0.5
      30 1.0
      31 0.5
      32 1.0
      33 0.5
      34 1.0
      35 0.5
      36 1.0
      37 0.5
      38 1.0
      39 0.5
      40 1.0

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1      i double FLOAT          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1   NA
      2  1.0
      3  0.5
      4  1.0
      5   NA
      6  1.0
      7  0.5
      8  1.0
      9  0.5
      10  NA
      11 0.5
      12 1.0
      13 0.5
      14 1.0
      15 0.5
      16 1.0
      17 0.5
      18 1.0
      19 0.5
      20  NA
      21 0.5
      22 1.0
      23 0.5
      24 1.0
      25 0.5
      26 1.0
      27 0.5
      28 1.0
      29 0.5
      30 1.0
      31 0.5
      32 1.0
      33 0.5
      34 1.0
      35 0.5
      36 1.0
      37 0.5
      38 1.0
      39  NA
      40 1.0

# DOUBLE

    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2  1.0
      3  1.5
      4  2.0
      5  2.5
      6  3.0
      7  3.5
      8  4.0
      9  4.5
      10 5.0

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2   NA
      3  1.5
      4  2.0
      5  2.5
      6   NA
      7  3.5
      8  4.0
      9  4.5
      10  NA

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1  0.5
      2  1.0
      3  0.5
      4  1.0
      5  0.5
      6  1.0
      7  0.5
      8  1.0
      9  0.5
      10 1.0
      11 0.5
      12 1.0
      13 0.5
      14 1.0
      15 0.5
      16 1.0
      17 0.5
      18 1.0
      19 0.5
      20 1.0
      21 0.5
      22 1.0
      23 0.5
      24 1.0
      25 0.5
      26 1.0
      27 0.5
      28 1.0
      29 0.5
      30 1.0
      31 0.5
      32 1.0
      33 0.5
      34 1.0
      35 0.5
      36 1.0
      37 0.5
      38 1.0
      39 0.5
      40 1.0

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet(tmp))
    Output
           i
      1   NA
      2  1.0
      3  0.5
      4  1.0
      5   NA
      6  1.0
      7  0.5
      8  1.0
      9  0.5
      10  NA
      11 0.5
      12 1.0
      13 0.5
      14 1.0
      15 0.5
      16 1.0
      17 0.5
      18 1.0
      19 0.5
      20  NA
      21 0.5
      22 1.0
      23 0.5
      24 1.0
      25 0.5
      26 1.0
      27 0.5
      28 1.0
      29 0.5
      30 1.0
      31 0.5
      32 1.0
      33 0.5
      34 1.0
      35 0.5
      36 1.0
      37 0.5
      38 1.0
      39  NA
      40 1.0

# BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      l   blob BYTE_ARRAY          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                             l
      1 01, 02, 03, 04, 05, 06, 07, 08, 09, 0a
      2 0b, 0c, 0d, 0e, 0f, 10, 11, 12, 13, 14
      3                             01, 02, 03
      4                                     01

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type       type type_length repetition_type converted_type
      1    NA schema   <NA>       <NA>          NA            <NA>           <NA>
      2     1      l   blob BYTE_ARRAY          NA        OPTIONAL           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                             l
      1 01, 02, 03, 04, 05, 06, 07, 08, 09, 0a
      2 0b, 0c, 0d, 0e, 0f, 10, 11, 12, 13, 14
      3                                   NULL
      4                                     01

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name    r_type       type type_length repetition_type converted_type
      1    NA schema      <NA>       <NA>          NA            <NA>           <NA>
      2     1      s character BYTE_ARRAY          NA        REQUIRED           UTF8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2       STRING           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name    r_type       type type_length repetition_type converted_type
      1    NA schema      <NA>       <NA>          NA            <NA>           <NA>
      2     1      s character BYTE_ARRAY          NA        OPTIONAL           UTF8
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2       STRING           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

# FIXED_LEN_BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      l   blob FIXED_LEN_BYTE_ARRAY          10        REQUIRED
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                             l
      1 01, 02, 03, 04, 05, 06, 07, 08, 09, 0a
      2 0b, 0c, 0d, 0e, 0f, 10, 11, 12, 13, 14
      3 01, 02, 03, 04, 05, 06, 07, 08, 09, 0a
      4 15, 16, 17, 18, 19, 1a, 1b, 1c, 1d, 1e

---

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      l   blob FIXED_LEN_BYTE_ARRAY          10        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                             l
      1 01, 02, 03, 04, 05, 06, 07, 08, 09, 0a
      2 0b, 0c, 0d, 0e, 0f, 10, 11, 12, 13, 14
      3                                   NULL
      4 15, 16, 17, 18, 19, 1a, 1b, 1c, 1d, 1e

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      s   blob FIXED_LEN_BYTE_ARRAY           3        REQUIRED
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3 61, 61, 61

---

    Code
      as.data.frame(read_parquet_pages(tmp))[["page_type"]]
    Output
      [1] "DICTIONARY_PAGE" "DATA_PAGE"      
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type                 type type_length repetition_type
      1    NA schema   <NA>                 <NA>          NA            <NA>
      2     1      s   blob FIXED_LEN_BYTE_ARRAY           3        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet(tmp))
    Output
                 s
      1 66, 6f, 6f
      2 62, 61, 72
      3       NULL
      4 61, 61, 61

# as_count

    Code
      as_count(1:2)
    Condition
      Error in `as_count()`:
      ! x must be a count, i.e. an integer scalar
    Code
      as_count(0)
    Condition
      Error in `as_count()`:
      ! x must be a count, i.e. an integer scalar
    Code
      as_count(NA_real_)
    Condition
      Error in `as_count()`:
      ! x must be a count, i.e. an integer scalar
    Code
      as_count(-100)
    Condition
      Error in `as_count()`:
      ! x must be a count, i.e. an integer scalar
    Code
      as_count(-100L)
    Condition
      Error in `as_count()`:
      ! x must be a count, i.e. an integer scalar

# read a subset, edge cases

    Code
      read_parquet(tmp, col_select = integer())
    Output
      # A data frame: 32 x 0
    Code
      read_parquet(tmp, col_select = character())
    Output
      # A data frame: 32 x 0

# error if a column is requested multiple times

    Code
      read_parquet(tmp, col_select = c(1, 1))
    Condition
      Error in `read_parquet()`:
      ! Column 1 selected multiple times in `read_parquet()`.
    Code
      read_parquet(tmp, col_select = c(3, 4, 5, 3))
    Condition
      Error in `read_parquet()`:
      ! Column 3 selected multiple times in `read_parquet()`.
    Code
      read_parquet(tmp, col_select = c(3:4, 4:3))
    Condition
      Error in `read_parquet()`:
      ! Columns 3, 4 selected multiple times in `read_parquet()`.
    Code
      read_parquet(tmp, col_select = "foo")
    Condition
      Error in `read_parquet()`:
      ! Column foo does not exist in Parquet file
    Code
      read_parquet(tmp, col_select = c("foo", "bar"))
    Condition
      Error in `read_parquet()`:
      ! Columns foo, bar do not exist in Parquet file
    Code
      read_parquet(tmp, col_select = c("nam", "nam"))
    Condition
      Error in `read_parquet()`:
      ! Column nam selected multiple times in `read_parquet()`.
    Code
      read_parquet(tmp, col_select = c("cyl", "disp", "hp", "cyl"))
    Condition
      Error in `read_parquet()`:
      ! Column cyl selected multiple times in `read_parquet()`.
    Code
      read_parquet(tmp, col_select = c("cyl", "disp", "disp", "cyl"))
    Condition
      Error in `read_parquet()`:
      ! Columns cyl, disp selected multiple times in `read_parquet()`.

# mixing RLE_DICTIONARY and PLAIN

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
              type repetition_type
      1       <NA>        REQUIRED
      2      INT32        REQUIRED
      3      INT64        REQUIRED
      4 BYTE_ARRAY        REQUIRED
      5      FLOAT        REQUIRED
      6     DOUBLE        REQUIRED
      7      INT96        REQUIRED
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
               page_type num_values       encoding
      1  DICTIONARY_PAGE        400          PLAIN
      2        DATA_PAGE       1024 RLE_DICTIONARY
      3        DATA_PAGE       1024          PLAIN
      4        DATA_PAGE        352          PLAIN
      5  DICTIONARY_PAGE        400          PLAIN
      6        DATA_PAGE       1024 RLE_DICTIONARY
      7        DATA_PAGE       1024          PLAIN
      8        DATA_PAGE        352          PLAIN
      9  DICTIONARY_PAGE        400          PLAIN
      10       DATA_PAGE       1024 RLE_DICTIONARY
      11       DATA_PAGE       1024          PLAIN
      12       DATA_PAGE        352          PLAIN
      13 DICTIONARY_PAGE        400          PLAIN
      14       DATA_PAGE       1024 RLE_DICTIONARY
      15       DATA_PAGE       1024          PLAIN
      16       DATA_PAGE        352          PLAIN
      17 DICTIONARY_PAGE        400          PLAIN
      18       DATA_PAGE       1024 RLE_DICTIONARY
      19       DATA_PAGE       1024          PLAIN
      20       DATA_PAGE        352          PLAIN
      21 DICTIONARY_PAGE        400          PLAIN
      22       DATA_PAGE       1024 RLE_DICTIONARY
      23       DATA_PAGE       1024          PLAIN
      24       DATA_PAGE        352          PLAIN

---

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
              type repetition_type
      1       <NA>        REQUIRED
      2      INT32        REQUIRED
      3      INT64        REQUIRED
      4 BYTE_ARRAY        REQUIRED
      5      FLOAT        REQUIRED
      6     DOUBLE        REQUIRED
      7      INT96        REQUIRED
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
               page_type num_values       encoding
      1  DICTIONARY_PAGE        400          PLAIN
      2        DATA_PAGE       1024 RLE_DICTIONARY
      3        DATA_PAGE       1024 RLE_DICTIONARY
      4        DATA_PAGE        352 RLE_DICTIONARY
      5  DICTIONARY_PAGE        400          PLAIN
      6        DATA_PAGE       1024 RLE_DICTIONARY
      7        DATA_PAGE       1024 RLE_DICTIONARY
      8        DATA_PAGE        352 RLE_DICTIONARY
      9  DICTIONARY_PAGE        400          PLAIN
      10       DATA_PAGE       1024 RLE_DICTIONARY
      11       DATA_PAGE       1024 RLE_DICTIONARY
      12       DATA_PAGE        352 RLE_DICTIONARY
      13 DICTIONARY_PAGE        400          PLAIN
      14       DATA_PAGE       1024 RLE_DICTIONARY
      15       DATA_PAGE       1024 RLE_DICTIONARY
      16       DATA_PAGE        352 RLE_DICTIONARY
      17 DICTIONARY_PAGE        400          PLAIN
      18       DATA_PAGE       1024 RLE_DICTIONARY
      19       DATA_PAGE       1024 RLE_DICTIONARY
      20       DATA_PAGE        352 RLE_DICTIONARY
      21 DICTIONARY_PAGE        400          PLAIN
      22       DATA_PAGE       1024 RLE_DICTIONARY
      23       DATA_PAGE       1024 RLE_DICTIONARY
      24       DATA_PAGE        352 RLE_DICTIONARY

---

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
              type repetition_type
      1       <NA>        REQUIRED
      2      INT32        OPTIONAL
      3      INT64        OPTIONAL
      4 BYTE_ARRAY        OPTIONAL
      5      FLOAT        OPTIONAL
      6     DOUBLE        OPTIONAL
      7      INT96        OPTIONAL
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
               page_type num_values       encoding
      1  DICTIONARY_PAGE       1009          PLAIN
      2        DATA_PAGE       1024 RLE_DICTIONARY
      3        DATA_PAGE       1024          PLAIN
      4        DATA_PAGE        352          PLAIN
      5  DICTIONARY_PAGE       1018          PLAIN
      6        DATA_PAGE       1024 RLE_DICTIONARY
      7        DATA_PAGE       1024          PLAIN
      8        DATA_PAGE        352          PLAIN
      9  DICTIONARY_PAGE       1014          PLAIN
      10       DATA_PAGE       1024 RLE_DICTIONARY
      11       DATA_PAGE       1024          PLAIN
      12       DATA_PAGE        352          PLAIN
      13 DICTIONARY_PAGE       1013          PLAIN
      14       DATA_PAGE       1024 RLE_DICTIONARY
      15       DATA_PAGE       1024          PLAIN
      16       DATA_PAGE        352          PLAIN
      17 DICTIONARY_PAGE       1018          PLAIN
      18       DATA_PAGE       1024 RLE_DICTIONARY
      19       DATA_PAGE       1024          PLAIN
      20       DATA_PAGE        352          PLAIN
      21 DICTIONARY_PAGE       1016          PLAIN
      22       DATA_PAGE       1024 RLE_DICTIONARY
      23       DATA_PAGE       1024          PLAIN
      24       DATA_PAGE        352          PLAIN

# mixing RLE_DICTIONARY and PLAIN, DECIMAL

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
                        type repetition_type
      1                 <NA>        REQUIRED
      2 FIXED_LEN_BYTE_ARRAY        REQUIRED
      3 FIXED_LEN_BYTE_ARRAY        OPTIONAL
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
              page_type num_values       encoding
      1 DICTIONARY_PAGE        400          PLAIN
      2       DATA_PAGE       1024 RLE_DICTIONARY
      3       DATA_PAGE        176          PLAIN
      4 DICTIONARY_PAGE        400          PLAIN
      5       DATA_PAGE       1024 RLE_DICTIONARY
      6       DATA_PAGE        176          PLAIN

---

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
         type repetition_type
      1  <NA>        REQUIRED
      2 INT32        REQUIRED
      3 INT32        OPTIONAL
      4 INT64        REQUIRED
      5 INT64        OPTIONAL
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
               page_type num_values       encoding
      1  DICTIONARY_PAGE        400          PLAIN
      2        DATA_PAGE       1024 RLE_DICTIONARY
      3        DATA_PAGE        176          PLAIN
      4  DICTIONARY_PAGE        400          PLAIN
      5        DATA_PAGE       1024 RLE_DICTIONARY
      6        DATA_PAGE        176          PLAIN
      7  DICTIONARY_PAGE        400          PLAIN
      8        DATA_PAGE       1024 RLE_DICTIONARY
      9        DATA_PAGE        176          PLAIN
      10 DICTIONARY_PAGE        400          PLAIN
      11       DATA_PAGE       1024 RLE_DICTIONARY
      12       DATA_PAGE        176          PLAIN

# mixing RLE_DICTIONARY and PLAIN, BYTE_ARRAY

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
              type repetition_type
      1       <NA>        REQUIRED
      2 BYTE_ARRAY        REQUIRED
      3 BYTE_ARRAY        OPTIONAL
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
              page_type num_values       encoding
      1 DICTIONARY_PAGE        400          PLAIN
      2       DATA_PAGE       1024 RLE_DICTIONARY
      3       DATA_PAGE        176          PLAIN
      4 DICTIONARY_PAGE        400          PLAIN
      5       DATA_PAGE       1024 RLE_DICTIONARY
      6       DATA_PAGE        176          PLAIN

# mixing RLE_DICTIONARY and PLAIN, FLOAT16

    Code
      as.data.frame(read_parquet_schema(pf)[, c("type", "repetition_type")])
    Output
                        type repetition_type
      1                 <NA>        REQUIRED
      2 FIXED_LEN_BYTE_ARRAY        REQUIRED
      3 FIXED_LEN_BYTE_ARRAY        OPTIONAL
    Code
      as.data.frame(read_parquet_pages(pf)[, c("page_type", "num_values", "encoding")])
    Output
              page_type num_values       encoding
      1 DICTIONARY_PAGE        400          PLAIN
      2       DATA_PAGE       1024 RLE_DICTIONARY
      3       DATA_PAGE        176          PLAIN
      4 DICTIONARY_PAGE        401          PLAIN
      5       DATA_PAGE       1024 RLE_DICTIONARY
      6       DATA_PAGE        176          PLAIN

# DECIMAL converted type

    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 100
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        REQUIRED        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA     2         5       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1 100
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1      d integer INT32          NA        REQUIRED        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA     2         5       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      

# DECIMAL in BA dict

    Code
      as.data.frame(read_parquet(pf))
    Output
          l1   l2   l3   l4
      1  0.1  0.1  0.1  0.1
      2 -0.1 -0.1 -0.1 -0.1

# deprecated warnings

    Code
      sch <- parquet_schema(pf)
    Condition
      Warning in `parquet_schema()`:
      Using `parquet_schema()` to read the schema from a file is deprecated. Use `read_parquet_schema()` instead.

---

    Code
      pct <- parquet_column_types(mtcars)
    Condition
      Warning in `parquet_column_types()`:
      `parquet_column_types()` is deprecated, please use `read_parquet_schema()` or `parquet_schema()` instead.

---

    Code
      mtd <- parquet_metadata(pf)
    Condition
      Warning in `parquet_metadata()`:
      `parquet_metadata()` is deprecated. Please use `read_parquet_metadata()` instead.

---

    Code
      info <- parquet_info(pf)
    Condition
      Warning in `parquet_info()`:
      `parquet_info()` is deprecated, please use `read_parquet_info() instead.

# TIMESTAMP

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

# TIMESTAMP, dictionary

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIMESTAM....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

# TIMESTAMP_MILLIS, TIMESTAMP_MICROS

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type   converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts double INT64          NA        REQUIRED TIMESTAMP_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         

# TIMESTAMP_MILLIS, TIMESTAMP_MICROS, dictionary

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type   converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts double INT64          NA        REQUIRED TIMESTAMP_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
                         ts
      1 2024-08-20 20:37:35
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type   converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>             <NA>
      2     1     ts POSIXct INT64          NA        REQUIRED TIMESTAMP_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

# TIME

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT32          NA        REQUIRED    TIME_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED    TIME_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

# TIME_MILLIS, TIME_MICROS

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT32          NA        REQUIRED    TIME_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED    TIME_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      

# TIME, dictionary

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT32          NA        REQUIRED    TIME_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MILLIS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED    TIME_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED           <NA>
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2 TIME, TR....           NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "NANOS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

# TIME_MILLIS, TIME_MICROS, dictionary

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT32          NA        REQUIRED    TIME_MILLIS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
              dt
      1 03:02:01
    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name r_type  type type_length repetition_type converted_type
      1    NA schema   <NA>  <NA>          NA            <NA>           <NA>
      2     1     dt    hms INT64          NA        REQUIRED    TIME_MICROS
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA    NA        NA       NA         
    Code
      as.data.frame(read_parquet_schema(tmp))[["logical_type"]]
    Output
      [[1]]
      NULL
      
      [[2]]
      NULL
      
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

# parquet_metadata

    Code
      as.data.frame(mtd$file_meta_data)
    Output
           file_name version num_rows mtd$file_meta_data$key_value_metadata
      1 test.parquet       1       32                          ARROW:sc....
                                        created_by
      1 https://github.com/gaborcsardi/nanoparquet
    Code
      as.data.frame(mtd$schema)
    Output
            file_name r_col   name    r_type       type type_length repetition_type
      1  test.parquet    NA schema      <NA>       <NA>          NA            <NA>
      2  test.parquet     1    nam character BYTE_ARRAY          NA        REQUIRED
      3  test.parquet     2    mpg    double     DOUBLE          NA        REQUIRED
      4  test.parquet     3    cyl   integer      INT32          NA        REQUIRED
      5  test.parquet     4   disp    double     DOUBLE          NA        REQUIRED
      6  test.parquet     5     hp    double     DOUBLE          NA        REQUIRED
      7  test.parquet     6   drat    double     DOUBLE          NA        REQUIRED
      8  test.parquet     7     wt    double     DOUBLE          NA        REQUIRED
      9  test.parquet     8   qsec    double     DOUBLE          NA        REQUIRED
      10 test.parquet     9     vs    double     DOUBLE          NA        REQUIRED
      11 test.parquet    10     am    double     DOUBLE          NA        REQUIRED
      12 test.parquet    11   gear    double     DOUBLE          NA        REQUIRED
      13 test.parquet    12   carb    double     DOUBLE          NA        REQUIRED
      14 test.parquet    13  large   logical    BOOLEAN          NA        REQUIRED
         converted_type logical_type num_children scale precision field_id children
      1            <NA>                        13    NA        NA       NA         
      2            UTF8       STRING           NA    NA        NA       NA         
      3            <NA>                        NA    NA        NA       NA         
      4          INT_32 INT, 32,....           NA    NA        NA       NA         
      5            <NA>                        NA    NA        NA       NA         
      6            <NA>                        NA    NA        NA       NA         
      7            <NA>                        NA    NA        NA       NA         
      8            <NA>                        NA    NA        NA       NA         
      9            <NA>                        NA    NA        NA       NA         
      10           <NA>                        NA    NA        NA       NA         
      11           <NA>                        NA    NA        NA       NA         
      12           <NA>                        NA    NA        NA       NA         
      13           <NA>                        NA    NA        NA       NA         
      14           <NA>                        NA    NA        NA       NA         
    Code
      as.data.frame(mtd$row_groups)
    Output
           file_name id total_byte_size num_rows file_offset total_compressed_size
      1 test.parquet  0            3446       32          NA                    NA
        ordinal
      1      NA
    Code
      as.data.frame(mtd$column_chunks[, 1:20])
    Output
            file_name row_group column file_path file_offset offset_index_offset
      1  test.parquet         0      0      <NA>           4                  NA
      2  test.parquet         0      1      <NA>         532                  NA
      3  test.parquet         0      2      <NA>         807                  NA
      4  test.parquet         0      3      <NA>         954                  NA
      5  test.parquet         0      4      <NA>        1229                  NA
      6  test.parquet         0      5      <NA>        1504                  NA
      7  test.parquet         0      6      <NA>        1779                  NA
      8  test.parquet         0      7      <NA>        2054                  NA
      9  test.parquet         0      8      <NA>        2329                  NA
      10 test.parquet         0      9      <NA>        2604                  NA
      11 test.parquet         0     10      <NA>        2879                  NA
      12 test.parquet         0     11      <NA>        3154                  NA
      13 test.parquet         0     12      <NA>        3429                  NA
         offset_index_length column_index_offset column_index_length       type
      1                   NA                  NA                  NA BYTE_ARRAY
      2                   NA                  NA                  NA     DOUBLE
      3                   NA                  NA                  NA      INT32
      4                   NA                  NA                  NA     DOUBLE
      5                   NA                  NA                  NA     DOUBLE
      6                   NA                  NA                  NA     DOUBLE
      7                   NA                  NA                  NA     DOUBLE
      8                   NA                  NA                  NA     DOUBLE
      9                   NA                  NA                  NA     DOUBLE
      10                  NA                  NA                  NA     DOUBLE
      11                  NA                  NA                  NA     DOUBLE
      12                  NA                  NA                  NA     DOUBLE
      13                  NA                  NA                  NA    BOOLEAN
         encodings path_in_schema        codec num_values total_uncompressed_size
      1      PLAIN            nam UNCOMPRESSED         32                     528
      2      PLAIN            mpg UNCOMPRESSED         32                     275
      3      PLAIN            cyl UNCOMPRESSED         32                     147
      4      PLAIN           disp UNCOMPRESSED         32                     275
      5      PLAIN             hp UNCOMPRESSED         32                     275
      6      PLAIN           drat UNCOMPRESSED         32                     275
      7      PLAIN             wt UNCOMPRESSED         32                     275
      8      PLAIN           qsec UNCOMPRESSED         32                     275
      9      PLAIN             vs UNCOMPRESSED         32                     275
      10     PLAIN             am UNCOMPRESSED         32                     275
      11     PLAIN           gear UNCOMPRESSED         32                     275
      12     PLAIN           carb UNCOMPRESSED         32                     275
      13     PLAIN          large UNCOMPRESSED         32                      21
         total_compressed_size data_page_offset index_page_offset
      1                    528                4                NA
      2                    275              532                NA
      3                    147              807                NA
      4                    275              954                NA
      5                    275             1229                NA
      6                    275             1504                NA
      7                    275             1779                NA
      8                    275             2054                NA
      9                    275             2329                NA
      10                   275             2604                NA
      11                   275             2879                NA
      12                   275             3154                NA
      13                    21             3429                NA
         dictionary_page_offset null_count
      1                      NA          0
      2                      NA          0
      3                      NA          0
      4                      NA          0
      5                      NA          0
      6                      NA          0
      7                      NA          0
      8                      NA          0
      9                      NA          0
      10                     NA          0
      11                     NA          0
      12                     NA          0
      13                     NA          0

---

    Code
      as.data.frame(sch)
    Output
            file_name r_col   name    r_type       type type_length repetition_type
      1  test.parquet    NA schema      <NA>       <NA>          NA            <NA>
      2  test.parquet     1    nam character BYTE_ARRAY          NA        REQUIRED
      3  test.parquet     2    mpg    double     DOUBLE          NA        REQUIRED
      4  test.parquet     3    cyl   integer      INT32          NA        REQUIRED
      5  test.parquet     4   disp    double     DOUBLE          NA        REQUIRED
      6  test.parquet     5     hp    double     DOUBLE          NA        REQUIRED
      7  test.parquet     6   drat    double     DOUBLE          NA        REQUIRED
      8  test.parquet     7     wt    double     DOUBLE          NA        REQUIRED
      9  test.parquet     8   qsec    double     DOUBLE          NA        REQUIRED
      10 test.parquet     9     vs    double     DOUBLE          NA        REQUIRED
      11 test.parquet    10     am    double     DOUBLE          NA        REQUIRED
      12 test.parquet    11   gear    double     DOUBLE          NA        REQUIRED
      13 test.parquet    12   carb    double     DOUBLE          NA        REQUIRED
      14 test.parquet    13  large   logical    BOOLEAN          NA        REQUIRED
         converted_type logical_type num_children scale precision field_id children
      1            <NA>                        13    NA        NA       NA         
      2            UTF8       STRING           NA    NA        NA       NA         
      3            <NA>                        NA    NA        NA       NA         
      4          INT_32 INT, 32,....           NA    NA        NA       NA         
      5            <NA>                        NA    NA        NA       NA         
      6            <NA>                        NA    NA        NA       NA         
      7            <NA>                        NA    NA        NA       NA         
      8            <NA>                        NA    NA        NA       NA         
      9            <NA>                        NA    NA        NA       NA         
      10           <NA>                        NA    NA        NA       NA         
      11           <NA>                        NA    NA        NA       NA         
      12           <NA>                        NA    NA        NA       NA         
      13           <NA>                        NA    NA        NA       NA         
      14           <NA>                        NA    NA        NA       NA         

# ENUM type

    Code
      as.data.frame(sch)
    Output
                 file_name r_col                   name
      1  data/enum.parquet    NA             trace.Span
      2  data/enum.parquet     1                     id
      3  data/enum.parquet     2              parent_id
      4  data/enum.parquet     3               trace_id
      5  data/enum.parquet     4                   name
      6  data/enum.parquet     5 start_timestamp_micros
      7  data/enum.parquet     6        duration_micros
      8  data/enum.parquet     7                   tags
      9  data/enum.parquet     7                    key
      10 data/enum.parquet     7                 v_type
      11 data/enum.parquet     7                  v_str
      12 data/enum.parquet     7                 v_bool
      13 data/enum.parquet     7                v_int64
      14 data/enum.parquet     7              v_float64
      15 data/enum.parquet     7               v_binary
                                                                             r_type
      1                                                                        <NA>
      2                                                                        blob
      3                                                                        blob
      4                                                                        blob
      5                                                                   character
      6                                                                      double
      7                                                                      double
      8  list(list(character, character, character, logical, double, double, blob))
      9                                                                        <NA>
      10                                                                       <NA>
      11                                                                       <NA>
      12                                                                       <NA>
      13                                                                       <NA>
      14                                                                       <NA>
      15                                                                       <NA>
               type type_length repetition_type converted_type logical_type
      1        <NA>          NA            <NA>           <NA>             
      2  BYTE_ARRAY          NA        OPTIONAL           <NA>             
      3  BYTE_ARRAY          NA        OPTIONAL           <NA>             
      4  BYTE_ARRAY          NA        OPTIONAL           <NA>             
      5  BYTE_ARRAY          NA        OPTIONAL           UTF8       STRING
      6       INT64          NA        OPTIONAL           <NA>             
      7       INT64          NA        OPTIONAL           <NA>             
      8        <NA>          NA        REPEATED           <NA>             
      9  BYTE_ARRAY          NA        OPTIONAL           UTF8       STRING
      10 BYTE_ARRAY          NA        OPTIONAL           ENUM         ENUM
      11 BYTE_ARRAY          NA        OPTIONAL           UTF8       STRING
      12    BOOLEAN          NA        OPTIONAL           <NA>             
      13      INT64          NA        OPTIONAL           <NA>             
      14     DOUBLE          NA        OPTIONAL           <NA>             
      15 BYTE_ARRAY          NA        OPTIONAL           <NA>             
         num_children scale precision field_id                  children
      1             7    NA        NA       NA                          
      2            NA    NA        NA        1                          
      3            NA    NA        NA        2                          
      4            NA    NA        NA        3                          
      5            NA    NA        NA        4                          
      6            NA    NA        NA        5                          
      7            NA    NA        NA        6                          
      8             7    NA        NA        7 9, 10, 11, 12, 13, 14, 15
      9            NA    NA        NA        1                          
      10           NA    NA        NA        2                          
      11           NA    NA        NA        3                          
      12           NA    NA        NA        4                          
      13           NA    NA        NA        5                          
      14           NA    NA        NA        6                          
      15           NA    NA        NA        7                          

# UUID type

    Code
      as.data.frame(sch)
    Output
                      file_name r_col          name    r_type                 type
      1 data/uuid-arrow.parquet    NA duckdb_schema      <NA>                 <NA>
      2 data/uuid-arrow.parquet     1             u character FIXED_LEN_BYTE_ARRAY
        type_length repetition_type converted_type logical_type num_children scale
      1          NA        REQUIRED           <NA>                         1    NA
      2          16        OPTIONAL           <NA>         UUID            0    NA
        precision field_id children
      1        NA       NA         
      2        NA       NA         

# DATE type

    Code
      as.data.frame(sch)
    Output
                file_name r_col  name r_type  type type_length repetition_type
      1 data/date.parquet    NA dates   <NA>  <NA>          NA            <NA>
      2 data/date.parquet     1     d   Date INT32          NA        OPTIONAL
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         1    NA        NA       NA         
      2           DATE         DATE           NA    NA        NA       NA         

# DECIMAL type

    Code
      as.data.frame(sch)
    Output
                    file_name r_col   name r_type                 type type_length
      1 data/decimals.parquet    NA schema   <NA>                 <NA>          NA
      2 data/decimals.parquet     1     l1 double FIXED_LEN_BYTE_ARRAY           2
      3 data/decimals.parquet     2     l2 double FIXED_LEN_BYTE_ARRAY           4
      4 data/decimals.parquet     3     l3 double FIXED_LEN_BYTE_ARRAY           7
      5 data/decimals.parquet     4     l4 double FIXED_LEN_BYTE_ARRAY          13
        repetition_type converted_type logical_type num_children scale precision
      1        REQUIRED           <NA>                         4    NA        NA
      2        OPTIONAL        DECIMAL DECIMAL,....           NA     2         3
      3        OPTIONAL        DECIMAL DECIMAL,....           NA     2         8
      4        OPTIONAL        DECIMAL DECIMAL,....           NA     2        15
      5        OPTIONAL        DECIMAL DECIMAL,....           NA     2        30
        field_id children
      1       NA         
      2       NA         
      3       NA         
      4       NA         
      5       NA         

---

    Code
      sch$logical_type
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 2
      
      $precision
      [1] 3
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      [[3]]
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 2
      
      $precision
      [1] 8
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      [[4]]
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 2
      
      $precision
      [1] 15
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      
      [[5]]
      $type
      [1] "DECIMAL"
      
      $scale
      [1] 2
      
      $precision
      [1] 30
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

# TIME type

    Code
      as.data.frame(sch)
    Output
                  file_name r_col          name r_type  type type_length
      1 data/timetz.parquet    NA duckdb_schema   <NA>  <NA>          NA
      2 data/timetz.parquet     1            tt    hms INT64          NA
        repetition_type converted_type logical_type num_children scale precision
      1        REQUIRED           <NA>                         1    NA        NA
      2        OPTIONAL    TIME_MICROS TIME, TR....           NA    NA        NA
        field_id children
      1       NA         
      2       NA         

---

    Code
      sch$logical_type
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIME"
      
      $is_adjusted_to_utc
      [1] TRUE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

# TIMESTAMP type

    Code
      as.data.frame(sch)
    Output
                     file_name r_col   name  r_type  type type_length repetition_type
      1 data/timestamp.parquet    NA schema    <NA>  <NA>          NA        REQUIRED
      2 data/timestamp.parquet     1   Time POSIXct INT64          NA        OPTIONAL
          converted_type logical_type num_children scale precision field_id children
      1             <NA>                         1    NA        NA       NA         
      2 TIMESTAMP_MICROS TIMESTAM....           NA    NA        NA       NA         

---

    Code
      sch$logical_type
    Output
      [[1]]
      NULL
      
      [[2]]
      $type
      [1] "TIMESTAMP"
      
      $is_adjusted_to_utc
      [1] FALSE
      
      $unit
      [1] "MICROS"
      
      attr(,"class")
      [1] "nanoparquet_logical_type"
      

# LIST type

    Code
      as.data.frame(sch)
    Output
                               file_name r_col         name
      1 data/nested_lists.snappy.parquet    NA spark_schema
      2 data/nested_lists.snappy.parquet     1            a
      3 data/nested_lists.snappy.parquet     1         list
      4 data/nested_lists.snappy.parquet     1      element
      5 data/nested_lists.snappy.parquet     1         list
      6 data/nested_lists.snappy.parquet     1      element
      7 data/nested_lists.snappy.parquet     1         list
      8 data/nested_lists.snappy.parquet     1      element
      9 data/nested_lists.snappy.parquet     2            b
                             r_type       type type_length repetition_type
      1                        <NA>       <NA>          NA            <NA>
      2 list(list(list(character)))       <NA>          NA        OPTIONAL
      3                        <NA>       <NA>          NA        REPEATED
      4                        <NA>       <NA>          NA        OPTIONAL
      5                        <NA>       <NA>          NA        REPEATED
      6                        <NA>       <NA>          NA        OPTIONAL
      7                        <NA>       <NA>          NA        REPEATED
      8                        <NA> BYTE_ARRAY          NA        OPTIONAL
      9                     integer      INT32          NA        REQUIRED
        converted_type logical_type num_children scale precision field_id children
      1           <NA>                         2    NA        NA       NA         
      2           LIST                         1    NA        NA       NA        3
      3           <NA>                         1    NA        NA       NA        4
      4           LIST                         1    NA        NA       NA        5
      5           <NA>                         1    NA        NA       NA        6
      6           LIST                         1    NA        NA       NA        7
      7           <NA>                         1    NA        NA       NA        8
      8           UTF8                        NA    NA        NA       NA         
      9           <NA>                        NA    NA        NA       NA         

# MAP type

    Code
      as.data.frame(sch)
    Output
               file_name r_col        name                                 r_type
      1 data/map.parquet    NA hive_schema                                   <NA>
      2 data/map.parquet     1  raw_header list(list(list(character, character)))
      3 data/map.parquet     1         map                                   <NA>
      4 data/map.parquet     1         key                                   <NA>
      5 data/map.parquet     1       value                                   <NA>
              type type_length repetition_type converted_type logical_type
      1       <NA>          NA            <NA>           <NA>             
      2       <NA>          NA        OPTIONAL            MAP             
      3       <NA>          NA        REPEATED  MAP_KEY_VALUE             
      4 BYTE_ARRAY          NA        REQUIRED           UTF8             
      5 BYTE_ARRAY          NA        OPTIONAL           UTF8             
        num_children scale precision field_id children
      1            1    NA        NA       NA         
      2            1    NA        NA       NA        3
      3            2    NA        NA       NA     4, 5
      4           NA    NA        NA       NA         
      5           NA    NA        NA       NA         

# key-value metadata

    Code
      as.data.frame(mtd$file_meta_data$key_value_metadata[[1]])
    Output
                 key
      1            r
      2 ARROW:schema
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                               value
      1                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      A\n3\n263168\n197888\n5\nUTF-8\n531\n1\n531\n13\n254\n254\n254\n254\n254\n254\n254\n254\n254\n254\n254\n254\n254\n1026\n1\n262153\n5\nnames\n16\n13\n262153\n4\nname\n262153\n3\nmpg\n262153\n3\ncyl\n262153\n4\ndisp\n262153\n2\nhp\n262153\n4\ndrat\n262153\n2\nwt\n262153\n4\nqsec\n262153\n2\nvs\n262153\n2\nam\n262153\n4\ngear\n262153\n4\ncarb\n262153\n3\nfac\n254\n1026\n511\n16\n1\n262153\n7\ncolumns\n254\n
      2  

# parquet_info

    Code
      read_parquet_info(test_path("data/enum.parquet"))
    Output
      # A data frame: 1 x 7
        file_name         num_cols num_rows num_row_groups file_size parquet_version
        <chr>                <int>    <dbl>          <int>     <dbl>           <int>
      1 data/enum.parquet        7        2              1      3930               1
      # i 1 more variable: created_by <chr>
    Code
      read_parquet_info(test_path("data/factor.parquet"))
    Output
      # A data frame: 1 x 7
        file_name           num_cols num_rows num_row_groups file_size parquet_version
        <chr>                  <int>    <dbl>          <int>     <dbl>           <int>
      1 data/factor.parquet       14       32              1      7469               2
      # i 1 more variable: created_by <chr>
    Code
      read_parquet_info(test_path("data/decimals.parquet"))
    Output
      # A data frame: 1 x 7
        file_name           num_cols num_rows num_row_groups file_size parquet_version
        <chr>                  <int>    <dbl>          <int>     <dbl>           <int>
      1 data/decimals.parq~        4        2              1      1653               1
      # i 1 more variable: created_by <chr>

# read_parquet_row_group

    Code
      as.data.frame(read_parquet_metadata(tmp)$row_groups)[-1]
    Output
        id total_byte_size num_rows file_offset total_compressed_size ordinal
      1  0             937       10          NA                    NA      NA
      2  1             947       10          NA                    NA      NA
      3  2            1011       10          NA                    NA      NA
      4  3             528        2          NA                    NA      NA

---

    Code
      as.data.frame(read_parquet_metadata(tmp)$row_groups)[-1]
    Output
        id total_byte_size num_rows file_offset total_compressed_size ordinal
      1  0             992       10          NA                    NA      NA
      2  1            1029       10          NA                    NA      NA
      3  2            1091       10          NA                    NA      NA
      4  3             606        2          NA                    NA      NA

---

    Code
      as.data.frame(read_parquet_metadata(tmp)$row_groups)[-1]
    Output
        id total_byte_size num_rows file_offset total_compressed_size ordinal
      1  0             937       10          NA                    NA      NA
      2  1             947       10          NA                    NA      NA
      3  2            1011       10          NA                    NA      NA
      4  3             528        2          NA                    NA      NA

---

    Code
      as.data.frame(read_parquet_metadata(tmp)$row_groups)[-1]
    Output
        id total_byte_size num_rows file_offset total_compressed_size ordinal
      1  0             992       10          NA                    NA      NA
      2  1            1029       10          NA                    NA      NA
      3  2            1091       10          NA                    NA      NA
      4  3             606        2          NA                    NA      NA

# LIST schema

    Code
      as.data.frame(read_parquet_schema(pf))
    Output
                file_name r_col    name       r_type   type type_length
      1 data/list.parquet    NA  schema         <NA>   <NA>          NA
      2 data/list.parquet     1       a      integer  INT32          NA
      3 data/list.parquet     2       l list(double)   <NA>          NA
      4 data/list.parquet     2    list         <NA>   <NA>          NA
      5 data/list.parquet     2 element         <NA> DOUBLE          NA
        repetition_type converted_type logical_type num_children scale precision
      1        REQUIRED           <NA>                         2    NA        NA
      2        OPTIONAL           <NA>                        NA    NA        NA
      3        OPTIONAL           LIST         LIST            1    NA        NA
      4        REPEATED           <NA>                         1    NA        NA
      5        OPTIONAL           <NA>                        NA    NA        NA
        field_id children
      1       NA         
      2       NA         
      3       NA        4
      4       NA        5
      5       NA         

# character -> ENUM

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           s
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
           s
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
           s
      1  foo
      2  foo
      3  foo
      4  foo
      5  foo
      6  foo
      7  foo
      8  foo
      9  foo
      10 foo

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
            s
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo
      11  foo

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
            s
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo
      11  foo

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
            s
      1   foo
      2   foo
      3   foo
      4   foo
      5  <NA>
      6   foo
      7   foo
      8   foo
      9   foo
      10  foo
      11  foo

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for BYTE_ARRAY column: RLE

# factor -> ENUM

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3 foobar

---

    Code
      schema
    Output
      [1] "ENUM"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
             s
      1    foo
      2    bar
      3   <NA>
      4 foobar

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_LENGTH_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_LENGTH_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BYTE_ARRAY")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for BYTE_ARRAY column: DELTA_BYTE_ARRAY
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for BYTE_ARRAY column: RLE

# integer -> DECOMAL INT32

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# integer -> DECIMAL INT64

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# double -> DECOMAL INT32

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT32"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> DECIMAL INT64

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [[1]]
      [1] "DECIMAL"
      
      $precision
      [1] 2
      
      $scale
      [1] 1
      
      $primitive_type
      [1] "INT64"
      
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# integer -> INT(8, *)

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# integer -> INT(16, *)

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# integer -> INT(64, *)

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# double -> INT(8, TRUE)

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(8, FALSE)

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_8"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(16, TRUE)

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(16, FALSE)

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_16"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(32, TRUE)

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(32, FALSE)

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_32"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT32 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT32 column: RLE

# double -> INT(64, TRUE)

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "INT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# double -> INT(64, FALSE)

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
        d
      1 1
      2 2
      3 3
      4 4
      5 5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  2
      3 NA
      4  3
      5  4
      6  5

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
         d
      1  1
      2  1
      3  1
      4  1
      5  1
      6  1
      7  1
      8  1
      9  1
      10 1

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      schema
    Output
      [1] "UINT_64"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
          d
      1   1
      2   1
      3   1
      4   1
      5   1
      6  NA
      7   1
      8   1
      9   1
      10  1
      11  1

---

    Code
      write_parquet(d, tmp, schema = schema, encoding = "DELTA_BINARY_PACKED")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: DELTA_BINARY_PACKED
    Code
      write_parquet(d, tmp, schema = schema, encoding = "BYTE_STREAM_SPLIT")
    Condition
      Error in `write_parquet()`:
      ! Unimplemented encoding for INT64 column: BYTE_STREAM_SPLIT
    Code
      write_parquet(d, tmp, schema = schema, encoding = "RLE")
    Condition
      Error in `write_parquet()`:
      ! Unsupported encoding for INT64 column: RLE

# character -> UUID

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3                                 <NA>
      4 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3                                 <NA>
      4 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                           u
      1 00112233-4455-6677-8899-aabbccddeeff
      2 01112233-4455-6677-8899-aabbccddeeff
      3                                 <NA>
      4 02112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5  00112233-4455-6677-8899-aabbccddeeff
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5  00112233-4455-6677-8899-aabbccddeeff
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5  00112233-4455-6677-8899-aabbccddeeff
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      NULL
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5                                  <NA>
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "PLAIN"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"   "PLAIN"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
        page_type encoding
      1 DATA_PAGE    PLAIN
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5                                  <NA>
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

---

    Code
      schema
    Output
      [1] "UUID"
    Code
      encoding
    Output
      [1] "RLE_DICTIONARY"
    Code
      read_parquet_metadata(tmp)[["column_chunks"]][["encodings"]]
    Output
      [[1]]
      [1] "RLE"            "PLAIN"          "RLE_DICTIONARY"
      
    Code
      as.data.frame(read_parquet_pages(tmp))[, c("page_type", "encoding")]
    Output
              page_type       encoding
      1 DICTIONARY_PAGE          PLAIN
      2       DATA_PAGE RLE_DICTIONARY
    Code
      as.data.frame(read_parquet(tmp))
    Output
                                            u
      1  00112233-4455-6677-8899-aabbccddeeff
      2  00112233-4455-6677-8899-aabbccddeeff
      3  00112233-4455-6677-8899-aabbccddeeff
      4  00112233-4455-6677-8899-aabbccddeeff
      5                                  <NA>
      6  00112233-4455-6677-8899-aabbccddeeff
      7  00112233-4455-6677-8899-aabbccddeeff
      8  00112233-4455-6677-8899-aabbccddeeff
      9  00112233-4455-6677-8899-aabbccddeeff
      10 00112233-4455-6677-8899-aabbccddeeff

# appending all-NA rows to REQUIRED columns gives a clear error

    Code
      append_parquet(df_nas, tmp)
    Condition
      Error in `check_schema_required_cols()`:
      ! Parquet schema does not allow missing values for columns:id, value, label

# base64

    Code
      base64_encode(1:10)
    Condition
      Error in `base64_encode()`:
      ! Invalid input in base64 encoder

---

    Code
      base64_decode(1:10)
    Condition
      Error in `base64_decode()`:
      ! Invalid input in base64 decoder

---

    Code
      base64_decode("foobar;;;")
    Condition
      Error in `base64_decode()`:
      ! Base64 decoding error at position 6

# temporal types

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type type nullable dictionary custom_metadata
      1    x      Date  DAY     TRUE               characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      [[1]]$date_unit
      [1] "DAY"
      
      

---

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type       type nullable dictionary custom_metadata
      1    x      Time SECOND, 32     TRUE               characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      [[1]]$time_unit
      [1] "SECOND"
      
      [[1]]$bit_width
      [1] 32
      
      

---

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type       type nullable dictionary custom_metadata
      1    x      Time SECOND, 32     TRUE               characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      [[1]]$time_unit
      [1] "SECOND"
      
      [[1]]$bit_width
      [1] 32
      
      

---

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type       type nullable dictionary custom_metadata
      1    x  Duration NANOSECOND     TRUE               characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      [[1]]$unit
      [1] "NANOSECOND"
      
      

---

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type         type nullable dictionary custom_metadata
      1    x Timestamp MICROSEC....     TRUE               characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      [[1]]$unit
      [1] "MICROSECOND"
      
      [[1]]$timezone
      [1] "UTC"
      
      

---

    Code
      as.data.frame(np[["columns"]])
    Output
        name type_type type nullable   dictionary custom_metadata
      1    x      Utf8          TRUE 0, list(....    characte....
    Code
      np[["columns"]][["type"]]
    Output
      [[1]]
      NULL
      

# obj_sum.nanoparquet_logical_type

    Code
      lapply(sch$logical_type, obj_sum.nanoparquet_logical_type)
    Output
      [[1]]
      [1] "INT(8, TRUE)"
      
      [[2]]
      [1] "DECIMAL(1, 3)"
      
      [[3]]
      [1] "TIME"
      
      [[4]]
      [1] "TIMESTAMP"
      

# REQ PLAIN

    Code
      tapply(pgs$num_values, pgs$column + 1, sum)
    Output
          1     2     3     4 
      10000 10000 10000 10000 

---

    Code
      tapply(pgs$num_values, pgs$column + 1, sum)
    Output
          1     2     3     4 
      10000 10000 10000 10000 

# OPT PLAIN

    Code
      tapply(pgs$num_values, pgs$column + 1, sum)
    Output
          1     2     3     4 
      10000 10000 10000 10000 

---

    Code
      tapply(pgs$num_values, pgs$column + 1, sum)
    Output
          1     2     3     4 
      10000 10000 10000 10000 

# infer_parquet_schema

    Code
      as.data.frame(infer_parquet_schema(test_df(missing = FALSE)))
    Output
         file_name r_col  name    r_type       type type_length repetition_type
      1       <NA>     1   nam character BYTE_ARRAY          NA        REQUIRED
      2       <NA>     2   mpg    double     DOUBLE          NA        REQUIRED
      3       <NA>     3   cyl   integer      INT32          NA        REQUIRED
      4       <NA>     4  disp    double     DOUBLE          NA        REQUIRED
      5       <NA>     5    hp    double     DOUBLE          NA        REQUIRED
      6       <NA>     6  drat    double     DOUBLE          NA        REQUIRED
      7       <NA>     7    wt    double     DOUBLE          NA        REQUIRED
      8       <NA>     8  qsec    double     DOUBLE          NA        REQUIRED
      9       <NA>     9    vs    double     DOUBLE          NA        REQUIRED
      10      <NA>    10    am    double     DOUBLE          NA        REQUIRED
      11      <NA>    11  gear    double     DOUBLE          NA        REQUIRED
      12      <NA>    12  carb    double     DOUBLE          NA        REQUIRED
      13      <NA>    13 large   logical    BOOLEAN          NA        REQUIRED
         converted_type logical_type num_children scale precision field_id
      1            UTF8       STRING           NA    NA        NA       NA
      2            <NA>                        NA    NA        NA       NA
      3          INT_32 INT, 32,....           NA    NA        NA       NA
      4            <NA>                        NA    NA        NA       NA
      5            <NA>                        NA    NA        NA       NA
      6            <NA>                        NA    NA        NA       NA
      7            <NA>                        NA    NA        NA       NA
      8            <NA>                        NA    NA        NA       NA
      9            <NA>                        NA    NA        NA       NA
      10           <NA>                        NA    NA        NA       NA
      11           <NA>                        NA    NA        NA       NA
      12           <NA>                        NA    NA        NA       NA
      13           <NA>                        NA    NA        NA       NA

---

    Code
      as.data.frame(infer_parquet_schema(test_df(missing = TRUE)))
    Output
         file_name r_col  name    r_type       type type_length repetition_type
      1       <NA>     1   nam character BYTE_ARRAY          NA        OPTIONAL
      2       <NA>     2   mpg    double     DOUBLE          NA        OPTIONAL
      3       <NA>     3   cyl   integer      INT32          NA        OPTIONAL
      4       <NA>     4  disp    double     DOUBLE          NA        OPTIONAL
      5       <NA>     5    hp    double     DOUBLE          NA        OPTIONAL
      6       <NA>     6  drat    double     DOUBLE          NA        OPTIONAL
      7       <NA>     7    wt    double     DOUBLE          NA        OPTIONAL
      8       <NA>     8  qsec    double     DOUBLE          NA        OPTIONAL
      9       <NA>     9    vs    double     DOUBLE          NA        OPTIONAL
      10      <NA>    10    am    double     DOUBLE          NA        OPTIONAL
      11      <NA>    11  gear    double     DOUBLE          NA        OPTIONAL
      12      <NA>    12  carb    double     DOUBLE          NA        OPTIONAL
      13      <NA>    13 large   logical    BOOLEAN          NA        OPTIONAL
         converted_type logical_type num_children scale precision field_id
      1            UTF8       STRING           NA    NA        NA       NA
      2            <NA>                        NA    NA        NA       NA
      3          INT_32 INT, 32,....           NA    NA        NA       NA
      4            <NA>                        NA    NA        NA       NA
      5            <NA>                        NA    NA        NA       NA
      6            <NA>                        NA    NA        NA       NA
      7            <NA>                        NA    NA        NA       NA
      8            <NA>                        NA    NA        NA       NA
      9            <NA>                        NA    NA        NA       NA
      10           <NA>                        NA    NA        NA       NA
      11           <NA>                        NA    NA        NA       NA
      12           <NA>                        NA    NA        NA       NA
      13           <NA>                        NA    NA        NA       NA

# logical_to_converted

    Code
      logical_to_converted(list("DECIMAL"))
    Condition
      Error:
      ! Parquet decimal logical type needs scale and precision
    Code
      logical_to_converted(list("TIME", TRUE, "SECS"))
    Condition
      Error:
      ! Unknown TIME time unit: SECS
    Code
      logical_to_converted(list("TIMESTAMP", TRUE, "SECS"))
    Condition
      Error:
      ! Unknown TIMESTAMP time unit: SECS
    Code
      logical_to_converted(list("INT"))
    Condition
      Error:
      ! Parquet integer logical type needs bit width and signedness
    Code
      logical_to_converted(list("FOOBAR"))
    Condition
      Error:
      ! Unknown Parquet logical type: FOOBAR

# REPEATED columns, no LIST

    Code
      as.data.frame(read_parquet(pf))
    Output
        Int32_list             String_list
      1 0, 1, 2, 3     foo, zero, one, two
      2                              three
      3          4                    four
      4 5, 6, 7, 8 five, six, seven, eight

# Only 3-layer LIST is supported for now

    Code
      as.data.frame(read_parquet(pf))
    Condition
      Error in `read_parquet()`:
      ! Only three-layer LIST columns are supported, could not read Parquet file at 'data/old_list_structure.parquet' @ lib/ParquetReader.cpp:202

# LIST

    Code
      as.data.frame(read_parquet(test_path("data/list-req-req.parquet")))
    Output
              a
      1 1, 2, 3
      2        
      3       4
    Code
      as.data.frame(read_parquet(test_path("data/list-req-opt.parquet")))
    Output
               a
      1 1, NA, 3
      2         
      3        4
    Code
      as.data.frame(read_parquet(test_path("data/list-opt-req.parquet")))
    Output
              a
      1 1, 2, 3
      2        
      3    NULL
      4       4
    Code
      as.data.frame(read_parquet(test_path("data/list-opt-opt.parquet")))
    Output
               a
      1 1, NA, 3
      2         
      3     NULL
      4        4

---

    Code
      as.data.frame(read_parquet(test_path("data/list-v2-req-req.parquet")))
    Output
              a
      1 1, 2, 3
      2        
      3       4
    Code
      as.data.frame(read_parquet(test_path("data/list-v2-req-opt.parquet")))
    Output
               a
      1 1, NA, 3
      2         
      3        4
    Code
      as.data.frame(read_parquet(test_path("data/list-v2-opt-req.parquet")))
    Output
              a
      1 1, 2, 3
      2        
      3    NULL
      4       4
    Code
      as.data.frame(read_parquet(test_path("data/list-v2-opt-opt.parquet")))
    Output
               a
      1 1, NA, 3
      2         
      3     NULL
      4        4

---

    Code
      read_parquet_page(pf, pgoff)[elts]
    Output
      $has_repetition_levels
      [1] TRUE
      
      $has_definition_levels
      [1] TRUE
      
      $num_values
      [1] 10
      
      $num_rows
      [1] NA
      

---

    Code
      read_parquet_page(pf2, pgoff2)[elts]
    Output
      $has_repetition_levels
      [1] TRUE
      
      $has_definition_levels
      [1] TRUE
      
      $num_values
      [1] 5
      
      $num_rows
      [1] NA
      

---

    Code
      read_parquet_page(pf3, pgoff3)[elts]
    Output
      $has_repetition_levels
      [1] TRUE
      
      $has_definition_levels
      [1] TRUE
      
      $num_values
      [1] 5
      
      $num_rows
      [1] NA
      

---

    Code
      read_parquet_page(pf4, pgoff4)[elts]
    Output
      $has_repetition_levels
      [1] TRUE
      
      $has_definition_levels
      [1] TRUE
      
      $num_values
      [1] 6
      
      $num_rows
      [1] NA
      

---

    Code
      read_parquet_page(pf5, pgoff5)[elts]
    Output
      $has_repetition_levels
      [1] TRUE
      
      $has_definition_levels
      [1] TRUE
      
      $num_values
      [1] 6
      
      $num_rows
      [1] NA
      

# errors

    Code
      parquet_options(num_rows_per_row_group = "foobar")
    Condition
      Error in `as_count()`:
      ! num_rows_per_row_group must be a count, i.e. an integer scalar

---

    Code
      write_parquet(df, tmp, row_groups = "foobar")
    Condition
      Error in `parse_row_groups()`:
      ! Row groups must be specified as a growing positive integer vector, starting with 1.
    Code
      write_parquet(df, tmp, row_groups = c(100L, 1L))
    Condition
      Error in `parse_row_groups()`:
      ! Row groups must be specified as a growing positive integer vector, starting with 1.
    Code
      write_parquet(df, tmp, row_groups = c(1L, 100L))
    Condition
      Error in `write_parquet()`:
      ! Internal nanoparquet error, row index too large

# non-factors write local dictionary

    Code
      for (do in dict_ofs) {
        print(read_parquet_page(tmp, do)[["data"]])
      }
    Output
      [1] 01 00 00 00 61
      [1] 01 00 00 00 61
       [1] 01 00 00 00 61 01 00 00 00 62
      [1] 01 00 00 00 62
      [1] 01 00 00 00 62
      [1] 01 00 00 00 63
      [1] 01 00 00 00 63
      [1] 01 00 00 00 63

# errors

    Code
      .Call(rf_nanoparquet_write, mtcars, tempfile(), dim(mtcars), 0L, list(character(),
      character()), rep(FALSE, ncol(mtcars)), options, map_schema_to_df(NULL, mtcars),
      rep(10L, ncol(mtcars)), 1L, sys.call())
    Condition
      Error:
      ! Unknown Praquet encoding code: 10

# force PLAIN / RLE

    Code
      read_parquet_pages(tmp)[["encoding"]]
    Output
      [1] "PLAIN" "PLAIN"

---

    Code
      read_parquet_pages(tmp)[["encoding"]]
    Output
      [1] "PLAIN"          "RLE_DICTIONARY" "PLAIN"          "RLE_DICTIONARY"

# cutoff for dict encoding decision

    Code
      read_parquet_pages(tmp)[["encoding"]]
    Output
      [1] "PLAIN"

---

    Code
      read_parquet_pages(tmp)[["encoding"]]
    Output
      [1] "PLAIN"

# write broken DECIMAL INT32

    Code
      as.data.frame(read_parquet_schema(tmp))[, -1]
    Output
        r_col   name  r_type  type type_length repetition_type converted_type
      1    NA schema    <NA>  <NA>          NA            <NA>           <NA>
      2     1    dec integer INT32          NA        REQUIRED        DECIMAL
        logical_type num_children scale precision field_id children
      1                         1    NA        NA       NA         
      2                        NA     2         5       NA         

---

    Code
      write_parquet(d, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Invalid Parquet file: scale is not set for DECIMAL converted type

---

    Code
      write_parquet(d, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Invalid Parquet file: precision is not set for DECIMAL converted type

---

    Code
      write_parquet(d, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Internal nanoparquet error, precision to high for INT32 DECIMAL

---

    Code
      write_parquet(d2, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Internal nanoparquet error, precision to high for INT32 DECIMAL

# write broken DECIMAL INT64

    Code
      write_parquet(d, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Internal nanoparquet error, precision to high for INT64 DECIMAL

---

    Code
      write_parquet(d2, tmp, schema = schema2)
    Condition
      Error in `write_parquet()`:
      ! Internal nanoparquet error, precision to high for INT64 DECIMAL

# write broken INT32

    Code
      write_parquet(d, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

# write broken UINT32

    Code
      write_parquet(d, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

# write broken INT64

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT64 INT type: 32

# INT96 errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT96"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet INT96 type.

# FLOAT errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("FLOAT"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet FLOAT type.

# DOUBLE errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("DOUBLE"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet DOUBLE type.

# BYTE_ARRAY errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("BYTE_ARRAY"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet BYTE_ARRAY element when writing a list column of RAW vectors.

---

    Code
      write_parquet(d2, tmp, schema = parquet_schema("BYTE_ARRAY"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet BYTE_ARRAY type.

# FIXED_LEN_BYTE_ARRAY errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("UUID"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a list as a Parquet UUID (FIXED_LEN_BYTE_ARRAY) type.

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("FLOAT16"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write a list as a Parquet FLOAT16 type.

---

    Code
      write_parquet(d2, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid string length: 6, expenting 3 for FIXED_LEN_TYPE_ARRAY

---

    Code
      write_parquet(d, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet BYTE_ARRAY element when writing a list column of RAW vectors.

---

    Code
      write_parquet(d3, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Invalid string length: 4, expenting 3 for FIXED_LEN_TYPE_ARRAY

---

    Code
      write_parquet(d4, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Cannot write a logical vector as a Parquet FIXED_LEN_BYTE_ARRAY type.

# BOOLEAN errors

    Code
      write_parquet(d, tmp, schema = parquet_schema("BOOLEAN"))
    Condition
      Error in `write_parquet()`:
      ! Cannot write an integer vector as a Parquet BOOLEAN type.

---

    Code
      write_parquet(d, tmp, schema = schema)
    Condition
      Error in `write_parquet()`:
      ! Cannot write an integer vector as a Parquet BOOLEAN type.

# Errors when writing a dictionary

    Code
      write_parquet(d, tmp, schema = parquet_schema("DOUBLE"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert an integer vector to Parquet type DOUBLE.

---

    Code
      write_parquet(d2, tmp, schema = parquet_schema("BYTE_ARRAY"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a double vector to Parquet type BYTE_ARRAY.

---

    Code
      write_parquet(d3, tmp, schema = parquet_schema("DOUBLE"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a factor to Parquet type DOUBLE.

---

    Code
      write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 2, scale 0: -101

---

    Code
      write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 2, scale 0: 101

---

    Code
      write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 2, scale 0: -101

---

    Code
      write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 2, scale 0: 101

---

    Code
      write_parquet(d5, tmp, schema = parquet_schema("DOUBLE"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert an integer vector to Parquet type DOUBLE.

# POSIXct dictionary

    Code
      as.data.frame(read_parquet(tmp))
    Output
                          d
      1 2024-08-20 12:38:00
    Code
      read_parquet_metadata(tmp)[["column_chunks"]]$encodings
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      write_parquet(d, tmp, schema = parquet_schema("INT32"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a POSIXct vector to Parquet type INT32.

---

    Code
      as.data.frame(read_parquet(tmp))
    Output
             x
      1 1 secs
    Code
      read_parquet_metadata(tmp)[["column_chunks"]]$encodings
    Output
      [[1]]
      [1] "PLAIN"          "RLE_DICTIONARY"
      

---

    Code
      write_parquet(d2, tmp, schema = parquet_schema("INT32"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a difftime vector to Parquet type INT32.

# more dictionaries

    Code
      write_parquet(d4, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT32 DECIMAL with precision 2, scale 0: -101.000000

---

    Code
      write_parquet(d5, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT32 DECIMAL with precision 2, scale 0: 101.000000

---

    Code
      write_parquet(d5, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

---

    Code
      write_parquet(d5, tmp, schema = schema, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Invalid bit width for INT32: 64

---

    Code
      write_parquet(d6, tmp, schema = parquet_schema("INT_8"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 8: 128.000000 at column 1

---

    Code
      write_parquet(d7, tmp, schema = parquet_schema("UINT_8"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Integer value too large for INT with bit width 8: 256.000000 at column 1.

---

    Code
      write_parquet(d8, tmp, schema = parquet_schema("UINT_8"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Negative values are not allowed in unsigned INT column: -1.000000 at column 1.

---

    Code
      write_parquet(d9, tmp, schema = schema4, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too small for INT64 DECIMAL with precision 2, scale 0: -101.000000

---

    Code
      write_parquet(d10, tmp, schema = schema5, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Value too large for INT64 DECIMAL with precision 2, scale 0: 101.000000

---

    Code
      write_parquet(d10, tmp, schema = parquet_schema("BYTE_ARRAY"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a double vector to Parquet type BYTE_ARRAY.

# Even more dictionaries

    Code
      write_parquet(d, tmp, schema = parquet_schema("UUID"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Invalid UUID value in column 1

---

    Code
      write_parquet(d2, tmp, schema = sch2, encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Invalid string length in FIXED_LEN_BYTE_ARRAY column: 9, should be 3.

---

    Code
      write_parquet(d2, tmp, schema = parquet_schema("DOUBLE"), encoding = "RLE_DICTIONARY")
    Condition
      Error in `write_parquet()`:
      ! Cannot convert a character vector to Parquet type DOUBLE.

# R -> Parquet mapping error

    Code
      infer_parquet_schema(d)
    Condition
      Error in `write_parquet()`:
      ! Cannot map a raw vector to any Parquet type

# argument errors

    Code
      write_parquet(mtcars, 1:10)
    Condition
      Error in `path.expand()`:
      ! invalid 'path' argument
    Code
      write_parquet(mtcars, letters)
    Condition
      Error in `write_parquet()`:
      ! nanoparquet_write: filename must be a string

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parquet-metadata.R
\name{read_parquet_metadata}
\alias{read_parquet_metadata}
\alias{parquet_metadata}
\title{Read the metadata of a Parquet file}
\usage{
read_parquet_metadata(file, options = parquet_options())

parquet_metadata(file)
}
\arguments{
\item{file}{Path to a Parquet file.}

\item{options}{Options that potentially alter the default Parquet to R
type mappings, see \code{\link[=parquet_options]{parquet_options()}}.}
}
\value{
A named list with entries:
\itemize{
\item \code{file_meta_data}: a data frame with file meta data:
\itemize{
\item \code{file_name}: file name.
\item \code{version}: Parquet version, an integer.
\item \code{num_rows}: total number of rows.
\item \code{key_value_metadata}: list column of a data frames with two
character columns called \code{key} and \code{value}. This is the key-value
metadata of the file. Arrow stores its schema here.
\item \code{created_by}: A string scalar, usually the name of the software
that created the file.
}
\item \code{schema}:
data frame, the schema of the file. It has one row for
each node (inner node or leaf node). For flat files this means one
root node (inner node), always the first one, and then one row for
each "real" column. For nested schemas, the rows are in depth-first
search order. Most important columns are:
\itemize{
\item \code{file_name}: file name.
\item \code{r_col}: integer, the column index in R, starting from one.
\code{NA} for the root node. For Parquet files without list columns
this is simply the column index in the file. A list column has
multiple schema rows, and they all have the same \code{r_col}.
\item \code{name}: column name.
\item \code{r_type}: the R type that corresponds to the Parquet type.
Might be \code{NA} if \code{\link[=read_parquet]{read_parquet()}} cannot read this column. See
\link{nanoparquet-types} for the type mapping rules.
\item \code{r_type}:
\item \code{type}: data type. One of the low level data types.
\item \code{type_length}: length for fixed length byte arrays.
\item \code{repettion_type}: character, one of \code{REQUIRED}, \code{OPTIONAL} or
\code{REPEATED}.
\item \code{logical_type}: a list column, the logical types of the columns.
An element has at least an entry called \code{type}, and potentially
additional entries, e.g. \code{bit_width}, \code{is_signed}, etc.
\item \code{num_children}: number of child nodes. Should be a non-negative
integer for the root node, and \code{NA} for a leaf node.
}
\item \verb{$row_groups}: a data frame, information about the row groups.
Some important columns:
\itemize{
\item \code{file_name}: file name.
\item \code{id}: row group id, integer from zero to number of row groups
minus one.
\item \code{total_byte_size}: total uncompressed size of all column data.
\item \code{num_rows}: number of rows.
\item \code{file_offset}: where the row group starts in the file. This is
optional, so it might be \code{NA}.
\item \code{total_compressed_size}: total byte size of all compressed
(and potentially encrypted) column data in this row group.
This is optional, so it might be \code{NA}.
\item \code{ordinal}: ordinal position of the row group in the file, starting
from zero. This is optional, so it might be \code{NA}. If \code{NA}, then
the order of the row groups is as they appear in the metadata.
}
\item \verb{$column_chunks}: a data frame, information about all column chunks,
across all row groups. Some important columns:
\itemize{
\item \code{file_name}: file name.
\item \code{row_group}: which row group this chunk belongs to.
\item \code{column}: which leaf column this chunks belongs to. The order is
the same as in \verb{$schema}, but only leaf columns (i.e. columns with
\code{NA} children) are counted.
\item \code{file_path}: which file the chunk is stored in. \code{NA} means the
same file.
\item \code{file_offset}: where the column chunk begins in the file.
\item \code{type}: low level parquet data type.
\item \code{encodings}: encodings used to store this chunk. It is a list
column of character vectors of encoding names. Current possible
encodings: "PLAIN", "GROUP_VAR_INT", "PLAIN_DICTIONARY", "RLE", "BIT_PACKED", "DELTA_BINARY_PACKED", "DELTA_LENGTH_BYTE_ARRAY", "DELTA_BYTE_ARRAY", "RLE_DICTIONARY", "BYTE_STREAM_SPLIT".
\item \code{path_in_scema}: list column of character vectors. It is simply
the path from the root node. It is simply the column name for
flat schemas.
\item \code{codec}: compression codec used for the column chunk. Possible
values are: "UNCOMPRESSED", "SNAPPY", "GZIP", "LZO", "BROTLI", "LZ4", "ZSTD".
\item \code{num_values}: number of values in this column chunk.
\item \code{total_uncompressed_size}: total uncompressed size in bytes.
\item \code{total_compressed_size}: total compressed size in bytes.
\item \code{data_page_offset}: absolute position of the first data page of
the column chunk in the file.
\item \code{index_page_offset}: absolute position of the first index page of
the column chunk in the file, or \code{NA} if there are no index pages.
\item \code{dictionary_page_offset}: absolute position of the first
dictionary page of the column chunk in the file, or \code{NA} if there
are no dictionary pages.
\item \code{null_count}: the number of missing values in the column chunk.
It may be \code{NA}.
\item \code{min_value}: list column of raw vectors, the minimum value of the
column, in binary. If \code{NULL}, then then it is not specified.
This column is experimental.
\item \code{max_value}: list column of raw vectors, the maximum value of the
column, in binary. If \code{NULL}, then then it is not specified.
This column is experimental.
\item \code{is_min_value_exact}: whether the minimum value is an actual
value of a column, or a bound. It may be \code{NA}.
\item \code{is_max_value_exact}: whether the maximum value is an actual
value of a column, or a bound. It may be \code{NA}.
}
}
}
\description{
This function should work on all files, even if \code{\link[=read_parquet]{read_parquet()}} is
unable to read them, because of an unsupported schema, encoding,
compression or other reason.
}
\examples{
file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
nanoparquet::read_parquet_metadata(file_name)
}
\seealso{
\code{\link[=read_parquet_info]{read_parquet_info()}} for a much shorter summary.
\code{\link[=read_parquet_schema]{read_parquet_schema()}} for column information.
\code{\link[=read_parquet]{read_parquet()}} to read, \code{\link[=write_parquet]{write_parquet()}} to write Parquet files,
\link{nanoparquet-types} for the R <-> Parquet type mappings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/porcelain.R
\name{rle_encode_int}
\alias{rle_encode_int}
\title{RLE encode integers}
\usage{
rle_encode_int(x)
}
\arguments{
\item{x}{Integer vector.}
}
\value{
Raw vector, the encoded integers. It has two attributes:
\itemize{
\item \code{bit_length}: the number of bits needed to encode the input, and
\item \code{length}: length of the original integer input.
}
}
\description{
RLE encode integers
}
\seealso{
\code{\link[=rle_decode_int]{rle_decode_int()}}

Other encodings:
\code{\link[=rle_decode_int]{rle_decode_int()}}
}
\concept{encodings}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-parquet.R
\name{write_parquet}
\alias{write_parquet}
\title{Write a data frame to a Parquet file}
\usage{
write_parquet(
  x,
  file,
  schema = NULL,
  compression = c("snappy", "gzip", "zstd", "uncompressed"),
  encoding = NULL,
  metadata = NULL,
  row_groups = NULL,
  options = parquet_options()
)
}
\arguments{
\item{x}{Data frame to write.}

\item{file}{Path to the output file. If this is the string \code{":raw:"},
then the data frame is written to a memory buffer, and the memory
buffer is returned as a raw vector. If \code{":stdout:"}, then it is
written to the standard output. (When writing to the standard output,
special care is needed to make sure no regular R output gets mixed
up with the Parquet bytes!)}

\item{schema}{Parquet schema. Specify a schema to tweak the default
nanoparquet R -> Parquet type mappings. Use \code{\link[=parquet_schema]{parquet_schema()}} to
create a schema that you can use here, or \code{\link[=read_parquet_schema]{read_parquet_schema()}} to
use the schema of a Parquet file.}

\item{compression}{Compression algorithm to use. Currently \code{"snappy"}
(the default), \code{"gzip"}, \code{"zstd"}, and \code{"uncompressed"} are supported.}

\item{encoding}{Encoding to use. Possible values:
\itemize{
\item If \code{NULL}, the appropriate encoding is selected automatically:
\code{RLE} or \code{PLAIN} for \code{BOOLEAN} columns, \code{RLE_DICTIONARY} for other
columns with many repeated values, and \code{PLAIN} otherwise.
\item If It is a single (unnamed) character string, then it'll be used
for all columns.
\item If it is an unnamed character vector of encoding names of the same
length as the number of columns in the data frame, then those
encodings will be used for each column.
\item If it is a named character vector, then the named must be unique
and each name must match a column name, to specify the encoding of
that column. The special empty name (\code{""}) applies to the rest of
the columns. If there is no empty name, the rest of the columns
will use the default encoding.
}

If \code{NA_character_} is specified for a column, the default encoding is
used for the column.

If a specified encoding is invalid for a certain column type,
or nanoparquet does not implement it, \code{write_parquet()} throws an
error.

Currently \code{write_parquet()} supports the following encodings:
\itemize{
\item \code{PLAIN} for all column types,
\item \code{PLAIN_DICTIONARY} and \code{RLE_DICTIONARY} for all column types,
\item \code{RLE} for BOOLEAN columns.
}

See \link{parquet-encodings} for more about encodings.}

\item{metadata}{Additional key-value metadata to add to the file.
This must be a named character vector, or a data frame with columns
character columns called \code{key} and \code{value}.}

\item{row_groups}{Row groups of the Parquet file. If \code{NULL}, then the
\code{num_rows_per_row_group} option is used from the \code{options} argument,
see \code{\link[=parquet_options]{parquet_options()}}. Otherwise it must be an integer vector,
specifying the starts of the row groups.}

\item{options}{Nanoparquet options, see \code{\link[=parquet_options]{parquet_options()}}.}
}
\value{
\code{NULL}, unless \code{file} is \code{":raw:"}, in which case the Parquet
file is returned as a raw vector.
}
\description{
Writes the contents of an R data frame into a Parquet file.
}
\details{
\code{write_parquet()} converts string columns to UTF-8 encoding by calling
\code{\link[base:enc2utf8]{base::enc2utf8()}}. It does the same for factor levels.
}
\examples{
\dontshow{if (FALSE) withAutoprint(\{ # examplesIf}
# add row names as a column, because `write_parquet()` ignores them.
mtcars2 <- cbind(name = rownames(mtcars), mtcars)
write_parquet(mtcars2, "mtcars.parquet")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=read_parquet_metadata]{read_parquet_metadata()}}, \code{\link[=read_parquet]{read_parquet()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write-parquet.R
\name{append_parquet}
\alias{append_parquet}
\title{Append a data frame to an existing Parquet file}
\usage{
append_parquet(
  x,
  file,
  compression = c("snappy", "gzip", "zstd", "uncompressed"),
  encoding = NULL,
  row_groups = NULL,
  options = parquet_options()
)
}
\arguments{
\item{x}{Data frame to append.}

\item{file}{Path to the output file.}

\item{compression}{Compression algorithm to use for the newly written
data. See \code{\link[=write_parquet]{write_parquet()}}.}

\item{encoding}{Encoding to use for the newly written data. It does not
have to be the same as the encoding of data in \code{file}. See
\code{\link[=write_parquet]{write_parquet()}} for possible values.}

\item{row_groups}{Row groups of the new, extended Parquet file.
\code{\link[=append_parquet]{append_parquet()}} can only change the last existing row group, and
if \code{row_groups} is specified, it has respect this. I.e. if the
existing file has \code{n} rows, and the last row group starts at \code{k}
(\code{k <= n}), then the first row group in \code{row_groups} that refers to
the new data must start at \code{k} or \code{n+1}.
(It is simpler to specify \code{num_rows_per_row_group} in \code{options}, see
\code{\link[=parquet_options]{parquet_options()}} instead of \code{row_groups}. Only use \code{row_groups} if
you need complete control.)}

\item{options}{Nanoparquet options, for the new data, see
\code{\link[=parquet_options]{parquet_options()}}. The \code{keep_row_groups} option also affects whether
\code{append_parquet()} overwrites existing row groups in \code{file}.}
}
\description{
The schema of the data frame must be compatible with the schema of
the file.
}
\section{Warning}{

This function is \strong{not} atomic! If it is interrupted, it may leave
the file in a corrupt state. To work around this create a copy of the
original file, append the new data to the copy, and then rename the
new, extended file to the original one.
}

\section{About row groups}{

A Parquet file may be partitioned into multiple row groups, and indeed
most large Parquet files are. \code{append_parquet()} is only able to update
the existing file along the row group boundaries. There are two
possibilities:
\itemize{
\item \code{append_parquet()} keeps all existing row groups in \code{file}, and
creates new row groups for the new data. This mode can be forced by
the \code{keep_row_groups} option in \code{options}, see \code{\link[=parquet_options]{parquet_options()}}.
\item Alternatively, \code{write_parquet} will overwrite the \emph{last} row group in
file, with its existing contents plus the (beginning of) the new data.
This mode makes more sense if the last row group is small, because
many small row groups are inefficient.
}

By default \code{append_parquet} chooses between the two modes automatically,
aiming to create row groups with at least \code{num_rows_per_row_group}
(see \code{\link[=parquet_options]{parquet_options()}}) rows. You can customize this behavior with
the \code{keep_row_groups} options and the \code{row_groups} argument.
}

\seealso{
\code{\link[=write_parquet]{write_parquet()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parquet-metadata.R
\name{read_parquet_info}
\alias{read_parquet_info}
\alias{parquet_info}
\title{Short summary of a Parquet file}
\usage{
read_parquet_info(file)

parquet_info(file)
}
\arguments{
\item{file}{Path to a Parquet file.}
}
\value{
Data frame with columns:
\itemize{
\item \code{file_name}: file name.
\item \code{num_cols}: number of columns. (The number of child nodes of the
root node in the schema.)
\item \code{num_rows}: number of rows.
\item \code{num_row_groups}: number of row groups.
\item \code{file_size}: file size in bytes.
\item \code{parquet_version}: Parquet version.
\item \code{created_by}: A string scalar, usually the name of the software
that created the file. \code{NA} if not available.
}
}
\description{
Short summary of a Parquet file
}
\seealso{
\code{\link[=read_parquet_metadata]{read_parquet_metadata()}} to read more metadata,
\code{\link[=read_parquet_schema]{read_parquet_schema()}} for column information.
\code{\link[=read_parquet]{read_parquet()}}, \code{\link[=write_parquet]{write_parquet()}}, \link{nanoparquet-types}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read-parquet.R
\name{read_parquet}
\alias{read_parquet}
\title{Read a Parquet file into a data frame}
\usage{
read_parquet(file, col_select = NULL, options = parquet_options())
}
\arguments{
\item{file}{Path to a Parquet file. It may also be an R connection,
in which case it first reads all data from the connection, writes
it into a temporary file, then reads the temporary file, and
deletes it. The connection might be open, it which case it must be
a binary connection. If it is not open, then \code{read_parquet()} will
open it and also close it in the end.}

\item{col_select}{Columns to read. It can be a numeric vector of column
indices, or a character vector of column names. It is an error to
select the same column multiple times. The order of the columns in
the result is the same as the order in \code{col_select}.}

\item{options}{Nanoparquet options, see \code{\link[=parquet_options]{parquet_options()}}.}
}
\value{
A \code{data.frame} with the file's contents.
}
\description{
Converts the contents of the named Parquet file to a R data frame.
}
\examples{
file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
parquet_df <- nanoparquet::read_parquet(file_name)
print(str(parquet_df))
}
\seealso{
See \code{\link[=write_parquet]{write_parquet()}} to write Parquet files,
\link{nanoparquet-types} for the R <-> Parquet type mapping.
See \code{\link[=read_parquet_info]{read_parquet_info()}}, for general information,
\code{\link[=read_parquet_schema]{read_parquet_schema()}} for information about the
columns, and \code{\link[=read_parquet_metadata]{read_parquet_metadata()}} for the complete metadata.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parquet-column-types.R
\name{parquet_column_types}
\alias{parquet_column_types}
\title{Map between R and Parquet data types}
\usage{
parquet_column_types(x, options = parquet_options())
}
\arguments{
\item{x}{Path to a Parquet file, or a data frame.}

\item{options}{Nanoparquet options, see \code{\link[=parquet_options]{parquet_options()}}.}
}
\value{
Data frame with columns:
\itemize{
\item \code{file_name}: file name.
\item \code{name}: column name.
\item \code{type}: (low level) Parquet data type.
\item \code{r_type}: the R type that corresponds to the Parquet type.
Might be \code{NA} if \code{\link[=read_parquet]{read_parquet()}} cannot read this column. See
\link{nanoparquet-types} for the type mapping rules.
\item \code{repetition_type}: whether the column in \code{REQUIRED} (cannot be
\code{NA}) or \code{OPTIONAL} (may be \code{NA}). \code{REPEATED} columns are not
currently supported by nanoparquet.
\item \code{logical_type}: Parquet logical type in a list column.
An element has at least an entry called \code{type}, and potentially
additional entries, e.g. \code{bit_width}, \code{is_signed}, etc.
}
}
\description{
Note that this function is now deprecated. Please use
\code{\link[=read_parquet_schema]{read_parquet_schema()}} for files, and \code{\link[=infer_parquet_schema]{infer_parquet_schema()}} for
data frames.
}
\details{
This function works two ways. It can map the R types of a data frame to
Parquet types, to see how \code{\link[=write_parquet]{write_parquet()}} would write out the data
frame. It can also map the types of a Parquet file to R types, to see
how \code{\link[=read_parquet]{read_parquet()}} would read the file into R.
}
\seealso{
\code{\link[=read_parquet_metadata]{read_parquet_metadata()}} to read more metadata,
\code{\link[=read_parquet_info]{read_parquet_info()}} for a very short summary.
\code{\link[=read_parquet_schema]{read_parquet_schema()}} for the complete Parquet schema.
\code{\link[=read_parquet]{read_parquet()}}, \code{\link[=write_parquet]{write_parquet()}}, \link{nanoparquet-types}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/porcelain.R
\name{rle_decode_int}
\alias{rle_decode_int}
\title{RLE decode integers}
\usage{
rle_decode_int(
  x,
  bit_width = attr(x, "bit_width"),
  length = attr(x, "length") \%||\% NA
)
}
\arguments{
\item{x}{Raw vector of the encoded integers.}

\item{bit_width}{Bit width used for the encoding.}

\item{length}{Length of the output. If \code{NA} then we assume that \code{x}
starts with length of the output, encoded as a 4 byte integer.}
}
\value{
The decoded integer vector.
}
\description{
RLE decode integers
}
\seealso{
\code{\link[=rle_encode_int]{rle_encode_int()}}

Other encodings:
\code{\link[=rle_encode_int]{rle_encode_int()}}
}
\concept{encodings}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nanoparquet-package.R
\name{parquet-encodings}
\alias{parquet-encodings}
\title{Parquet encodings}
\description{
Various Parquet encodings
}
\section{Nanoparquet defaults}{
Currently the defaults are decided based on the R types. This might
change in the future. In general, the defaults will likely change until
nanoparquet reaches version 1.0.0.

Current encoding defaults:
\itemize{
\item Definition levels always use \code{RLE}. (Nanoparquet does not currently
write repetition levels, but they'll also use \code{RLE}, once implemented.)
\item \code{factor} columns use \code{RLE_DICTIONARY}.
\item \code{logical} columns use \code{RLE} if the average run length of the first
10,000 values is at least 15. Otherwise they use the \code{PLAIN} encoding.
\item \code{integer}, \code{double} and \code{character} columns use \code{RLE_DICTIONARY} if at
least two third of their values are repeated. Otherwise they use
\code{PLAIN} encoding.
\item \code{list} columns of \code{raw} vectors always use the \code{PLAIN} encoding
currently.
}
}

\section{Parquet encodings}{
See \url{https://github.com/apache/parquet-format/blob/master/Encodings.md}
for more details on Parquet encodings.
\subsection{\code{PLAIN} encoding}{

Supported types: all.

In general values are written back to back:
\itemize{
\item Integer types are little endian.
\item Floating point types follow the IEEE standard.
\item \code{BYTE_ARRAY}: for each element, there is a little endian 4-byte length
and then the bytes themselves.
\item \code{FIXED_LEN_BYTE_ARRAY}: bytes are written back to back.
}

Nanoparquet can read and write this encoding for all primitive types.
}

\subsection{\code{RLE_DICTIONARY} encoding}{

Supported types: dictionary indices in data pages.

This encoding combines run-length encoding and bit-packing.
Repeated sequences of the same value can be run-length encoded, and
non-repeated parts are bit packed.
It is used for data pages of dictionaries.
The dictionary pages themselves are \code{PLAIN} encoded.

The deprecated \code{PLAIN_DICTIONARY} name is treated the same as
\code{RLE_DICTIONARY}.

Nanoparquet can read and write this encoding.
}

\subsection{\code{RLE} encoding}{

Supported types: \code{BOOLEAN}. Also for definition and repetition levels.

This is the same encoding as \code{RLE_DICTIONARY}, with a slightly different
header. It combines run-length encoding and bit packing.
It is used for \code{BOOLEAN} columns, and also for definition and
repetition levels.

Nanoparquet can read and write this encoding.
}

\subsection{\code{BIT_PACKED} encoding (deprecated in favor of \code{RLE})}{

Supported types: none. Only for definition and repetition levels, but
\code{RLE} should be used instead.

This is a simple bit packing encoding for integers, that was previously
used for encoding definition and repetition levels. It is not used in new
Parquet files because the the \code{RLE} encoding includes it and it is better.

Nanoparquet currently cannot read or write the \code{BIT_PACKED} encoding.
}

\subsection{\code{DELTA_BINARY_PACKED} encoding}{

Supported types: \code{INT32}, \code{INT64}.

This encoding efficiently encodes integer columns if the differences
between consecutive elements are often the same, and/or the differences
between consecutive elements are small. The extreme case of an arithmetic
sequence can be encoded in O(1) space.

Nanoparquet can read this encoding, but cannot currently write it.
}

\subsection{\code{DELTA_LENGTH_BYTE_ARRAY} encoding}{

Supported types: \code{BYTE_ARRAY}.

This encoding uses \code{DELTA_BINARY_PACKED} to encode the length of all
byte array elements. It is especially efficient for short byte array
elements, i.e. a column of short strings.

Nanoparquet can read this encoding, but cannot currently write it.
}

\subsection{\code{DELTA_BYTE_ARRAY} encoding}{

Supported types: \code{BYTE_ARRAY}, \code{FIXED_LEN_BYTE_ARRAY}.

This encoding is efficient if consecutive byte array elements share the
same prefix, because each element can reuse a prefix of the previous
element.

Nanoparquet can read this encoding, but cannot currently write it.
}

\subsection{\code{BYTE_STREAM_SPLIT} encoding}{

Supported types: \code{FLOAT}, \code{DOUBLE}, \code{INT32}, \code{INT64},
\code{FIXED_LEN_BYTE_ARRAY}.

This encoding stores the first bytes of the elements first, then the
second bytes, etc. It does not reduce the size in itself, but may allow
more efficient compression.

Nanoparquet can read this encoding, but cannot currently write it.
}
}

\seealso{
\code{\link[=write_parquet]{write_parquet()}} on how to select a non-default encoding when
writing Parquet files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/porcelain.R
\name{read_parquet_page}
\alias{read_parquet_page}
\title{Read a page from a Parquet file}
\usage{
read_parquet_page(file, offset)
}
\arguments{
\item{file}{Path to a Parquet file.}

\item{offset}{Integer offset of the start of the page in the file.
See \code{\link[=read_parquet_pages]{read_parquet_pages()}} for a list of all pages and their offsets.}
}
\value{
Named list. Many entries correspond to the columns of
the result of \code{\link[=read_parquet_pages]{read_parquet_pages()}}. Additional entries are:
\itemize{
\item \code{codec}: compression codec. Possible values:
\item \code{has_repetition_levels}: whether the page has repetition levels.
\item \code{has_definition_levels}: whether the page has definition levels.
\item \code{schema_column}: which schema column the page corresponds to. Note
that only leaf columns have pages.
\item \code{data_type}: low level Parquet data type. Possible values:
\item \code{repetition_type}: whether the column the page belongs to is
\code{REQUIRED}, \code{OPTIONAL} or \code{REPEATED}.
\item \code{page_header}: the bytes of the page header in a raw vector.
\item \code{num_null}: number of missing (\code{NA}) values. Only set in V2 data
pages.
\item \code{num_rows}: this is the same as \code{num_values} for flat tables, i.e.
files without repetition levels.
\item \code{compressed_data}: the data of the page in a raw vector. It includes
repetition and definition levels, if any.
\item \code{data}: the uncompressed data, if nanoparquet supports the
compression codec of the file (GZIP and SNAPPY at the time of
writing), or if the file is not compressed. In the latter case it
is the same as \code{compressed_data}.
}
}
\description{
Read a page from a Parquet file
}
\examples{
\dontshow{if (Sys.getenv("IN_PKGDOWN") == "true") withAutoprint(\{ # examplesIf}
file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
nanoparquet:::read_parquet_pages(file_name)
options(max.print = 100)  # otherwise long raw vector
nanoparquet:::read_parquet_page(file_name, 4L)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[=read_parquet_pages]{read_parquet_pages()}} for a summary of all pages.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/schema.R
\name{parquet_schema}
\alias{parquet_schema}
\title{Create a Parquet schema}
\usage{
parquet_schema(...)
}
\arguments{
\item{...}{Parquet type specifications, see below.
For backwards compatibility, you can supply a file name
here, and then \code{parquet_schema} behaves as \code{\link[=read_parquet_schema]{read_parquet_schema()}}.}
}
\value{
Data frame with the same columns as \code{\link[=read_parquet_schema]{read_parquet_schema()}}:
\code{file_name}, \code{r_col}, \code{name}, \code{r_type}, \code{type}, \code{type_length}, \code{repetition_type}, \code{converted_type}, \code{logical_type}, \code{num_children}, \code{scale}, \code{precision}, \code{field_id}.
}
\description{
You can use this schema to specify how to write out a data frame to
a Parquet file with \code{\link[=write_parquet]{write_parquet()}}.
}
\details{
A schema is a list of potentially named type specifications. A schema
is stored in a data frame. Each (potentially named) argument of
\code{parquet_schema} may be a character scalar, or a list. Parameterized
types need to be specified as a list. Primitive Parquet types may be
specified as a string or a list.
}
\section{Possible types:}{
Special type:
\itemize{
\item \code{"AUTO"}: this is not a Parquet type, but it tells \code{\link[=write_parquet]{write_parquet()}}
to map the R type to Parquet automatically, using the default mapping
rules.
}

Primitive Parquet types:
\itemize{
\item \code{"BOOLEAN"}
\item \code{"INT32"}
\item \code{"INT64"}
\item \code{"INT96"}
\item \code{"FLOAT"}
\item \code{"DOUBLE"}
\item \code{"BYTE_ARRAY"}
\item \code{"FIXED_LEN_BYTE_ARRAY"}: fixed-length byte array. It needs a
\code{type_length} parameter, an integer between 0 and 2^31-1.
}

Parquet logical types:
\itemize{
\item \code{"STRING"}
\item \code{"ENUM"}
\item \code{"UUID"}
\item \code{"INTEGER"}: signed or unsigned integer. It needs a \code{bit_width} and
an \code{is_signed} parameter. \code{bit_width} must be 8, 16, 32 or 64.
\code{is_signed} must be \code{TRUE} or \code{FALSE}.
\item \code{"INT"}: same as \code{"INTEGER"}. The Parquet documentation uses \code{"INT"},
but the actual specification uses \code{"INTEGER"}. Both are supported in
nanoparquet.
\item \code{"DECIMAL"}: decimal number of specified scale and precision.
It needs the \code{precision} and \code{primitive_type} parameters. Also
supports the \code{scale} parameter, it defaults to zero if not specified.
\item \code{"FLOAT16"}
\item \code{"DATE"}
\item \code{"TIME"}: needs an \code{is_adjusted_utc} (\code{TRUE} or \code{FALSE}) and a
\code{unit} parameter. \code{unit} must be \code{"MILLIS"}, \code{"MICROS"} or \code{"NANOS"}.
\item \code{"TIMESTAMP"}: needs an \code{is_adjusted_utc} (\code{TRUE} or \code{FALSE}) and a
\code{unit} parameter. \code{unit} must be \code{"MILLIS"}, \code{"MICROS"} or \code{"NANOS"}.
\item \code{"JSON"}
\item \code{"BSON"}
\item \code{"LIST"}: list of some other type. It needs an \code{element} parameter,
which is a type specification for the list elements. Currently
\code{element} can be only \code{"INT32"}, \code{"DOUBLE"} or \code{"STRING"}.
Also, currently nanoparquet always write LIST columns that are
\code{"OPTIONAL"}, both for the list elements and the list itself.
}

Logical types \code{MAP}, and \code{UNKNOWN} are not supported currently.

Converted types are deprecated in the Parquet specification in favor of
logical types, but \code{parquet_schema()} accepts some converted types as a
syntactic shortcut for the corresponding logical types:
\itemize{
\item \code{INT_8} mean \code{list("INT", bit_width = 8, is_signed = TRUE)}.
\item \code{INT_16} mean \code{list("INT", bit_width = 16, is_signed = TRUE)}.
\item \code{INT_32} mean \code{list("INT", bit_width = 32, is_signed = TRUE)}.
\item \code{INT_64} mean \code{list("INT", bit_width = 64, is_signed = TRUE)}.
\item \code{TIME_MICROS} means \code{list("TIME", is_adjusted_utc = TRUE, unit = "MICROS")}.
\item \code{TIME_MILLIS} means \code{list("TIME", is_adjusted_utc = TRUE, unit = "MILLIS")}.
\item \code{TIMESTAMP_MICROS} means \code{list("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MICROS")}.
\item \code{TIMESTAMP_MILLIS} means \code{list("TIMESTAMP", is_adjusted_utc = TRUE, unit = "MILLIS")}.
\item \code{UINT_8} means \code{list("INT", bit_width = 8, is_signed = FALSE)}.
\item \code{UINT_16} means \code{list("INT", bit_width = 16, is_signed = FALSE)}.
\item \code{UINT_32} means \code{list("INT", bit_width = 32, is_signed = FALSE)}.
\item \code{UINT_64} means \code{list("INT", bit_width = 64, is_signed = FALSE)}.
}
\subsection{Missing values}{

Each type might also have a \code{repetition_type} parameter, with possible
values \code{"REQUIRED"}, \code{"OPTIONAL"} or \code{"REPEATED"}. \code{"REQUIRED"} columns
do not allow missing values. Missing values are allowed in \code{"OPTIONAL"}
columns. \code{"REPEATED"} columns are currently not supported in
\code{\link[=write_parquet]{write_parquet()}}.
}
}

\examples{
parquet_schema(
  c1 = "INT32",
  c2 = list("INT", bit_width = 64, is_signed = TRUE),
  c3 = list("STRING", repetition_type = "OPTIONAL"),
  l = list("LIST", element = "DOUBLE")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parquet-metadata.R
\name{read_parquet_schema}
\alias{read_parquet_schema}
\title{Read the schema of a Parquet file}
\usage{
read_parquet_schema(file, options = parquet_options())
}
\arguments{
\item{file}{Path to a Parquet file.}

\item{options}{Return value of \code{\link[=parquet_options]{parquet_options()}}, options that
potentially modify the Parquet to R type mappings.}
}
\value{
\if{html}{\out{<div class="sourceCode">}}\preformatted{Data frame, the schema of the file. It has one row for
each node (inner node or leaf node). For flat files this means one
root node (inner node), always the first one, and then one row for
each "real" column. For nested schemas, the rows are in depth-first
search order. Most important columns are:
- `file_name`: file name.
- `r_col`: integer, the column index in R, starting from one.
  `NA` for the root node. For Parquet files without list columns
  this is simply the column index in the file. A list column has
  multiple schema rows, and they all have the same `r_col`.
- `name`: column name.
- `r_type`: the R type that corresponds to the Parquet type.
  Might be `NA` if [read_parquet()] cannot read this column. See
  [nanoparquet-types] for the type mapping rules.
- `type`: data type. One of the low level data types.
- `type_length`: length for fixed length byte arrays.
- `repettion_type`: character, one of `REQUIRED`, `OPTIONAL` or
  `REPEATED`.
- `logical_type`: a list column, the logical types of the columns.
  An element has at least an entry called `type`, and potentially
  additional entries, e.g. `bit_width`, `is_signed`, etc.
- `num_children`: number of child nodes. Should be a non-negative
  integer for the root node, and `NA` for a leaf node.
}\if{html}{\out{</div>}}
}
\description{
This function should work on all files, even if \code{\link[=read_parquet]{read_parquet()}} is
unable to read them, because of an unsupported schema, encoding,
compression or other reason.
}
\seealso{
\code{\link[=read_parquet_metadata]{read_parquet_metadata()}} to read more metadata,
\code{\link[=read_parquet_info]{read_parquet_info()}} to show only basic information.
\code{\link[=read_parquet]{read_parquet()}}, \code{\link[=write_parquet]{write_parquet()}}, \link{nanoparquet-types}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{parquet_options}
\alias{parquet_options}
\title{Nanoparquet options}
\usage{
parquet_options(
  class = getOption("nanoparquet.class", "tbl"),
  compression_level = getOption("nanoparquet.compression_level", NA_integer_),
  read_int64_type = getOption("nanoparquet.read_int64_type", "double"),
  keep_row_groups = FALSE,
  num_rows_per_row_group = getOption("nanoparquet.num_rows_per_row_group", 10000000L),
  use_arrow_metadata = getOption("nanoparquet.use_arrow_metadata", TRUE),
  write_arrow_metadata = getOption("nanoparquet.write_arrow_metadata", TRUE),
  write_data_page_version = getOption("nanoparquet.write_data_page_version", 1L),
  write_minmax_values = getOption("nanoparquet.write_minmax_values", TRUE)
)
}
\arguments{
\item{class}{The extra class or classes to add to data frames created
in \code{\link[=read_parquet]{read_parquet()}}. By default nanoparquet adds the \code{"tbl"} class,
so data frames are printed differently if the pillar package is
loaded.}

\item{compression_level}{The compression level in \code{\link[=write_parquet]{write_parquet()}}.
\code{NA} is the default, and it specifies the default
compression level of each method. \code{Inf} always selects the highest
possible compression level. More details:
\itemize{
\item Snappy does not support compression levels currently.
\item GZIP supports levels from 0 (uncompressed), 1 (fastest), to 9 (best).
The default is 6.
\item ZSTD allows positive levels up to 22 currently. 20 and above require
more memory. Negative levels are also allowed, the lower the level,
the faster the speed, at the cost of compression. Currently the
smallest level is -131072. The default level is 3.
}}

\item{read_int64_type}{How to represent INT64 columns with a 64-bit integer
logical type in \code{\link[=read_parquet]{read_parquet()}} and \code{\link[=read_parquet_schema]{read_parquet_schema()}}.
Possible values:
\itemize{
\item \code{"double"} (the default): read as a regular R \code{double} vector,
which may lose precision for large values.
\item \code{"integer64"} or \code{"bit64::integer64"}: read as a \code{bit64::integer64}
vector. Requires the bit64 package to be installed.
}}

\item{keep_row_groups}{This option is used when appending to a Parquet
file with \code{\link[=append_parquet]{append_parquet()}}. If \code{TRUE} then the existing row groups
of the file are always kept as is and nanoparquet creates new row
groups for the new data. If \code{FALSE} (the default), then the last row
group of the file will be overwritten if it is smaller than the
default row group size, i.e. \code{num_rows_per_row_group}.}

\item{num_rows_per_row_group}{The number of rows to put into a row
group, if row groups are not specified explicitly. It should be
an integer scalar. Defaults to 10 million.}

\item{use_arrow_metadata}{\code{TRUE} or \code{FALSE}. If \code{TRUE}, then
\code{\link[=read_parquet]{read_parquet()}} and \code{\link[=read_parquet_schema]{read_parquet_schema()}} will make use of the Apache
Arrow metadata to assign R classes to Parquet columns.
This is currently used to detect factor columns, and to detect
"difftime" columns.

If this option is \code{FALSE}:
\itemize{
\item "factor" columns are read as character vectors.
\item "difftime" columns are read as real numbers, meaning one
of seconds, milliseconds, microseconds or nanoseconds. Impossible
to tell which without using the Arrow metadata.
}}

\item{write_arrow_metadata}{Whether to add the Apache Arrow types as
metadata to the file \code{\link[=write_parquet]{write_parquet()}}.}

\item{write_data_page_version}{Data version to write by default.
Possible values are 1 and 2. Default is 1.}

\item{write_minmax_values}{Whether to write minimum and maximum values
per row group, for data types that support this in \code{\link[=write_parquet]{write_parquet()}}.
However, nanoparquet currently does not support minimum and maximum
values for the \code{DECIMAL}, \code{UUID} and \code{FLOAT16} logical types and the
\code{BOOLEAN}, \code{BYTE_ARRAY} and \code{FIXED_LEN_BYTE_ARRAY} primitive types
if they are writing without a logical type. Currently the default
is \code{TRUE}.}
}
\value{
List of nanoparquet options.
}
\description{
Create a list of nanoparquet options.
}
\examples{
\dontshow{if (FALSE) withAutoprint(\{ # examplesIf}
# the effect of using Arrow metadata
tmp <- tempfile(fileext = ".parquet")
d <- data.frame(
  fct = as.factor("a"),
  dft = as.difftime(10, units = "secs")
)
write_parquet(d, tmp)
read_parquet(tmp, options = parquet_options(use_arrow_metadata = TRUE))
read_parquet(tmp, options = parquet_options(use_arrow_metadata = FALSE))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nanoparquet-package.R
\docType{package}
\name{nanoparquet-package}
\alias{nanoparquet}
\alias{nanoparquet-package}
\title{nanoparquet: Read and Write 'Parquet' Files}
\description{
Self-sufficient reader and writer for flat 'Parquet' files. Can read most 'Parquet' data types. Can write many 'R' data types, including factors and temporal types. See docs for limitations.
}
\details{
\code{nanoparquet} is a reader and writer for a common subset of Parquet files.
\subsection{Features:}{
\itemize{
\item Read and write flat (i.e. non-nested) Parquet files.
\item Can read most \href{https://nanoparquet.r-lib.org/reference/nanoparquet-types.html}{Parquet data types}.
\item Can read a subset of columns from a Parquet file.
\item Can write many R data types, including factors and temporal types
to Parquet.
\item Can append a data frame to a Parquet file without first reading and then
rewriting the whole file.
\item Completely dependency free.
\item Supports Snappy, Gzip and Zstd compression.
\item \href{https://nanoparquet.r-lib.org/dev/articles/benchmarks.html}{Competitive} with other
tools in terms of speed, memory use and file size.
}
}

\subsection{Limitations:}{
\itemize{
\item Nested Parquet types are not supported.
\item Some Parquet logical types are not supported: \code{INTERVAL},
\code{UNKNOWN}.
\item Only Snappy, Gzip and Zstd compression is supported.
\item Encryption is not supported.
\item Reading files from URLs is not supported.
\item nanoparquet always reads the data (or the selected subset of it) into
memory. It does not work with out-of-memory data in Parquet files like
Apache Arrow and DuckDB does.
}
}

\subsection{Installation}{

Install the R package from CRAN:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{install.packages("nanoparquet")
}\if{html}{\out{</div>}}
}

\subsection{Usage}{
\subsection{Read}{

Call \code{read_parquet()} to read a Parquet file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{df <- nanoparquet::read_parquet("example.parquet")
}\if{html}{\out{</div>}}

To see the columns of a Parquet file and how their types are mapped to
R types by \code{read_parquet()}, call \code{read_parquet_schema()} first:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nanoparquet::read_parquet_schema("example.parquet")
}\if{html}{\out{</div>}}

Folders of similar-structured Parquet files (e.g. produced by Spark)
can be read like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{df <- data.table::rbindlist(lapply(
  Sys.glob("some-folder/part-*.parquet"),
  nanoparquet::read_parquet
))
}\if{html}{\out{</div>}}
}

\subsection{Write}{

Call \code{write_parquet()} to write a data frame to a Parquet file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nanoparquet::write_parquet(mtcars, "mtcars.parquet")
}\if{html}{\out{</div>}}

To see how the columns of the data frame will be mapped to Parquet types
by \code{write_parquet()}, call \code{infer_parquet_schema()} first:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nanoparquet::infer_parquet_schema(mtcars)
}\if{html}{\out{</div>}}
}

\subsection{Inspect}{

Call \code{read_parquet_info()}, \code{read_parquet_schema()}, or
\code{read_parquet_metadata()} to see various kinds of metadata from a Parquet
file:
\itemize{
\item \code{read_parquet_info()} shows a basic summary of the file.
\item \code{read_parquet_schema()} shows all columns, including non-leaf columns,
and how they are mapped to R types by \code{read_parquet()}.
\item \code{read_parquet_metadata()} shows the most complete metadata information:
file meta data, the schema, the row groups and column chunks of the
file.
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{nanoparquet::read_parquet_info("mtcars.parquet")
nanoparquet::read_parquet_schema("mtcars.parquet")
nanoparquet::read_parquet_metadata("mtcars.parquet")
}\if{html}{\out{</div>}}

If you find a file that should be supported but isn't, please open an
issue here with a link to the file.
}

}

\subsection{Options}{

See also \code{?parquet_options()} for further details.
\itemize{
\item \code{nanoparquet.class}: extra class to add to data frames returned by
\code{read_parquet()}. If it is not defined, the default is \code{"tbl"},
which changes how the data frame is printed if the pillar package is
loaded.
\item \code{nanoparquet.compression_level}: See \code{?parquet_options()} for the
defaults and the possible values for each compression method. \code{Inf}
selects maximum compression for each method.
\item \code{nanoparquet.num_rows_per_row_group}: The number of rows to put into a
row group by \code{write_parquet()}, if row groups are not specified
explicitly. It should be an integer scalar. Defaults to 10 million.
\item \code{nanoparquet.use_arrow_metadata}: unless this is set to \code{FALSE},
\code{read_parquet()} will make use of Arrow metadata in the Parquet file.
Currently this is used to detect factor columns.
\item \code{nanoparquet.write_arrow_metadata}: unless this is set to \code{FALSE},
\code{write_parquet()} will add Arrow metadata to the Parquet file.
This helps preserving classes of columns, e.g. factors will be read
back as factors, both by nanoparquet and Arrow.
\item \code{nanoparquet.write_data_page_version}: Data version to write by default.
Possible values are 1 and 2. Default is 1.
\item \code{nanoparquet.write_minmax_values}: Whether to write minimum and maximum
values per row group, for data types that support this in
\code{write_parquet()}.
}
}

\subsection{License}{

MIT
}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/r-lib/nanoparquet}
  \item \url{https://nanoparquet.r-lib.org/}
  \item Report bugs at \url{https://github.com/r-lib/nanoparquet/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Authors:
\itemize{
  \item Gábor Csárdi \email{csardi.gabor@gmail.com}
  \item Hannes Mühleisen (\href{https://orcid.org/0000-0001-8552-0029}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Google Inc. [copyright holder]
  \item Apache Software Foundation [copyright holder]
  \item Posit Software, PBC [copyright holder]
  \item RAD Game Tools [copyright holder]
  \item Valve Software [copyright holder]
  \item Tenacious Software LLC [copyright holder]
  \item Facebook, Inc. [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nanoparquet-package.R
\name{nanoparquet-types}
\alias{nanoparquet-types}
\title{nanoparquet's type maps}
\description{
How nanoparquet maps R types to Parquet types.
}
\section{R's data types}{
When writing out a data frame, nanoparquet maps R's data types to Parquet
logical types. The following table is a summary of the mapping. For the
details see below.\tabular{llcl}{
   R type \tab Parquet type \tab Default \tab Notes \cr
   bit64::integer64 \tab INT64 \tab x \tab \code{NA_integer64_} marks missing values. \cr
   blob::blob \tab BYTE_ARRAY \tab x \tab Missing values are \code{NULL}. \cr
   " \tab FIXED_LEN_BYTE_ARRAY \tab  \tab All entries must have the same length. Missing values are \code{NULL}. \cr
   character \tab STRING(BYTE_ARRAY) \tab x \tab I.e. STRSXP. Converted to UTF-8. \cr
   " \tab BYTE_ARRAY \tab  \tab  \cr
   " \tab FIXED_LEN_BYTE_ARRAY \tab  \tab  \cr
   " \tab ENUM \tab  \tab  \cr
   " \tab UUID \tab  \tab  \cr
   Date \tab DATE \tab x \tab  \cr
   difftime \tab INT64 \tab x \tab If not hms::hms. Arrow metadata marks it as Duration(NS). \cr
   factor \tab STRING \tab x \tab Arrow metadata marks it as a factor. \cr
   " \tab ENUM \tab  \tab  \cr
   hms::hms \tab TIME(true, MILLIS) \tab x \tab Sub-milliseconds precision is lost. \cr
   integer \tab INT(32, true) \tab x \tab I.e. INTSXP. \cr
   " \tab INT64 \tab  \tab  \cr
   " \tab INT96 \tab  \tab  \cr
   " \tab DECIMAL(INT32) \tab  \tab  \cr
   " \tab DECIMAL(INT64) \tab  \tab  \cr
   " \tab INT(8, *) \tab  \tab  \cr
   " \tab INT(16, *) \tab  \tab  \cr
   " \tab INT(32, signed) \tab  \tab  \cr
   list \tab LIST(INT32 elements) \tab x \tab List of integer vectors. \code{NULL} entries and \code{NA} elements are supported. \cr
   " \tab LIST(DOUBLE elements) \tab x \tab List of double vectors. \code{NULL} entries and \code{NA} elements are supported. \cr
   " \tab LIST(STRING elements) \tab x \tab List of character vectors. \code{NULL} entries and \code{NA} elements are supported. \cr
   " \tab BYTE_ARRAY \tab  \tab Must be a list of raw vectors. Missing values are \code{NULL}. \cr
   " \tab FIXED_LEN_BYTE_ARRAY \tab  \tab Must be a list of raw vectors of the same length. Missing values are \code{NULL}. \cr
   logical \tab BOOLEAN \tab x \tab I.e. LGLSXP. \cr
   numeric \tab DOUBLE \tab x \tab I.e. REALSXP. \cr
   " \tab INT96 \tab  \tab  \cr
   " \tab FLOAT \tab  \tab  \cr
   " \tab DECIMAL(INT32) \tab  \tab  \cr
   " \tab DECIMAL(INT64) \tab  \tab  \cr
   " \tab INT(*, *) \tab  \tab  \cr
   " \tab FLOAT16 \tab  \tab  \cr
   POSIXct \tab TIMESTAMP(true, MICROS) \tab x \tab Sub-microsecond precision is lost. \cr
}


The non-default mappings can be selected via the \code{schema} argument. E.g.
to write out a factor column called 'name' as \code{ENUM}, use

\if{html}{\out{<div class="sourceCode r">}}\preformatted{write_parquet(..., schema = parquet_schema(name = "ENUM"))
}\if{html}{\out{</div>}}

The detailed mapping rules are listed below, in order of preference.
These rules will likely change until nanoparquet reaches version 1.0.0.
\enumerate{
\item \code{bit64::integer64} objects (from the bit64 package) are written as
\code{INT64}. nanoparquet handles any object that inherits the \code{integer64}
class this way. \code{NA_integer64_} (i.e. \code{INT64_MIN}) marks missing values.
\item \code{blob::blob} objects (from the blob package) are written as \code{BYTE_ARRAY}.
\code{blob::blob} is a list of raw vectors, and nanoparquet handles any object
that inherits the \code{blob} class this way, even if the blob package is not
installed. Missing values (i.e. \code{NULL} list entries) are supported.
\item Factors (i.e. vectors that inherit the \emph{factor} class) are converted
to character vectors using \code{as.character()}, then written as a
\code{STRSXP} (character vector) type. The fact that a column is a factor
is stored in the Arrow metadata (see below), unless the
\code{nanoparquet.write_arrow_metadata} option is set to \code{FALSE}.
\item Dates (i.e. the \code{Date} class) is written as \code{DATE} logical type, which
is an \code{INT32} type internally.
\item \code{hms} objects (from the hms package) are written as \code{TIME(true, MILLIS)}.
logical type, which is internally the \code{INT32} Parquet type.
Sub-milliseconds precision is lost.
\item \code{POSIXct} objects are written as \code{TIMESTAMP(true, MICROS)} logical type,
which is internally the \code{INT64} Parquet type.
Sub-microsecond precision is lost.
\item \code{difftime} objects (that are not \code{hms} objects, see above), are
written as an \code{INT64} Parquet type, and noting in the Arrow metadata
(see below) that this column has type \code{Duration} with \code{NANOSECONDS}
unit.
\item Integer vectors (\code{INTSXP}) are written as \code{INT(32, true)} logical type,
which corresponds to the \code{INT32} type.
\item Real vectors (\code{REALSXP}) are written as the \code{DOUBLE} type.
\item Character vectors (\code{STRSXP}) are written as the \code{STRING} logical type,
which has the \code{BYTE_ARRAY} type. They are always converted to UTF-8
before writing.
\item Logical vectors (\code{LGLSXP}) are written as the \code{BOOLEAN} type.
\item Other vectors error currently.
}

You can use \code{\link[=infer_parquet_schema]{infer_parquet_schema()}} on a data frame to map R data types
to Parquet data types.

To change the default R to Parquet mapping, use \code{\link[=parquet_schema]{parquet_schema()}} and
the \code{schema} argument of \code{\link[=write_parquet]{write_parquet()}}. Currently supported
non-default mappings are:
\itemize{
\item \code{integer} to \code{INT64},
\item \code{integer} to \code{INT96},
\item \code{double} to \code{INT96},
\item \code{double} to \code{FLOAT},
\item \code{character} to \code{BYTE_ARRAY},
\item \code{character} to \code{FIXED_LEN_BYTE_ARRAY},
\item \code{character} to \code{ENUM},
\item \code{factor} to \code{ENUM},
\item \code{integer} to \code{DECIAML} & \code{INT32},
\item \code{integer} to \code{DECIAML} & \code{INT64},
\item \code{double} to \code{DECIAML} & \code{INT32},
\item \code{double} to \code{DECIAML} & \code{INT64},
\item \code{integer} to \verb{INT(8, *)}, \verb{INT(16, *)}, \code{INT(32, signed)},
\item \code{double} to \verb{INT(*, *)},
\item \code{character} to \code{UUID},
\item \code{double} to \code{FLOAT16},
\item \code{list} of \code{integer} vectors to \code{LIST} with \code{INT32} elements,
\item \code{list} of \code{double} vectors to \code{LIST} with \code{DOUBLE} elements,
\item \code{list} of \code{character} vectors to \code{LIST} with \code{STRING} elements,
\item \code{list} of \code{raw} vectors to \code{BYTE_ARRAY},
\item \code{list} of \code{raw} vectors to \code{FIXED_LEN_BYTE_ARRAY},
\item \code{blob::blob} to \code{FIXED_LEN_BYTE_ARRAY}.
}
}

\section{Parquet's data types}{
When reading a Parquet file nanoparquet also relies on logical types and
the Arrow metadata (if present, see below) in addition to the low level
data types. The following table summarizes the mappings. See more details
below.\tabular{lll}{
   Parquet type \tab R type \tab Notes \cr
   \emph{Logical types} \tab  \tab  \cr
   BSON \tab character \tab  \cr
   DATE \tab Date \tab  \cr
   DECIMAL \tab numeric \tab REALSXP, potentially losing precision. \cr
   ENUM \tab character \tab  \cr
   FLOAT16 \tab numeric \tab REALSXP \cr
   INT(8, *) \tab integer \tab  \cr
   INT(16, *) \tab integer \tab  \cr
   INT(32, *) \tab integer \tab Large unsigned values may overflow! \cr
   INT(64, *) \tab numeric \tab REALSXP, or \code{integer64} if \code{read_int64_type} option is set. \cr
   INTERVAL \tab list(raw) \tab Missing values are \code{NULL}. \cr
   JSON \tab character \tab  \cr
   LIST \tab list \tab Elements are read as their corresponding R type. \cr
   MAP \tab  \tab Not supported. \cr
   STRING \tab factor \tab If Arrow metadata says it is a factor. Also UTF8. \cr
   " \tab character \tab Otherwise. Also UTF8. \cr
   TIME \tab hms::hms \tab Also TIME_MILLIS and TIME_MICROS. \cr
   TIMESTAMP \tab POSIXct \tab Also TIMESTAMP_MILLIS and TIMESTAMP_MICROS. \cr
   UUID \tab character \tab In \code{00112233-4455-6677-8899-aabbccddeeff} form. \cr
   UNKNOWN \tab  \tab Not supported. \cr
   \emph{Primitive types} \tab  \tab  \cr
   BOOLEAN \tab logical \tab  \cr
   BYTE_ARRAY \tab factor \tab If Arrow metadata says it is a factor. \cr
   " \tab blob::blob \tab Otherwise. Missing values are \code{NULL}. \cr
   DOUBLE \tab numeric \tab REALSXP \cr
   FIXED_LEN_BYTE_ARRAY \tab blob::blob \tab Missing values are \code{NULL}. \cr
   FLOAT \tab numeric \tab REALSXP \cr
   INT32 \tab integer \tab  \cr
   INT64 \tab numeric \tab REALSXP, or \code{integer64} if \code{read_int64_type} option is set. \cr
   INT96 \tab POSIXct \tab  \cr
}


The exact rules are below. These rules will likely change until nanoparquet
reaches version 1.0.0.
\enumerate{
\item The \code{BOOLEAN} type is read as a logical vector (\code{LGLSXP}).
\item The \code{STRING} logical type and the \code{UTF8} converted type is read as
a character vector with UTF-8 encoding.
\item The \code{DATE} logical type and the \code{DATE} converted type are read as a
\code{Date} R object.
\item The \code{TIME} logical type and the \code{TIME_MILLIS} and \code{TIME_MICROS}
converted types are read as an \code{hms} object, see the hms package.
\item The \code{TIMESTAMP} logical type and the \code{TIMESTAMP_MILLIS} and
\code{TIMESTAMP_MICROS} converted types are read as \code{POSIXct} objects.
If the logical type has the \code{UTC} flag set, then the time zone of the
\code{POSIXct} object is set to \code{UTC}.
\item \code{INT32} is read as an integer vector (\code{INTSXP}).
\item \code{INT64} is read as a real vector (\code{REALSXP}) by default. If the
\code{read_int64_type} option in \code{parquet_options()} is set to \code{"integer64"}
or \code{"bit64::integer64"}, it is read as a \code{bit64::integer64} vector
instead. \code{NA_integer64_} (i.e. \code{INT64_MIN}) marks missing values.
\item \code{DOUBLE} and \code{FLOAT} are read as real vectors (\code{REALSXP}).
\item \code{INT96} is read as a \code{POSIXct} read vector with the \code{tzone} attribute
set to \code{"UTC"}. It was an old convention to store time stamps as
\code{INT96} objects.
\item The \code{DECIMAL} converted type (\code{FIXED_LEN_BYTE_ARRAY} or \code{BYTE_ARRAY}
type) is read as a real vector (\code{REALSXP}), potentially losing
precision.
\item The \code{ENUM} logical type is read as a character vector.
\item The \code{UUID} logical type is read as a character vector that uses the
\code{00112233-4455-6677-8899-aabbccddeeff} form.
\item The \code{FLOAT16} logical type is read as a real vector (\code{REALSXP}).
\item \code{BYTE_ARRAY} is read as a \emph{factor} object if the file was written
by Arrow and the original data type of the column was a factor.
(See 'The Arrow metadata below.)
\item Otherwise \code{BYTE_ARRAY} and \code{FIXED_LEN_BYTE_ARRAY} are read as a
\code{blob::blob} object. \code{blob::blob} is a list of raw vectors with class
\code{blob} (and related vctrs classes). The blob package is not required to
use this object; it is simply a list of raw vectors. Missing values are
denoted by \code{NULL}.
}

Other logical and converted types are read as their annotated low level
types:
\enumerate{
\item \code{INT(8, true)}, \code{INT(16, true)} and \code{INT(32, true)} are read as
integer vectors because they are \code{INT32} internally in Parquet.
\item \code{INT(64, true)} is read as a real vector (\code{REALSXP}), unless the
\code{read_int64_type} option is set (see above).
\item Unsigned integer types \code{INT(8, false)}, \code{INT(16, false)} and
\code{INT(32, false)} are read as integer vectors (\code{INTSXP}). Large
positive values may overflow into negative values, this is a known
issue that we will fix.
\item \code{INT(64, false)} is read as a real vector (\code{REALSXP}), unless the
\code{read_int64_type} option is set (see above). Large positive values may
overflow into negative values, this is a known issue that we will fix.
\item \code{INTERVAL} is a fixed length byte array, and nanoparquet reads it as
a list of raw vectors. Missing values are denoted by \code{NULL}.
\item \code{JSON} columns are read as character vectors (\code{STRSXP}).
\item \code{BSON} columns are read as raw vectors (\code{RAWSXP}).
}

These types are not yet supported:
\enumerate{
\item Nested \code{LIST} types (lists of lists) are not supported.
\item The \code{MAP} logical type is not supported.
\item The \code{UNKNOWN} logical type is not supported.
}

You can use the \code{\link[=read_parquet_schema]{read_parquet_schema()}} function to see how R would read
the columns of a Parquet file. Look at the \code{r_type} column.
}

\section{The Arrow metadata}{
Apache Arrow (i.e. the arrow R package) adds additional metadata to
Parquet files when writing them in \code{arrow::write_parquet()}. Then,
when reading the file in \code{arrow::read_parquet()}, it uses this metadata
to recreate the same Arrow and R data types as before writing.

\code{nanoparquet::write_parquet()} also adds the Arrow metadata to Parquet
files, unless the \code{nanoparquet.write_arrow_metadata} option is set to
\code{FALSE}.

Similarly, \code{nanoparquet::read_parquet()} uses the Arrow metadata in the
Parquet file (if present), unless the \code{nanoparquet.use_arrow_metadata}
option is set to FALSE.

The Arrow metadata is stored in the file level key-value metadata, with
key \code{ARROW:schema}.

Currently nanoparquet uses the Arrow metadata for two things:
\itemize{
\item It uses it to detect factors. Without the Arrow metadata factors are
read as string vectors.
\item It uses it to detect \code{difftime} objects. Without the arrow metadata
these are read as \code{INT64} columns, containing the time difference in
nanoseconds.
}
}

\seealso{
\link{nanoparquet-package} for options that modify the type
mappings.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infer-parquet-schema.R
\name{infer_parquet_schema}
\alias{infer_parquet_schema}
\title{Infer Parquet schema of a data frame}
\usage{
infer_parquet_schema(df, options = parquet_options())
}
\arguments{
\item{df}{Data frame.}

\item{options}{Return value of \code{\link[=parquet_options]{parquet_options()}}, may modify the
R to Parquet type mappings.}
}
\value{
Data frame, the inferred schema. It has the same columns as
the return value of \code{\link[=read_parquet_schema]{read_parquet_schema()}}:
\code{file_name}, \code{r_col}, \code{name}, \code{r_type}, \code{type}, \code{type_length}, \code{repetition_type}, \code{converted_type}, \code{logical_type}, \code{num_children}, \code{scale}, \code{precision}, \code{field_id}.
}
\description{
Infer Parquet schema of a data frame
}
\seealso{
\code{\link[=read_parquet_schema]{read_parquet_schema()}} to read the schema of a Parquet file,
\code{\link[=parquet_schema]{parquet_schema()}} to create a Parquet schema from scratch.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/porcelain.R
\name{read_parquet_pages}
\alias{read_parquet_pages}
\title{Metadata of all pages of a Parquet file}
\usage{
read_parquet_pages(file)
}
\arguments{
\item{file}{Path to a Parquet file.}
}
\value{
Data frame with columns:
\itemize{
\item \code{file_name}: file name.
\item \code{row_group}: id of the row group the page belongs to,
an integer between 0 and the number of row groups
minus one.
\item \code{column}: id of the column. An integer between the
number of leaf columns minus one. Note that only leaf
columns are considered, as non-leaf columns do not
have any pages.
\item \code{page_type}: \code{DATA_PAGE}, \code{INDEX_PAGE}, \code{DICTIONARY_PAGE} or
\code{DATA_PAGE_V2}.
\item \code{page_header_offset}: offset of the data page (its header) in the
file.
\item \code{uncompressed_page_size}: does not include the page header, as per
Parquet spec.
\item \code{compressed_page_size}: without the page header.
\item \code{crc}: integer, checksum, if present in the file, can be \code{NA}.
\item \code{num_values}: number of data values in this page, include
\code{NULL} (\code{NA} in R) values.
\item \code{encoding}: encoding of the page, current possible encodings:
"PLAIN", "GROUP_VAR_INT", "PLAIN_DICTIONARY", "RLE", "BIT_PACKED", "DELTA_BINARY_PACKED", "DELTA_LENGTH_BYTE_ARRAY", "DELTA_BYTE_ARRAY", "RLE_DICTIONARY", "BYTE_STREAM_SPLIT".
\item \code{definition_level_encoding}: encoding of the definition levels,
see \code{encoding} for possible values. This can be missing in V2 data
pages, where they are always RLE encoded.
\item \code{repetition_level_encoding}: encoding of the repetition levels,
see \code{encoding} for possible values. This can be missing in V2 data
pages, where they are always RLE encoded.
\item \code{data_offset}: offset of the actual data in the file.
\item \code{page_header_length}: size of the page header, in bytes.
}
}
\description{
Metadata of all pages of a Parquet file
}
\details{
Reading all the page headers might be slow for large files, especially
if the file has many small pages.
}
\examples{
file_name <- system.file("extdata/userdata1.parquet", package = "nanoparquet")
nanoparquet:::read_parquet_pages(file_name)
}
\seealso{
\code{\link[=read_parquet_page]{read_parquet_page()}} to read a page.
}
\keyword{internal}
